
static char RcsId[] =
"$Id: main.c,v 3.67 1997/04/10 12:30:24 nakashim Exp nakashim $";

/* main.c 1995/7/18 */

/* M-on-RISC Dynamically Translating Simulator */
/*       Copyright (C) 1995 by FUJITSU LIMITED */
/*                 Primary writer: Y.Nakashima */
/*          nakashim@pine.mfd.cs.fujitsu.co.jp */

#include <stdio.h>
#include <errno.h>
#include <fcntl.h>
#include <sgtty.h>
#include <setjmp.h>
#include <sys/signal.h>
#include "sdef.h"

#ifdef SPARC
#include <sys/ttold.h>
unsigned int swap2(a) unsigned short a;
{ return (a); }

unsigned int swap3(a) unsigned int a;
{ return (a); }

unsigned int swap4(a) unsigned int a;
{ return (a); }

llu swap8(a) llu a;
{ return (a); }
#endif

#ifdef INTEL
#include <termcap.h>
unsigned int swap2(a) unsigned short a;
{ return ((((a)&0xff)<<8)|((a)>>8)); }

unsigned int swap3(a) unsigned int a;
{ return ((((a)&0xff)<<16)|((a)&0xff00)|(a>>16)); }

unsigned int swap4(a) unsigned int a;
{ return (((a)<<24)|(((a)&0xff00)<<8)|(((a)&0xff0000)>>8)|((a)>>24)); }

llu swap8(a) llu a;
{ return (( a                    <<56)|((a&0x00000000ff00LL)<<40)
         |((a&0x00000000ff0000LL)<<24)|((a&0x0000ff000000LL)<< 8)
         |((a&0x0000ff00000000LL)>> 8)|((a&0xff0000000000LL)>>24)
         |((a&0xff000000000000LL)>>40)|( a                  >>56)); }
#endif

void	    onintr_sigwinch();
void	    onintr_cmd();
void	    onintr_sys();
long	    gettime();
char	    *getenv();
char	    *version();
char	    logfile[32]; /* name of log_file */
FILE	    *log_fp;
FILE	    *m_fp;       /* for M object */
FILE	    *sparc_fp;   /* for SPARC object */
int         sparc_size;  /* program size */
int	    sc_lines;	 /* # of lines of terminal */
int         sc_noupdate; /* redraw switch */
int	    sigwinint;	 /* SIGWINCH interrupt */
int         cmdintr;     /* command interrupt */
jmp_buf	    sjbuf;
int         dim_version;
struct m_op *optable;
struct m_op *optb2xx;

#include "../common/mdef.h"
#include "../common/util.c"

main(argc, argv)
  int argc;
  char **argv;
{
#define	BUFSIZE	1024
  char	buf[BUFSIZE];
  char	*cmdP;
  char	*termtype;
  struct tchars TCHARS;
#ifdef IGNOREEOF
  char	save_eofc;
#endif
  unsigned sw_disasm, sw_step, sw_interp, sw_strace;
  unsigned addr, step, data, i, j;
  unsigned char *sp, *ep;
  unsigned gindex0;
  struct guest* guest0;
  unsigned char* gmem0;
  unsigned char* gemm0;

  termtype = getenv("TERM");
  switch(tgetent(buf, termtype)) {
  case -1:
    fprintf(stderr, "can't read termcap\n");
    exit(1);
    break;
  case 0:
    fprintf(stderr, "can't find your terminal type (%s) in termcap\n", termtype);
    exit(1);
    break;
  default:
    sc_lines = tgetnum("li");
    break;
  }

  if (argc != 3) {
    fprintf(stderr, "usage: %s <SPARC-object> <M-object>\n", argv[0]);
    exit(1);
  }

  if ((sparc_fp = fopen(argv[1], "r")) == NULL) {
    fprintf(stderr, "can't read \"%s\"\n", argv[1]);
    exit(1);
  }

  if ((m_fp = fopen(argv[2], "r")) == NULL) {
    fprintf(stderr, "can't read \"%s\"\n", argv[2]);
    exit(1);
  }

  sprintf(logfile ,"/tmp/.dimlog%u", getpid());
  if ((log_fp = fopen(logfile, "w")) == NULL) {
    fprintf(stderr, "can't create \"%s\"\n", logfile);
    exit(1);
  }

/* SPARC V8 memory structure
!
! %fp-->  +-----------------------------------------+ ---
! %fp-off.|automatic var. not used  (%fp not used)  |  ^
!	  |-----------------------------------------|  |
! %sp+100 |%i7 is saved here			    |  size = 12
! %sp+96  |  (reserved)				    |  |
! %sp+92  |%i0 is saved here (struct *guest)	    |  v
!	  |-----------------------------------------| ---
! %sp+88  |1 word for calee			    |  ^
! %sp+84  |1 word for calee			    |  |
! %sp+80  |1 word for calee			    |  |
! %sp+76  |1 word for calee			    |  |
! %sp+72  |1 word for calee			    |  |
! %sp+68  |1 word for calee			    |  |
!	  |-----------------------------------------|  |
! %sp+off.|1 word hidden param			    |  size = 92
!	  |-----------------------------------------|  |
! %sp+off.|16 words for in,local (used by OS)	    |  v
! %sp-->  +-----------------------------------------+ ---
!
*/

  if ((mem = (unsigned char*)malloc(SPARC_MEMSIZE)) == NULL){
    message("Can't malloc SPARC mem\n");
    exit(1);
  }
  gindex0 = 0x00975800;
  gmem0   = (unsigned char*)&mem[0x00a00000];
  gemm0   = (unsigned char*)&mem[0x00975000];
  guest0  = (struct guest*)&mem[gindex0];
  guest0->guest_mem     = (unsigned char*) swap4(0x00a00000);
  guest0->host_imap     = (unsigned*)      swap4(0x00c00000);
  guest0->host_ibufpp   = (unsigned char**)swap4(0x00976000);
  guest0->ibuf          = (unsigned char*) swap4(0x00980000);
  guest0->freep         = (unsigned char*) swap4(0x00980000);
  guest0->interp_ibuf   = (unsigned char*) swap4(0x00978000);
  guest0->guest_memsize = swap4(GUEST_MEMSIZE);
  guest0->host_imapsize = swap4((GUEST_MEMSIZE/2)*4);
  guest0->host_bufpsize = swap4((GUEST_MEMSIZE/GUEST_PAGESIZE)*16);
  guest0->g1            = swap4(0x00000000);
  guest0->g2            = swap4(0x00000000);
  guest0->g3            = swap4(0x00ffffff);
  guest0->g4            = (unsigned)(guest0->guest_mem);
  guest0->g5            = (unsigned)(guest0->host_imap);
  guest0->g7            = swap4(0x00000800); /* IM=0,AM=0,PROG=0 */
  guest0->pa_trans      = swap4(0x00000000);
  guest0->cslocka       = swap4(0x009757fc);
  guest0->emmptr        = swap4(0x00975000);

  /* initialize m[],%sp,%i7 */
  for (i=0; i<8; i++) {
    m[i] = &g[i];
    m[8+i] = &o[i]; /* out reg */
  }
  o[6] = SPARC_STKBOTM; /* stack pointer */
  o[7] = 0x00000000; /* ret addr */
  
  message("Reading SPARC program\n");
  sparc_size = read_sparc(sparc_fp);

  dim_version         = swap4(*(unsigned*)&mem[8]);
  optable = (struct m_op*)&mem[swap4(*(unsigned*)&mem[12])];
  optb2xx = (struct m_op*)&mem[swap4(*(unsigned*)&mem[16])];
  guest0->pa_sinsn    = *(unsigned*)&mem[20];
  guest0->pa_pinsn    = *(unsigned*)&mem[24];
  guest0->br_hist_top = *(unsigned*)&mem[28];
  guest0->br_hist_cp  = *(unsigned*)&mem[28];

/* M memory structure
   åե졼  åХեե졼

    ^   SP(R13)                           EP(R12)
  |   +--------------------------+        +--------------------------+
  +-|-->|  0x10񤭹ޤ      |+ 0     |                          |+0
  | |   |--------------------------|        |--------------------------|
  | +---|   SP               |+ 4     |                          |+4
  |   |--------------------------|        |--------------------------|
  |   |                          |+ 8     |                          |+8
  |   |--------------------------|        |--------------------------|
  |   |  M14(ret-addr)       |+12     |  SP                |+12
  |   |--------------------------|        |--------------------------|
  |   |  M15(callƬ)       |+16     |                          |
  |   |--------------------------|        |                          |
  |   |  M0                  |+20     |                          |
  |   |--------------------------|      |                          |
  |   |  M12                 |+68     |                          |
  |   |--------------------------|        |                          |
  |   |SPΥԡ(any)   |+72     |                          |
  |   |--------------------------|        |                          |
  |  +--| SP                   |+76     |                          |
  |  |  |--------------------------|      |--------------------------|
  |  |  |                          |    +---|  Ķ      |+116
  |  |  |                          |    |   |--------------------------|
  |  |  |                          |    +-->|LA 0,0xFFE(0,0)41000FFE   |+120
  |  |  +--------------------------+        |--------------------------|
  |  +->|                          |        |BC 0,0xFED(0,0)47000FED   |+124
  |     |--------------------------|        |--------------------------|
  +-----|                          |   R14->|LA 0,0xFFF(0,0)41000FFF   |+128
        |--------------------------|        |--------------------------|
        |                          |        |BC 0,0xFED(0,0)47000FED   |+132
                                            |--------------------------|
                                            |          ͽ            |+136
                                            |--------------------------|
  Ͼ̴ؿ¦Ѱ                      |    |
                                            |--------------------------|
                                            |          ͽ            |+252
                                            +--------------------------+

  R11  Ƭɥ쥹
  R12  SPĶΤι¤   dimư0x1FFF00Ȥ
                                        0x1FFF000x1FFFFF˱ޤǼ
                                        (+124)dimĥ̿SPOV񤤤Ƥ
  R13  SP                           dimư0x100000(1MB)
  R14  return-addr(̿Ƭ)  dimư0x1FFF80     
  R15  callƬɥ쥹             dimư0x000000(gccmain)
*/

  guest0->m[12] = swap4(GUEST_MEMSIZE-256);
  guest0->m[13] = swap4(GUEST_MEMSIZE-GUEST_STKSIZE);
  guest0->m[14] = swap4(GUEST_MEMSIZE-128);
  sp = gmem0+GUEST_MEMSIZE-GUEST_STKSIZE;
  *(unsigned*)(sp+ 72) = swap4(0x12345678);
  *(unsigned*)(sp+ 76) = swap4(GUEST_MEMSIZE-GUEST_STKSIZE+80);
  ep = gmem0+GUEST_MEMSIZE-256;
  *(unsigned*)(ep+ 12) = swap4(GUEST_MEMSIZE-256);
  *(unsigned*)(ep+116) = swap4(GUEST_MEMSIZE-136);
  *(unsigned*)(ep+120) = swap4(0x41000FFE);
  *(unsigned*)(ep+124) = swap4(0x47000FED);
  *(unsigned*)(ep+128) = swap4(0x41000FFF);
  *(unsigned*)(ep+132) = swap4(0x47000FED);

  message("Reading M program\n");
  addr = -1;
  while ((i = getc(m_fp)) != EOF) {
    if (++addr > GUEST_MEMSIZE-GUEST_STKSIZE-1) {
      message("object too large (size>%d)\n", GUEST_MEMSIZE-GUEST_STKSIZE-1);
      exit(1);
    }
    gmem0[addr] = i;
  }
  fclose(m_fp);
  message(" M object=0x00000000-0x%08.8x\n", addr);

#ifdef IGNOREEOF
  ioctl(0, TIOCGETC, &TCHARS);
  save_eofc = TCHARS.t_eofc;
  TCHARS.t_eofc = 'w'-0x60;
  ioctl(0, TIOCSETC, &TCHARS);
#endif
  /* system("stty tabs"); */
  scr_setr(5);
  signal(SIGWINCH, onintr_sigwinch);

  message("\nM-on-SPARC-on-ANY Dynamically Translating Simulator\n");
  message("Version %s\n", version());
  message(" SPARC memory=0x00000000-0x%08.8x\n", SPARC_MEMSIZE-1);
  message(" SPARC object=0x00000000-0x%08.8x\n", sparc_size-1);
  message(" SPARC stack =0x00000000-0x%08.8x\n", SPARC_STKBOTM-1);
  message(" M gcpu_s=0x%08.8x\n", gindex0);
  message(" M memory=0x%08.8x memsize=%u\n",
            swap4(guest0->g4), GUEST_MEMSIZE);
  message(" M imap  =0x%08.8x mapsize=%u\n",
            swap4(guest0->g5), swap4(guest0->host_imapsize));
  message(" M ibufpp=0x%08.8x\n", swap4(guest0->host_ibufpp));
  message(" M ibuf  =0x%08.8x i_ibuf=0x%08.8x",
            swap4(guest0->ibuf), swap4(guest0->interp_ibuf));
  message(" ibuf_psize=%u ibuf_size=%u\n", IBUF_PAGESIZE, IBUF_SIZE);
  message(" M object=0x%08.8x-0x%08.8x (0x00000000-0x%08.8x)\n",
            swap4(guest0->g4), swap4(guest0->g4)+addr, addr);
  message(" M stack =0x%08.8x-0x%08.8x (0x%08.8x-0x%08.8x)\n", 
            swap4(guest0->g4)+GUEST_MEMSIZE-GUEST_STKSIZE,
            swap4(guest0->g4)+GUEST_MEMSIZE-257,
            GUEST_MEMSIZE-GUEST_STKSIZE, GUEST_MEMSIZE-257);
  message(" M stkov =0x%08.8x-0x%08.8x (0x%08.8x-0x%08.8x)\n",
            swap4(guest0->g4)+GUEST_MEMSIZE-256,
            swap4(guest0->g4)+GUEST_MEMSIZE-1,
            GUEST_MEMSIZE-256, GUEST_MEMSIZE-1);
  message(" logfile =\"%s\"\n", logfile);

  addr = 0;
  step = 10;
  for (;;) {
    signal(SIGINT, SIG_IGN);
    setjmp(sjbuf);
    signal(SIGINT, onintr_cmd);

    sc_noupdate = 0;
    disp_stat(guest0);
    message("-> ");
    buf[0] = '\0';
    sigwinint = 0;
    if (fgets(cmdP = buf, BUFSIZE, stdin) == NULL) {
      if (!sigwinint)
        goto encounter_eof;
    }
    log_output(buf);

    switch (*cmdP++) {
    case '\0':
    case '\r':
    case '\n':
    case ' ':
      break; /* do nothing */

    case 'h':
      message("^C ............... stop\n");
      message("h ................ help\n");
      message("q ................ quit\n");
      message("l ................ display log\n");
      message("d [addr [step]] .. disassemble\n");
      message("m [addr [data]] .. dump/set guest memory\n");
      message("z [addr] ......... dump host memory\n");
      message("a [addr] ......... set IAR\n");
      message("g[@] [addr] ...... go\n");
      message("s[@] [step] ...... step exec with no disasm\n");
      message("t[@] [step] ...... step exec with M disasm\n");
      message("G[@] [addr] ...... interpretive go\n");
      message("S[@] [step] ...... interpretive step exec with no disasm\n");
      message("T[@] [step] ...... interpretive step exec with M disasm\n");
      message("  @: means SPARC trace\n");
      break;

    case 'q':
encounter_eof:
      scr_reset();
#ifdef IGNOREEOF
      TCHARS.t_eofc = save_eofc;
      ioctl(0, TIOCSETC, &TCHARS);
#endif
      /* system("stty -tabs"); */
      unlink(logfile);
      exit(0);
      break;

    case 'l':
      sprintf(buf, "lv '-P%%f(%%i of %%m) [line %%lb/%%L %%pB\\%%]$' +G %s", logfile);
      req_system(buf);
      break;

    case 'd':
      sscanf(cmdP, "%x %d", &addr, &step);
      if (!step)
        step = 1;
      for (i=0; i<step; i++) {
        if (addr > GUEST_MEMSIZE-1) {
          message("address too large (addr>0x%06.6x)\n", GUEST_MEMSIZE-1);
          goto exit_dumpmem;
        }
        addr += disasm(guest0, gmem0, addr);
      }
exit_disassemble:
      break;

    case 'm':
      if (sscanf(cmdP, "%x %x", &addr, &data) == 2) {
	if (addr > GUEST_MEMSIZE-1 || addr+3 > GUEST_MEMSIZE-1) {
	  message("address to large (addr,addr+3>0x%06.6x)\n", GUEST_MEMSIZE-1);
	  goto exit_dumpmem;
	}
	message("%06.6x: %08.8x -> %08.8x\n",
		addr,
		gmem0[addr  ]<<24
	       |gmem0[addr+1]<<16
	       |gmem0[addr+2]<< 8
	       |gmem0[addr+3],
		data);
	gmem0[addr  ] = data>>24;
	gmem0[addr+1] = data>>16;
	gmem0[addr+2] = data>> 8;
	gmem0[addr+3] = data    ;
      }
      else {
	message("  addr +0 +1 +2 +3 +4 +5 +6 +7  ");
	message("+8 +9 +a +b +c +d +e +f  ");
	message("0123456789abcdef\n");
	for (i = 0; i < 8; i++) {
	  message("%06.6x", addr);
	  for (j = 0; j < 16; j++) {
	    if (j == 8) message(" ");
	    if (addr+j > GUEST_MEMSIZE-1) {
	      message("address to large (addr>0x%06.6x)\n", GUEST_MEMSIZE-1);
	      goto exit_dumpmem;
	    }
	    message(" %02.2x", gmem0[addr+j]);
	  }
	  message("  ");
	  for (j = 0; j < 16; j++)
	    message("%c", e2adump[gmem0[addr+j]]);
	  message("\n");
	  addr += 16;
	}
      }
exit_dumpmem:
      break;

    case 'z':
      sscanf(cmdP, "%x", &addr);
      message("    addr +0 +1 +2 +3 +4 +5 +6 +7  ");
      message("+8 +9 +a +b +c +d +e +f\n");
      for (i = 0; i < 8; i++) {
        message("%08.8x", addr);
        for (j = 0; j < 16; j++) {
          if (j == 8) message(" ");
          message(" %02.2x", mem[addr+j]);
        }
        message("\n");
        addr += 16;
      }
      break;

    case 'a':
      guest0->g2 = swap4(guest0->g2);
      sscanf(cmdP, "%x", &(guest0->g2));
      guest0->g2 = swap4(guest0->g2);
      break;

    case 'g':
    case 's':
    case 't':
    case 'G':
    case 'S':
    case 'T':
      if (*(cmdP-1) == 'G' || *(cmdP-1) == 'S'
       || *(cmdP-1) == 'T')
	sw_interp = 1;
      else
	sw_interp = 0;
      if (*(cmdP-1) == 's' || *(cmdP-1) == 'S'
       || *(cmdP-1) == 't' || *(cmdP-1) == 'T')
	sw_step = 1;
      else
	sw_step = 0;
      if (*(cmdP-1) == 't' || *(cmdP-1) == 'T')
	sw_disasm = 1;
      else
	sw_disasm = 0;
      if (*cmdP == '@') {
	cmdP++;
	sw_strace = 1;
      }
      else
	sw_strace = 0;

      if (sw_step) {
	sscanf(cmdP, "%d", &step);
	if (!step)
	  step = 1;
      }
      else {
        guest0->g2 = swap4(guest0->g2);
	sscanf(cmdP, "%x", &(guest0->g2));
        guest0->g2 = swap4(guest0->g2);
      }

      cmdintr = 0;
      signal(SIGINT, SIG_IGN);
      signal(SIGINT, onintr_sys);
      if (sw_interp)
        guest0->g7 |= swap4( 0x00000020);
      else
        guest0->g7 &= swap4(~0x00000020);

      clrtime();
#ifdef SPARC
      asm("ta 6"); /* activate alignment-free */
#endif
      for (i=0; !sw_step || i<step; i++) {
        m_step++;
        if (sw_disasm)
          disasm(guest0, gmem0, swap4(guest0->g2));
	guest0->g7 &= swap4(~0xc1ff0000);
	if (sw_step)
	  guest0->g7 |= swap4(0x01000000);
	g[7] = swap4(guest0->g7);/*asm("ld %0,%%g7":"=m"(guest0->g7)::"%g7");*/
        o[0] = gindex0;
        ssim(sw_strace); /* retval = o[0] */
	guest0->g7 = swap4(g[7]);/*asm("st %%g7,%0":"=m"(guest0->g7)::"%g7");*/
        if (o[0]) {
          message("SPARC retval=%08.8x\n", o[0]);
          goto exit_loop;
        }

#include "../common/control.c"

        if (cmdintr) { /* command break */
          message("command break detected!\n");
          break;
        }
        if (sw_step)
	  disp_stat(guest0);
      }
exit_loop:
      message("M stopped (time=%umsec sparc_step=0x%08.8x_%08.8x)\n",
               gettime(),
               (unsigned)(sparc_step>>32), (unsigned)(sparc_step));
      break;

    default:
      message("unknown command: \"%c\"\n", *(cmdP-1));
      break;
    }
  }
  exit(0);
}

void
onintr_sigwinch()
{
  struct winsize ws;

  if (ioctl (1, TIOCGWINSZ, &ws) != -1) {
    if (ws.ws_row != 0)
      sc_lines = ws.ws_row;
  }
  scr_reset();
  scr_setr(5);
  sigwinint = 1;
}

void
onintr_cmd()
{
  message("!\n");
  longjmp(sjbuf, 1);
}

void
onintr_sys()
{
  message("!\n");
  cmdintr = 1;
  g[7] |= 0x01000000;
}

