
/* def.h 1995/7/18 */

/* M-on-RISC Dynamically Translating Simulator */
/*       Copyright (C) 1995 by FUJITSU LIMITED */
/*                 Primary writer: Y.Nakashima */
/*          nakashim@pine.mfd.cs.fujitsu.co.jp */

#ifndef LLU_DEFINED
typedef long long unsigned llu;
typedef long long int      ll;
#define LLU_DEFINED
#endif

#define ULTRA_TLB_SIZE 64
llu  i_tlb_tag0[ULTRA_TLB_SIZE];
llu  i_tlb_data0[ULTRA_TLB_SIZE];
llu  d_tlb_tag0[ULTRA_TLB_SIZE];
llu  d_tlb_data0[ULTRA_TLB_SIZE];

unsigned (*call_obp)();
unsigned serial_ih;
struct cif_struct {
    llu		name;
    llu		nargs;
    llu		nreturns;
    llu		param[16];
} cif_struct0;

#define DEV_TREE_SIZE 1024
union dev_tree {
  struct node {
    union dev_tree *prop;
    union dev_tree *parent;
    union dev_tree *peer;
    union dev_tree *child;
  } node;
  struct prop {
    unsigned short flag;  /* 0:empty,   1:node,    2:prop     3:lastprop */
    unsigned short ptype; /* 0:byte,    1:int32,   2:int64,   3:string   */
                          /* 4:range    5:trans                          */
    char           *name;
    char           *value;
    int            len;
  } prop;
} dev_tree0[DEV_TREE_SIZE];
union dev_tree *dev_treeP;

#define DEV_DATA_SIZE 32768
char dev_data0[DEV_DATA_SIZE];
char *dev_dataP;

unsigned dtree_system_clock_frequency;
unsigned dtree_mem_page_size;
unsigned dtree_cpu_manufacturer;
unsigned dtree_cpu_implementation;
unsigned dtree_cpu_ecache_associativity;
unsigned dtree_cpu_ecache_line_size;
unsigned dtree_cpu_ecache_size;
unsigned dtree_cpu_dtlb_entries;
unsigned dtree_cpu_dcache_associativity;
unsigned dtree_cpu_dcache_line_size;
unsigned dtree_cpu_dcache_size;
unsigned dtree_cpu_itlb_entries;
unsigned dtree_cpu_icache_associativity;
unsigned dtree_cpu_icache_line_size;
unsigned dtree_cpu_icache_size;
unsigned dtree_cpu_upa_portid;
unsigned dtree_cpu_clock_frequency;

#define MEM_AVAIL_SIZE 32
struct dtree_mem_avail {
  llu start;
  llu len;
} dtree_mem_avail0[MEM_AVAIL_SIZE];

#ifndef GTLB_SIZE
#define GTLB_SIZE	1024
#endif

#ifndef IBUF_PAGESIZE
#define IBUF_PAGESIZE	32768
#endif

#ifndef IBUF_SIZE
#define IBUF_SIZE	32768*16
#endif

#ifndef GUEST_PAGESIZE
#define GUEST_PAGESIZE  4096
#endif

#ifndef GUEST_MEMSIZE
#define GUEST_MEMSIZE   1048576 /* default size is 1MB */
#endif

#ifndef GUEST_STKSIZE
#define GUEST_STKSIZE   524288 /* default size is 512KB */
#endif

struct m_op {
    int		mlen;  /* length          */
    unsigned	ifunc; /* int (*ifunc)(); */
    unsigned	name;  /* char *name;     */
    unsigned	args;  /* char *args;     */
    unsigned	cfunc; /* int (*cfunc)(); */
    unsigned	tfunc; /* int (*tfunc)(); */
    int		type;  /* trans_type      */
    unsigned	rsv1;  /* reserved        */
};

struct gtlb {
    llu    gs_if;
    llu    ds_la;
    llu    gs_ld;
    llu    gs_st;
};

struct gcpu_s { /* must be aligned to 128B boundary */
    unsigned	gr[16];		/* +  0 ref by dim/emm */
    llu		fr[8];		/* + 64 ref by dim/emm */
    unsigned	ar[16];		/* +128 ref by dim/emm */
    unsigned	cr[16];		/* +192 ref by dim/emm */

    llu		xcr[16];	/* +256 ref by dim/emm */
    unsigned	ecr[16];	/* +384 ref by dim/emm */
    llu		psw;		/* +448 ref by dim/emm */
    unsigned	fcr;		/* +456 ref by dim/emm */
    unsigned	iomr;		/* +460 ref by dim/emm */
    unsigned	rsv1[12];	/* +464 - +508	       */

    llu		rr[16];		/* +512 ref by dim/emm */

    llu         g0; /* reserved /* +640 (ignoded)      */
    llu         g1; /* M-CC     /* +648 ref by dim/emm */
    llu         g2; /* IAR      /* +656 ref by dim/emm */
    llu         g3; /* A-MASK   /* +664 ref by dim/emm */
    llu         g4; /* OP-BASE  /* +672 ref by dim/emm */
    llu         g5; /* IF-BASE  /* +680 ref by dim/emm */
    llu         g6; /* reserved /* +688 (ignored)      */
    llu         g7; /* reserved /* +696 (ignored)      */

    llu         minsn;          /* +704 ref by dim/emm */
    unsigned    csaddr;         /* +712 ref by dim/emm */
    unsigned    cslocka;        /* +716 ref by dim/emm */
    unsigned    rsv2[7];        /* +720 - +744         */
    unsigned	pa_minsn;	/* +748 ref by dim/emm */
    unsigned	pa_sinsn;	/* +752 ref by dim/emm */
    unsigned	pa_pinsn;	/* +756 ref by dim/emm */
    unsigned	br_hist_top;	/* +760 ref by dim/emm */
    unsigned	br_hist_cp;	/* +764 ref by dim/emm */

    llu		param00[2];	/* +768 ref by dim     */
    llu		param01[2];	/* +784 ref by dim     */
    llu		param10[2];	/* +800 ref by dim     */
    llu		param11[2];	/* +816 ref by dim     */

    unsigned    mlen;           /* +832 ref by (dim)   */
    unsigned    cfunc;          /* +836 ref by (dim)   */
    unsigned    tfunc;          /* +840 ref by (dim)   */
    unsigned    type;           /* +844 ref by (dim)   */
    unsigned    slen;           /* +848 ref by (dim)   */
    unsigned    flusha;         /* +852 ref by (dim)   */
    unsigned    ibufbpa;        /* +856 ref by (dim)   */
    unsigned    rsv4;           /* +860                */

    unsigned    guest_memsize;  /* +864 ->g3   (dim)   */
    unsigned    guest_mem;      /* +868 ->g4   (dim)   */
    unsigned    host_imapsize;  /* +872 ->g3   (dim)   */
    unsigned    host_imap;      /* +876 ->g5   (dim)   */
    unsigned    host_bufpsize;  /* +880 (alloc)(dim)   */
    unsigned    host_ibufpp;    /* +884 (alloc)(dim)   */
    unsigned    ibuf;           /* +888 (alloc)(dim)   */
    unsigned    freep;          /* +892 (alloc)(dim)   */
    unsigned    interp_ibuf;    /* +896 ref by (dim)   */
    unsigned	rsv5;		/* +900                */

    unsigned	sta_key;	/* +904 ref by dim/emm */
    unsigned	acc_mask;	/* +908 ref by dim/emm */
    unsigned	rsv6[28];	/* +912 - +1020	       */

    struct gtlb tlb[GTLB_SIZE];	/* +1K - +1K+32K */
};

