
static char RcsId[] =
"$Id: fltlib.c,v 1.11 1997/06/11 02:39:51 nakashim Exp nakashim $";

/* fltlib.c 1995/7/17 */

/* M-on-RISC DAT equipped Instruction Simulator            */
/* All Rights Reserved, Copyright (C) Fujitsu Limited 1997 */
/*   Licensed Material Program Property of Fujitsu Limited */
/*		               Primary writer: Y.Nakashima */
/*	                         nakashim@cs.fujitsu.co.jp */

#ifdef WITHOS
#include <stdio.h>
#include <sys/times.h>
#endif

#define FIX_OVF_MASK 8
#define DEC_OVF_MASK 4
#define EXP_UNF_MASK 2
#define SIGNIFI_MASK 1

#define FIX_OVF 0x00080000
#define EXP_OVF 0x000c0000
#define EXP_UNF 0x000d0000
#define SIGNIFI 0x000e0000
#define FLO_DIV 0x000f0000

typedef long long unsigned llu;

llu
__lshrdi3(src, sft)
  llu      src;
  unsigned sft;
{
  union { llu ll; unsigned i[2]; } dst;

  if (sft == 0)
    return (src);

  dst.ll = src;
  if (sft >= 32) {
    dst.i[1] = dst.i[0] >> (sft - 32);
    dst.i[0] = 0;
  }
  else {
    dst.i[1] = dst.i[0] << (32 - sft) | dst.i[1] >> sft;
    dst.i[0] >>= sft;
  }
  return (dst.ll);
}

llu
__ashldi3(src, sft)
  llu      src;
  unsigned sft;
{
  union { llu ll; unsigned i[2]; } dst;

  if (sft == 0)
    return (src);

  dst.ll = src;
  if (sft >= 32) {
    dst.i[0] = dst.i[1] << (sft - 32);
    dst.i[1] = 0;
  }
  else {
    dst.i[0] = dst.i[0] << sft | dst.i[1] >> (32 - sft);
    dst.i[1] <<= sft;
  }
  return (dst.ll);
}

unsigned
fl_addsub_short_norm(dummy, r1, r2, mask, addsub)
  unsigned dummy;
  unsigned char *r1, *r2;
  unsigned mask;   /* program mask */
  unsigned addsub; /* 0:add, 1:sub */
  /* return INTR-CODE|CC */
{
  unsigned src1, src2;
  unsigned s1, s2;
  unsigned c1, c2;
  unsigned f1, f2;
  unsigned sft;
  unsigned intr = 0, cc = 0;

  src1 = *(unsigned*)r1; /* always aligned */
  src2 = *(unsigned*)r2; /* always aligned */
  src2^= addsub << 31;

  s1 = src1     & 0x80000000; /* s------- -------- -------- -------- */
  s2 = src2     & 0x80000000; /* s------- -------- -------- -------- */
  c1 = src1>>24 & 0x0000007f; /* -------- -------- -------- -ccccccc */
  c2 = src2>>24 & 0x0000007f; /* -------- -------- -------- -ccccccc */
  f1 = src1<< 4 & 0x0ffffff0; /* ----ffff ffffffff ffffffff ffff---- */
  f2 = src2<< 4 & 0x0ffffff0; /* ----ffff ffffffff ffffffff ffff---- */
                              /*                    guard digit ^^^^ */
#ifdef DEBUG_FLOAT
  umes("fl_addsub_short_norm: s1=%08.8x c1=%08.8x f1=%08.8x\n", s1, c1, f1);
  umes("fl_addsub_short_norm: s2=%08.8x c2=%08.8x f2=%08.8x\n", s2, c2, f2);
#endif

  /* fraction alignment */
  if (c1 > c2) { /* shift f2 right */
    sft = (c1 - c2) << 2;
    c2 = c1;
    f2 = (sft >= 32) ? 0 : f2 >> sft;
  }
  else if (c1 < c2) { /* shift f1 right */
    sft = (c2 - c1) << 2;
    c1 = c2;
    f1 = (sft >= 32) ? 0 : f1 >> sft;
  }
  else {
    /* do nothing if c1 == c2 */
  }

#ifdef DEBUG_FLOAT
  umes("fl_addsub_short_norm: s1=%08.8x c1=%08.8x f1=%08.8x\n", s1, c1, f1);
  umes("fl_addsub_short_norm: s2=%08.8x c2=%08.8x f2=%08.8x\n", s2, c2, f2);
#endif

  /* add algebraically */
  if (s1 != s2) {
    if (f1 >= f2)   f1 = f1 - f2;
    else { s1 = s2; f1 = f2 - f1; }
  }
  else { /* s1 == s2 */
    f1 += f2;
  }

#ifdef DEBUG_FLOAT
  umes("fl_addsub_short_norm: s1=%08.8x c1=%08.8x f1=%08.8x\n", s1, c1, f1);
#endif

  /* test carry, normalize */
  if (f1 & 0xf0000000) { /* test carry */
    f1 >>= 4;
    c1++;
    if (c1 > 127) { /* if characteristic overflow */
      c1 -= 128;
      intr = EXP_OVF; /* exponent overflow */
    }
  }
  else {
    if (f1) { /* normalize */
      if      (f1 & 0x0f000000) { /* do nothing */ }
      else if (f1 & 0x00f00000) { f1 <<=  4; c1 -= 1; }
      else if (f1 & 0x000f0000) { f1 <<=  8; c1 -= 2; }
      else if (f1 & 0x0000f000) { f1 <<= 12; c1 -= 3; }
      else if (f1 & 0x00000f00) { f1 <<= 16; c1 -= 4; }
      else if (f1 & 0x000000f0) { f1 <<= 20; c1 -= 5; }
      else                      { f1 <<= 24; c1 -= 6; }

      if (c1 & 0x80000000) { /* if characteristic underflow */
        if (mask & EXP_UNF_MASK) {
          c1 += 128;
          intr = EXP_UNF; /* exponent underflow */
        }
        else {
          s1 = c1 = f1 = 0; /* true zero */
        }
      }
    }
    else { /* fraction == 0 */
      s1 = 0;
      if (mask & SIGNIFI_MASK) {
        intr = SIGNIFI; /* significance */
      }
      else {
        s1 = c1 = f1 = 0; /* true zero */
      }
    }
  }

  /* format cc */
  if (f1) {
    if (s1 > 0) cc = 1; /* neg */
    else        cc = 2; /* pos */
  }

  /* format result */
  src1 = s1 | c1 << 24 | f1 >> 4;
  *(unsigned*)r1 = src1; /* always aligned */

#ifdef DEBUG_FLOAT
  umes("fl_addsub_short_norm: s1=%08.8x c1=%08.8x f1=%08.8x\n", s1, c1, f1);
#endif

  return (intr|cc);
}

unsigned
fl_addsub_short_unnorm(dummy, r1, r2, mask, addsub)
  unsigned dummy;
  unsigned char *r1, *r2;
  unsigned mask;   /* program mask */
  unsigned addsub; /* 0:add, 1:sub */
  /* return INTR-CODE|CC */
{
  unsigned src1, src2;
  unsigned s1, s2;
  unsigned c1, c2;
  unsigned f1, f2;
  unsigned sft;
  unsigned intr = 0, cc = 0;

  src1 = *(unsigned*)r1; /* always aligned */
  src2 = *(unsigned*)r2; /* always aligned */
  src2^= addsub << 31;

  s1 = src1     & 0x80000000; /* s------- -------- -------- -------- */
  s2 = src2     & 0x80000000; /* s------- -------- -------- -------- */
  c1 = src1>>24 & 0x0000007f; /* -------- -------- -------- -ccccccc */
  c2 = src2>>24 & 0x0000007f; /* -------- -------- -------- -ccccccc */
  f1 = src1<< 4 & 0x0ffffff0; /* ----ffff ffffffff ffffffff ffff---- */
  f2 = src2<< 4 & 0x0ffffff0; /* ----ffff ffffffff ffffffff ffff---- */
                              /*                    guard digit ^^^^ */
#ifdef DEBUG_FLOAT
  umes("fl_addsub_short_unnorm: s1=%08.8x c1=%08.8x f1=%08.8x\n", s1, c1, f1);
  umes("fl_addsub_short_unnorm: s2=%08.8x c2=%08.8x f2=%08.8x\n", s2, c2, f2);
#endif

  /* fraction alignment */
  if (c1 > c2) { /* shift f2 right */
    sft = (c1 - c2) << 2;
    c2 = c1;
    f2 = (sft >= 32) ? 0 : f2 >> sft;
  }
  else if (c1 < c2) { /* shift f1 right */
    sft = (c2 - c1) << 2;
    c1 = c2;
    f1 = (sft >= 32) ? 0 : f1 >> sft;
  }
  else {
    /* do nothing if c1 == c2 */
  }

#ifdef DEBUG_FLOAT
  umes("fl_addsub_short_unnorm: s1=%08.8x c1=%08.8x f1=%08.8x\n", s1, c1, f1);
  umes("fl_addsub_short_unnorm: s2=%08.8x c2=%08.8x f2=%08.8x\n", s2, c2, f2);
#endif

  /* add algebraically */
  if (s1 != s2) {
    if (f1 >= f2)   f1 = f1 - f2;
    else { s1 = s2; f1 = f2 - f1; }
  }
  else { /* s1 == s2 */
    f1 += f2;
  }

#ifdef DEBUG_FLOAT
  umes("fl_addsub_short_unnorm: s1=%08.8x c1=%08.8x f1=%08.8x\n", s1, c1, f1);
#endif

  /* test carry, normalize */
  if (f1 & 0xf0000000) { /* test carry */
    f1 >>= 8;
    c1++;
    if (c1 > 127) { /* if characteristic overflow */
      c1 -= 128;
      intr = EXP_OVF; /* exponent overflow */
    }
  }
  else {
    f1 >>= 4;
    if (!f1) { /* fraction == 0 */
      s1 = 0;
      if (mask & SIGNIFI_MASK) {
        intr = SIGNIFI; /* significance */
      }
      else {
        s1 = c1 = f1 = 0; /* true zero */
      }
    }
  }

  /* format cc */
  if (f1) {
    if (s1 > 0) cc = 1; /* neg */
    else        cc = 2; /* pos */
  }

  /* format result */
  src1 = s1 | c1 << 24 | f1;
  *(unsigned*)r1 = src1; /* always aligned */

#ifdef DEBUG_FLOAT
  umes("fl_addsub_short_unnorm: s1=%08.8x c1=%08.8x f1=%08.8x\n", s1, c1, f1);
#endif

  return (intr|cc);
}

unsigned
fl_addsub_long_norm(dummy, r1, r2, mask, addsub)
  unsigned dummy;
  unsigned char *r1, *r2;
  unsigned mask;   /* program mask */
  unsigned addsub; /* 0:add, 1:sub */
  /* return INTR-CODE|CC */
{
  llu      src1, src2;
  unsigned s1, s2;
  unsigned c1, c2;
  llu      f1, f2;
  unsigned sft;
  unsigned intr = 0, cc = 0;

  src1 = *(llu*)r1;        /* always aligned */
  src2 = *(llu*)r2;        /* always aligned */
  src2^= (llu)addsub << 63;

  s1 = src1>>32 & 0x80000000;           /* s------- -------- -------- -------- */
  s2 = src2>>32 & 0x80000000;           /* s------- -------- -------- -------- */
  c1 = src1>>56 & 0x0000007f;           /* -------- -------- -------- -ccccccc */
  c2 = src2>>56 & 0x0000007f;           /* -------- -------- -------- -ccccccc */
  f1 = src1<< 4 & 0x0ffffffffffffff0LL; /* ----ffff ffffff.. ffffffff ffff---- */
  f2 = src2<< 4 & 0x0ffffffffffffff0LL; /* ----ffff ffffff.. ffffffff ffff---- */
                                        /*                    guard digit ^^^^ */
#ifdef DEBUG_FLOAT
  umes("fl_addsub_long_norm: s1=%08.8x c1=%08.8x f1=%08.8x_%08.8x\n", s1, c1, (unsigned)(f1>>32), (unsigned)(f1));
  umes("fl_addsub_long_norm: s2=%08.8x c2=%08.8x f2=%08.8x_%08.8x\n", s2, c2, (unsigned)(f2>>32), (unsigned)(f2));
#endif

  /* fraction alignment */
  if (c1 > c2) { /* shift f2 right */
    sft = (c1 - c2) << 2;
    c2 = c1;
    f2 = (sft >= 64) ? 0LL : f2 >> sft;
  }
  else if (c1 < c2) { /* shift f1 right */
    sft = (c2 - c1) << 2;
    c1 = c2;
    f1 = (sft >= 64) ? 0LL : f1 >> sft;
  }
  else {
    /* do nothing if c1 == c2 */
  }

#ifdef DEBUG_FLOAT
  umes("fl_addsub_long_norm: s1=%08.8x c1=%08.8x f1=%08.8x_%08.8x\n", s1, c1, (unsigned)(f1>>32), (unsigned)(f1));
  umes("fl_addsub_long_norm: s2=%08.8x c2=%08.8x f2=%08.8x_%08.8x\n", s2, c2, (unsigned)(f2>>32), (unsigned)(f2));
#endif

  /* add algebraically */
  if (s1 != s2) {
    if (f1 >= f2)   f1 = f1 - f2;
    else { s1 = s2; f1 = f2 - f1; }
  }
  else { /* s1 == s2 */
    f1 += f2;
  }

#ifdef DEBUG_FLOAT
  umes("fl_addsub_long_norm: s1=%08.8x c1=%08.8x f1=%08.8x_%08.8x\n", s1, c1, (unsigned)(f1>>32), (unsigned)(f1));
#endif

  /* test carry, normalize */
  if (f1 & 0xf000000000000000LL) { /* test carry */
    f1 >>= 4;
    c1++;
    if (c1 > 127) { /* if characteristic overflow */
      c1 -= 128;
      intr = EXP_OVF; /* exponent overflow */
    }
  }
  else {
    if (f1) { /* normalize */
      if      (f1 & 0x0f00000000000000LL) { /* do nothing */ }
      else if (f1 & 0x00f0000000000000LL) { f1 <<=  4; c1 -=  1; }
      else if (f1 & 0x000f000000000000LL) { f1 <<=  8; c1 -=  2; }
      else if (f1 & 0x0000f00000000000LL) { f1 <<= 12; c1 -=  3; }
      else if (f1 & 0x00000f0000000000LL) { f1 <<= 16; c1 -=  4; }
      else if (f1 & 0x000000f000000000LL) { f1 <<= 20; c1 -=  5; }
      else if (f1 & 0x0000000f00000000LL) { f1 <<= 24; c1 -=  6; }
      else if (f1 & 0x00000000f0000000LL) { f1 <<= 28; c1 -=  7; }
      else if (f1 & 0x000000000f000000LL) { f1 <<= 32; c1 -=  8; }
      else if (f1 & 0x0000000000f00000LL) { f1 <<= 36; c1 -=  9; }
      else if (f1 & 0x00000000000f0000LL) { f1 <<= 40; c1 -= 10; }
      else if (f1 & 0x000000000000f000LL) { f1 <<= 44; c1 -= 11; }
      else if (f1 & 0x0000000000000f00LL) { f1 <<= 48; c1 -= 12; }
      else if (f1 & 0x00000000000000f0LL) { f1 <<= 52; c1 -= 13; }
      else                                { f1 <<= 56; c1 -= 14; }

      if (c1 & 0x80000000) { /* if characteristic underflow */
        if (mask & EXP_UNF_MASK) {
          c1 += 128;
          intr = EXP_UNF; /* exponent underflow */
        }
        else {
          s1 = c1 = 0; /* true zero */
          f1 = 0LL;    /* true zero */
        }
      }
    }
    else { /* fraction == 0 */
      s1 = 0;
      if (mask & SIGNIFI_MASK) {
        intr = SIGNIFI; /* significance */
      }
      else {
        s1 = c1 = 0; /* true zero */
        f1 = 0LL;    /* true zero */
      }
    }
  }

  /* format cc */
  if (f1) {
    if (s1 > 0) cc = 1; /* neg */
    else        cc = 2; /* pos */
  }

  /* format result */
  src1 = (llu)s1 << 32 | (llu)c1 << 56 | f1 >> 4;
  *(llu*)r1 = src1; /* always aligned */

#ifdef DEBUG_FLOAT
  umes("fl_addsub_long_norm: s1=%08.8x c1=%08.8x f1=%08.8x_%08.8x\n", s1, c1, (unsigned)(f1>>32), (unsigned)(f1));
#endif

  return (intr|cc);
}

unsigned
fl_addsub_long_unnorm(dummy, r1, r2, mask, addsub)
  unsigned dummy;
  unsigned char *r1, *r2;
  unsigned mask;   /* program mask */
  unsigned addsub; /* 0:add, 1:sub */
  /* return INTR-CODE|CC */
{
  llu      src1, src2;
  unsigned s1, s2;
  unsigned c1, c2;
  llu      f1, f2;
  unsigned sft;
  unsigned intr = 0, cc = 0;

  src1 = *(llu*)r1;        /* always aligned */
  src2 = *(llu*)r2;        /* always aligned */
  src2^= (llu)addsub << 63;

  s1 = src1>>32 & 0x80000000;           /* s------- -------- -------- -------- */
  s2 = src2>>32 & 0x80000000;           /* s------- -------- -------- -------- */
  c1 = src1>>56 & 0x0000007f;           /* -------- -------- -------- -ccccccc */
  c2 = src2>>56 & 0x0000007f;           /* -------- -------- -------- -ccccccc */
  f1 = src1<< 4 & 0x0ffffffffffffff0LL; /* ----ffff ffffff.. ffffffff ffff---- */
  f2 = src2<< 4 & 0x0ffffffffffffff0LL; /* ----ffff ffffff.. ffffffff ffff---- */
                                        /*                    guard digit ^^^^ */
#ifdef DEBUG_FLOAT
  umes("fl_addsub_long_unnorm: s1=%08.8x c1=%08.8x f1=%08.8x_%08.8x\n", s1, c1, (unsigned)(f1>>32), (unsigned)(f1));
  umes("fl_addsub_long_unnorm: s2=%08.8x c2=%08.8x f2=%08.8x_%08.8x\n", s2, c2, (unsigned)(f2>>32), (unsigned)(f2));
#endif

  /* fraction alignment */
  if (c1 > c2) { /* shift f2 right */
    sft = (c1 - c2) << 2;
    c2 = c1;
    f2 = (sft >= 64) ? 0LL : f2 >> sft;
  }
  else if (c1 < c2) { /* shift f1 right */
    sft = (c2 - c1) << 2;
    c1 = c2;
    f1 = (sft >= 64) ? 0LL : f1 >> sft;
  }
  else {
    /* do nothing if c1 == c2 */
  }

#ifdef DEBUG_FLOAT
  umes("fl_addsub_long_unnorm: s1=%08.8x c1=%08.8x f1=%08.8x_%08.8x\n", s1, c1, (unsigned)(f1>>32), (unsigned)(f1));
  umes("fl_addsub_long_unnorm: s2=%08.8x c2=%08.8x f2=%08.8x_%08.8x\n", s2, c2, (unsigned)(f2>>32), (unsigned)(f2));
#endif

  /* add algebraically */
  if (s1 != s2) {
    if (f1 >= f2)   f1 = f1 - f2;
    else { s1 = s2; f1 = f2 - f1; }
  }
  else { /* s1 == s2 */
    f1 += f2;
  }

#ifdef DEBUG_FLOAT
  umes("fl_addsub_long_unnorm: s1=%08.8x c1=%08.8x f1=%08.8x_%08.8x\n", s1, c1, (unsigned)(f1>>32), (unsigned)(f1));
#endif

  /* test carry, normalize */
  if (f1 & 0xf000000000000000LL) { /* test carry */
    f1 >>= 8;
    c1++;
    if (c1 > 127) { /* if characteristic overflow */
      c1 -= 128;
      intr = EXP_OVF; /* exponent overflow */
    }
  }
  else {
    f1 >>= 4;
    if (!f1) { /* fraction == 0 */
      s1 = 0;
      if (mask & SIGNIFI_MASK) {
        intr = SIGNIFI; /* significance */
      }
      else {
        s1 = c1 = 0; /* true zero */
        f1 = 0LL;    /* true zero */
      }
    }
  }

  /* format cc */
  if (f1) {
    if (s1 > 0) cc = 1; /* neg */
    else        cc = 2; /* pos */
  }

  /* format result */
  src1 = (llu)s1 << 32 | (llu)c1 << 56 | f1;
  *(llu*)r1 = src1; /* always aligned */

#ifdef DEBUG_FLOAT
  umes("fl_addsub_long_unnorm: s1=%08.8x c1=%08.8x f1=%08.8x_%08.8x\n", s1, c1, (unsigned)(f1>>32), (unsigned)(f1));
#endif

  return (intr|cc);
}

unsigned
fl_addsub_ext_norm(dummy, r1, r2, mask, addsub)
  unsigned dummy;
  unsigned char *r1, *r2;
  unsigned mask;   /* program mask */
  unsigned addsub; /* 0:add, 1:sub */
  /* return INTR-CODE|CC */
{
  llu      src1u, src1l, src2u, src2l;
  unsigned s1, s2;
  unsigned c1, c2, c1l;
  llu      f1u, f1l, f2u, f2l;
  unsigned sft;
  unsigned intr = 0, cc = 0;

  src1u = *(llu*)(r1  );        /* always aligned */
  src1l = *(llu*)(r1+8);        /* always aligned */
  src2u = *(llu*)(r2  );        /* always aligned */
  src2l = *(llu*)(r2+8);        /* always aligned */
  src2u^= (llu)addsub << 63;

  s1  = src1u>>32 & 0x80000000;           /* s------- -------- -------- -------- */
  s2  = src2u>>32 & 0x80000000;           /* s------- -------- -------- -------- */
  c1  = src1u>>56 & 0x0000007f;           /* -------- -------- -------- -ccccccc */
  c2  = src2u>>56 & 0x0000007f;           /* -------- -------- -------- -ccccccc */
  f1u = src1u     & 0x00ffffffffffffffLL; /* -------- ffffff.. ffffffff ffffffff */
  f2u = src2u     & 0x00ffffffffffffffLL; /* -------- ffffff.. ffffffff ffffffff */
  f1l = src1l<< 4 & 0x0ffffffffffffff0LL; /* ----ffff ffffff.. ffffffff ffff---- */
  f2l = src2l<< 4 & 0x0ffffffffffffff0LL; /* ----ffff ffffff.. ffffffff ffff---- */
                                          /*                    guard digit ^^^^ */
#ifdef DEBUG_FLOAT
  umes("fl_addsub_ext_norm: s1=%08.8x c1=%08.8x ", s1, c1);
  umes("f1=%08.8x_%08.8x %08.8x_%08.8x\n", (unsigned)(f1u>>32), (unsigned)(f1u), (unsigned)(f1l>>32), (unsigned)(f1l));
  umes("fl_addsub_ext_norm: s2=%08.8x c2=%08.8x ", s2, c2);
  umes("f2=%08.8x_%08.8x %08.8x_%08.8x\n", (unsigned)(f2u>>32), (unsigned)(f2u), (unsigned)(f2l>>32), (unsigned)(f2l));
#endif

  /* fraction alignment */
  if (c1 > c2) { /* shift f2 right */
    sft = (c1 - c2) << 2;
    c2 = c1;
    if (sft >= 60) {
      f2l = f2u >> (sft - 60);
      f2u = 0LL;
    }
    else {
      f2l = (f2u << (60 - sft) | f2l >> sft) & 0x0fffffffffffffffLL;
      f2u >>= sft;
    }
  }
  else if (c1 < c2) { /* shift f1 right */
    sft = (c2 - c1) << 2;
    c1 = c2;
    if (sft >= 60) {
      f1l = f1u >> (sft - 60);
      f1u = 0LL;
    }
    else {
      f1l = (f1u << (60 - sft) | f1l >> sft) & 0x0fffffffffffffffLL;
      f1u >>= sft;
    }
  }
  else {
    /* do nothing if c1 == c2 */
  }

#ifdef DEBUG_FLOAT
  umes("fl_addsub_ext_norm: s1=%08.8x c1=%08.8x ", s1, c1);
  umes("f1=%08.8x_%08.8x %08.8x_%08.8x\n", (unsigned)(f1u>>32), (unsigned)(f1u), (unsigned)(f1l>>32), (unsigned)(f1l));
  umes("fl_addsub_ext_norm: s2=%08.8x c2=%08.8x ", s2, c2);
  umes("f2=%08.8x_%08.8x %08.8x_%08.8x\n", (unsigned)(f2u>>32), (unsigned)(f2u), (unsigned)(f2l>>32), (unsigned)(f2l));
#endif

  /* add algebraically */
  if (s1 != s2) {
    if (f1u == f2u) {
      if (f1l >= f2l) {
        f1l = f1l - f2l;
        f1u = f1u - f2u;
      }
      else {
        s1 = s2;
        f1l = f2l - f1l;
        f1u = f2u - f1u;
      }
    }
    else if (f1u > f2u) {
      if (f1l >= f2l) {
        f1l = f1l - f2l;
        f1u = f1u - f2u;
      }
      else {
        f1l = (f1l - f2l) & 0x0fffffffffffffffLL;
        f1u = f1u - f2u - 1;
      }
    }
    else {
      if (f1l >= f2l) {
	s1 = s2;
        f1l = (f2l - f1l) & 0x0fffffffffffffffLL;
        f1u = f2u - f1u - 1;
      }
      else {
	s1 = s2;
        f1l = f2l - f1l;
        f1u = f2u - f1u;
      }
    }
  }
  else { /* s1 == s2 */
    f1l = f1l + f2l;
    f1u = f1u + f2u;
    if (f1l & 0xf000000000000000LL) {
      f1l &= 0x0fffffffffffffffLL;
      f1u++;
    }
  }

#ifdef DEBUG_FLOAT
  umes("fl_addsub_ext_norm: s1=%08.8x c1=%08.8x ", s1, c1);
  umes("f1=%08.8x_%08.8x %08.8x_%08.8x\n", (unsigned)(f1u>>32), (unsigned)(f1u), (unsigned)(f1l>>32), (unsigned)(f1l));
#endif

  /* test carry, normalize */
  if (f1u & 0x0f00000000000000LL) { /* test carry */
    f1l = (f1u << 56 | f1l >> 4) & 0x0fffffffffffffffLL;
    f1u >>= 4;
    c1++;
    if (c1 > 127) { /* if characteristic overflow */
      c1 -= 128;
      intr = EXP_OVF; /* exponent overflow */
    }
    c1l = (c1 - 14) & 0x7f;
  }
  else {
    if (f1u || f1l) { /* normalize */
      if      (f1u & 0x00f0000000000000LL) { /* do nothing */ }
      else if (f1u & 0x000f000000000000LL) { f1u = f1u <<  4 | f1l >> 56; f1l = f1l <<  4 & 0x0fffffffffffffffLL; c1 -=  1; }
      else if (f1u & 0x0000f00000000000LL) { f1u = f1u <<  8 | f1l >> 52; f1l = f1l <<  8 & 0x0fffffffffffffffLL; c1 -=  2; }
      else if (f1u & 0x00000f0000000000LL) { f1u = f1u << 12 | f1l >> 48; f1l = f1l << 12 & 0x0fffffffffffffffLL; c1 -=  3; }
      else if (f1u & 0x000000f000000000LL) { f1u = f1u << 16 | f1l >> 44; f1l = f1l << 16 & 0x0fffffffffffffffLL; c1 -=  4; }
      else if (f1u & 0x0000000f00000000LL) { f1u = f1u << 20 | f1l >> 40; f1l = f1l << 20 & 0x0fffffffffffffffLL; c1 -=  5; }
      else if (f1u & 0x00000000f0000000LL) { f1u = f1u << 24 | f1l >> 36; f1l = f1l << 24 & 0x0fffffffffffffffLL; c1 -=  6; }
      else if (f1u & 0x000000000f000000LL) { f1u = f1u << 28 | f1l >> 32; f1l = f1l << 28 & 0x0fffffffffffffffLL; c1 -=  7; }
      else if (f1u & 0x0000000000f00000LL) { f1u = f1u << 32 | f1l >> 28; f1l = f1l << 32 & 0x0fffffffffffffffLL; c1 -=  8; }
      else if (f1u & 0x00000000000f0000LL) { f1u = f1u << 36 | f1l >> 24; f1l = f1l << 36 & 0x0fffffffffffffffLL; c1 -=  9; }
      else if (f1u & 0x000000000000f000LL) { f1u = f1u << 40 | f1l >> 20; f1l = f1l << 40 & 0x0fffffffffffffffLL; c1 -= 10; }
      else if (f1u & 0x0000000000000f00LL) { f1u = f1u << 44 | f1l >> 16; f1l = f1l << 44 & 0x0fffffffffffffffLL; c1 -= 11; }
      else if (f1u & 0x00000000000000f0LL) { f1u = f1u << 48 | f1l >> 12; f1l = f1l << 48 & 0x0fffffffffffffffLL; c1 -= 12; }
      else if (f1u & 0x000000000000000fLL) { f1u = f1u << 52 | f1l >>  8; f1l = f1l << 52 & 0x0fffffffffffffffLL; c1 -= 13; }
      else if (f1l & 0x0f00000000000000LL) { f1u =             f1l >>  4; f1l = f1l << 56 & 0x0fffffffffffffffLL; c1 -= 14; }
      else if (f1l & 0x00f0000000000000LL) { f1u =             f1l      ; f1l =             0x0000000000000000LL; c1 -= 15; }
      else if (f1l & 0x000f000000000000LL) { f1u =             f1l <<  4; f1l =             0x0000000000000000LL; c1 -= 16; }
      else if (f1l & 0x0000f00000000000LL) { f1u =             f1l <<  8; f1l =             0x0000000000000000LL; c1 -= 17; }
      else if (f1l & 0x00000f0000000000LL) { f1u =             f1l << 12; f1l =             0x0000000000000000LL; c1 -= 18; }
      else if (f1l & 0x000000f000000000LL) { f1u =             f1l << 16; f1l =             0x0000000000000000LL; c1 -= 19; }
      else if (f1l & 0x0000000f00000000LL) { f1u =             f1l << 20; f1l =             0x0000000000000000LL; c1 -= 20; }
      else if (f1l & 0x00000000f0000000LL) { f1u =             f1l << 24; f1l =             0x0000000000000000LL; c1 -= 21; }
      else if (f1l & 0x000000000f000000LL) { f1u =             f1l << 28; f1l =             0x0000000000000000LL; c1 -= 22; }
      else if (f1l & 0x0000000000f00000LL) { f1u =             f1l << 32; f1l =             0x0000000000000000LL; c1 -= 23; }
      else if (f1l & 0x00000000000f0000LL) { f1u =             f1l << 36; f1l =             0x0000000000000000LL; c1 -= 24; }
      else if (f1l & 0x000000000000f000LL) { f1u =             f1l << 40; f1l =             0x0000000000000000LL; c1 -= 25; }
      else if (f1l & 0x0000000000000f00LL) { f1u =             f1l << 44; f1l =             0x0000000000000000LL; c1 -= 26; }
      else if (f1l & 0x00000000000000f0LL) { f1u =             f1l << 48; f1l =             0x0000000000000000LL; c1 -= 27; }
      else                                 { f1u =             f1l << 52; f1l =             0x0000000000000000LL; c1 -= 28; }

      if (c1 & 0x80000000) { /* if characteristic underflow */
        if (mask & EXP_UNF_MASK) {
          c1 += 128;
          c1l = (c1 - 14) & 0x7f;
          intr = EXP_UNF; /* exponent underflow */
        }
        else {
          s1 = c1 = c1l = 0; /* true zero */
          f1u = f1l = 0LL;   /* true zero */
        }
      }
      else {
        c1l = (c1 - 14) & 0x7f;
      }
    }
    else { /* fraction == 0 */
      s1 = 0;
      if (mask & SIGNIFI_MASK) {
        c1l = (c1 - 14) & 0x7f;
        intr = SIGNIFI; /* significance */
      }
      else {
        s1 = c1 = c1l = 0; /* true zero */
        f1u = f1l = 0LL;   /* true zero */
      }
    }
  }

  /* format cc */
  if (f1u || f1l) {
    if (s1 > 0) cc = 1; /* neg */
    else        cc = 2; /* pos */
  }

  /* format result */
  src1u = (llu)s1 << 32 | (llu)c1  << 56 | f1u;
  src1l = (llu)s1 << 32 | (llu)c1l << 56 | f1l >> 4;
  *(llu*)(r1  ) = src1u; /* always aligned */
  *(llu*)(r1+8) = src1l; /* always aligned */

#ifdef DEBUG_FLOAT
  umes("fl_addsub_ext_norm: s1=%08.8x c1=%08.8x ", s1, c1);
  umes("f1=%08.8x_%08.8x %08.8x_%08.8x\n", (unsigned)(f1u>>32), (unsigned)(f1u), (unsigned)(f1l>>32), (unsigned)(f1l));
#endif

  return (intr|cc);
}

unsigned
fl_comp_short(dummy, r1, r2, mask)
  unsigned dummy;
  unsigned char *r1, *r2;
  unsigned mask;   /* program mask */
  /* return CC */
{
  unsigned src1, src2;
  unsigned s1, s2;
  unsigned c1, c2;
  unsigned f1, f2;
  unsigned sft;
  unsigned cc = 0;

  src1 = *(unsigned*)r1; /* always aligned */
  src2 = *(unsigned*)r2; /* always aligned */
  src2^= 0x80000000;

  s1 = src1     & 0x80000000; /* s------- -------- -------- -------- */
  s2 = src2     & 0x80000000; /* s------- -------- -------- -------- */
  c1 = src1>>24 & 0x0000007f; /* -------- -------- -------- -ccccccc */
  c2 = src2>>24 & 0x0000007f; /* -------- -------- -------- -ccccccc */
  f1 = src1<< 4 & 0x0ffffff0; /* ----ffff ffffffff ffffffff ffff---- */
  f2 = src2<< 4 & 0x0ffffff0; /* ----ffff ffffffff ffffffff ffff---- */
                              /*                    guard digit ^^^^ */
#ifdef DEBUG_FLOAT
  umes("fl_comp_short: s1=%08.8x c1=%08.8x f1=%08.8x\n", s1, c1, f1);
  umes("fl_comp_short: s2=%08.8x c2=%08.8x f2=%08.8x\n", s2, c2, f2);
#endif

  /* fraction alignment */
  if (c1 > c2) { /* shift f2 right */
    sft = (c1 - c2) << 2;
    c2 = c1;
    f2 = (sft >= 32) ? 0 : f2 >> sft;
  }
  else if (c1 < c2) { /* shift f1 right */
    sft = (c2 - c1) << 2;
    c1 = c2;
    f1 = (sft >= 32) ? 0 : f1 >> sft;
  }
  else {
    /* do nothing if c1 == c2 */
  }

#ifdef DEBUG_FLOAT
  umes("fl_comp_short: s1=%08.8x c1=%08.8x f1=%08.8x\n", s1, c1, f1);
  umes("fl_comp_short: s2=%08.8x c2=%08.8x f2=%08.8x\n", s2, c2, f2);
#endif

  /* add algebraically */
  if (s1 != s2) {
    if (f1 >= f2)   f1 = f1 - f2;
    else { s1 = s2; f1 = f2 - f1; }
  }
  else { /* s1 == s2 */
    f1 += f2;
  }

#ifdef DEBUG_FLOAT
  umes("fl_comp_short: s1=%08.8x c1=%08.8x f1=%08.8x\n", s1, c1, f1);
#endif

  /* test carry, normalize */
  if (f1 & 0xf0000000) { /* test carry */
    f1 >>= 4;
    c1++;
  }
  else {
    if (f1) { /* normalize */
      if      (f1 & 0x0f000000) { /* do nothing */ }
      else if (f1 & 0x00f00000) { f1 <<=  4; c1 -= 1; }
      else if (f1 & 0x000f0000) { f1 <<=  8; c1 -= 2; }
      else if (f1 & 0x0000f000) { f1 <<= 12; c1 -= 3; }
      else if (f1 & 0x00000f00) { f1 <<= 16; c1 -= 4; }
      else if (f1 & 0x000000f0) { f1 <<= 20; c1 -= 5; }
      else                      { f1 <<= 24; c1 -= 6; }
    }
  }

  /* format cc */
  if (f1) {
    if (s1 > 0) cc = 1; /* neg */
    else        cc = 2; /* pos */
  }

#ifdef DEBUG_FLOAT
  umes("fl_comp_short: s1=%08.8x c1=%08.8x f1=%08.8x\n", s1, c1, f1);
#endif

  return (cc);
}

unsigned
fl_comp_long(dummy, r1, r2, mask)
  unsigned dummy;
  unsigned char *r1, *r2;
  unsigned mask;   /* program mask */
  /* return CC */
{
  llu      src1, src2;
  unsigned s1, s2;
  unsigned c1, c2;
  llu      f1, f2;
  unsigned sft;
  unsigned cc = 0;

  src1 = *(llu*)r1;        /* always aligned */
  src2 = *(llu*)r2;        /* always aligned */
  src2^= 0x8000000000000000LL;

  s1 = src1>>32 & 0x80000000;           /* s------- -------- -------- -------- */
  s2 = src2>>32 & 0x80000000;           /* s------- -------- -------- -------- */
  c1 = src1>>56 & 0x0000007f;           /* -------- -------- -------- -ccccccc */
  c2 = src2>>56 & 0x0000007f;           /* -------- -------- -------- -ccccccc */
  f1 = src1<< 4 & 0x0ffffffffffffff0LL; /* ----ffff ffffff.. ffffffff ffff---- */
  f2 = src2<< 4 & 0x0ffffffffffffff0LL; /* ----ffff ffffff.. ffffffff ffff---- */
                                        /*                    guard digit ^^^^ */
#ifdef DEBUG_FLOAT
  umes("fl_comp_long: s1=%08.8x c1=%08.8x f1=%08.8x_%08.8x\n", s1, c1, (unsigned)(f1>>32), (unsigned)(f1));
  umes("fl_comp_long: s2=%08.8x c2=%08.8x f2=%08.8x_%08.8x\n", s2, c2, (unsigned)(f2>>32), (unsigned)(f2));
#endif

  /* fraction alignment */
  if (c1 > c2) { /* shift f2 right */
    sft = (c1 - c2) << 2;
    c2 = c1;
    f2 = (sft >= 64) ? 0LL : f2 >> sft;
  }
  else if (c1 < c2) { /* shift f1 right */
    sft = (c2 - c1) << 2;
    c1 = c2;
    f1 = (sft >= 64) ? 0LL : f1 >> sft;
  }
  else {
    /* do nothing if c1 == c2 */
  }

#ifdef DEBUG_FLOAT
  umes("fl_comp_long: s1=%08.8x c1=%08.8x f1=%08.8x_%08.8x\n", s1, c1, (unsigned)(f1>>32), (unsigned)(f1));
  umes("fl_comp_long: s2=%08.8x c2=%08.8x f2=%08.8x_%08.8x\n", s2, c2, (unsigned)(f2>>32), (unsigned)(f2));
#endif

  /* add algebraically */
  if (s1 != s2) {
    if (f1 >= f2)   f1 = f1 - f2;
    else { s1 = s2; f1 = f2 - f1; }
  }
  else { /* s1 == s2 */
    f1 += f2;
  }

#ifdef DEBUG_FLOAT
  umes("fl_comp_long: s1=%08.8x c1=%08.8x f1=%08.8x_%08.8x\n", s1, c1, (unsigned)(f1>>32), (unsigned)(f1));
#endif

  /* test carry, normalize */
  if (f1 & 0xf000000000000000LL) { /* test carry */
    f1 >>= 4;
    c1++;
  }
  else {
    if (f1) { /* normalize */
      if      (f1 & 0x0f00000000000000LL) { /* do nothing */ }
      else if (f1 & 0x00f0000000000000LL) { f1 <<=  4; c1 -=  1; }
      else if (f1 & 0x000f000000000000LL) { f1 <<=  8; c1 -=  2; }
      else if (f1 & 0x0000f00000000000LL) { f1 <<= 12; c1 -=  3; }
      else if (f1 & 0x00000f0000000000LL) { f1 <<= 16; c1 -=  4; }
      else if (f1 & 0x000000f000000000LL) { f1 <<= 20; c1 -=  5; }
      else if (f1 & 0x0000000f00000000LL) { f1 <<= 24; c1 -=  6; }
      else if (f1 & 0x00000000f0000000LL) { f1 <<= 28; c1 -=  7; }
      else if (f1 & 0x000000000f000000LL) { f1 <<= 32; c1 -=  8; }
      else if (f1 & 0x0000000000f00000LL) { f1 <<= 36; c1 -=  9; }
      else if (f1 & 0x00000000000f0000LL) { f1 <<= 40; c1 -= 10; }
      else if (f1 & 0x000000000000f000LL) { f1 <<= 44; c1 -= 11; }
      else if (f1 & 0x0000000000000f00LL) { f1 <<= 48; c1 -= 12; }
      else if (f1 & 0x00000000000000f0LL) { f1 <<= 52; c1 -= 13; }
      else                                { f1 <<= 56; c1 -= 14; }
    }
  }

  /* format cc */
  if (f1) {
    if (s1 > 0) cc = 1; /* neg */
    else        cc = 2; /* pos */
  }

#ifdef DEBUG_FLOAT
  umes("fl_comp_long: s1=%08.8x c1=%08.8x f1=%08.8x_%08.8x\n", s1, c1, (unsigned)(f1>>32), (unsigned)(f1));
#endif

  return (cc);
}

unsigned
fl_mult_short2long(dummy, r1, r2, mask)
  unsigned dummy;
  unsigned char *r1, *r2;
  unsigned mask;   /* program mask */
  /* return INTR-CODE */
{
  unsigned src1, src2;
  llu      srcw;
  unsigned s1, s2;
  unsigned c1, c2;
  unsigned f1, f2;
  llu      fw;
  unsigned intr = 0;

  src1 = *(unsigned*)r1; /* always aligned */
  src2 = *(unsigned*)r2; /* always aligned */

  s1 = src1     & 0x80000000; /* s------- -------- -------- -------- */
  s2 = src2     & 0x80000000; /* s------- -------- -------- -------- */
  c1 = src1>>24 & 0x0000007f; /* -------- -------- -------- -ccccccc */
  c2 = src2>>24 & 0x0000007f; /* -------- -------- -------- -ccccccc */
  f1 = src1     & 0x00ffffff; /* -------- ffffffff ffffffff ffffffff */
  f2 = src2     & 0x00ffffff; /* -------- ffffffff ffffffff ffffffff */

#ifdef DEBUG_FLOAT
  umes("fl_mult_short2long: s1=%08.8x c1=%08.8x f1=%08.8x\n", s1, c1, f1);
  umes("fl_mult_short2long: s2=%08.8x c2=%08.8x f2=%08.8x\n", s2, c2, f2);
#endif

  /* normalize */
  if (f1) { /* normalize */
    if      (f1 & 0x00f00000) { /* do nothing */ }
    else if (f1 & 0x000f0000) { f1 <<=  4; c1 -= 1; }
    else if (f1 & 0x0000f000) { f1 <<=  8; c1 -= 2; }
    else if (f1 & 0x00000f00) { f1 <<= 12; c1 -= 3; }
    else if (f1 & 0x000000f0) { f1 <<= 16; c1 -= 4; }
    else                      { f1 <<= 20; c1 -= 5; }
  }
  if (f2) { /* normalize */
    if      (f2 & 0x00f00000) { /* do nothing */ }
    else if (f2 & 0x000f0000) { f2 <<=  4; c2 -= 1; }
    else if (f2 & 0x0000f000) { f2 <<=  8; c2 -= 2; }
    else if (f2 & 0x00000f00) { f2 <<= 12; c2 -= 3; }
    else if (f2 & 0x000000f0) { f2 <<= 16; c2 -= 4; }
    else                      { f2 <<= 20; c2 -= 5; }
  }

#ifdef DEBUG_FLOAT
  umes("fl_mult_short2long: s1=%08.8x c1=%08.8x f1=%08.8x\n", s1, c1, f1);
  umes("fl_mult_short2long: s2=%08.8x c2=%08.8x f2=%08.8x\n", s2, c2, f2);
#endif

  /* mult algebraically */
  s1 = s1 ^ s2;
  c1 = c1 + c2 - 64;
  fw = (llu)f1 * (llu)f2;

#ifdef DEBUG_FLOAT
  umes("fl_mult_short2long: s1=%08.8x c1=%08.8x fw=%08.8x_%08.8x\n", s1, c1, (unsigned)(fw>>32), (unsigned)(fw));
#endif

  /* normalize */
  if (fw) { /* normalize */
    if (fw & 0x0000f00000000000LL) { fw <<=  8;       }
    else                           { fw <<= 12; c1--; }

    if (c1 & 0x80000000) { /* if characteristic underflow */
      if (mask & EXP_UNF_MASK) {
        c1 += 128;
        intr = EXP_UNF; /* exponent underflow */
      }
      else {
        s1 = c1 = 0; /* true zero */
        fw = 0LL;    /* true zero */
      }
    }
    else if (c1 > 127) { /* if characteristic overflow */
      c1 -= 128;
      intr = EXP_OVF; /* exponent overflow */
    }
  }
  else { /* fraction == 0 */
    s1 = c1 = 0; /* true zero */
    fw = 0LL;    /* true zero */
  }

  /* format result */
  srcw = (llu)s1 << 32 | (llu)c1 << 56 | fw;
  *(llu*)r1 = srcw; /* always aligned */

#ifdef DEBUG_FLOAT
  umes("fl_mult_short2long: s1=%08.8x c1=%08.8x fw=%08.8x_%08.8x\n", s1, c1, (unsigned)(fw>>32), (unsigned)(fw));
#endif

  return (intr);
}

unsigned
fl_mult_long2long(dummy, r1, r2, mask)
  unsigned dummy;
  unsigned char *r1, *r2;
  unsigned mask;   /* program mask */
  /* return INTR-CODE */
{
  llu      src1, src2;
  llu      srcw;
  unsigned s1, s2;
  unsigned c1, c2;
  llu      f1, f2;
  llu      fwu, fwm, fwl;
  unsigned intr = 0;

  src1 = *(llu*)r1;        /* always aligned */
  src2 = *(llu*)r2;        /* always aligned */

  s1 = src1>>32 & 0x80000000;           /* s------- -------- -------- -------- */
  s2 = src2>>32 & 0x80000000;           /* s------- -------- -------- -------- */
  c1 = src1>>56 & 0x0000007f;           /* -------- -------- -------- -ccccccc */
  c2 = src2>>56 & 0x0000007f;           /* -------- -------- -------- -ccccccc */
  f1 = src1     & 0x00ffffffffffffffLL; /* -------- ffffff.. ffffffff ffffffff */
  f2 = src2     & 0x00ffffffffffffffLL; /* -------- ffffff.. ffffffff ffffffff */

#ifdef DEBUG_FLOAT
  umes("fl_mult_long2long: s1=%08.8x c1=%08.8x f1=%08.8x_%08.0x\n", s1, c1, (unsigned)(f1>>32), (unsigned)(f1));
  umes("fl_mult_long2long: s2=%08.8x c2=%08.8x f2=%08.8x_%08.8x\n", s2, c2, (unsigned)(f2>>32), (unsigned)(f2));
#endif

  /* normalize */
  if (f1) { /* normalize */
    if      (f1 & 0x00f0000000000000LL) { /* do nothing */ }
    else if (f1 & 0x000f000000000000LL) { f1 <<=  4; c1 -=  1; }
    else if (f1 & 0x0000f00000000000LL) { f1 <<=  8; c1 -=  2; }
    else if (f1 & 0x00000f0000000000LL) { f1 <<= 12; c1 -=  3; }
    else if (f1 & 0x000000f000000000LL) { f1 <<= 16; c1 -=  4; }
    else if (f1 & 0x0000000f00000000LL) { f1 <<= 20; c1 -=  5; }
    else if (f1 & 0x00000000f0000000LL) { f1 <<= 24; c1 -=  6; }
    else if (f1 & 0x000000000f000000LL) { f1 <<= 28; c1 -=  7; }
    else if (f1 & 0x0000000000f00000LL) { f1 <<= 32; c1 -=  8; }
    else if (f1 & 0x00000000000f0000LL) { f1 <<= 36; c1 -=  9; }
    else if (f1 & 0x000000000000f000LL) { f1 <<= 40; c1 -= 10; }
    else if (f1 & 0x0000000000000f00LL) { f1 <<= 44; c1 -= 11; }
    else if (f1 & 0x00000000000000f0LL) { f1 <<= 48; c1 -= 12; }
    else                                { f1 <<= 52; c1 -= 13; }
  }
  if (f2) { /* normalize */
    if      (f2 & 0x00f0000000000000LL) { /* do nothing */ }
    else if (f2 & 0x000f000000000000LL) { f2 <<=  4; c2 -=  1; }
    else if (f2 & 0x0000f00000000000LL) { f2 <<=  8; c2 -=  2; }
    else if (f2 & 0x00000f0000000000LL) { f2 <<= 12; c2 -=  3; }
    else if (f2 & 0x000000f000000000LL) { f2 <<= 16; c2 -=  4; }
    else if (f2 & 0x0000000f00000000LL) { f2 <<= 20; c2 -=  5; }
    else if (f2 & 0x00000000f0000000LL) { f2 <<= 24; c2 -=  6; }
    else if (f2 & 0x000000000f000000LL) { f2 <<= 28; c2 -=  7; }
    else if (f2 & 0x0000000000f00000LL) { f2 <<= 32; c2 -=  8; }
    else if (f2 & 0x00000000000f0000LL) { f2 <<= 36; c2 -=  9; }
    else if (f2 & 0x000000000000f000LL) { f2 <<= 40; c2 -= 10; }
    else if (f2 & 0x0000000000000f00LL) { f2 <<= 44; c2 -= 11; }
    else if (f2 & 0x00000000000000f0LL) { f2 <<= 48; c2 -= 12; }
    else                                { f2 <<= 52; c2 -= 13; }
  }

#ifdef DEBUG_FLOAT
  umes("fl_mult_long2long: s1=%08.8x c1=%08.8x f1=%08.8x_%08.0x\n", s1, c1, (unsigned)(f1>>32), (unsigned)(f1));
  umes("fl_mult_long2long: s2=%08.8x c2=%08.8x f2=%08.8x_%08.8x\n", s2, c2, (unsigned)(f2>>32), (unsigned)(f2));
#endif

  /* mult algebraically */
  s1 = s1 ^ s2;
  c1 = c1 + c2 - 64;

  fwl  = (llu)(unsigned)(f1&0x000000000fffffffLL) * (llu)(unsigned)(f2&0x000000000fffffffLL);
  fwm  = (llu)(unsigned)(f1>>28)                  * (llu)(unsigned)(f2&0x000000000fffffffLL);
  fwm += (llu)(unsigned)(f1&0x000000000fffffffLL) * (llu)(unsigned)(f2>>28);
  fwu  = (llu)(unsigned)(f1>>28)                  * (llu)(unsigned)(f2>>28);
  fwm += fwl>>28;
  fwu += fwm>>28;
  fwl  = fwu<<56 | (fwm<<28)&0x00fffffff0000000LL | fwl&0x000000000fffffffLL;
  fwu  = fwu>> 8;

#ifdef DEBUG_FLOAT
  umes("fl_mult_long2long: s1=%08.8x c1=%08.8x ", s1, c1);
  umes("fw=%08.8x_%08.8x %08.8x_%08.8x\n", (unsigned)(fwu>>32), (unsigned)(fwu), (unsigned)(fwl>>32), (unsigned)(fwl));
#endif

  /* normalize */
  if (fwu) { /* normalize */
    if (fwu & 0x0000f00000000000LL) {
      fwu = fwu <<  8 | fwl >> 56;
      fwl <<=  8;
    }
    else {
      fwu = fwu << 12 | fwl >> 52;
      fwl <<= 12;
      c1--;
    }

    if (c1 & 0x80000000) { /* if characteristic underflow */
      if (mask & EXP_UNF_MASK) {
        c1 += 128;
        intr = EXP_UNF; /* exponent underflow */
      }
      else {
        s1 = c1 = 0;     /* true zero */
        fwu = fwl = 0LL; /* true zero */
      }
    }
    else if (c1 > 127) { /* if characteristic overflow */
      c1 -= 128;
      intr = EXP_OVF; /* exponent overflow */
    }
  }
  else { /* fraction == 0 */
    s1 = c1 = 0;     /* true zero */
    fwu = fwl = 0LL; /* true zero */
  }

  /* format result */
  srcw = (llu)s1 << 32 | (llu)c1 << 56 | fwu;
  *(llu*)r1 = srcw; /* always aligned */

#ifdef DEBUG_FLOAT
  umes("fl_mult_long2long: s1=%08.8x c1=%08.8x ", s1, c1);
  umes("fw=%08.8x_%08.8x %08.8x_%08.8x\n", (unsigned)(fwu>>32), (unsigned)(fwu), (unsigned)(fwl>>32), (unsigned)(fwl));
#endif

  return (intr);
}

unsigned
fl_mult_long2ext(dummy, r1, r2, mask)
  unsigned dummy;
  unsigned char *r1, *r2;
  unsigned mask;   /* program mask */
  /* return INTR-CODE */
{
  llu      src1, src2;
  llu      srcwu, srcwl;
  unsigned s1, s2;
  unsigned c1, c2, c1l;
  llu      f1, f2;
  llu      fwu, fwl, fwm;
  unsigned intr = 0;

  src1 = *(llu*)r1;        /* always aligned */
  src2 = *(llu*)r2;        /* always aligned */

  s1 = src1>>32 & 0x80000000;           /* s------- -------- -------- -------- */
  s2 = src2>>32 & 0x80000000;           /* s------- -------- -------- -------- */
  c1 = src1>>56 & 0x0000007f;           /* -------- -------- -------- -ccccccc */
  c2 = src2>>56 & 0x0000007f;           /* -------- -------- -------- -ccccccc */
  f1 = src1     & 0x00ffffffffffffffLL; /* -------- ffffff.. ffffffff ffffffff */
  f2 = src2     & 0x00ffffffffffffffLL; /* -------- ffffff.. ffffffff ffffffff */

#ifdef DEBUG_FLOAT
  umes("fl_mult_long2ext: s1=%08.8x c1=%08.8x f1=%08.8x_%08.0x\n", s1, c1, (unsigned)(f1>>32), (unsigned)(f1));
  umes("fl_mult_long2ext: s2=%08.8x c2=%08.8x f2=%08.8x_%08.8x\n", s2, c2, (unsigned)(f2>>32), (unsigned)(f2));
#endif

  /* normalize */
  if (f1) { /* normalize */
    if      (f1 & 0x00f0000000000000LL) { /* do nothing */ }
    else if (f1 & 0x000f000000000000LL) { f1 <<=  4; c1 -=  1; }
    else if (f1 & 0x0000f00000000000LL) { f1 <<=  8; c1 -=  2; }
    else if (f1 & 0x00000f0000000000LL) { f1 <<= 12; c1 -=  3; }
    else if (f1 & 0x000000f000000000LL) { f1 <<= 16; c1 -=  4; }
    else if (f1 & 0x0000000f00000000LL) { f1 <<= 20; c1 -=  5; }
    else if (f1 & 0x00000000f0000000LL) { f1 <<= 24; c1 -=  6; }
    else if (f1 & 0x000000000f000000LL) { f1 <<= 28; c1 -=  7; }
    else if (f1 & 0x0000000000f00000LL) { f1 <<= 32; c1 -=  8; }
    else if (f1 & 0x00000000000f0000LL) { f1 <<= 36; c1 -=  9; }
    else if (f1 & 0x000000000000f000LL) { f1 <<= 40; c1 -= 10; }
    else if (f1 & 0x0000000000000f00LL) { f1 <<= 44; c1 -= 11; }
    else if (f1 & 0x00000000000000f0LL) { f1 <<= 48; c1 -= 12; }
    else                                { f1 <<= 52; c1 -= 13; }
  }
  if (f2) { /* normalize */
    if      (f2 & 0x00f0000000000000LL) { /* do nothing */ }
    else if (f2 & 0x000f000000000000LL) { f2 <<=  4; c2 -=  1; }
    else if (f2 & 0x0000f00000000000LL) { f2 <<=  8; c2 -=  2; }
    else if (f2 & 0x00000f0000000000LL) { f2 <<= 12; c2 -=  3; }
    else if (f2 & 0x000000f000000000LL) { f2 <<= 16; c2 -=  4; }
    else if (f2 & 0x0000000f00000000LL) { f2 <<= 20; c2 -=  5; }
    else if (f2 & 0x00000000f0000000LL) { f2 <<= 24; c2 -=  6; }
    else if (f2 & 0x000000000f000000LL) { f2 <<= 28; c2 -=  7; }
    else if (f2 & 0x0000000000f00000LL) { f2 <<= 32; c2 -=  8; }
    else if (f2 & 0x00000000000f0000LL) { f2 <<= 36; c2 -=  9; }
    else if (f2 & 0x000000000000f000LL) { f2 <<= 40; c2 -= 10; }
    else if (f2 & 0x0000000000000f00LL) { f2 <<= 44; c2 -= 11; }
    else if (f2 & 0x00000000000000f0LL) { f2 <<= 48; c2 -= 12; }
    else                                { f2 <<= 52; c2 -= 13; }
  }

#ifdef DEBUG_FLOAT
  umes("fl_mult_long2ext: s1=%08.8x c1=%08.8x f1=%08.8x_%08.0x\n", s1, c1, (unsigned)(f1>>32), (unsigned)(f1));
  umes("fl_mult_long2ext: s2=%08.8x c2=%08.8x f2=%08.8x_%08.8x\n", s2, c2, (unsigned)(f2>>32), (unsigned)(f2));
#endif

  /* mult algebraically */
  s1 = s1 ^ s2;
  c1 = c1 + c2 - 64;
  fwl  = (llu)(unsigned)(f1&0x000000000fffffffLL) * (llu)(unsigned)(f2&0x000000000fffffffLL);
  fwm  = (llu)(unsigned)(f1>>28)                  * (llu)(unsigned)(f2&0x000000000fffffffLL);
  fwm += (llu)(unsigned)(f1&0x000000000fffffffLL) * (llu)(unsigned)(f2>>28);
  fwu  = (llu)(unsigned)(f1>>28)                  * (llu)(unsigned)(f2>>28);
  fwm += fwl>>28;
  fwu += fwm>>28;
  fwl  = fwu<<56 | (fwm<<28)&0x00fffffff0000000LL | fwl&0x000000000fffffffLL;
  fwu  = fwu>> 8;

#ifdef DEBUG_FLOAT
  umes("fl_mult_long2ext: s1=%08.8x c1=%08.8x ", s1, c1);
  umes("fw=%08.8x_%08.8x %08.8x_%08.8x\n", (unsigned)(fwu>>32), (unsigned)(fwu), (unsigned)(fwl>>32), (unsigned)(fwl));
#endif

  /* normalize */
  if (fwu) { /* normalize */
    if (fwu & 0x0000f00000000000LL) {
      fwu = fwu <<  8 | fwl >> 56;
      fwl <<=  8;
    }
    else {
      fwu = fwu << 12 | fwl >> 52;
      fwl <<= 12;
      c1--;
    }

    if (c1 & 0x80000000) { /* if characteristic underflow */
      if (mask & EXP_UNF_MASK) {
        c1 += 128;
        c1l = (c1 - 14) & 0x7f;
        intr = EXP_UNF; /* exponent underflow */
      }
      else {
        s1 = c1 = c1l = 0; /* true zero */
        fwu = fwl = 0LL;   /* true zero */
      }
    }
    else {
      if (c1 > 127) { /* if characteristic overflow */
        c1 -= 128;
        intr = EXP_OVF; /* exponent overflow */
      }
      c1l = (c1 - 14) & 0x7f;
    }
  }
  else { /* fraction == 0 */
    s1 = c1 = c1l = 0; /* true zero */
    fwu = fwl = 0LL;   /* true zero */
  }

  /* format result */
  srcwu = (llu)s1 << 32 | (llu)c1  << 56 | fwu;
  srcwl = (llu)s1 << 32 | (llu)c1l << 56 | fwl >> 8;
  *(llu*)(r1  ) = srcwu; /* always aligned */
  *(llu*)(r1+8) = srcwl; /* always aligned */

#ifdef DEBUG_FLOAT
  umes("fl_mult_long2ext: s1=%08.8x c1=%08.8x ", s1, c1);
  umes("fw=%08.8x_%08.8x %08.8x_%08.8x\n", (unsigned)(fwu>>32), (unsigned)(fwu), (unsigned)(fwl>>32), (unsigned)(fwl));
#endif

  return (intr);
}

unsigned
fl_mult_ext2ext(dummy, r1, r2, mask)
  unsigned dummy;
  unsigned char *r1, *r2;
  unsigned mask;   /* program mask */
  /* return INTR-CODE */
{
  llu      src1u, src1l, src2u, src2l;
  llu      srcwu, srcwl;
  unsigned s1, s2;
  unsigned c1, c2, c1l;
  llu      f1u, f1l, f2u, f2l;
  llu      fwu, fwmuu, fwmu, fwmm, fwml, fwmll, fwl;
  unsigned intr = 0;

  src1u = *(llu*)(r1  );        /* always aligned */
  src1l = *(llu*)(r1+8);        /* always aligned */
  src2u = *(llu*)(r2  );        /* always aligned */
  src2l = *(llu*)(r2+8);        /* always aligned */

  s1  = src1u>>32 & 0x80000000;           /* s------- -------- -------- -------- */
  s2  = src2u>>32 & 0x80000000;           /* s------- -------- -------- -------- */
  c1  = src1u>>56 & 0x0000007f;           /* -------- -------- -------- -ccccccc */
  c2  = src2u>>56 & 0x0000007f;           /* -------- -------- -------- -ccccccc */
  f1u = src1u     & 0x00ffffffffffffffLL; /* -------- ffffff.. ffffffff ffffffff */
  f2u = src2u     & 0x00ffffffffffffffLL; /* -------- ffffff.. ffffffff ffffffff */
  f1l = src1l     & 0x00ffffffffffffffLL; /* -------- ffffff.. ffffffff ffffffff */
  f2l = src2l     & 0x00ffffffffffffffLL; /* -------- ffffff.. ffffffff ffffffff */

#ifdef DEBUG_FLOAT
  umes("fl_mult_ext2ext: s1=%08.8x c1=%08.8x ", s1, c1);
  umes("f1=%08.8x_%08.8x %08.8x_%08.8x\n", (unsigned)(f1u>>32), (unsigned)(f1u), (unsigned)(f1l>>32), (unsigned)(f1l));
  umes("fl_mult_ext2ext: s2=%08.8x c2=%08.8x ", s2, c2);
  umes("f2=%08.8x_%08.8x %08.8x_%08.8x\n", (unsigned)(f2u>>32), (unsigned)(f2u), (unsigned)(f2l>>32), (unsigned)(f2l));
#endif

  /* normalize */
  if (f1u || f1l) { /* normalize */
    if      (f1u & 0x00f0000000000000LL) { /* do nothing */ }
    else if (f1u & 0x000f000000000000LL) { f1u = f1u <<  4 | f1l >> 52; f1l = f1l <<  4 & 0x00ffffffffffffffLL; c1 -=  1; }
    else if (f1u & 0x0000f00000000000LL) { f1u = f1u <<  8 | f1l >> 48; f1l = f1l <<  8 & 0x00ffffffffffffffLL; c1 -=  2; }
    else if (f1u & 0x00000f0000000000LL) { f1u = f1u << 12 | f1l >> 44; f1l = f1l << 12 & 0x00ffffffffffffffLL; c1 -=  3; }
    else if (f1u & 0x000000f000000000LL) { f1u = f1u << 16 | f1l >> 40; f1l = f1l << 16 & 0x00ffffffffffffffLL; c1 -=  4; }
    else if (f1u & 0x0000000f00000000LL) { f1u = f1u << 20 | f1l >> 36; f1l = f1l << 20 & 0x00ffffffffffffffLL; c1 -=  5; }
    else if (f1u & 0x00000000f0000000LL) { f1u = f1u << 24 | f1l >> 32; f1l = f1l << 24 & 0x00ffffffffffffffLL; c1 -=  6; }
    else if (f1u & 0x000000000f000000LL) { f1u = f1u << 28 | f1l >> 28; f1l = f1l << 28 & 0x00ffffffffffffffLL; c1 -=  7; }
    else if (f1u & 0x0000000000f00000LL) { f1u = f1u << 32 | f1l >> 24; f1l = f1l << 32 & 0x00ffffffffffffffLL; c1 -=  8; }
    else if (f1u & 0x00000000000f0000LL) { f1u = f1u << 36 | f1l >> 20; f1l = f1l << 36 & 0x00ffffffffffffffLL; c1 -=  9; }
    else if (f1u & 0x000000000000f000LL) { f1u = f1u << 40 | f1l >> 16; f1l = f1l << 40 & 0x00ffffffffffffffLL; c1 -= 10; }
    else if (f1u & 0x0000000000000f00LL) { f1u = f1u << 44 | f1l >> 12; f1l = f1l << 44 & 0x00ffffffffffffffLL; c1 -= 11; }
    else if (f1u & 0x00000000000000f0LL) { f1u = f1u << 48 | f1l >>  8; f1l = f1l << 48 & 0x00ffffffffffffffLL; c1 -= 12; }
    else if (f1u & 0x000000000000000fLL) { f1u = f1u << 52 | f1l >>  4; f1l = f1l << 52 & 0x00ffffffffffffffLL; c1 -= 13; }
    else if (f1l & 0x00f0000000000000LL) { f1u =             f1l      ; f1l =             0x0000000000000000LL; c1 -= 14; }
    else if (f1l & 0x000f000000000000LL) { f1u =             f1l <<  4; f1l =             0x0000000000000000LL; c1 -= 15; }
    else if (f1l & 0x0000f00000000000LL) { f1u =             f1l <<  8; f1l =             0x0000000000000000LL; c1 -= 16; }
    else if (f1l & 0x00000f0000000000LL) { f1u =             f1l << 12; f1l =             0x0000000000000000LL; c1 -= 17; }
    else if (f1l & 0x000000f000000000LL) { f1u =             f1l << 16; f1l =             0x0000000000000000LL; c1 -= 18; }
    else if (f1l & 0x0000000f00000000LL) { f1u =             f1l << 20; f1l =             0x0000000000000000LL; c1 -= 19; }
    else if (f1l & 0x00000000f0000000LL) { f1u =             f1l << 24; f1l =             0x0000000000000000LL; c1 -= 20; }
    else if (f1l & 0x000000000f000000LL) { f1u =             f1l << 28; f1l =             0x0000000000000000LL; c1 -= 21; }
    else if (f1l & 0x0000000000f00000LL) { f1u =             f1l << 32; f1l =             0x0000000000000000LL; c1 -= 22; }
    else if (f1l & 0x00000000000f0000LL) { f1u =             f1l << 36; f1l =             0x0000000000000000LL; c1 -= 23; }
    else if (f1l & 0x000000000000f000LL) { f1u =             f1l << 40; f1l =             0x0000000000000000LL; c1 -= 24; }
    else if (f1l & 0x0000000000000f00LL) { f1u =             f1l << 44; f1l =             0x0000000000000000LL; c1 -= 25; }
    else if (f1l & 0x00000000000000f0LL) { f1u =             f1l << 48; f1l =             0x0000000000000000LL; c1 -= 26; }
    else                                 { f1u =             f1l << 52; f1l =             0x0000000000000000LL; c1 -= 27; }
  }
  if (f2u || f2l) { /* normalize */
    if      (f2u & 0x00f0000000000000LL) { /* do nothing */ }
    else if (f2u & 0x000f000000000000LL) { f2u = f2u <<  4 | f2l >> 52; f2l = f2l <<  4 & 0x00ffffffffffffffLL; c2 -=  1; }
    else if (f2u & 0x0000f00000000000LL) { f2u = f2u <<  8 | f2l >> 48; f2l = f2l <<  8 & 0x00ffffffffffffffLL; c2 -=  2; }
    else if (f2u & 0x00000f0000000000LL) { f2u = f2u << 12 | f2l >> 44; f2l = f2l << 12 & 0x00ffffffffffffffLL; c2 -=  3; }
    else if (f2u & 0x000000f000000000LL) { f2u = f2u << 16 | f2l >> 40; f2l = f2l << 16 & 0x00ffffffffffffffLL; c2 -=  4; }
    else if (f2u & 0x0000000f00000000LL) { f2u = f2u << 20 | f2l >> 36; f2l = f2l << 20 & 0x00ffffffffffffffLL; c2 -=  5; }
    else if (f2u & 0x00000000f0000000LL) { f2u = f2u << 24 | f2l >> 32; f2l = f2l << 24 & 0x00ffffffffffffffLL; c2 -=  6; }
    else if (f2u & 0x000000000f000000LL) { f2u = f2u << 28 | f2l >> 28; f2l = f2l << 28 & 0x00ffffffffffffffLL; c2 -=  7; }
    else if (f2u & 0x0000000000f00000LL) { f2u = f2u << 32 | f2l >> 24; f2l = f2l << 32 & 0x00ffffffffffffffLL; c2 -=  8; }
    else if (f2u & 0x00000000000f0000LL) { f2u = f2u << 36 | f2l >> 20; f2l = f2l << 36 & 0x00ffffffffffffffLL; c2 -=  9; }
    else if (f2u & 0x000000000000f000LL) { f2u = f2u << 40 | f2l >> 16; f2l = f2l << 40 & 0x00ffffffffffffffLL; c2 -= 10; }
    else if (f2u & 0x0000000000000f00LL) { f2u = f2u << 44 | f2l >> 12; f2l = f2l << 44 & 0x00ffffffffffffffLL; c2 -= 11; }
    else if (f2u & 0x00000000000000f0LL) { f2u = f2u << 48 | f2l >>  8; f2l = f2l << 48 & 0x00ffffffffffffffLL; c2 -= 12; }
    else if (f2u & 0x000000000000000fLL) { f2u = f2u << 52 | f2l >>  4; f2l = f2l << 52 & 0x00ffffffffffffffLL; c2 -= 13; }
    else if (f2l & 0x00f0000000000000LL) { f2u =             f2l      ; f2l =             0x0000000000000000LL; c2 -= 14; }
    else if (f2l & 0x000f000000000000LL) { f2u =             f2l <<  4; f2l =             0x0000000000000000LL; c2 -= 15; }
    else if (f2l & 0x0000f00000000000LL) { f2u =             f2l <<  8; f2l =             0x0000000000000000LL; c2 -= 16; }
    else if (f2l & 0x00000f0000000000LL) { f2u =             f2l << 12; f2l =             0x0000000000000000LL; c2 -= 17; }
    else if (f2l & 0x000000f000000000LL) { f2u =             f2l << 16; f2l =             0x0000000000000000LL; c2 -= 18; }
    else if (f2l & 0x0000000f00000000LL) { f2u =             f2l << 20; f2l =             0x0000000000000000LL; c2 -= 19; }
    else if (f2l & 0x00000000f0000000LL) { f2u =             f2l << 24; f2l =             0x0000000000000000LL; c2 -= 20; }
    else if (f2l & 0x000000000f000000LL) { f2u =             f2l << 28; f2l =             0x0000000000000000LL; c2 -= 21; }
    else if (f2l & 0x0000000000f00000LL) { f2u =             f2l << 32; f2l =             0x0000000000000000LL; c2 -= 22; }
    else if (f2l & 0x00000000000f0000LL) { f2u =             f2l << 36; f2l =             0x0000000000000000LL; c2 -= 23; }
    else if (f2l & 0x000000000000f000LL) { f2u =             f2l << 40; f2l =             0x0000000000000000LL; c2 -= 24; }
    else if (f2l & 0x0000000000000f00LL) { f2u =             f2l << 44; f2l =             0x0000000000000000LL; c2 -= 25; }
    else if (f2l & 0x00000000000000f0LL) { f2u =             f2l << 48; f2l =             0x0000000000000000LL; c2 -= 26; }
    else                                 { f2u =             f2l << 52; f2l =             0x0000000000000000LL; c2 -= 27; }
  }

#ifdef DEBUG_FLOAT
  umes("fl_mult_ext2ext: s1=%08.8x c1=%08.8x ", s1, c1);
  umes("f1=%08.8x_%08.8x %08.8x_%08.8x\n", (unsigned)(f1u>>32), (unsigned)(f1u), (unsigned)(f1l>>32), (unsigned)(f1l));
  umes("fl_mult_ext2ext: s2=%08.8x c2=%08.8x ", s2, c2);
  umes("f2=%08.8x_%08.8x %08.8x_%08.8x\n", (unsigned)(f2u>>32), (unsigned)(f2u), (unsigned)(f2l>>32), (unsigned)(f2l));
#endif

  /* mult algebraically */
  s1 = s1 ^ s2;
  c1 = c1 + c2 - 64;
  fwl    = (llu)(unsigned)(f1l&0x000000000fffffffLL) * (llu)(unsigned)(f2l&0x000000000fffffffLL);
  fwmll  = (llu)(unsigned)(f1l>>28)                  * (llu)(unsigned)(f2l&0x000000000fffffffLL);
  fwml   = (llu)(unsigned)(f1u&0x000000000fffffffLL) * (llu)(unsigned)(f2l&0x000000000fffffffLL);
  fwmm   = (llu)(unsigned)(f1u>>28)                  * (llu)(unsigned)(f2l&0x000000000fffffffLL);
  fwmll += (llu)(unsigned)(f1l&0x000000000fffffffLL) * (llu)(unsigned)(f2l>>28);
  fwml  += (llu)(unsigned)(f1l>>28)                  * (llu)(unsigned)(f2l>>28);
  fwmm  += (llu)(unsigned)(f1u&0x000000000fffffffLL) * (llu)(unsigned)(f2l>>28);
  fwmu   = (llu)(unsigned)(f1u>>28)                  * (llu)(unsigned)(f2l>>28);
  fwml  += (llu)(unsigned)(f1l&0x000000000fffffffLL) * (llu)(unsigned)(f2u&0x000000000fffffffLL);
  fwmm  += (llu)(unsigned)(f1l>>28)                  * (llu)(unsigned)(f2u&0x000000000fffffffLL);
  fwmu  += (llu)(unsigned)(f1u&0x000000000fffffffLL) * (llu)(unsigned)(f2u&0x000000000fffffffLL);
  fwmuu  = (llu)(unsigned)(f1u>>28)                  * (llu)(unsigned)(f2u&0x000000000fffffffLL);
  fwmm  += (llu)(unsigned)(f1l&0x000000000fffffffLL) * (llu)(unsigned)(f2u>>28);
  fwmu  += (llu)(unsigned)(f1l>>28)                  * (llu)(unsigned)(f2u>>28);
  fwmuu += (llu)(unsigned)(f1u&0x000000000fffffffLL) * (llu)(unsigned)(f2u>>28);
  fwu    = (llu)(unsigned)(f1u>>28)                  * (llu)(unsigned)(f2u>>28);
  fwmll += fwl  >>28;
  fwml  += fwmll>>28;
  fwmm  += fwml >>28;
  fwmu  += fwmm >>28;
  fwmuu += fwmu >>28;
  fwu   += fwmuu>>28;
  fwl    = fwml<<56 | (fwmll<<28)&0x00fffffff0000000LL | (fwl     )&0x000000000fffffffLL;
  fwml   = fwmu<<48 | (fwmm <<20)&0x0000fffffff00000LL | (fwml>> 8)&0x00000000000fffffLL;
  fwmu   = fwu <<40 | (fwmuu<<12)&0x000000fffffff000LL | (fwmu>>16)&0x0000000000000fffLL;
  fwu    = fwu >>24;

#ifdef DEBUG_FLOAT
  umes("fl_mult_ext2ext: s1=%08.8x c1=%08.8x ", s1, c1);
  umes("fw=%08.8x_%08.8x %08.8x_%08.8x ", (unsigned)(fwu>>32), (unsigned)(fwu),(unsigned)(fwmu>>32), (unsigned)(fwmu));
  umes("%08.8x_%08.8x %08.8x_%08.8x\n", (unsigned)(fwml>>32), (unsigned)(fwml), (unsigned)(fwl>>32), (unsigned)(fwl));
#endif

  /* normalize */
  if (fwu) { /* normalize */
    if (fwu & 0x00000000f0000000LL) {
      fwu  = fwu  << 24 | fwmu >> 40;
      fwmu = fwmu << 24 | fwml >> 40;
      fwml = fwml << 24 | fwl  >> 40;
      fwl  <<= 24;
    }
    else {
      fwu  = fwu  << 28 | fwmu >> 36;
      fwmu = fwmu << 28 | fwml >> 36;
      fwml = fwml << 28 | fwl  >> 36;
      fwl  <<= 28;
      c1--;
    }

    if (c1 & 0x80000000) { /* if characteristic underflow */
      if (mask & EXP_UNF_MASK) {
        c1 += 128;
        c1l = (c1 - 14) & 0x7f;
        intr = EXP_UNF; /* exponent underflow */
      }
      else {
        s1 = c1 = c1l = 0;             /* true zero */
        fwu = fwmu = fwml = fwl = 0LL; /* true zero */
      }
    }
    else {
      if (c1 > 127) { /* if characteristic overflow */
        c1 -= 128;
        intr = EXP_OVF; /* exponent overflow */
      }
      c1l = (c1 - 14) & 0x7f;
    }
  }
  else { /* fraction == 0 */
    s1 = c1 = c1l = 0;             /* true zero */
    fwu = fwmu = fwml = fwl = 0LL; /* true zero */
  }

  /* format result */
  srcwu = (llu)s1 << 32 | (llu)c1  << 56 | fwu;
  srcwl = (llu)s1 << 32 | (llu)c1l << 56 | fwmu >> 8;
  *(llu*)(r1  ) = srcwu; /* always aligned */
  *(llu*)(r1+8) = srcwl; /* always aligned */

#ifdef DEBUG_FLOAT
  umes("fl_mult_ext2ext: s1=%08.8x c1=%08.8x ", s1, c1);
  umes("fw=%08.8x_%08.8x %08.8x_%08.8x ", (unsigned)(fwu>>32), (unsigned)(fwu),(unsigned)(fwmu>>32), (unsigned)(fwmu));
  umes("%08.8x_%08.8x %08.8x_%08.8x\n", (unsigned)(fwml>>32), (unsigned)(fwml), (unsigned)(fwl>>32), (unsigned)(fwl));
#endif

  return (intr);
}

unsigned
fl_div_short(dummy, r1, r2, mask)
  unsigned dummy;
  unsigned char *r1, *r2;
  unsigned mask;   /* program mask */
  /* return INTR-CODE */
{
  unsigned src1, src2;
  unsigned s1, s2;
  unsigned c1, c2;
  unsigned f1, f2, fw;
  unsigned i;
  unsigned intr = 0;

  src1 = *(unsigned*)r1; /* always aligned */
  src2 = *(unsigned*)r2; /* always aligned */

  s1 = src1     & 0x80000000; /* s------- -------- -------- -------- */
  s2 = src2     & 0x80000000; /* s------- -------- -------- -------- */
  c1 = src1>>24 & 0x0000007f; /* -------- -------- -------- -ccccccc */
  c2 = src2>>24 & 0x0000007f; /* -------- -------- -------- -ccccccc */
  f1 = src1     & 0x00ffffff; /* -------- ffffffff ffffffff ffffffff */
  f2 = src2     & 0x00ffffff; /* -------- ffffffff ffffffff ffffffff */

#ifdef DEBUG_FLOAT
  umes("fl_div_short: s1=%08.8x c1=%08.8x f1=%08.8x\n", s1, c1, f1);
  umes("fl_div_short: s2=%08.8x c2=%08.8x f2=%08.8x\n", s2, c2, f2);
#endif

  /* normalize */
  if (f1) { /* normalize */
    if      (f1 & 0x00f00000) { /* do nothing */ }
    else if (f1 & 0x000f0000) { f1 <<=  4; c1 -= 1; }
    else if (f1 & 0x0000f000) { f1 <<=  8; c1 -= 2; }
    else if (f1 & 0x00000f00) { f1 <<= 12; c1 -= 3; }
    else if (f1 & 0x000000f0) { f1 <<= 16; c1 -= 4; }
    else                      { f1 <<= 20; c1 -= 5; }
  }
  if (f2) { /* normalize */
    if      (f2 & 0x00f00000) { f2 <<=  3; c2 -= 1; }
    else if (f2 & 0x000f0000) { f2 <<=  7; c2 -= 2; }
    else if (f2 & 0x0000f000) { f2 <<= 11; c2 -= 3; }
    else if (f2 & 0x00000f00) { f2 <<= 15; c2 -= 4; }
    else if (f2 & 0x000000f0) { f2 <<= 19; c2 -= 5; }
    else                      { f2 <<= 23; c2 -= 6; }
  }

#ifdef DEBUG_FLOAT
  umes("fl_div_short: s1=%08.8x c1=%08.8x f1=%08.8x\n", s1, c1, f1);
  umes("fl_div_short: s2=%08.8x c2=%08.8x f2=%08.8x\n", s2, c2, f2);
#endif

  /* check divisor */
  if (!f2) {
    intr = FLO_DIV; /* floating point divide */
    return (intr);
  }

  /* divide algebraically */
  s1 = s1 ^ s2;
  c1 = c1 - c2 + 64;
  fw = 0;
  for (i=0; i<28; i++) {
    fw <<= 1;
    if (f1 >= f2) {
      f1 -= f2;
      fw++;
    }
    f1 <<= 1;
#ifdef DEBUG_FLOAT
  umes("fl_div_short: f1=%08.8x f2=%08.8x fw=%08.8x\n", f1, f2, fw);
#endif
  }

#ifdef DEBUG_FLOAT
  umes("fl_div_short: s1=%08.8x c1=%08.8x fw=%08.8x\n", s1, c1, fw);
#endif

  /* normalize */
  if (fw) { /* normalize */
    if (fw & 0x0f000000) { fw >>= 4; }
    else                 { c1--; }

    if (c1 & 0x80000000) { /* if characteristic underflow */
      if (mask & EXP_UNF_MASK) {
        c1 += 128;
        intr = EXP_UNF; /* exponent underflow */
      }
      else {
        s1 = c1 = fw = 0; /* true zero */
      }
    }
    else if (c1 > 127) { /* if characteristic overflow */
      c1 -= 128;
      intr = EXP_OVF; /* exponent overflow */
    }
  }
  else { /* fraction == 0 */
    s1 = c1 = fw = 0; /* true zero */
  }

  /* format result */
  src1 = s1 | c1 << 24 | fw;
  *(unsigned*)r1 = src1; /* always aligned */

#ifdef DEBUG_FLOAT
  umes("fl_div_short: s1=%08.8x c1=%08.8x fw=%08.8x\n", s1, c1, fw);
#endif

  return (intr);
}

unsigned
fl_div_long(dummy, r1, r2, mask)
  unsigned dummy;
  unsigned char *r1, *r2;
  unsigned mask;   /* program mask */
  /* return INTR-CODE */
{
  llu      src1, src2;
  unsigned s1, s2;
  unsigned c1, c2;
  llu      f1, f2, fw;
  unsigned i;
  unsigned intr = 0;

  src1 = *(llu*)r1;        /* always aligned */
  src2 = *(llu*)r2;        /* always aligned */

  s1 = src1>>32 & 0x80000000;           /* s------- -------- -------- -------- */
  s2 = src2>>32 & 0x80000000;           /* s------- -------- -------- -------- */
  c1 = src1>>56 & 0x0000007f;           /* -------- -------- -------- -ccccccc */
  c2 = src2>>56 & 0x0000007f;           /* -------- -------- -------- -ccccccc */
  f1 = src1     & 0x00ffffffffffffffLL; /* -------- ffffff.. ffffffff ffffffff */
  f2 = src2     & 0x00ffffffffffffffLL; /* -------- ffffff.. ffffffff ffffffff */

#ifdef DEBUG_FLOAT
  umes("fl_div_long: s1=%08.8x c1=%08.8x f1=%08.8x_%08.8x\n", s1, c1, (unsigned)(f1>>32), (unsigned)(f1));
  umes("fl_div_long: s2=%08.8x c2=%08.8x f2=%08.8x_%08.8x\n", s2, c2, (unsigned)(f2>>32), (unsigned)(f2));
#endif

  /* normalize */
  if (f1) { /* normalize */
    if      (f1 & 0x00f0000000000000LL) { /* do nothing */ }
    else if (f1 & 0x000f000000000000LL) { f1 <<=  4; c1 -=  1; }
    else if (f1 & 0x0000f00000000000LL) { f1 <<=  8; c1 -=  2; }
    else if (f1 & 0x00000f0000000000LL) { f1 <<= 12; c1 -=  3; }
    else if (f1 & 0x000000f000000000LL) { f1 <<= 16; c1 -=  4; }
    else if (f1 & 0x0000000f00000000LL) { f1 <<= 20; c1 -=  5; }
    else if (f1 & 0x00000000f0000000LL) { f1 <<= 24; c1 -=  6; }
    else if (f1 & 0x000000000f000000LL) { f1 <<= 28; c1 -=  7; }
    else if (f1 & 0x0000000000f00000LL) { f1 <<= 32; c1 -=  8; }
    else if (f1 & 0x00000000000f0000LL) { f1 <<= 36; c1 -=  9; }
    else if (f1 & 0x000000000000f000LL) { f1 <<= 40; c1 -= 10; }
    else if (f1 & 0x0000000000000f00LL) { f1 <<= 44; c1 -= 11; }
    else if (f1 & 0x00000000000000f0LL) { f1 <<= 48; c1 -= 12; }
    else                                { f1 <<= 52; c1 -= 13; }
  }
  if (f2) { /* normalize */
    if      (f2 & 0x00f0000000000000LL) { f2 <<=  3; c2 -=  1; }
    else if (f2 & 0x000f000000000000LL) { f2 <<=  7; c2 -=  2; }
    else if (f2 & 0x0000f00000000000LL) { f2 <<= 11; c2 -=  3; }
    else if (f2 & 0x00000f0000000000LL) { f2 <<= 15; c2 -=  4; }
    else if (f2 & 0x000000f000000000LL) { f2 <<= 19; c2 -=  5; }
    else if (f2 & 0x0000000f00000000LL) { f2 <<= 23; c2 -=  6; }
    else if (f2 & 0x00000000f0000000LL) { f2 <<= 27; c2 -=  7; }
    else if (f2 & 0x000000000f000000LL) { f2 <<= 31; c2 -=  8; }
    else if (f2 & 0x0000000000f00000LL) { f2 <<= 35; c2 -=  9; }
    else if (f2 & 0x00000000000f0000LL) { f2 <<= 39; c2 -= 10; }
    else if (f2 & 0x000000000000f000LL) { f2 <<= 43; c2 -= 11; }
    else if (f2 & 0x0000000000000f00LL) { f2 <<= 47; c2 -= 12; }
    else if (f2 & 0x00000000000000f0LL) { f2 <<= 51; c2 -= 13; }
    else                                { f2 <<= 55; c2 -= 14; }
  }

#ifdef DEBUG_FLOAT
  umes("fl_div_long: s1=%08.8x c1=%08.8x f1=%08.8x_%08.8x\n", s1, c1, (unsigned)(f1>>32), (unsigned)(f1));
  umes("fl_div_long: s2=%08.8x c2=%08.8x f2=%08.8x_%08.8x\n", s2, c2, (unsigned)(f2>>32), (unsigned)(f2));
#endif

  /* check divisor */
  if (!f2) {
    intr = FLO_DIV; /* floating point divide */
    return (intr);
  }

  /* divide algebraically */
  s1 = s1 ^ s2;
  c1 = c1 - c2 + 64;
  fw = 0LL;
  for (i=0; i<60; i++) {
    fw <<= 1;
    if (f1 >= f2) {
      f1 -= f2;
      fw++;
    }
    f1 <<= 1;
#ifdef DEBUG_FLOAT
  umes("fl_div_long: f1=%08.8x_%08.8x f2=%08.8x_%08.8x ", (unsigned)(f1>>32), (unsigned)(f1), (unsigned)(f2>>32), (unsigned)(f2));
  umes("fw=%08.8x_%08.8x\n", (unsigned)(fw>>32), (unsigned)(fw));
#endif
  }

#ifdef DEBUG_FLOAT
  umes("fl_div_long: s1=%08.8x c1=%08.8x fw=%08.8x_%08.8x\n", s1, c1, (unsigned)(fw>>32), (unsigned)(fw));
#endif

  /* normalize */
  if (fw) { /* normalize */
    if (fw & 0x0f00000000000000LL) { fw >>= 4; }
    else                           { c1--; }

    if (c1 & 0x80000000) { /* if characteristic underflow */
      if (mask & EXP_UNF_MASK) {
        c1 += 128;
        intr = EXP_UNF; /* exponent underflow */
      }
      else {
        s1 = c1 = 0; /* true zero */
        fw = 0LL;    /* true zero */
      }
    }
    else if (c1 > 127) { /* if characteristic overflow */
      c1 -= 128;
      intr = EXP_OVF; /* exponent overflow */
    }
  }
  else { /* fraction == 0 */
    s1 = c1 = 0; /* true zero */
    fw = 0LL;    /* true zero */
  }

  /* format result */
  src1 = (llu)s1 << 32 | (llu)c1 << 56 | fw;
  *(llu*)r1 = src1; /* always aligned */

#ifdef DEBUG_FLOAT
  umes("fl_div_long: s1=%08.8x c1=%08.8x fw=%08.8x_%08.8x\n", s1, c1, (unsigned)(fw>>32), (unsigned)(fw));
#endif

  return (intr);
}

unsigned
fl_div_ext(dummy, r1, r2, mask)
  unsigned dummy;
  unsigned char *r1, *r2;
  unsigned mask;   /* program mask */
  /* return INTR-CODE */
{
  llu      src1u, src1l, src2u, src2l;
  unsigned s1, s2;
  unsigned c1, c2, c1l;
  llu      f1u, f1l, f2u, f2l, fwu, fwl;
  unsigned i;
  unsigned intr = 0;

  src1u = *(llu*)(r1  );        /* always aligned */
  src1l = *(llu*)(r1+8);        /* always aligned */
  src2u = *(llu*)(r2  );        /* always aligned */
  src2l = *(llu*)(r2+8);        /* always aligned */

  s1  = src1u>>32 & 0x80000000;           /* s------- -------- -------- -------- */
  s2  = src2u>>32 & 0x80000000;           /* s------- -------- -------- -------- */
  c1  = src1u>>56 & 0x0000007f;           /* -------- -------- -------- -ccccccc */
  c2  = src2u>>56 & 0x0000007f;           /* -------- -------- -------- -ccccccc */
  f1u = src1u     & 0x00ffffffffffffffLL; /* -------- ffffff.. ffffffff ffffffff */
  f2u = src2u     & 0x00ffffffffffffffLL; /* -------- ffffff.. ffffffff ffffffff */
  f1l = src1l     & 0x00ffffffffffffffLL; /* -------- ffffff.. ffffffff ffffffff */
  f2l = src2l     & 0x00ffffffffffffffLL; /* -------- ffffff.. ffffffff ffffffff */

#ifdef DEBUG_FLOAT
  umes("fl_div_ext: s1=%08.8x c1=%08.8x ", s1, c1);
  umes("f1=%08.8x_%08.8x %08.8x_%08.8x\n", (unsigned)(f1u>>32), (unsigned)(f1u), (unsigned)(f1l>>32), (unsigned)(f1l));
  umes("fl_div_ext: s2=%08.8x c2=%08.8x ", s2, c2);
  umes("f2=%08.8x_%08.8x %08.8x_%08.8x\n", (unsigned)(f2u>>32), (unsigned)(f2u), (unsigned)(f2l>>32), (unsigned)(f2l));
#endif

  /* normalize */
  if (f1u || f1l) { /* normalize */
    if      (f1u & 0x00f0000000000000LL) { /* do nothing */ }
    else if (f1u & 0x000f000000000000LL) { f1u = f1u <<  4 | f1l >> 52; f1l = f1l <<  4 & 0x00ffffffffffffffLL; c1 -=  1; }
    else if (f1u & 0x0000f00000000000LL) { f1u = f1u <<  8 | f1l >> 48; f1l = f1l <<  8 & 0x00ffffffffffffffLL; c1 -=  2; }
    else if (f1u & 0x00000f0000000000LL) { f1u = f1u << 12 | f1l >> 44; f1l = f1l << 12 & 0x00ffffffffffffffLL; c1 -=  3; }
    else if (f1u & 0x000000f000000000LL) { f1u = f1u << 16 | f1l >> 40; f1l = f1l << 16 & 0x00ffffffffffffffLL; c1 -=  4; }
    else if (f1u & 0x0000000f00000000LL) { f1u = f1u << 20 | f1l >> 36; f1l = f1l << 20 & 0x00ffffffffffffffLL; c1 -=  5; }
    else if (f1u & 0x00000000f0000000LL) { f1u = f1u << 24 | f1l >> 32; f1l = f1l << 24 & 0x00ffffffffffffffLL; c1 -=  6; }
    else if (f1u & 0x000000000f000000LL) { f1u = f1u << 28 | f1l >> 28; f1l = f1l << 28 & 0x00ffffffffffffffLL; c1 -=  7; }
    else if (f1u & 0x0000000000f00000LL) { f1u = f1u << 32 | f1l >> 24; f1l = f1l << 32 & 0x00ffffffffffffffLL; c1 -=  8; }
    else if (f1u & 0x00000000000f0000LL) { f1u = f1u << 36 | f1l >> 20; f1l = f1l << 36 & 0x00ffffffffffffffLL; c1 -=  9; }
    else if (f1u & 0x000000000000f000LL) { f1u = f1u << 40 | f1l >> 16; f1l = f1l << 40 & 0x00ffffffffffffffLL; c1 -= 10; }
    else if (f1u & 0x0000000000000f00LL) { f1u = f1u << 44 | f1l >> 12; f1l = f1l << 44 & 0x00ffffffffffffffLL; c1 -= 11; }
    else if (f1u & 0x00000000000000f0LL) { f1u = f1u << 48 | f1l >>  8; f1l = f1l << 48 & 0x00ffffffffffffffLL; c1 -= 12; }
    else if (f1u & 0x000000000000000fLL) { f1u = f1u << 52 | f1l >>  4; f1l = f1l << 52 & 0x00ffffffffffffffLL; c1 -= 13; }
    else if (f1l & 0x00f0000000000000LL) { f1u =             f1l      ; f1l =             0x0000000000000000LL; c1 -= 14; }
    else if (f1l & 0x000f000000000000LL) { f1u =             f1l <<  4; f1l =             0x0000000000000000LL; c1 -= 15; }
    else if (f1l & 0x0000f00000000000LL) { f1u =             f1l <<  8; f1l =             0x0000000000000000LL; c1 -= 16; }
    else if (f1l & 0x00000f0000000000LL) { f1u =             f1l << 12; f1l =             0x0000000000000000LL; c1 -= 17; }
    else if (f1l & 0x000000f000000000LL) { f1u =             f1l << 16; f1l =             0x0000000000000000LL; c1 -= 18; }
    else if (f1l & 0x0000000f00000000LL) { f1u =             f1l << 20; f1l =             0x0000000000000000LL; c1 -= 19; }
    else if (f1l & 0x00000000f0000000LL) { f1u =             f1l << 24; f1l =             0x0000000000000000LL; c1 -= 20; }
    else if (f1l & 0x000000000f000000LL) { f1u =             f1l << 28; f1l =             0x0000000000000000LL; c1 -= 21; }
    else if (f1l & 0x0000000000f00000LL) { f1u =             f1l << 32; f1l =             0x0000000000000000LL; c1 -= 22; }
    else if (f1l & 0x00000000000f0000LL) { f1u =             f1l << 36; f1l =             0x0000000000000000LL; c1 -= 23; }
    else if (f1l & 0x000000000000f000LL) { f1u =             f1l << 40; f1l =             0x0000000000000000LL; c1 -= 24; }
    else if (f1l & 0x0000000000000f00LL) { f1u =             f1l << 44; f1l =             0x0000000000000000LL; c1 -= 25; }
    else if (f1l & 0x00000000000000f0LL) { f1u =             f1l << 48; f1l =             0x0000000000000000LL; c1 -= 26; }
    else                                 { f1u =             f1l << 52; f1l =             0x0000000000000000LL; c1 -= 27; }
  }
  if (f2u || f2l) { /* normalize */
    if      (f2u & 0x00f0000000000000LL) { f2u = f2u <<  3 | f2l >> 53; f2l = f2l <<  3 & 0x00ffffffffffffffLL; c2 -=  1; }
    else if (f2u & 0x000f000000000000LL) { f2u = f2u <<  7 | f2l >> 49; f2l = f2l <<  7 & 0x00ffffffffffffffLL; c2 -=  2; }
    else if (f2u & 0x0000f00000000000LL) { f2u = f2u << 11 | f2l >> 45; f2l = f2l << 11 & 0x00ffffffffffffffLL; c2 -=  3; }
    else if (f2u & 0x00000f0000000000LL) { f2u = f2u << 15 | f2l >> 41; f2l = f2l << 15 & 0x00ffffffffffffffLL; c2 -=  4; }
    else if (f2u & 0x000000f000000000LL) { f2u = f2u << 19 | f2l >> 37; f2l = f2l << 19 & 0x00ffffffffffffffLL; c2 -=  5; }
    else if (f2u & 0x0000000f00000000LL) { f2u = f2u << 23 | f2l >> 33; f2l = f2l << 23 & 0x00ffffffffffffffLL; c2 -=  6; }
    else if (f2u & 0x00000000f0000000LL) { f2u = f2u << 27 | f2l >> 29; f2l = f2l << 27 & 0x00ffffffffffffffLL; c2 -=  7; }
    else if (f2u & 0x000000000f000000LL) { f2u = f2u << 31 | f2l >> 25; f2l = f2l << 31 & 0x00ffffffffffffffLL; c2 -=  8; }
    else if (f2u & 0x0000000000f00000LL) { f2u = f2u << 35 | f2l >> 21; f2l = f2l << 35 & 0x00ffffffffffffffLL; c2 -=  9; }
    else if (f2u & 0x00000000000f0000LL) { f2u = f2u << 39 | f2l >> 17; f2l = f2l << 39 & 0x00ffffffffffffffLL; c2 -= 10; }
    else if (f2u & 0x000000000000f000LL) { f2u = f2u << 43 | f2l >> 13; f2l = f2l << 43 & 0x00ffffffffffffffLL; c2 -= 11; }
    else if (f2u & 0x0000000000000f00LL) { f2u = f2u << 47 | f2l >>  9; f2l = f2l << 47 & 0x00ffffffffffffffLL; c2 -= 12; }
    else if (f2u & 0x00000000000000f0LL) { f2u = f2u << 51 | f2l >>  5; f2l = f2l << 51 & 0x00ffffffffffffffLL; c2 -= 13; }
    else if (f2u & 0x000000000000000fLL) { f2u = f2u << 55 | f2l >>  1; f2l = f2l << 55 & 0x00ffffffffffffffLL; c2 -= 14; }
    else if (f2l & 0x00f0000000000000LL) { f2u =             f2l <<  3; f2l =             0x0000000000000000LL; c2 -= 15; }
    else if (f2l & 0x000f000000000000LL) { f2u =             f2l <<  7; f2l =             0x0000000000000000LL; c2 -= 16; }
    else if (f2l & 0x0000f00000000000LL) { f2u =             f2l << 11; f2l =             0x0000000000000000LL; c2 -= 17; }
    else if (f2l & 0x00000f0000000000LL) { f2u =             f2l << 15; f2l =             0x0000000000000000LL; c2 -= 18; }
    else if (f2l & 0x000000f000000000LL) { f2u =             f2l << 19; f2l =             0x0000000000000000LL; c2 -= 19; }
    else if (f2l & 0x0000000f00000000LL) { f2u =             f2l << 23; f2l =             0x0000000000000000LL; c2 -= 20; }
    else if (f2l & 0x00000000f0000000LL) { f2u =             f2l << 27; f2l =             0x0000000000000000LL; c2 -= 21; }
    else if (f2l & 0x000000000f000000LL) { f2u =             f2l << 31; f2l =             0x0000000000000000LL; c2 -= 22; }
    else if (f2l & 0x0000000000f00000LL) { f2u =             f2l << 35; f2l =             0x0000000000000000LL; c2 -= 23; }
    else if (f2l & 0x00000000000f0000LL) { f2u =             f2l << 39; f2l =             0x0000000000000000LL; c2 -= 24; }
    else if (f2l & 0x000000000000f000LL) { f2u =             f2l << 43; f2l =             0x0000000000000000LL; c2 -= 25; }
    else if (f2l & 0x0000000000000f00LL) { f2u =             f2l << 47; f2l =             0x0000000000000000LL; c2 -= 26; }
    else if (f2l & 0x00000000000000f0LL) { f2u =             f2l << 51; f2l =             0x0000000000000000LL; c2 -= 27; }
    else                                 { f2u =             f2l << 55; f2l =             0x0000000000000000LL; c2 -= 28; }
  }

#ifdef DEBUG_FLOAT
  umes("fl_div_ext: s1=%08.8x c1=%08.8x ", s1, c1);
  umes("f1=%08.8x_%08.8x %08.8x_%08.8x\n", (unsigned)(f1u>>32), (unsigned)(f1u), (unsigned)(f1l>>32), (unsigned)(f1l));
  umes("fl_div_ext: s2=%08.8x c2=%08.8x ", s2, c2);
  umes("f2=%08.8x_%08.8x %08.8x_%08.8x\n", (unsigned)(f2u>>32), (unsigned)(f2u), (unsigned)(f2l>>32), (unsigned)(f2l));
#endif

  /* check divisor */
  if (!f2u && !f2l) {
    intr = FLO_DIV; /* floating point divide */
    return (intr);
  }

  /* divide algebraically */
  s1 = s1 ^ s2;
  c1 = c1 - c2 + 64;
  fwu = fwl = 0LL;
  for (i=0; i<60; i++) {
    fwu <<= 1;
    if (f1u >= f2u || (f1u == f2u && f1l >= f2l)) {
      if ((f1l -= f2l) & 0x8000000000000000LL) {
	f1l &= 0x00ffffffffffffffLL;
        f1u -= (f2u+1);
      }
      else
        f1u -= (f2u);
      fwu++;
    }
    f1u = f1u << 1 | f1l >> 55;
    f1l = f1l << 1 & 0x00ffffffffffffffLL;
#ifdef DEBUG_FLOAT
  umes("fl_div_ext: s1=%08.8x c1=%08.8x ", s1, c1);
  umes("fw=%08.8x_%08.8x %08.8x_%08.8x\n", (unsigned)(fwu>>32), (unsigned)(fwu), (unsigned)(fwl>>32), (unsigned)(fwl));
#endif
  }
  for (i=0; i<60; i++) {
    fwl <<= 1;
    if (f1u >= f2u || (f1u == f2u && f1l >= f2l)) {
      if ((f1l -= f2l) & 0x8000000000000000LL) {
	f1l &= 0x00ffffffffffffffLL;
        f1u -= (f2u+1);
      }
      else
        f1u -= (f2u);
      fwl++;
    }
    f1u = f1u << 1 | f1l >> 55;
    f1l = f1l << 1 & 0x00ffffffffffffffLL;
#ifdef DEBUG_FLOAT
  umes("fl_div_ext: s1=%08.8x c1=%08.8x ", s1, c1);
  umes("fw=%08.8x_%08.8x %08.8x_%08.8x\n", (unsigned)(fwu>>32), (unsigned)(fwu), (unsigned)(fwl>>32), (unsigned)(fwl));
#endif
  }

#ifdef DEBUG_FLOAT
  umes("fl_div_ext: s1=%08.8x c1=%08.8x ", s1, c1);
  umes("fw=%08.8x_%08.8x %08.8x_%08.8x\n", (unsigned)(fwu>>32), (unsigned)(fwu), (unsigned)(fwl>>32), (unsigned)(fwl));
#endif

  /* normalize */
  if (fwu) { /* normalize */
    if (fwu & 0x0f00000000000000LL) {
      fwl = fwu << 56 & 0x0fffffffffffffffLL | fwl >> 4;
      fwu >>= 4;
    }
    else {
      c1--;
    }

    if (c1 & 0x80000000) { /* if characteristic underflow */
      if (mask & EXP_UNF_MASK) {
        c1 += 128;
        c1l = (c1 - 14) & 0x7f;
        intr = EXP_UNF; /* exponent underflow */
      }
      else {
        s1 = c1 = c1l = 0; /* true zero */
        fwu = fwl = 0LL;   /* true zero */
      }
    }
    else {
      if (c1 > 127) { /* if characteristic overflow */
        c1 -= 128;
        intr = EXP_OVF; /* exponent overflow */
      }
      c1l = (c1 - 14) & 0x7f;
    }
  }
  else { /* fraction == 0 */
    s1 = c1 = c1l = 0; /* true zero */
    fwu = fwl = 0LL;   /* true zero */
  }

  /* format result */
  src1u = (llu)s1 << 32 | (llu)c1  << 56 | fwu;
  src1l = (llu)s1 << 32 | (llu)c1l << 56 | fwl >> 4;
  *(llu*)(r1  ) = src1u; /* always aligned */
  *(llu*)(r1+8) = src1l; /* always aligned */

#ifdef DEBUG_FLOAT
  umes("fl_div_ext: s1=%08.8x c1=%08.8x ", s1, c1);
  umes("fw=%08.8x_%08.8x %08.8x_%08.8x\n", (unsigned)(fwu>>32), (unsigned)(fwu),(unsigned)(fwl>>32), (unsigned)(fwl));
#endif

  return (intr);
}

unsigned
fl_halve_short(dummy, r1, r2, mask)
  unsigned dummy;
  unsigned char *r1, *r2;
  unsigned mask;   /* program mask */
  /* return INTR-CODE */
{
  unsigned src2;
  unsigned s2;
  unsigned c2;
  unsigned f2;
  unsigned intr = 0;

  src2 = *(unsigned*)r2; /* always aligned */

  s2 = src2     & 0x80000000; /* s------- -------- -------- -------- */
  c2 = src2>>24 & 0x0000007f; /* -------- -------- -------- -ccccccc */
  f2 = src2<<4  & 0x0ffffff0; /* ----ffff ffffffff ffffffff ffff---- */
                              /*                    guard digit ^^^^ */
#ifdef DEBUG_FLOAT
  umes("fl_halve_short: s2=%08.8x c2=%08.8x f2=%08.8x\n", s2, c2, f2);
#endif

  /* shift right 1bit */
  f2 >>= 1;

#ifdef DEBUG_FLOAT
  umes("fl_halve_short: s2=%08.8x c2=%08.8x f2=%08.8x\n", s2, c2, f2);
#endif

  /* normalize */
  if (f2) { /* normalize */
    if      (f2 & 0x0f000000) { f2 >>=  4;          }
    else if (f2 & 0x00f00000) {            c2 -= 1; }
    else if (f2 & 0x000f0000) { f2 <<=  4; c2 -= 2; }
    else if (f2 & 0x0000f000) { f2 <<=  8; c2 -= 3; }
    else if (f2 & 0x00000f00) { f2 <<= 12; c2 -= 4; }
    else if (f2 & 0x000000f0) { f2 <<= 16; c2 -= 5; }
    else                      { f2 <<= 20; c2 -= 6; }

    if (c2 & 0x80000000) { /* if characteristic underflow */
      if (mask & EXP_UNF_MASK) {
        c2 += 128;
        intr = EXP_UNF; /* exponent underflow */
      }
      else {
        s2 = c2 = f2 = 0; /* true zero */
      }
    }
  }
  else { /* fraction == 0 */
    s2 = c2 = f2 = 0; /* true zero */
  }

  /* format result */
  src2 = s2 | c2 << 24 | f2;
  *(unsigned*)r1 = src2; /* always aligned */

#ifdef DEBUG_FLOAT
  umes("fl_halve_short: s2=%08.8x c2=%08.8x f2=%08.8x\n", s2, c2, f2);
#endif

  return (intr);
}

unsigned
fl_halve_long(dummy, r1, r2, mask)
  unsigned dummy;
  unsigned char *r1, *r2;
  unsigned mask;   /* program mask */
  /* return INTR-CODE */
{
  llu      src2;
  unsigned s2;
  unsigned c2;
  llu      f2;
  unsigned intr = 0;

  src2 = *(llu*)r2;     /* always aligned */

  s2 = src2>>32 & 0x80000000;           /* s------- -------- -------- -------- */
  c2 = src2>>56 & 0x0000007f;           /* -------- -------- -------- -ccccccc */
  f2 = src2<<4  & 0x0ffffffffffffff0LL; /* ----ffff ffffff.. ffffffff ffff---- */
                                        /*                    guard digit ^^^^ */
#ifdef DEBUG_FLOAT
  umes("fl_halve_long: s2=%08.8x c2=%08.8x f2=%08.8x_%08.8x\n", s2, c2, (unsigned)(f2>>32), (unsigned)(f2));
#endif

  /* shift right 1bit */
  f2 >>= 1;

#ifdef DEBUG_FLOAT
  umes("fl_halve_long: s2=%08.8x c2=%08.8x f2=%08.8x_%08.8x\n", s2, c2, (unsigned)(f2>>32), (unsigned)(f2));
#endif

  /* normalize */
  if (f2) { /* normalize */
    if      (f2 & 0x0f00000000000000LL) { f2 >>=  4;           }
    else if (f2 & 0x00f0000000000000LL) {            c2 -=  1; }
    else if (f2 & 0x000f000000000000LL) { f2 <<=  4; c2 -=  2; }
    else if (f2 & 0x0000f00000000000LL) { f2 <<=  8; c2 -=  3; }
    else if (f2 & 0x00000f0000000000LL) { f2 <<= 12; c2 -=  4; }
    else if (f2 & 0x000000f000000000LL) { f2 <<= 16; c2 -=  5; }
    else if (f2 & 0x0000000f00000000LL) { f2 <<= 20; c2 -=  6; }
    else if (f2 & 0x00000000f0000000LL) { f2 <<= 24; c2 -=  7; }
    else if (f2 & 0x000000000f000000LL) { f2 <<= 28; c2 -=  8; }
    else if (f2 & 0x0000000000f00000LL) { f2 <<= 32; c2 -=  9; }
    else if (f2 & 0x00000000000f0000LL) { f2 <<= 36; c2 -= 10; }
    else if (f2 & 0x000000000000f000LL) { f2 <<= 40; c2 -= 11; }
    else if (f2 & 0x0000000000000f00LL) { f2 <<= 44; c2 -= 12; }
    else if (f2 & 0x00000000000000f0LL) { f2 <<= 48; c2 -= 13; }
    else                                { f2 <<= 52; c2 -= 14; }

    if (c2 & 0x80000000) { /* if characteristic underflow */
      if (mask & EXP_UNF_MASK) {
        c2 += 128;
        intr = EXP_UNF; /* exponent underflow */
      }
      else {
        s2 = c2 = 0; /* true zero */
        f2 = 0LL;    /* true zero */
      }
    }
  }
  else { /* fraction == 0 */
    s2 = c2 = 0; /* true zero */
    f2 = 0LL;    /* true zero */
  }

  /* format result */
  src2 = (llu)s2 << 32 | (llu)c2 << 56 | f2;
  *(llu*)r1 = src2; /* always aligned */

#ifdef DEBUG_FLOAT
  umes("fl_halve_long: s2=%08.8x c2=%08.8x f2=%08.8x_%08.8x\n", s2, c2, (unsigned)(f2>>32), (unsigned)(f2));
#endif

  return (intr);
}

unsigned
fl_conv_float2int(dummy, r2, r1u, r1l, mask)
  unsigned dummy;
  unsigned *r2;
  unsigned r1u, r1l;
  unsigned mask;
  /* return INTR-CODE|CC */
{
  unsigned s;
  unsigned c;
  llu      f;
  int      i;
  unsigned intr = 0, cc = 0;

  s = r1u     & 0x80000000;                     /* s------- -------- -------- -------- */
  c = r1u>>24 & 0x0000007f;                     /* -------- -------- -------- -ccccccc */
  f = (llu)(r1u & 0x00ffffff)<<32 | (llu)(r1l); /* -------- ffffff.. ffffffff ffffffff */

#ifdef DEBUG_FLOAT
  umes("fl_conv_float2int: s=%08.8x c=%08.8x f=%08.8x_%08.8x\n", s, c, (unsigned)(f>>32), (unsigned)(f));
#endif

  if (f == 0LL) /* fraction == 0 */
    i = 0;
  else if (c <= 64) /* too small */
    i = 0;
  else if (c <= 78) { /* shift right f */
    f >>= (78 - c) << 2;
    i = f;
    if (f == 0LL) /* zero */
      cc = 0;
    else if (!s) { /* positive */
      if (f <= 0x7fffffffLL)
	cc = 2;
      else { /* positive overflow */
	i &= 0x7fffffff;
	cc = 3;
      }
    }
    else { /* negative */
      i = -i;
      if (f <= 0x80000000LL)
	cc = 1;
      else { /* negative overflow */
	i |= 0x80000000;
	cc = 3;
      }
    }
  }
  else if (c <= 85) { /* shift left f */
    if (f & (0xffffffffffffffffLL << ((86 - c) << 2))) { /* overflow */
      f <<= (c - 78) << 2;
      i = f;
      if (!s) /* positive */
	i &= 0x7fffffff;
      else { /* negative */
	i = -i;
	i |= 0x80000000;
      }
      cc = 3;
    }
    else {
      f <<= (c - 78) << 2;
      i = f;
      if (!s) { /* positive */
	if (f <= 0x7fffffffLL)
	  cc = 2;
	else { /* positive overflow */
	  i &= 0x7fffffff;
	  cc = 3;
	}
      }
      else { /* negative */
	i = -i;
	if (f <= 0x80000000LL)
	  cc = 1;
	else { /* negative overflow */
	  i |= 0x80000000;
	  cc = 3;
	}
      }
    }
  }
  else { /* overflow */
    if (!s) /* positive */
      i = 0;
    else /* negative */
      i = 0x80000000;
    cc = 3;
  }

  /* write result */
  *r2 = i;
  if (cc == 3 && mask & FIX_OVF_MASK)
    intr = FIX_OVF;

#ifdef DEBUG_FLOAT
  umes("fl_conv_float2int: i=%08.8x cc=%b\n", i, cc);
#endif

  return (intr|cc);
}

unsigned
fl_conv_int2float(dummy, r1, r2)
  unsigned dummy;
  unsigned r1;
  unsigned char *r2;
  /* return CC */
{
  llu      src;
  unsigned s;
  unsigned c;
  llu      f;
  unsigned cc = 0;

  s = r1 & 0x80000000; /* s------- -------- -------- -------- */
  c = 78;              /* -------- -------- -------- -ccccccc */
  f = (llu)(s?-r1:r1); /* -------- ffffff.. ffffffff ffffffff */

#ifdef DEBUG_FLOAT
  umes("fl_conv_int2float: i=%08.8x\n", r1);
  umes("fl_conv_int2float: s=%08.8x c=%08.8x f=%08.8x_%08.8x\n", s, c, (unsigned)(f>>32), (unsigned)(f));
#endif

  if (f) { /* normalize */
    if      (f & 0x00000000f0000000LL) { f <<= 24; c -=  6; }
    else if (f & 0x000000000f000000LL) { f <<= 28; c -=  7; }
    else if (f & 0x0000000000f00000LL) { f <<= 32; c -=  8; }
    else if (f & 0x00000000000f0000LL) { f <<= 36; c -=  9; }
    else if (f & 0x000000000000f000LL) { f <<= 40; c -= 10; }
    else if (f & 0x0000000000000f00LL) { f <<= 44; c -= 11; }
    else if (f & 0x00000000000000f0LL) { f <<= 48; c -= 12; }
    else                               { f <<= 52; c -= 13; }
    if (!s) /* positive */
      cc = 2;
    else /* negative */
      cc = 1;
  }
  else {
    s = 0;
    c = 0;
    f = 0LL;
  }

  /* format result */
  src = (llu)s << 32 | (llu)c << 56 | f;
  *(llu*)r2 = src; /* always aligned */

#ifdef DEBUG_FLOAT
  umes("fl_conv_int2float: s=%08.8x c=%08.8x f=%08.8x_%08.8x cc=%b\n", s, c, (unsigned)(f>>32), (unsigned)(f), cc);
#endif

  return (cc);
}

