
static char RcsId[] =
"$Id: file32.c,v 1.2 1997/04/01 07:52:04 nakashim Exp nakashim $";

/* file.c 1996/3/22 */

/* SPARC Integer Instruction Simulator         */
/*       Copyright (C) 1995 by FUJITSU LIMITED */
/*                 Primary writer: Y.Nakashima */
/*          nakashim@pine.mfd.cs.fujitsu.co.jp */

#include <stdio.h>
#include "sdef.h"

typedef unsigned long   Elf32_Addr;
typedef unsigned short  Elf32_Half;
typedef unsigned long   Elf32_Off;
typedef long            Elf32_Sword;
typedef unsigned long   Elf32_Word;

#define ELF32_FSZ_ADDR	4
#define ELF32_FSZ_HALF	2
#define ELF32_FSZ_OFF	4
#define ELF32_FSZ_SWORD	4
#define ELF32_FSZ_WORD	4

/* ELF header */

#define EI_NIDENT	16

typedef struct {
	unsigned char	e_ident[EI_NIDENT];	/* ident bytes */
	Elf32_Half	e_type;			/* file type */
	Elf32_Half	e_machine;		/* target machine */
	Elf32_Word	e_version;		/* file version */
	Elf32_Addr	e_entry;		/* start address */
	Elf32_Off	e_phoff;		/* phdr file offset */
	Elf32_Off	e_shoff;		/* shdr file offset */
	Elf32_Word	e_flags;		/* file flags */
	Elf32_Half	e_ehsize;		/* sizeof ehdr */
	Elf32_Half	e_phentsize;		/* sizeof phdr */
	Elf32_Half	e_phnum;		/* number phdrs */
	Elf32_Half	e_shentsize;		/* sizeof shdr */
	Elf32_Half	e_shnum;		/* number shdrs */
	Elf32_Half	e_shstrndx;		/* shdr string index */
} Elf32_Ehdr;

/* Program header */

typedef struct {
	Elf32_Word	p_type;		/* entry type */
	Elf32_Off	p_offset;	/* file offset */
	Elf32_Addr	p_vaddr;	/* virtual address */
	Elf32_Addr	p_paddr;	/* physical address */
	Elf32_Word	p_filesz;	/* file size */
	Elf32_Word	p_memsz;	/* memory size */
	Elf32_Word	p_flags;	/* entry flags */
	Elf32_Word	p_align;	/* memory/file alignment */
} Elf32_Phdr;

/* Section header */

typedef struct {
	Elf32_Word	sh_name;	/* section name */
	Elf32_Word	sh_type;	/* SHT_... */
	Elf32_Word	sh_flags;	/* SHF_... */
	Elf32_Addr	sh_addr;	/* virtual address */
	Elf32_Off	sh_offset;	/* file offset */
	Elf32_Word	sh_size;	/* section size */
	Elf32_Word	sh_link;	/* misc info */
	Elf32_Word	sh_info;	/* misc info */
	Elf32_Word	sh_addralign;	/* memory alignment */
	Elf32_Word	sh_entsize;	/* entry size if table */
} Elf32_Shdr;

/* Symbol table */

typedef struct {
	Elf32_Word	st_name;
	Elf32_Addr	st_value;
	Elf32_Word	st_size;
	unsigned char	st_info;	/* bind, type: ELF_32_ST_... */
	unsigned char	st_other;
	Elf32_Half	st_shndx;	/* SHN_... */
} Elf32_Sym;

/* Relocation */

typedef struct {
	Elf32_Addr	r_offset;
	Elf32_Word	r_info;		/* sym, type: ELF32_R_... */
} Elf32_Rel;

typedef struct {
	Elf32_Addr	r_offset;
	Elf32_Word	r_info;		/* sym, type: ELF32_R_... */
	Elf32_Sword	r_addend;
} Elf32_Rela;

/* a.out struct */	

struct aout_head {
	unsigned short  a_info;
	unsigned short  a_magic;
	int 		a_text;
	int		a_data;
	int		a_bss;
	int		a_syms;
	int		a_entry;
	int		a_trsize;
	int		a_drsize;
};
struct aout_head Ahead;

struct rel {
	int		r_off;
	unsigned int    r_sym:   24,
			r_extern: 1,
			        : 2,
			r_type:   5;
	int 		r_addend;
};

struct sym {
	int		s_name;
	unsigned int	s_type:  8,
			      : 24;
	int		s_value;
};

struct rel *Drel;
struct rel *Trel;
struct sym *Sym;

/* elf struct */

Elf32_Ehdr	Ehead;

Elf32_Rel	*E_textrel;
Elf32_Rel	*E_datarel;
Elf32_Sym	*E_sym;

Elf32_Phdr	Phead[3];
Elf32_Shdr	Shead[16];

/* buffer */

char *Text_p, *Data_p, *Textrel_p;
char *Datarel_p, *Sym_p, *Str_p;
char *Shstr_p;
int  len, count;
int  trsz_num, drsz_num;
int  textndx   = 0;
int  datandx   = 0;
int  bssndx    = 0;
int  symtabndx = 0;
int  strtabndx = 0;

int  baseaddr = 0;

/* READ_ELF() */
int read_sparc(s_fp)
  FILE *s_fp;
{
  /* FILEOPEN */
  int dummy[16], i, j;
  
  /* READ ELF_HEADER */
  if (fread(&Ehead, 1, sizeof Ehead, s_fp) !=  (sizeof Ehead)){
    message("Can't read elf_header\n");
    exit(1);
  }

  Ehead.e_type      = swap2(Ehead.e_type);      /****/
  Ehead.e_machine   = swap2(Ehead.e_machine);   /****/
  Ehead.e_version   = swap4(Ehead.e_version);   /****/
  Ehead.e_entry     = swap4(Ehead.e_entry);     /****/
  Ehead.e_phoff     = swap4(Ehead.e_phoff);     /****/
  Ehead.e_shoff     = swap4(Ehead.e_shoff);     /****/
  Ehead.e_flags     = swap4(Ehead.e_flags);     /****/
  Ehead.e_ehsize    = swap2(Ehead.e_ehsize);    /****/
  Ehead.e_phentsize = swap2(Ehead.e_phentsize); /****/
  Ehead.e_phnum     = swap2(Ehead.e_phnum);     /****/
  Ehead.e_shentsize = swap2(Ehead.e_shentsize); /****/
  Ehead.e_shnum     = swap2(Ehead.e_shnum);     /****/
  Ehead.e_shstrndx  = swap2(Ehead.e_shstrndx);  /****/
  
  fseek(s_fp, Ehead.e_shoff, 0);
  for (i=0; i<16 && i<Ehead.e_shnum; i++) {
    fread(&Shead[i], 1, Ehead.e_shentsize, s_fp);
    Shead[i].sh_name      = swap4(Shead[i].sh_name);      /****/
    Shead[i].sh_type      = swap4(Shead[i].sh_type);      /****/
    Shead[i].sh_flags     = swap4(Shead[i].sh_flags);     /****/
    Shead[i].sh_addr      = swap4(Shead[i].sh_addr);      /****/
    Shead[i].sh_offset    = swap4(Shead[i].sh_offset);    /****/
    Shead[i].sh_size      = swap4(Shead[i].sh_size);      /****/
    Shead[i].sh_link      = swap4(Shead[i].sh_link);      /****/
    Shead[i].sh_info      = swap4(Shead[i].sh_info);      /****/
    Shead[i].sh_addralign = swap4(Shead[i].sh_addralign); /****/
    Shead[i].sh_entsize   = swap4(Shead[i].sh_entsize);   /****/
  }
  
  Shstr_p = (char *)malloc(Shead[Ehead.e_shstrndx].sh_size);
  fseek(s_fp, Shead[Ehead.e_shstrndx].sh_offset, 0);
  if (fread(Shstr_p, 1, Shead[Ehead.e_shstrndx].sh_size, s_fp) != Shead[Ehead.e_shstrndx].sh_size) {
    message("Can't read shstr\n");
    exit(1);
  }
  
  for (i=0; i<16 && i<Ehead.e_shnum; i++) {
    if      (!strcmp(Shstr_p+Shead[i].sh_name, ".text"))
      textndx = i;
    else if (!strcmp(Shstr_p+Shead[i].sh_name, ".data"))
      datandx = i;
    else if (!strcmp(Shstr_p+Shead[i].sh_name, ".bss"))
      bssndx = i;
    else if (!strcmp(Shstr_p+Shead[i].sh_name, ".symtab"))
      symtabndx = i;
    else if (!strcmp(Shstr_p+Shead[i].sh_name, ".strtab"))
      strtabndx = i;
  }
  
  message(" ELF:text=%08.8x ", Shead[textndx].sh_size);
  message("data=%08.8x ", Shead[datandx].sh_size);
  message("bss=%08.8x ", Shead[bssndx].sh_size);
  message("symt=%08.8x ", Shead[symtabndx].sh_size);
  message("strt=%08.8x\n", Shead[strtabndx].sh_size-1);
  
  /* READ PROGRAM_HEADER */
  fseek(s_fp, Ehead.e_phoff, 0);
  if (fread(Phead, Ehead.e_phnum, sizeof *Phead, s_fp) !=  (sizeof *Phead)){
    message("Can't read program_header\n");
    exit(1);
  }
  
  Phead[0].p_type   = swap4(Phead[0].p_type);   /****/
  Phead[0].p_offset = swap4(Phead[0].p_offset); /****/
  Phead[0].p_vaddr  = swap4(Phead[0].p_vaddr);  /****/
  Phead[0].p_paddr  = swap4(Phead[0].p_paddr);  /****/
  Phead[0].p_filesz = swap4(Phead[0].p_filesz); /****/
  Phead[0].p_memsz  = swap4(Phead[0].p_memsz);  /****/
  Phead[0].p_flags  = swap4(Phead[0].p_flags);  /****/
  Phead[0].p_align  = swap4(Phead[0].p_align);  /****/
  Phead[1].p_type   = swap4(Phead[1].p_type);   /****/
  Phead[1].p_offset = swap4(Phead[1].p_offset); /****/
  Phead[1].p_vaddr  = swap4(Phead[1].p_vaddr);  /****/
  Phead[1].p_paddr  = swap4(Phead[1].p_paddr);  /****/
  Phead[1].p_filesz = swap4(Phead[1].p_filesz); /****/
  Phead[1].p_memsz  = swap4(Phead[1].p_memsz);  /****/
  Phead[1].p_flags  = swap4(Phead[1].p_flags);  /****/
  Phead[1].p_align  = swap4(Phead[1].p_align);  /****/
  
  if (Shead[datandx].sh_size == 0) {
    /* MALLOC */
    if ((Text_p = (char *)malloc(Phead[0].p_memsz)) == NULL){
      message("Can't malloc elf_text\n");
      exit(1);
    }
    memset(Text_p, 0, Phead[0].p_memsz-1);
  }
  else {
    /* MALLOC */
    if ((Text_p = (char *)malloc(Phead[1].p_vaddr-Phead[0].p_vaddr)) == NULL){
      message("Can't malloc elf_text\n");
      exit(1);
    }
    memset(Text_p, 0, Phead[1].p_vaddr-Phead[0].p_vaddr);
    if ((Data_p = (char *)malloc((Phead[1].p_memsz-1|7)+1)) == NULL){
      message("Can't malloc elf_data\n");
      exit(1);
    }
    memset(Data_p, 0, (Phead[1].p_memsz-1|7)+1);
  }
  
  /* READ (text, data) */
  if (Shead[textndx].sh_size) {
    fseek(s_fp, Phead[0].p_offset, 0);
    if (fread(Text_p, 1, Phead[0].p_filesz, s_fp) != Phead[0].p_filesz){
      message("Can't read elf_text\n");
      exit(1);
    }
  }
  if (Shead[datandx].sh_size) {
    fseek(s_fp, Phead[1].p_offset, 0);
    if (fread(Data_p, 1, Phead[1].p_filesz, s_fp) != Phead[1].p_filesz){
      message("Can't read elf_data\n");
      exit(1);
    }
  }
  
  fclose(s_fp);

/* CONV_DATA() */
  
  /* ELF_HEADER --> A.OUT_HEADER */
  if (Shead[datandx].sh_size == 0) {
    Ahead.a_info   = 0x00c8;
    Ahead.a_magic  = 0x0107;
    Ahead.a_text   = Shead[textndx].sh_size ? Phead[0].p_memsz:0;
    Ahead.a_data   = 0;
    Ahead.a_bss    = 0;
    Ahead.a_syms   = 0;
    Ahead.a_entry  = Phead[0].p_vaddr;
    Ahead.a_trsize = 0;
    Ahead.a_drsize = 0;
  }
  else {
    Ahead.a_info   = 0x00c8;
    Ahead.a_magic  = 0x0107;
    Ahead.a_text   = Phead[1].p_vaddr-Phead[0].p_vaddr;
    Ahead.a_data   = (Phead[1].p_memsz-1|7)+1;
    Ahead.a_bss    = 0;
    Ahead.a_syms   = 0;
    Ahead.a_entry  = Phead[0].p_vaddr;
    Ahead.a_trsize = 0;
    Ahead.a_drsize = 0;
  }
  
  message(" BSD:text=%08.8x ", Ahead.a_text);
  message("data=%08.8x ", Ahead.a_data);
  message("bss=-------- ");
  message("symt=-------- ");
  message("strt=%08.8x\n", Shead[strtabndx].sh_size-1);

/* WRITE_AOUT() */

  message(" SPARC object=0x%08.8x-0x%08.8x\n", Ahead.a_entry, Ahead.a_text + Ahead.a_data - 1);
  if (Ahead.a_text + Ahead.a_data > SPARC_MEMSIZE) {
    message("Program too large\n");
    exit(1);
  }
  if (Ahead.a_entry != 0) {
    message("SPARC start address is 0x%08.8x (must be 0)\n", Ahead.a_entry);
    exit(1);
  }
  memcpy(&mem[           0], Text_p, Ahead.a_text);
  memcpy(&mem[Ahead.a_text], Data_p, Ahead.a_data);

  free(Text_p);
  free(Data_p);
  free(Shstr_p);

  return(Ahead.a_text + Ahead.a_data);
}

