
static char tn6680gw_RcsId[] =
"$Id: tn6680gw.c,v 1.25 1996/12/02 01:18:58 nakashim Exp nakashim $";

/* M-on-RISC Dynamically Translating Simulator */
/*                              TN6680 GATEWAY */
/*       Copyright (c) 1996 by FUJITSU LIMITED */
/*                 Primary writer: Y.Nakashima */
/*               nakashim@mfd.cs.fujitsu.co.jp */

#include <stdio.h>
#include <errno.h>
#include <fcntl.h>
#include <sgtty.h>
#include <time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

/* tn6680gw.c 1996/10/2 by Y.Nakashima */

#define BUFFSIZE (8192)
unsigned char cbuf[BUFFSIZE];  /* com buffer */
unsigned char nbuf[BUFFSIZE];  /* net buffer */

int           cfd1, cfd2;      /* com fds  */
int           nfd1, nfd2;      /* net fds  */
int           dfd1;            /* debug fd */
unsigned char aid;

main(argc, argv)
     int argc;
     char **argv;
{
  int     port;         /* port #   */
  struct  sockaddr_in serv_addr;
  struct  sockaddr_in cli_addr;
  int     cli_len;
  char    logfile[32];  /* name of log_file */
  char    *version();
  struct  sgttyb tArg;
  fd_set  fdset;

#ifdef DEBUG_SEPARATE_TTY
  if (argc != 4) {
    fprintf(stderr, "usage: %s <in_file> <out_file> <port#>\n", argv[0]);
    exit(1);
  }
  if ((cfd1=open(argv[1], O_RDONLY)) == -1) {
    fprintf(stderr, "can't open %s\n", argv[1]);
    exit(1);
  }
  if ((cfd2=open(argv[2], O_WRONLY)) == -1) {
    fprintf(stderr, "can't open %s\n", argv[2]);
    exit(1);
  }
  if (sscanf(argv[3], "%d", &port) != 1) {
    fprintf(stderr, "illegal port: %s\n", argv[3]);
    exit(1);
  }
#else
  if (argc != 3) {
    fprintf(stderr, "usage: %s <in/out_file> <port#>\n", argv[0]);
    exit(1);
  }
  if ((cfd1=cfd2=open(argv[1], O_RDWR)) == -1) {
    fprintf(stderr, "can't open %s\n", argv[1]);
    exit(1);
  }
  if (sscanf(argv[2], "%d", &port) != 1) {
    fprintf(stderr, "illegal port: %s\n", argv[2]);
    exit(1);
  }
#endif

#ifdef DEBUG
  sprintf(logfile ,"/tmp/.tn6680gwlog%u", getpid());
  if ((dfd1 = open(logfile, O_CREAT|O_WRONLY, 0777)) == -1) {
    fprintf(stderr, "can't create \"%s\"\n", logfile);
    exit(1);
  }
#endif

  if ((nfd1 = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
    fprintf(stderr, "can't open stream socket\n");
    exit(1);
  }

  memset((char*)&serv_addr, 0, sizeof(serv_addr));
  serv_addr.sin_family = AF_INET;
  serv_addr.sin_addr.s_addr = htonl(INADDR_ANY);
  serv_addr.sin_port = htons(port);

  if (bind(nfd1, (struct sockaddr*)&serv_addr, sizeof(serv_addr)) < 0) {
    fprintf(stderr, "can't bind address\n");
    exit(1);
  }

  listen(nfd1, 5);

  cli_len = sizeof(cli_addr);
  if ((nfd2 = accept(nfd1, (struct sockaddr*)&cli_addr, &cli_len)) < 0) {
    fprintf(stderr, "can't accept\n");
    exit(1);
  }
  close(nfd1);
 
  ioctl(cfd1, TIOCGETP, &tArg);
  tArg.sg_flags |= RAW;
  tArg.sg_flags &= ~ECHO;
  tArg.sg_ispeed = LINESPEED;
  ioctl(cfd1, TIOCSETP, &tArg);

  ioctl(cfd2, TIOCGETP, &tArg);
  tArg.sg_flags |= RAW;
  tArg.sg_flags &= ~ECHO;
  tArg.sg_ospeed = LINESPEED;
  ioctl(cfd2, TIOCSETP, &tArg);

  /*ioctl(nfd2, TIOCPKT, &tArg);*/

#ifdef RTSCTS
  rtscts(cfd1);
  rtscts(cfd2);
#endif

  /* send IAC DO TTYPE */
  uwrite(nfd2, "\377\375\030", 3);

  /* recv IAC WILL/WONT TTYPE */
  uread(nfd2, nbuf, 3);
  if (nbuf[1] != 0xfb) { /* not WILL */
    fprintf(stderr, "IAC WILL TTYPE->unexpected reply:%02.2x %02.2x %02.2x\n",
	    nbuf[0], nbuf[1], nbuf[2]);
    exit(1);
  }

  /* negotiate TTYPE */
  {  int i; unsigned char prev[256];
     prev[0] = 0;

     while (1) {
       /* send IAC SB TTYPE TELQUAL_SEND IAC SE */
       uwrite(nfd2, "\377\372\030\001\377\360", 6);
     
       /* recv IAC SB TTYPE TELQUAL_IS "TTYPE_LIST" IAC SE */
       uread(nfd2, nbuf, 4);
       if (nbuf[3] != 0) { /* not TELQUAL_IS */
	 fprintf(stderr, "IAC SB TTYPE->unexpected reply:%02.2x %02.2x %02.2x\n",
		 nbuf[0], nbuf[1], nbuf[2]);
	 exit(1);
       }
       for (i=0; i<256; i++) {
	 uread(nfd2, nbuf+i, 1);
	 if (nbuf[i] == 0xf0)
	   break;
       }
       nbuf[i-1] = 0;
       fprintf(stderr, "IAC SB TTYPE->recv \"%s\"\n", nbuf);
       if (!strcmp("FUJITSU-6680-01", nbuf)) {
	 fprintf(stderr, "IAC SB TTYPE->find \"%s\"\n", nbuf);
	 break;
       }
       if (!strcmp(prev, nbuf)) {
	 fprintf(stderr, "IAC SB TTYPE->TTYPE_LIST terminated\n");
	 break;
       }
       strcpy(prev, nbuf);
     }
   }
  
  /* send IAC WILL BINARY */
  uwrite(nfd2, "\377\373\000", 3);

  /* send IAC DO BINARY */
  uwrite(nfd2, "\377\375\000", 3);

  /* send IAC WILL EOR */
  uwrite(nfd2, "\377\373\031", 3);

  /* send IAC DO EOR */
  uwrite(nfd2, "\377\375\031", 3);

  /* recv IAC DO/DONT BINARY */
  uread(nfd2, nbuf, 3);
  if (nbuf[1] != 0xfd) { /* not DO */
    fprintf(stderr, "IAC DO BINARY->unexpected reply:%02.2x %02.2x %02.2x\n",
	    nbuf[0], nbuf[1], nbuf[2]);
    exit(1);
  }

  /* recv IAC WILL/WONT BINARY */
  uread(nfd2, nbuf, 3);
  if (nbuf[1] != 0xfb) { /* not WILL */
    fprintf(stderr, "IAC WILL BINARY->unexpected reply:%02.2x %02.2x %02.2x\n",
	    nbuf[0], nbuf[1], nbuf[2]);
    exit(1);
  }

  /* recv IAC DO/DONT EOR */
  uread(nfd2, nbuf, 3);
  if (nbuf[1] != 0xfd) { /* not DO */
    fprintf(stderr, "IAC DO EOR->unexpected reply:%02.2x %02.2x %02.2x\n",
	    nbuf[0], nbuf[1], nbuf[2]);
    exit(1);
  }

  /* recv IAC WILL/WONT EOR */
  uread(nfd2, nbuf, 3);
  if (nbuf[1] != 0xfb) { /* not WILL */
    fprintf(stderr, "IAC WILL EOR->unexpected reply:%02.2x %02.2x %02.2x\n",
	    nbuf[0], nbuf[1], nbuf[2]);
    exit(1);
  }

  /* send ERASE/WRITE+LOCK */
  uwrite(nfd2, "\365\000\377\357", 4);

  /* send ERASE/WRITE+UNLOCK */
  uwrite(nfd2, "\365\303\377\357", 4);

  FD_ZERO(&fdset);
  FD_SET(cfd1, &fdset);
  FD_SET(nfd2, &fdset);

  fprintf(stderr, "TN6680GW %s connected\n", version());

  while (select(((cfd1>nfd2)?cfd1:nfd2)+1, &fdset, NULL, NULL, NULL) != -1) {
    if (FD_ISSET(nfd2, &fdset)) {
      if (net_read())
        break;
      FD_SET(cfd1, &fdset);
    }
    else if (FD_ISSET(cfd1, &fdset)) {
      if (com_read())
        break;
      FD_SET(nfd2, &fdset);
    }
  }

  fprintf(stderr, "TN6680GW disconnected\n");
  sleep(1);

  close(cfd1);
  close(cfd2);
  close(nfd2);
  exit(0);
}

net_read()
{
  int i, j;
  
  for (i=0;; i++) {
    if (uread(nfd2, nbuf+i, 1) <= 0)
      return(1);

    if (nbuf[i] == 0xff) {
      uread(nfd2, nbuf+i, 1);
      switch (nbuf[i]) {
      case 0xef: /* EOR */
	uwrite(cfd2, "\241\301", 2);
	cbuf[0] = (i>>24) & 0xff;
	cbuf[1] = (i>>16) & 0xff;
	cbuf[2] = (i>> 8) & 0xff;
	cbuf[3] = (i    ) & 0xff;
	uwrite(cfd2, cbuf, 4);
	uwrite(cfd2, nbuf, i);
	aid = nbuf[0];
	return(0);
	
      case 0xf6: /* AYT */
	/* reply "\r\n[Yes]\r\n" */
	uwrite(nfd2, "\015\012\133\131\145\163\135\015\012\377\357", 11);
	return(0);

      case 0xf3: /* Break */
	/* convert to PA1 */
	uwrite(cfd2, "\241\301", 2);
	cbuf[0] = 0;
	cbuf[1] = 0;
	cbuf[2] = 0;
	cbuf[3] = 1;
	uwrite(cfd2, cbuf, 4);
	uwrite(cfd2, "\154", 1);
	return(0);

      default:
	break;
      }
    }
  }
}

com_read()
{
#define SCRSIZE (80*24)
  int         i, j, k;
  int         cmd; /* 1:write, 2:readm, 3:readb, 0:other */
  static int  bufa, kanji;
  char        chain;
 
  if (uread(cfd1, cbuf, 1) <= 0)
    return(1);

  if (uread(cfd1, &chain, 1) <= 0)
    return(1);

  switch (*cbuf) {
  case 0x01: /* write */
    *cbuf = 0xf1;
    cmd = 1;
    break;
  case 0x05: /* erase/write */
    *cbuf = 0xf5;
    cmd = 1;
    break;
  case 0x0d: /* erase/write alternate */
    *cbuf = 0x7e;
    cmd = 1;
    break;
  case 0x06: /* read modified */
    *cbuf = 0xf6;
    cmd = 2;
    break;
  case 0x02: /* read buffer */
    *cbuf = 0xf2;
    cmd = 3;
    break;
  case 0x0f: /* erase all unprotected */
    *cbuf = 0x6f;
    cmd = 0; /* I can't simulate bufa in EAU */
    break;
  case 0x11: /* write structured field */
    *cbuf = 0xf3;
    cmd = 0; /* I can't simulate bufa in WSF */
    break;
  default:
    cmd = 0;
    break;
  }

  uwrite(nfd2, cbuf, 1); /* command */

  uread(cfd1, cbuf, 4); /* length */
  j = cbuf[0]<<24 | cbuf[1]<<16 | cbuf[2]<<8 | cbuf[3];

  if (uread(cfd1, cbuf, j) != j) {
    uwrite(cfd2, "\241\331\000\000\000\001\343", 7); /* TIMEOUT */
    return(0);
  }

  for (i=0; i<j; i++) {
    if (cbuf[i] == 0xff)
      uwrite(nfd2, "\377", 1);
    uwrite(nfd2, cbuf+i, 1);
  }
  uwrite(nfd2, "\377\357", 2); /* IAC EOR */

  switch (cmd) {
  case 1: /* write */
    uwrite(cfd2, "\241\331\000\000\000\001\326", 7); /* OK */
    for (i=0; i<j; i++) {
      switch (cbuf[i]) {
      case 0x11: /* SBA */
	bufa = (cbuf[i+1]<<6&0xfc0)|(cbuf[i+2]&0x3f);
	i += 2;
	break;
      case 0x13: /* IC */
	break;
      case 0x1D: /* SF */
	bufa = (bufa+1) % SCRSIZE;
	i++;
	break;
      case 0x1B: /* SFE/F */
	bufa = (bufa+1) % SCRSIZE;
	i += 2;
	break;
      case 0x0F: /* SFE */
	i += 1 + cbuf[i+1]*2; /* SFE ignored */
	break;
      case 0x3C: /* RA */
	bufa = (cbuf[i+1]<<6&0xfc0)|(cbuf[i+2]&0x3f);
	if (kanji)
	  i += 4;
	else
	  i += 3;
	break;
      case 0x28: /* kanji on */
	kanji = 1;
	break;
      case 0x29: /* kanji off */
	kanji = 0;
	break;
      case 0x2C: /* MF */
      case 0x0E: /* SA */
      case 0x05: /* PT */
      case 0x12: /* EUA */
      case 0x3A: /* LINS */
      case 0x39: /* LDEL */
	break;
      default: /* character */
	bufa = (bufa+1) % SCRSIZE;
	break;
      }
    }
    break;

  case 2: /* readm */
    for (i=0;; i++) {
      uread(nfd2, nbuf+i, 1);
      if (nbuf[i] == 0xff) {
	uread(nfd2, nbuf+i, 1);
	if (nbuf[i] == 0xef) /* EOR */
	  break;
      }
    }
    uwrite(cfd2, "\241\331", 2);
    if (chain == 0) { /* not chained read */
      cbuf[0] = (i>>24) & 0xff;
      cbuf[1] = (i>>16) & 0xff;
      cbuf[2] = (i>> 8) & 0xff;
      cbuf[3] = (i    ) & 0xff;
      uwrite(cfd2, cbuf, 4);
      if (nbuf[0] == 0x60 && aid != 0x60) {
	nbuf[0] = aid;
	aid = 0x60;
      }
      uwrite(cfd2, nbuf, i);
    }
    else { /* chained read */
      for (j=i,i=3; i<j; i++) {
	if (nbuf[i] == 0x11) { /* SBA */
	  if (((nbuf[i+1]<<6&0xfc0)|(nbuf[i+2]&0x3f)) >= bufa)
	    break;
	  i += 2;
	}
      }
      cbuf[0] = ((j-i+3)>>24) & 0xff;
      cbuf[1] = ((j-i+3)>>16) & 0xff;
      cbuf[2] = ((j-i+3)>> 8) & 0xff;
      cbuf[3] = ((j-i+3)    ) & 0xff;
      uwrite(cfd2, cbuf, 4);
      if (nbuf[0] == 0x60 && aid != 0x60) {
	nbuf[0] = aid;
	aid = 0x60;
      }
      uwrite(cfd2, nbuf, 3);
      if (i < j)
	uwrite(cfd2, nbuf+i, (j-i));
    }
    break;

  case 3: /* readb */
    for (i=0;; i++) {
      uread(nfd2, nbuf+i, 1);
      if (nbuf[i] == 0xff) {
	uread(nfd2, nbuf+i, 1);
	if (nbuf[i] == 0xef) /* EOR */
	  break;
      }
    }
    uwrite(cfd2, "\241\331", 2);
    if (chain == 0) { /* not chained read */
      cbuf[0] = (i>>24) & 0xff;
      cbuf[1] = (i>>16) & 0xff;
      cbuf[2] = (i>> 8) & 0xff;
      cbuf[3] = (i    ) & 0xff;
      uwrite(cfd2, cbuf, 4);
      if (nbuf[0] == 0x60 && aid != 0x60) {
	nbuf[0] = aid;
	aid = 0x60;
      }
      uwrite(cfd2, nbuf, i);
    }
    else { /* chained read */
      for (k=0,j=i,i=3; i<j; k++,i++) {
	if (k >= bufa)
	  break;
	switch (nbuf[i]) {
	case 0x1d: /* SF */
	  i++;
	  break;
	case 0x1b: /* SFE */
	  i += 2;
	  break;
	default:
	  break;
	}
      }
      cbuf[0] = ((j-i+3)>>24) & 0xff;
      cbuf[1] = ((j-i+3)>>16) & 0xff;
      cbuf[2] = ((j-i+3)>> 8) & 0xff;
      cbuf[3] = ((j-i+3)    ) & 0xff;
      uwrite(cfd2, cbuf, 4);
      if (nbuf[0] == 0x60 && aid != 0x60) {
	nbuf[0] = aid;
	aid = 0x60;
      }
      uwrite(cfd2, nbuf, 3);
      if (i < j)
	uwrite(cfd2, nbuf+i, (j-i));
    }
    break;

  default:
    uwrite(cfd2, "\241\331\000\000\000\001\325", 7); /* NG */
    break;
  }

  return(0);
}

int
uread(fd, p, len)
     int fd;
     char *p;
     int len;
{
  int n, val;

  clrtime();

  for (n=len; n>0; n-=val) {
  again:
    switch (val = read(fd, p, n)) {
    case 0: /* EOF */
      return(0);
    case -1: /* EAGAIN */
      if (gettime() > 5) /* 5sec */
        return(n);
      else
        goto again;
    default:
      p += val;
      break;
    }
  }
#ifdef DEBUG
  uwrite(dfd1, p-len, len);
#endif
  return(len);
}

int
uwrite(fd, p, len)
     int fd;
     char *p;
     int len;
{
  int n, val;

  for (n=len; n>0; n-=val) {
    while ((val = write(fd, p, n)) <= 0);
    p += val;
  }
  return(len);
}

time_t tloc;
time_t tmsave;

clrtime()
{
  time(&tloc);
  tmsave = tloc;     
}

gettime()
{
  time(&tloc);
  return(tloc-tmsave);
}

char *
version()
{
  char *i;

  for (i=tn6680gw_RcsId; *i && *i!=' '; i++);
  for (                ; *i && *i==' '; i++);
  for (                ; *i && *i!=' '; i++);
  for (                ; *i && *i==' '; i++);
  return(i);
}

