/*
 *  $Id: back_end.c,v 1.17 1995/03/15 17:05:19 forssell Exp $
 *  
 *  Copyright (c) 1994 HAL Computer Systems International, Ltd.
 *  All rights reserved.  Unpublished -- rights reserved under
 *  the Copyright Laws of the United States.  USE OF A COPYRIGHT
 *  NOTICE IS PRECAUTIONARY ONLY AND DOES NOT IMPLY PUBLICATION
 *  OR DISCLOSURE.
 *  
 *  THIS SOFTWARE CONTAINS CONFIDENTIAL INFORMATION AND TRADE
 *  SECRETS OF HAL COMPUTER SYSTEMS INTERNATIONAL, LTD.  USE,
 *  DISCLOSURE, OR REPRODUCTION IS PROHIBITED WITHOUT THE
 *  PRIOR EXPRESS WRITTEN PERMISSION OF HAL COMPUTER SYSTEMS
 *  INTERNATIONAL, LTD.
 *  
 *                          RESTRICTED RIGHTS LEGEND
 *  Use, duplication, or disclosure by the Government is subject
 *  to the restrictions as set forth in subparagraph (c)(l)(ii)
 *  of the Rights in Technical Data and Computer Software clause
 *  at DFARS 252.227-7013.
 *  
 *           HAL COMPUTER SYSTEMS INTERNATIONAL, LTD.
 *                   1315 Dell Avenue
 *                   Campbell, CA  95008
 */
/*==============================================================================*/ 
/* Hal Denali Trace Driven Simulator						*/
/* Copyright (c) 1994 HaL Computer Systems, Inc.                      		*/
/*==============================================================================*/
/* back_end.c									*/
/*										*/
/*==============================================================================*/
/* Revision history								*/
/*										*/
/* Date	     Who	   	Description					*/ 
/* --------  -----------      	-----------					*/
/* 10/30/94  Per Forssell  	Initial defintion				*/
/*										*/
/*==============================================================================*/

#include	<stdio.h>
#include	<stdlib.h>
#include	<math.h>
#include	<string.h>

#include	"globdef.h"
#include	"back_end.h"

/*==============================================================================*/

extern gstats	global_stats ;	/* global statistics				*/

extern ulong	clk_cycle ;	/* current clock cycle				*/

extern int	pstate_reg ;  	/* processor status register			*/
extern int	cwp_reg ;	/* current window pointer			*/

extern ulong	n_inst_no ;	/* instruction sequence counter	in next 
				   clock cycle					*/

extern qslot	*iq_head ;	/* first entry in the instruction que		*/
extern qslot	*iq_tail ;	/* last entry in the instruction que		*/
extern qslot	*lastv_qslot ;	/* last valid qslot				*/
extern qslot	*firstf_qslot ;	/* first free qslot				*/

extern gconf	gl_conf ;	/* back end configuration			*/
extern bconf	be_conf ;	/* back end configuration			*/

/*==============================================================================*/

/* The remap routing maps source and destination registers into a contigous
   register space. The processor state registers occupy the lowest 32 locations,
   the processor priviledged registers occupy the next 32 locations, the floating
   point registers occupy the next 64 locations, the integer global register
   occupy the following 16 register, and the integer register windows occupy the
   rest of the register space (with the NWINDOWS - 1 window occupying the
   registers above the globals, the NWINDOWS - 2 occupying the register above the
   NWINDOWS - 1 registers, etc). */

int
remap (int reg_no, 		/* architected register number			*/
       int reg_type, 		/* register type (fixed, floating, state, priv)	*/
       int imm)			/* immediate field rather than register value	*/
      	
{ 
  int	rm_reg_no ;		/* remapped register number		    	*/
  int	l_reg_no ;		/* fixed point register number when mapped to
				   a contiguous flat fixed point register space	*/
  
  if (!imm) {

    switch (reg_type) {

    case (SR) :			/* processor state registers 			*/
      rm_reg_no = reg_no ;
      break ;

    case (PR) :			/* processor priviledged registers		*/
      rm_reg_no = reg_no + 32 ;
      break ;

    case (FP) :			/* floating point registers			*/
      rm_reg_no = reg_no + 64 ;
      break ;

    case (FX) :			/* fixed point registers			*/
      if (reg_no == 0)
	rm_reg_no = NOT_VALID ;
      else if ((reg_no > 0) && (reg_no <= 7)) {
	if (!GETBIT(pstate_reg, AG))
	  rm_reg_no = reg_no + 128 ;
	else
	  rm_reg_no = reg_no + 136 ;
      }
      else {
	l_reg_no = ((be_conf.no_of_windows - cwp_reg - 1)*24 + reg_no) %
	           ( be_conf.no_of_windows*24 + 8) ;
	rm_reg_no = l_reg_no + 256 ; 
      }
      break ;

    default :
      rm_reg_no = NOT_VALID ;
      break ;

    }

  }
  else
    rm_reg_no = NOT_VALID ;

  return (rm_reg_no) ;

}

/*==============================================================================*/

/* The recode_table34 routine recodes the instruction opcode as defined in table
   34 of the SPARC v9 manual (page 270). */

void
recode_table34 (einst *ei, rinst *ri)
	
{
  switch (ei->opf) {

  case 0x001 : 	/* FMOVs	*/
  case 0x002 :	/* FMOVd	*/
  case 0x005 : 	/* FNEGs	*/
  case 0x006 :	/* FNEGd	*/
  case 0x009 : 	/* FABSs	*/
  case 0x00a :	/* FABSd	*/
    ri->type	= FMOVE ;
    ri->src1	= NOT_VALID ;
    ri->src2	= remap (ei->src2, FP, 0) ;
    ri->src3	= NOT_VALID ;
    ri->dst     = remap (ei->dst,  FP, 0) ;
    ri->spsrc   = NOT_VALID ;
    ri->spdst   = NOT_VALID ;
    ri->sync     = 0 ;
    break ;

  case 0x041 : 	/* FADDs	*/
  case 0x042 :	/* FADDd	*/
  case 0x045 : 	/* FSUBs	*/
  case 0x046 :	/* FSUBd	*/
    ri->type	= FADD ;
    ri->src1	= remap (ei->src1, FP, 0) ;
    ri->src2	= remap (ei->src2, FP, 0) ;
    ri->src3	= NOT_VALID ;
    ri->dst     = remap (ei->dst,  FP, 0) ;
    ri->spsrc   = NOT_VALID ;
    ri->spdst   = FCC ;
    ri->sync    = 0 ;
    break ;

  case 0x049 : 	/* FMULs	*/
  case 0x04a :	/* FMULd	*/
  case 0x069 :	/* FsMULd	*/
    ri->type	= FMULT ;
    ri->src1	= remap (ei->src1, FP, 0) ;
    ri->src2	= remap (ei->src2, FP, 0) ;
    ri->src3	= NOT_VALID ;
    ri->dst     = remap (ei->dst,  FP, 0) ;
    ri->spsrc   = NOT_VALID ;
    ri->spdst   = FCC ;
    ri->sync    = 0 ;
    break ;

  case 0x029 : 	/* FSQRTs	*/
  case 0x02a :	/* FSQRTd	*/
    ri->type	= FSQRT ;
    ri->src1	= remap (ei->src1, FP, 0) ;
    ri->src2	= remap (ei->src2, FP, 0) ;
    ri->src3	= NOT_VALID ;
    ri->dst     = remap (ei->dst,  FP, 0) ;
    ri->spsrc   = NOT_VALID ;
    ri->spdst   = FCC ;
    ri->sync    = 0 ;
    break ;

  case 0x04c : 	/* FDIVs	*/
  case 0x04d :	/* FDIVd	*/
    ri->type	= FDIV ;
    ri->src1	= remap (ei->src1, FP, 0) ;
    ri->src2	= remap (ei->src2, FP, 0) ;
    ri->src3	= NOT_VALID ;
    ri->dst     = remap (ei->dst,  FP, 0) ;
    ri->spsrc   = NOT_VALID ;
    ri->spdst   = FCC ;
    ri->sync    = 0 ;
    break ;

  case 0x081 : 	/* FsTOx	*/
  case 0x082 :	/* FdTOx	*/
  case 0x084 :	/* FxTOs	*/
  case 0x088 :	/* FxTOd	*/
  case 0x0c4 :	/* FiTOs	*/
  case 0x0c6 :  /* FdTOs	*/
  case 0x0c8 :  /* FiTOd	*/
  case 0x0c9 :  /* FsTOd	*/
  case 0x0d1 :	/* FsTOi	*/
  case 0x0d2 :	/* FdTOi	*/
    ri->type	= FCONV ;
    ri->src1	= NOT_VALID ;
    ri->src2	= remap (ei->src2, FP, 0) ;
    ri->src3	= NOT_VALID ;
    ri->dst     = remap (ei->dst,  FP, 0) ;
    ri->spsrc   = NOT_VALID ;
    ri->spdst   = FCC ;
    ri->sync    = 0 ;
    break ;

  /* The following instructions have not yet been implemented; Will be added
     later once the basic mechanism of the simulator works. */
    
  default :
    ri->type	= UNIMP ;
    ri->src1	= NOT_VALID ;
    ri->src2	= NOT_VALID ;
    ri->src3	= NOT_VALID ; 
    ri->dst     = NOT_VALID ;
    ri->spsrc   = NOT_VALID ;
    ri->spdst   = NOT_VALID ;
    ri->sync    = SYNCH_UNIMP ;
    break ;
  
  }

} /* end recode_table34 */

/*==============================================================================*/

/* The recode_table35 routine recodes the instruction opcode as defined in table
   35 of the SPARC v9 manual (page 271). */

void
recode_table35 (einst *ei, rinst *ri)
	
{
  switch (ei->opf) {

  case 0x001 : 	/* FMOVs (fcc0)	*/
  case 0x041 : 	/* FMOVs (fcc1)	*/
  case 0x081 : 	/* FMOVs (fcc2)	*/
  case 0x0c1 : 	/* FMOVs (fcc3)	*/
  case 0x002 : 	/* FMOVd (fcc0)	*/
  case 0x042 : 	/* FMOVd (fcc1)	*/
  case 0x082 : 	/* FMOVd (fcc2)	*/
  case 0x0c2 : 	/* FMOVd (fcc3)	*/
    ri->type	= FMOVE ;
    ri->src1	= remap (ei->dst,  FP, 0) ;
    ri->src2	= remap (ei->src2, FP, 0) ;
    ri->src3	= NOT_VALID ;
    ri->dst     = remap (ei->dst,  FP, 0) ;
    ri->spsrc   = FCC ;
    ri->spdst   = NOT_VALID ;
    ri->sync     = 0 ;
    break ;

  case 0x101 : 	/* FMOVs (icc)	*/
  case 0x181 : 	/* FMOVd (icc)	*/
  case 0x102 : 	/* FMOVs (xcc)	*/
  case 0x182 : 	/* FMOVd (xcc)	*/
    ri->type	= FMOVE ;
    ri->src1	= remap (ei->dst,  FP, 0) ;
    ri->src2	= remap (ei->src2, FP, 0) ;
    ri->src3	= NOT_VALID ;
    ri->dst     = remap (ei->dst,  FP, 0) ;
    ri->spsrc   = CCR ;
    ri->spdst   = NOT_VALID ;
    ri->sync     = 0 ;
    break ;

  case 0x025 : 	/* FMOVRsZ 	*/
  case 0x026 : 	/* FMOVRdZ 	*/
  case 0x045 : 	/* FMOVRsLEZ 	*/
  case 0x046 : 	/* FMOVRdLEZ 	*/
  case 0x065 : 	/* FMOVRsLZ 	*/
  case 0x066 : 	/* FMOVRdLZ 	*/
  case 0x0a5 : 	/* FMOVRsNZ 	*/
  case 0x0a6 : 	/* FMOVRdNZ 	*/
  case 0x0c5 : 	/* FMOVRsGZ 	*/
  case 0x0c6 : 	/* FMOVRdGZ 	*/
  case 0x0e5 : 	/* FMOVRsGEZ 	*/
  case 0x0e6 : 	/* FMOVRdGEZ 	*/
    ri->type	= FMOVE ;
    ri->src1	= remap (ei->dst,  FP, 0) ;
    ri->src2	= remap (ei->src2, FP, 0) ;
    ri->src3	= remap (ei->src1, FP, 0) ;
    ri->dst     = remap (ei->dst,  FP, 0) ;
    ri->spsrc   = NOT_VALID ;
    ri->spdst   = NOT_VALID ;
    ri->sync     = 0 ;
    break ;

  case 0x051 : 	/* FCMPs	*/
  case 0x052 :	/* FCMPd	*/
  case 0x055 : 	/* FCMPs	*/
  case 0x056 :	/* FCMPd	*/
    ri->type	= FADD ;
    ri->src1	= remap (ei->src1, FP, 0) ;
    ri->src2	= remap (ei->src2, FP, 0) ;
    ri->src3	= NOT_VALID ;
    ri->dst     = NOT_VALID ;
    ri->spsrc   = NOT_VALID ;
    ri->spdst   = FCC ;
    ri->sync    = 0 ;
    break ;

  /* The following instructions have not yet been implemented; Will be added
     later once the basic mechanism of the simulator works. */
    
  default :
    ri->type	= UNIMP ;
    ri->src1	= NOT_VALID ;
    ri->src2	= NOT_VALID ;
    ri->src3	= NOT_VALID ; 
    ri->dst     = NOT_VALID ;
    ri->spsrc   = NOT_VALID ;
    ri->spdst   = NOT_VALID ;
    ri->sync    = SYNCH_UNIMP ;
    break ;
  
  }

} /* end recode_table35 */

/*==============================================================================*/

/* The recode_table36 routine recodes the instruction opcode as defined in HaL
   implementation specific spec for IMPDEP2. */

void
recode_table36 (einst *ei, rinst *ri)
	
{
  switch ((ei->opf) & 0xf) {

  case 0x1 : 	/* FMADDs	*/
  case 0x2 :	/* FMADDd	*/
  case 0x5 :	/* FMSUBs	*/
  case 0x6 :	/* FMSUBd	*/
  case 0xd :	/* FNMADDs	*/
  case 0xe :	/* FNMADDd	*/
  case 0x9 :	/* FNMSUBs	*/
  case 0xa :	/* FNMSUBd	*/
    ri->type	= FMADD ;
    ri->src1	= remap (ei->src1, FP, 0) ;
    ri->src2	= remap (ei->src2, FP, 0) ;
    ri->src3	= remap (ei->src3, FP, 0) ; ;
    ri->dst     = remap (ei->dst,  FP, 0) ;
    ri->spsrc   = NOT_VALID ;
    ri->spdst   = FCC ;
    ri->sync    = 0 ;
    break ;

  /* The following instructions have not yet been implemented; Will be added
     later once the basic mechanism of the simulator works. */
    
  default :
    ri->type	= UNIMP ;
    ri->src1	= NOT_VALID ;
    ri->src2	= NOT_VALID ;
    ri->src3	= NOT_VALID ; 
    ri->dst     = NOT_VALID ;
    ri->spsrc   = NOT_VALID ;
    ri->spdst   = NOT_VALID ;
    ri->sync    = SYNCH_UNIMP ;
    break ;
  
  }

} /* end recode_table36 */

/*==============================================================================*/

/* The recode_table32 routine recodes the instruction opcode as defined in table
   32 of the SPARC v9 manual (page 268). */

void
recode_table32 (einst *ei, rinst *ri)
	
{
  switch (ei->op3) {

  case 0x00 :  	/* ADD		*/
  case 0x04 :	/* SUB		*/
    ri->type	= ADDSUB ;
    ri->src1	= remap (ei->src1, FX, 0) ;
    ri->src2	= remap (ei->src2, FX, ei->i) ;
    ri->src3	= NOT_VALID ;
    ri->dst     = remap (ei->dst,  FX, 0) ;
    ri->spsrc   = NOT_VALID ;
    ri->spdst   = NOT_VALID ;
    ri->sync     = 0 ;
    break ;

  case 0x08 :	/* ADDC		*/
  case 0x0c :	/* SUBC		*/
    ri->type	= ADDSUB ;
    ri->src1	= remap (ei->src1, FX, 0) ;
    ri->src2	= remap (ei->src2, FX, ei->i) ;
    ri->src3	= NOT_VALID ;
    ri->dst     = remap (ei->dst,  FX, 0) ;
    ri->spsrc   = CCR ;
    ri->spdst   = NOT_VALID ;
    ri->sync     = 0 ;
    break ;

  case 0x10 :  	/* ADDcc	*/
  case 0x14 :  	/* SUBcc	*/
  case 0x20 :	/* TADDcc	*/
  case 0x21 :	/* TSUBcc	*/
  case 0x22 :	/* TADDccTV	*/
  case 0x23 :	/* TSUBccTV	*/
    ri->type	= ADDSUB ;
    ri->src1	= remap (ei->src1, FX, 0) ;
    ri->src2	= remap (ei->src2, FX, ei->i) ;
    ri->src3	= NOT_VALID ;
    ri->dst     = remap (ei->dst,  FX, 0) ;
    ri->spsrc   = NOT_VALID ;
    ri->spdst   = CCR ;
    ri->sync     = 0 ;
    break ;

  case 0x18 :  	/* ADDCcc	*/
  case 0x1c :  	/* SUBCcc	*/
    ri->type	= ADDSUB ;
    ri->src1	= remap (ei->src1, FX, 0) ;
    ri->src2	= remap (ei->src2, FX, ei->i) ;
    ri->src3	= NOT_VALID ;
    ri->dst     = remap (ei->dst,  FX, 0) ;
    ri->spsrc   = CCR ;
    ri->spdst   = CCR ;
    ri->sync     = 0 ;
    break ;

  case 0x01 :  	/* AND		*/
  case 0x02 :  	/* OR		*/
  case 0x03 :  	/* XOR		*/
  case 0x05 :  	/* ANDN		*/
  case 0x06 :  	/* ORN		*/
  case 0x07 :  	/* XNOR		*/
    ri->type	= LOGIC ;
    ri->src1	= remap (ei->src1, FX, 0) ;
    ri->src2	= remap (ei->src2, FX, ei->i) ;
    ri->src3	= NOT_VALID ;
    ri->dst     = remap (ei->dst,  FX, 0) ;
    ri->spsrc   = NOT_VALID ;
    ri->spdst   = NOT_VALID ;
    ri->sync     = 0 ;
    break ;

  case 0x11 :  	/* ANDcc	*/
  case 0x12 :  	/* ORcc		*/
  case 0x13 :  	/* XORcc	*/
  case 0x15 :  	/* ANDNcc	*/
  case 0x16 :  	/* ORNcc	*/
  case 0x17 :  	/* XNORcc	*/
    ri->type	= LOGIC ;
    ri->src1	= remap (ei->src1, FX, 0) ;
    ri->src2	= remap (ei->src2, FX, ei->i) ;
    ri->src3	= NOT_VALID ;
    ri->dst     = remap (ei->dst,  FX, 0) ;
    ri->spsrc   = NOT_VALID ;
    ri->spdst   = CCR ;
    ri->sync     = 0 ;
    break ;

  case 0x25 :  	/* SLL, SLLX	*/
  case 0x26 :  	/* SRL, SRLX	*/
  case 0x27 :  	/* SRA, SRAX	*/
    ri->type	= SHIFT ;
    ri->src1	= remap (ei->src1, FX, 0) ;
    ri->src2	= remap (ei->src2, FX, ei->i) ;
    ri->src3	= NOT_VALID ;
    ri->dst     = remap (ei->dst,  FX, 0) ;
    ri->spsrc   = NOT_VALID ;
    ri->spdst   = NOT_VALID ;
    ri->sync     = 0 ;
    break ;

  case 0x09 :  	/* MULX		*/
    ri->type	= IMULT ;
    ri->src1	= remap (ei->src1, FX, 0) ;
    ri->src2	= remap (ei->src2, FX, ei->i) ;
    ri->src3	= NOT_VALID ;
    ri->dst     = remap (ei->dst,  FX, 0) ;
    ri->spsrc   = NOT_VALID ;
    ri->spdst   = NOT_VALID ;
    ri->sync    = 0 ;
    break ;

  case 0x0a :  	/* UMUL		*/
  case 0x0b :  	/* SMUL		*/
    ri->type	= IMULT ;
    ri->src1	= remap (ei->src1, FX, 0) ;
    ri->src2	= remap (ei->src2, FX, ei->i) ;
    ri->src3	= NOT_VALID ;
    ri->dst     = remap (ei->dst,  FX, 0) ;
    ri->spsrc   = NOT_VALID ;
    ri->spdst   = NOT_VALID ;
    ri->sync    = WSYNCH_Y ;
    break ;

  case 0x1a :  	/* UMULcc	*/
  case 0x1b :  	/* SMULcc	*/
    ri->type	= IMULT ;
    ri->src1	= remap (ei->src1, FX, 0) ;
    ri->src2	= remap (ei->src2, FX, ei->i) ;
    ri->src3	= NOT_VALID ;
    ri->dst     = remap (ei->dst,  FX, 0) ;
    ri->spsrc   = NOT_VALID ;
    ri->spdst   = CCR ;
    ri->sync    = WSYNCH_Y ;
    break ;

  case 0x24 :  	/* MULScc	*/
    ri->type	= IMULT ;
    ri->src1	= remap (ei->src1, FX, 0) ;
    ri->src2	= remap (ei->src2, FX, ei->i) ;
    ri->src3	= NOT_VALID ;
    ri->dst     = remap (ei->dst,  FX, 0) ;
    ri->spsrc   = NOT_VALID ;
    ri->spdst   = CCR ;
    ri->sync    = WSYNCH_Y ;
    break ;

  case 0x0d :  	/* UDIVX	*/
  case 0x2d :  	/* SDIVX	*/
    ri->type	= IDIV ;
    ri->src1	= remap (ei->src1, FX, 0) ;
    ri->src2	= remap (ei->src2, FX, ei->i) ;
    ri->src3	= NOT_VALID ;
    ri->dst     = remap (ei->dst,  FX, 0) ;
    ri->spsrc   = NOT_VALID ;
    ri->spdst   = NOT_VALID ;
    ri->sync    = 0 ;
    break ;

  case 0x0e :  	/* UDIV		*/
  case 0x0f :  	/* SDIV		*/
    ri->type	= IDIV ;
    ri->src1	= remap (ei->src1, FX, 0) ;
    ri->src2	= remap (ei->src2, FX, ei->i) ;
    ri->src3	= NOT_VALID ;
    ri->dst     = remap (ei->dst,  FX, 0) ;
    ri->spsrc   = NOT_VALID ;
    ri->spdst   = NOT_VALID ;
    ri->sync    = WSYNCH_Y ;
    break ;

  case 0x1e :  	/* UDIVcc	*/
  case 0x1f :  	/* SDIVcc	*/
    ri->type	= IDIV ;
    ri->src1	= remap (ei->src1, FX, 0) ;
    ri->src2	= remap (ei->src2, FX, ei->i) ;
    ri->src3	= NOT_VALID ;
    ri->dst     = remap (ei->dst,  FX, 0) ;
    ri->spsrc   = NOT_VALID ;
    ri->spdst   = CCR ;
    ri->sync    = WSYNCH_Y ;
    break ;

  case 0x30 :  	/* WR status registers	*/
    ri->type	= SYS ;
    ri->src1	= remap (ei->src1, FX, 0) ;
    ri->src2	= remap (ei->src2, FX, ei->i) ;
    ri->src3	= NOT_VALID ;
    ri->dst     = remap (ei->dst,  SR, 0) ;
    ri->spsrc   = NOT_VALID ;
    switch (ei->dst) {
    case 0 :	/* WRY			*/
      ri->spdst   = NOT_VALID ;
      ri->sync    = WSYNCH_Y ;
      break ;

    case 2 :	/* WRCCR		*/
      ri->spdst   = CCR ;
      ri->sync     = 0 ;
      break ;
    case 3 :	/* WRASI		*/
      ri->spdst   = NOT_VALID ;
      ri->sync    = 0 ;
      break ;
    case 6 :	/* WRFPSR		*/
      ri->spdst   = NOT_VALID ;
      ri->sync    = WSYNCH_FPRS ;
      break ;
    default :	/* WRASR		*/
      ri->spdst   = NOT_VALID ;
      ri->sync    = 0 ;
      break ;
    }

  case 0x28 :  	/* RD status registers	*/
    ri->type	= SYS ;
    ri->src1	= remap (ei->src1, SR, 0) ;
    ri->src2	= NOT_VALID ;
    ri->src3	= NOT_VALID ;
    ri->dst     = remap (ei->dst,  FX, 0) ;
    ri->spdst   = NOT_VALID ;
    switch (ei->src1) {
    case 0 :	/* RDY			*/
      ri->spsrc   = NOT_VALID ;
      ri->sync    = RSYNCH_Y ;
      break ;
    case 4 :	/* RDTICK		*/
      ri->spsrc   = NOT_VALID ;
      ri->sync    = RSYNCH_TICK ;
      break ;
    case 6 :	/* RDFPRS		*/
      ri->spsrc   = NOT_VALID ;
      ri->sync    = RSYNCH_FPRS ;
      break ;
    case 3 :	/* RDASI		*/
    case 5 :	/* RDPC			*/
      ri->spsrc   = NOT_VALID ;
      ri->sync    = 0 ;
      break ;
    case 2 :	/* RDCCR		*/
      ri->spsrc   = CCR ;
      ri->sync    = 0 ;
      break ;
    default :	/* RDASR		*/
      ri->spsrc   = NOT_VALID ;
      ri->sync    = 0 ;
      break ;
    }

  case 0x32 :  	/* WR priviledged registers	*/
    if (ei->dst != 15) {
      ri->type	= SYS ;
      ri->src1	= remap (ei->src1, FX, 0) ;
      ri->src2	= remap (ei->src2, FX, ei->i) ;
      ri->src3	= NOT_VALID ;
      ri->dst   = remap (ei->dst,  PR, 0) ;
      ri->spsrc = NOT_VALID ;
      ri->spdst = NOT_VALID ;
      switch (ei->dst) {
      case 0  : ri->sync  = WSYNCH_TPC ; break ;	/* WRPR TPC		*/
      case 1  : ri->sync  = WSYNCH_TNPC ; break ;	/* WRPR TNPC		*/
      case 2  : ri->sync  = WSYNCH_TSTATE ; break ;	/* WRPR TSTATE		*/
      case 3  : ri->sync  = WSYNCH_TT ; break ;		/* WRPR TT		*/
      case 4  : ri->sync  = WSYNCH_TICK ; break ;	/* WRPR TICK		*/
      case 5  : ri->sync  = WSYNCH_TBA ; break ;	/* WRPR TBA		*/
      case 6  : ri->sync  = WSYNCH_PSTATE ; break ;	/* WRPR PSTATE		*/
      case 7  : ri->sync  = WSYNCH_TL ; break ;		/* WRPR TL		*/
      case 8  : ri->sync  = WSYNCH_PIL ; break ;	/* WRPR PIL		*/
      case 9  : ri->sync  = WSYNCH_CWP ; break ;	/* WRPR CWP		*/
      case 10 : ri->sync  = WSYNCH_CANSAVE ; break ;	/* WRPR CANSAVE		*/
      case 11 : ri->sync  = WSYNCH_CANRESTORE ; break ;	/* WRPR CANRESTORE	*/
      case 12 : ri->sync  = WSYNCH_CLEANWIN ; break ;	/* WRPR CLEANWINDOW	*/
      case 13 : ri->sync  = WSYNCH_OTHERWIN ; break ;	/* WRPR OTHERWIN	*/
      case 14 : ri->sync  = WSYNCH_WSTATE ; break ;	/* WRPR WSTATE		*/
      default : ri->sync  = 0 ; 
      }
    }

    /* MEMBAR and STBAR have not yet been implemented; Will be added later once
       the basic mechanism of the simulator works. */

    else {
      ri->type	= UNIMP ;
      ri->src1	= NOT_VALID ;
      ri->src2	= NOT_VALID ;
      ri->src3	= NOT_VALID ; 
      ri->dst   = NOT_VALID ;
      ri->spsrc = NOT_VALID ;
      ri->spdst = NOT_VALID ;
      ri->sync  = SYNCH_UNIMP ;
      break ;
    }

  case 0x2a :  	/* RD priviledged registers	*/
    ri->type	= SYS ;
    ri->src1	= remap (ei->src1, PR, 0) ;
    ri->src2	= NOT_VALID ;
    ri->src3	= NOT_VALID ;
    ri->dst     = remap (ei->dst,  FX, 0) ;
    ri->spsrc   = NOT_VALID ;
    ri->spdst   = NOT_VALID ;
    switch (ei->src1) {
    case 0  : ri->sync  = RSYNCH_TPC ; break ;		/* RDPR TPC		*/
    case 1  : ri->sync  = RSYNCH_TNPC ; break ;		/* RDPR TNPC		*/
    case 2  : ri->sync  = RSYNCH_TSTATE ; break ;	/* RDPR TSTATE		*/
    case 3  : ri->sync  = RSYNCH_TT ; break ;		/* RDPR TT		*/
    case 4  : ri->sync  = RSYNCH_TICK ; break ;		/* RDPR TICK		*/
    case 5  : ri->sync  = RSYNCH_TBA ; break ;		/* RDPR TBA		*/
    case 6  : ri->sync  = RSYNCH_PSTATE ; break ;	/* RDPR PSTATE		*/
    case 7  : ri->sync  = RSYNCH_TL ; break ;		/* RDPR TL		*/
    case 8  : ri->sync  = RSYNCH_PIL ; break ;		/* RDPR PIL		*/
    case 9  : ri->sync  = RSYNCH_CWP ; break ;		/* RDPR CWP		*/
    case 10 : ri->sync  = RSYNCH_CANSAVE ; break ;	/* RDPR CANSAVE		*/
    case 11 : ri->sync  = RSYNCH_CANRESTORE ; break ;	/* RDPR CANRESTORE	*/
    case 12 : ri->sync  = RSYNCH_CLEANWIN ; break ;	/* RDPR CLEANWINDOW	*/
    case 13 : ri->sync  = RSYNCH_OTHERWIN ; break ;	/* RDPR OTHERWIN	*/
    case 14 : ri->sync  = RSYNCH_WSTATE ; break ;	/* RDPR WSTATE		*/
    default : ri->sync  = 0 ; 
    }

  case 0x2f :  	/* MOVr 	*/
    ri->type	= IMOVE ;
    ri->src1	= remap (ei->src1, FX, 0) ;
    ri->src2	= remap (ei->src2, FX, ei->i) ;
    ri->src3	= NOT_VALID ;
    ri->dst     = remap (ei->dst,  FX, 0) ;
    ri->spsrc   = NOT_VALID ;
    ri->spdst   = NOT_VALID ;
    ri->sync    = 0 ;
    break ;

  case 0x2c :  	/* MOVcc 	*/
    ri->type	= IMOVE ;
    ri->src1	= NOT_VALID ;
    ri->src2	= remap (ei->src1, FX, ei->i) ;
    ri->src3	= NOT_VALID ;
    ri->dst     = remap (ei->dst,  FX, 0) ;
    if (ei->cc2 == 0)
      ri->spsrc = FCC ;
    else
      ri->spsrc = CCR ;
    ri->spdst   = NOT_VALID ;
    ri->sync    = 0 ;
    break ;

  case 0x3c :  	/* SAVE		*/
    ri->type	= ADDSUB ;
    ri->src1	= remap (ei->src1, FX, 0) ;
    ri->src2	= remap (ei->src2, FX, ei->i) ;
    ri->src3	= NOT_VALID ;
    if (!ei->sqsh)
      cwp_reg     = (cwp_reg + 1) % be_conf.no_of_windows ;
    ri->dst     = remap (ei->dst, FX, 0) ;
    ri->spsrc   = NOT_VALID ;
    ri->spdst   = NOT_VALID ;
    ri->sync    = 0 ;
    break ;

  case 0x3d :  	/* RESTORE	*/
    ri->type	= ADDSUB ;
    ri->src1	= remap (ei->src1, FX, 0) ;
    ri->src2	= remap (ei->src2, FX, ei->i) ;
    ri->src3	= NOT_VALID ;
    if (!ei->sqsh)
      cwp_reg     = (cwp_reg - 1) % be_conf.no_of_windows ; 
    ri->dst     = remap (ei->dst, FX, 0) ;
    ri->spsrc   = NOT_VALID ;
    ri->spdst   = NOT_VALID ;
    ri->sync    = 0 ;
    break ;

  case 0x38 :  	/* JMPL	 	   */
    ri->type	= JMPL ;
    ri->src1	= remap (ei->src1, FX, 0) ;
    ri->src2	= remap (ei->src2, FX, ei->i) ;
    ri->src3	= NOT_VALID ;
    ri->dst     = remap (ei->dst,  FX, 0) ;
    ri->spsrc   = NOT_VALID ;
    ri->spdst   = NOT_VALID ;
    ri->sync    = 0 ;
    ri->pred	= ei->pred ;
    ri->annul_next = 0 ;
    break ;

  case 0x39 :  	/* RETURN 	   */
    ri->type	= RETURN ;
    ri->src1	= remap (ei->src1, FX, 0) ;
    ri->src2	= remap (ei->src2, FX, ei->i) ;
    ri->src3	= NOT_VALID ;
    if (!ei->sqsh)
      cwp_reg     = (cwp_reg - 1) % be_conf.no_of_windows ; 
    ri->dst     = NOT_VALID ;
    ri->spsrc   = NOT_VALID ;
    ri->spdst   = NOT_VALID ;
    ri->sync    = 0 ;
    ri->pred	= ei->pred ;
    ri->annul_next = 0 ;
    break ;

  case 0x31 :  	/* SAVED, RESTORED */
    ri->type	= SYS ;
    ri->src1	= NOT_VALID ;
    ri->src2	= NOT_VALID ;
    ri->src3	= NOT_VALID ;
    ri->dst     = NOT_VALID ;
    ri->spsrc   = NOT_VALID ;
    ri->spdst   = NOT_VALID ;
    ri->sync    = 0 ;
    break ;

  case 0x3e :  	/* DONE, RETRY 	   */
    ri->type	= SYS ;
    ri->src1	= NOT_VALID ;
    ri->src2	= NOT_VALID ;
    ri->src3	= NOT_VALID ;
    ri->dst     = NOT_VALID ;
    ri->spsrc   = NOT_VALID ;
    ri->spdst   = NOT_VALID ;
    ri->sync    = 0 ;
    ri->annul_next = 0 ;
    break ;

  case 0x34 :  	/* FPop1	   */
    recode_table34 (ei, ri) ;
    break ;

  case 0x35 :  	/* FPop2	   */
    recode_table35 (ei, ri) ;
    break ;

  case 0x36 :  	/* IMPDEP1	   */
  case 0x37 :  	/* IMPDEP2	   */
    recode_table36 (ei, ri) ;
    break ;

  /* The following instructions have not yet been implemented; Will be added
     later once the basic mechanism of the simulator works. */

  case 0x3a :  	/* Tcc	 	   */
  case 0x2e :  	/* POPC		   */
  case 0x2b :  	/* FLUSHW	   */
  case 0x3b :  	/* FLUSH	   */
    ri->type	= UNIMP ;
    ri->src1	= NOT_VALID ;
    ri->src2	= NOT_VALID ;
    ri->src3	= NOT_VALID ; 
    ri->dst     = NOT_VALID ;
    ri->spsrc   = NOT_VALID ;
    ri->spdst   = NOT_VALID ;
    ri->sync    = SYNCH_UNIMP ;
    break ;
    
  default :
    ri->type	= UNIMP ;
    ri->src1	= NOT_VALID ;
    ri->src2	= NOT_VALID ;
    ri->src3	= NOT_VALID ; 
    ri->dst     = NOT_VALID ;
    ri->spsrc   = NOT_VALID ;
    ri->spdst   = NOT_VALID ;
    ri->sync    = SYNCH_UNIMP ;
    break ;
  
  }

} /* end recode_table32 */

/*==============================================================================*/

/* The recode_table33 routine recodes the instruction opcode as defined in table
   33 of the SPARC v9 manual (page 269). */

void
recode_table33 (einst *ei, rinst *ri)
	
{
  switch (ei->op3) {

  case 0x0b :  	/* LDX	 	   */
  case 0x00 :  	/* LDUW	 	   */
  case 0x01 :  	/* LDUB	 	   */
  case 0x02 :  	/* LDUH	 	   */
  case 0x08 :  	/* LDSW	 	   */
  case 0x09 :  	/* LDSB	 	   */
  case 0x0a :  	/* LDSH	 	   */
  case 0x0d :  	/* LDSTUB 	   */
    ri->type	= ILOAD ;
    ri->src1	= remap (ei->src1, FX, 0) ;
    ri->src2	= remap (ei->src2, FX, ei->i) ;
    ri->src3	= NOT_VALID ;
    ri->dst     = remap (ei->dst,  FX, 0) ;
    ri->spsrc   = NOT_VALID ;
    ri->spdst   = NOT_VALID ;
    ri->sync     = 0 ;
    break ;

  case 0x1b :  	/* LDXA	 	   */
  case 0x10 :  	/* LDUWA 	   */
  case 0x11 :  	/* LDUBA 	   */
  case 0x12 :  	/* LDUHA 	   */
  case 0x18 :  	/* LDSWA 	   */
  case 0x19 :  	/* LDSBA 	   */
  case 0x1a :  	/* LDSHA 	   */
  case 0x1d :  	/* LDSTUBA 	   */
    ri->type	= ILOAD ;
    ri->src1	= remap (ei->src1, FX, 0) ;
    ri->src2	= remap (ei->src2, FX, ei->i) ;
    ri->src3	= NOT_VALID ;
    ri->dst     = remap (ei->dst,  FX, 0) ;
    ri->spsrc   = remap (ASI,      SR, (~(ei->i) & 0x1) ) ;
    ri->spdst   = NOT_VALID ;
    ri->sync    = 0 ;
    break ;

  case 0x0e :  	/* STX	 	   */
  case 0x04 :  	/* STW	 	   */
  case 0x05 :  	/* STB	 	   */
  case 0x06 :  	/* STH	 	   */
    ri->type	= ISTORE ;
    ri->src1	= remap (ei->src1, FX, 0) ;
    ri->src2	= remap (ei->src2, FX, ei->i) ;
    ri->src3	= remap (ei->dst,  FX, 0) ;
    ri->dst     = NOT_VALID ;
    ri->spsrc   = NOT_VALID ;
    ri->spdst   = NOT_VALID ;
    ri->sync     = 0 ;
    break ;

  case 0x1e :  	/* STXA	 	   */
  case 0x14 :  	/* STWA	 	   */
  case 0x15 :  	/* STBA	 	   */
  case 0x16 :  	/* STHA	 	   */
    ri->type	= ISTORE ;
    ri->src1	= remap (ei->src1, FX, 0) ;
    ri->src2	= remap (ei->src2, FX, ei->i) ;
    ri->src3	= remap (ei->dst,  FX, 0) ;
    ri->dst     = NOT_VALID ;
    ri->spsrc   = remap (ASI,      SR, (~(ei->i) & 0x1) ) ;
    ri->spdst   = NOT_VALID ;
    ri->sync    = 0 ;
    break ;

  case 0x20 :  	/* LDF	 	   */
  case 0x23 :  	/* LDDF	 	   */
    ri->type	= FLOAD ;
    ri->src1	= remap (ei->src1, FX, 0) ;
    ri->src2	= remap (ei->src2, FX, ei->i) ;
    ri->src3	= NOT_VALID ;
    ri->dst     = remap (ei->dst,  FP, 0) ;
    ri->spsrc   = NOT_VALID ;
    ri->spdst   = NOT_VALID ;
    ri->sync     = 0 ;
    break ;

  case 0x30 :  	/* LDFA	 	   */
  case 0x33 :  	/* LDDFA 	   */
    ri->type	= FLOAD ;
    ri->src1	= remap (ei->src1, FX, 0) ;
    ri->src2	= remap (ei->src2, FX, ei->i) ;
    ri->src3	= NOT_VALID ;
    ri->dst     = remap (ei->dst,  FP, 0) ;
    ri->spsrc   = remap (ASI,      SR, (~(ei->i) & 0x1) ) ;
    ri->spdst   = NOT_VALID ;
    ri->sync    = 0 ;
    break ;

  case 0x24 :  	/* STF	 	   */
  case 0x27 :  	/* STDF	 	   */
    ri->type	= FSTORE ;
    ri->src1	= remap (ei->src1, FX, 0) ;
    ri->src2	= remap (ei->src2, FX, ei->i) ;
    ri->src3	= remap (ei->dst,  FP, 0) ;
    ri->dst     = NOT_VALID ;
    ri->spsrc   = NOT_VALID ;
    ri->spdst   = NOT_VALID ;
    ri->sync     = 0 ;
    break ;

  case 0x34 :  	/* STFA	 	   */
  case 0x37 :  	/* STDFA 	   */
    ri->type	= FSTORE ;
    ri->src1	= remap (ei->src1, FX, 0) ;
    ri->src2	= remap (ei->src2, FX, ei->i) ;
    ri->src3	= remap (ei->dst,  FP, 0) ;
    ri->dst     = NOT_VALID ;
    ri->spsrc   = remap (ASI,      SR, (~(ei->i) & 0x1) ) ;
    ri->spdst   = NOT_VALID ;
    ri->sync    = 0 ;
    break ;

  case 0x21 :  	/* LDFSR, LDFXSR   */
    ri->type	= FLOAD ;
    ri->src1	= remap (ei->src1, FX, 0) ;
    ri->src2	= remap (ei->src2, FX, ei->i) ;
    ri->src3	= NOT_VALID ;
    ri->dst     = remap (FSR,  SR, 0) ;
    ri->spsrc   = NOT_VALID ;
    ri->spdst   = NOT_VALID ;
    ri->sync    = RSYNCH_FSR ;
    break ;

  case 0x25 :  	/* STFSR, STXFSR   */
    ri->type	= FSTORE ;
    ri->src1	= remap (ei->src1, FX, 0) ;
    ri->src2	= remap (ei->src2, FX, ei->i) ;
    ri->src3	= remap (FSR,  SR, 0) ;
    ri->dst     = NOT_VALID ;
    ri->spsrc   = NOT_VALID ;
    ri->spdst   = NOT_VALID ;
    ri->sync    = WSYNCH_FSR ;
    break ;


  /* The following instructions have not yet been implemented; Will be added
     later once the basic mechanism of the simulator works. */

  case 0x03 :  	/* LDD	 	   */
  case 0x07 :  	/* STD	 	   */
  case 0x0f :  	/* SWAP	 	   */

  case 0x13 :  	/* LDDA	 	   */
  case 0x17 :  	/* STDA	 	   */
  case 0x1f :  	/* SWAPA 	   */

  case 0x22 :  	/* LDQF	 	   */
  case 0x26 :  	/* STQF	 	   */
  case 0x2d :  	/* PREFETCH 	   */

  case 0x32 :  	/* LDQFA 	   */
  case 0x36 :  	/* STQFA 	   */
  case 0x3c :  	/* CASA 	   */
  case 0x3d :  	/* PREFETCHA 	   */
  case 0x3e :  	/* CASXA 	   */
    ri->type	= UNIMP ;
    ri->src1	= NOT_VALID ;
    ri->src2	= NOT_VALID ;
    ri->src3	= NOT_VALID ; 
    ri->dst     = NOT_VALID ;
    ri->spsrc   = NOT_VALID ;
    ri->spdst   = NOT_VALID ;
    ri->sync    = SYNCH_UNIMP ;
    break ;
    
  default :
    ri->type	= UNIMP ;
    ri->src1	= NOT_VALID ;
    ri->src2	= NOT_VALID ;
    ri->src3	= NOT_VALID ; 
    ri->dst     = NOT_VALID ;
    ri->spsrc   = NOT_VALID ;
    ri->spdst   = NOT_VALID ;
    ri->sync    = SYNCH_UNIMP ;
    break ;
  
  }

} /* end recode_table33 */

/*==============================================================================*/

/* The recode routine recodes the instruction opcode as defined in tables 30 and
   31 of the SPARC v9 manual (page 267). */

void
recode (einst *ei, rinst *ri)
	
{

  switch (ei->op) {

  case 0x0 :	/* Branches and SETHI	*/

    switch (ei->op2) {

    case 0x1 :  /* BPcc 	   */
    case 0x2 :  /* BPicc 	   */
      ri->type	= BRANCH ;
      ri->src1	= NOT_VALID ;
      ri->src2	= NOT_VALID ;
      ri->src3	= NOT_VALID ;
      ri->dst   = NOT_VALID ;
      if (GETFLD(ei->bcond, 0, 3) == 0) 	/* branch allways or never	*/
	ri->spsrc = NOT_VALID ;
      else
	ri->spsrc = CCR ;
      ri->spdst = NOT_VALID ;
      ri->sync  = 0 ;
      ri->pred	= ei->pred ;  
      if (!ei->taken && ei->annul)
	ri->annul_next = 1 ;
      else
	ri->annul_next = 0 ;
      break ;

    case 0x3 :  /* BPr	 	   */
      ri->type	= BRANCH ;
      ri->src1	= remap (ei->src1,  FX, 0) ;
      ri->src2	= NOT_VALID ;
      ri->src3	= NOT_VALID ;
      ri->dst   = NOT_VALID ;
      ri->spsrc = NOT_VALID ;
      ri->spdst = NOT_VALID ;
      ri->sync  = 0 ;
      ri->pred	= ei->pred ;
      if (ei->annul) 
	ri->annul_next = 1 ;
      else
	ri->annul_next = 0 ;
      break ;

    case 0x4 :  /* SETHI, NOP	   */
      ri->type	= LOGIC ;
      ri->src1	= NOT_VALID ;
      ri->src2	= NOT_VALID ;
      ri->src3	= NOT_VALID ;
      ri->dst   = remap (ei->dst,  FX, 0) ;
      ri->spsrc = NOT_VALID ;
      ri->spdst = NOT_VALID ;
      ri->sync  = 0 ;
      break ;

    case 0x5 :  /* FBPfcc	   */
    case 0x6 :  /* FBfcc	   */
      ri->type	= BRANCH ;
      ri->src1	= NOT_VALID ;
      ri->src2	= NOT_VALID ;
      ri->src3	= NOT_VALID ;
      ri->dst   = NOT_VALID ;
      if (GETFLD(ei->bcond, 0, 3) == 0) 	/* branch allways or never	*/
	ri->spsrc = NOT_VALID ;
      else
	ri->spsrc = FCC ;
      ri->spdst = NOT_VALID ;
      ri->sync  = 0 ;
      ri->pred	= ei->pred ;
      if (!ei->taken && ei->annul)
	ri->annul_next = 1 ;
      else
	ri->annul_next = 0 ;
      break ;
      
  /* The following instructions have not yet been implemented; Will be added
     later once the basic mechanism of the simulator works. */

    case 0x0 :  /* ILLTRAP 	   */
    default  :
      ri->type	  = UNIMP ;
      ri->src1	  = NOT_VALID ;
      ri->src2	  = NOT_VALID ;
      ri->src3	  = NOT_VALID ; 
      ri->dst     = NOT_VALID ;
      ri->spsrc   = NOT_VALID ;
      ri->spdst   = NOT_VALID ;
      ri->sync    = SYNCH_UNIMP ;
      break ;
    }
    break ;

  case 0x1 :	/* CALL		*/
    ri->type	  = CALL ;
    ri->src1	  = NOT_VALID ;
    ri->src2	  = NOT_VALID ;
    ri->src3	  = NOT_VALID ; 
    ri->dst       = remap (15,  FX,  0) ;
    ri->spsrc     = NOT_VALID ;
    ri->spdst     = NOT_VALID ;
    ri->sync      =  0 ;
    ri->annul_next = 0 ;
    break ;
    
  case 0x2 :	/* Arithmetic and misc.	*/
    recode_table32 (ei, ri) ;
    break ;

  case 0x3 :	/* Loads and stores	*/
    recode_table33 (ei, ri) ;
    break ;

 default :
   break ;
 
  }

} /* end recode */

/*==============================================================================*/

/* The fetch pipe generates instruction fetch requests based on the current state
   of the instruction que. The fetch requests are serviced by the branch
   processor. The branch processor determines the fetch completion cycles for the
   fetched instructions, and determines whether branches are predicted correcly.

   The fetch pipe does not handle separate branch direction and branch target
   misspredicts. An incorrect brediction will stall fetch until the condition
   codes have been resolved and the branch target calculated.  Future refinements
   to the simulator may differentiate between the two miss predict cases.

   The handling of fetch requests is done based on allready fetched instructions.
   The cases are handled as follows:

   (1) Sequential fetch: the fetch pipe will signal the branch unit for the next
       set of n instructions if there are no misspredicted branches in the
       instruction que, that have not yet completed execution, and if the current
       clock cycle equals the 'fetch' cycle of the last valid instruction in the
       instruction que.

   (2) Correctly predicted taken or not taken branch: the fetch unit will treat
       correctly predicted taken, or not taken branches as sequential
       instructions from a fetch point of view (ie will do as in (1)). The branch
       processor will determine any delays caused by taken branches. 

   (3) Incorrectly predicted taken or not taken branch: the fetch unit will wait
       until the misspredicted branch instruction has completed execution. It
       will then signal the branch processor for the next set of n instructions.

   The protocol between the fetch pipe and the branch processor is such that the
   fetch pipe indicates each cycle the maximum number of instructions that it can
   receive during the following cycle. The branch processor attempts to deliver
   the requested number of instructions (or less) during the following cycle (or
   cycles). The branch unit indicates the time when the instructions will be
   available by the 'fetch_cycle' field in the instruction fetch packets.

   The branch processor will handle misspredicted branches by waiting for the
   fetch pipe to request the next set of instructions following a misspredicted
   branch. The branch processor will then deliver the requested instructions with
   the 'fetch_cycle' field reflecting the delay it takes to fetch from the new
   target address.  */


void
fetch_pipe (ireq *fetch_r ,	/* instruction fetch request			*/
	    ipkt *fetch_p )	/* instruction fetch packet array		*/
{
	
  qslot	  *qs ;		/* que slot pointer					*/
  qentry  *qe ;		/* que entry pointer					*/
  ipkt    *fp ;		/* instruction fetch packet pointer			*/

  rinst	   ri ;		/* recoded instruction					*/

  int	   inst_cnt ;	/* number of instructions in fetch packet		*/
  int	   fetch_cnt ;	/* number of instructions that will be requested	*/
  int	   fpkt_cnt ;	/* fetch packet array counter				*/

  int	   iq_length ;	/* length of instruction que				*/

  /* Determine the number of instructions to be requested from the branch unit.
     The number of instructions is determined by counting the number of free
     qslots in the instruction que, then subtracting the number of instructions
     in the current instruction packet from the number of free qslots. If the
     remaining number of qslots is less than the maximum instruction fetch band
     width, then the number of instructions to be requested will be the number of
     free qslots. Otherwise, the number of instructions will be the maximum
     instruction fetch bandwidth. */

  inst_cnt = 0 ;
  fp = fetch_p ;
  for (fpkt_cnt = 0; fpkt_cnt < MAX_IFETCH; fpkt_cnt++) {
    if (VALID(fp->valid)) 
      inst_cnt++ ;
    fp++ ;
  }

  fetch_cnt = -inst_cnt ;
  qs = firstf_qslot ;
  while ((fetch_cnt < gl_conf.ifetch_cnt) && (qs != NULL)) {
    qs = qs->next_qslot ;
    fetch_cnt++ ;
  }

  /* Determine whether instructions are to be requested in the current cycle.
     Instructions are to be requested if (1) all incorrectly predicted control
     transfer instructions will complete execution in the current cycle, (2) the
     last instruction in the instruction que will complete fetch in the current
     cycle, and (3) all instructions (if any) in the instruction fetch packet
     will complete fetch in the current cycle. */

  fetch_r->cnt = fetch_cnt ;

  qs = iq_head ;
  while (qs != firstf_qslot) {
    qe = qs->c_entry ;
    ri = qe->rec_inst ;
    if (!ANLD(qe->state) && CT_OP(ri.type) && 
	!ri.pred && (qe->exec_cycle > clk_cycle)) {
      fetch_r->cnt = 0 ;

      (global_stats.cti_fetch_stall)++ ;
      if (lastv_qslot == NULL)
	iq_length = 0 ;
      else {
	iq_length = lastv_qslot->no - iq_head->no + 1;
	if (iq_length <= 0)
	  iq_length = be_conf.no_of_qslots + iq_length;
      }
      (global_stats.cti_fetch_stall_iq_stats[iq_length])++ ;

    }
    qs = qs->next_qslot ;
  }

  if (lastv_qslot != NULL)
    if (lastv_qslot->c_entry->fetch_cycle > clk_cycle) 
      fetch_r->cnt = 0 ;

  fp = fetch_p ;
  for (fpkt_cnt = 0; fpkt_cnt < MAX_IFETCH; fpkt_cnt++) {
    if (VALID(fp->valid) && (fp->fetch_cycle > clk_cycle)) 
      fetch_r->cnt = 0 ;    
  }

} /* end fetch_pipe */

/*==============================================================================*/

/* The issue pipe receives a set of instructions from the branch processor. The
   number of instructions is allways less than the number of free instruction que
   entries (this is guaranteed by the fetch pipe). The issue pipe issues the
   instructions into the instruction que starting from the first free instruction
   que entry. Input register and input condition code dependencies are resolved
   at issue time.

   The issue pipe also handles control transfer instruction completion. It will
   update the execution wait cycle of the youngest instruction in the instruction
   que if the instruction is a control transfer instruction. */

void
issue_pipe (ireq *fetch_r ,	/* instruction fetch request			*/
	    ipkt  *fetch_p ) {
  
  ipkt	  *fp ;			/* instruction packet				*/
  qentry  *qe ;			/* que entry					*/
  qslot   *qs ;			/* que slot					*/

  einst	   ei ;			/* exploded instruction				*/
  rinst    ri ;			/* recoded instruction				*/

  int	   fpkt_cnt ;		/* fetch packet count				*/

  int	   iq_length ;		/* length of instruction que			*/
  int	   fetch_cnt = 0 ;	/* fetch count					*/

  static int annul_next = 0 ;	/* annull next instruction			*/ 

  /* Inspect all entries in the fetch packet. Issue all valid entries into the
     instruction que. */

  fp = fetch_p ;
  for (fpkt_cnt = 0; fpkt_cnt < MAX_IFETCH; fpkt_cnt++) {
    if (VALID(fp->valid)) {

      /* explode the recoded instruction fields into sub components relevant to
         the backend recoder. */

      ei.op      = GETFLD(fp->recode[0],  6,  2) ;
      ei.op2     = GETFLD(fp->recode[0],  3,  3) ;
      ei.op3     = GETFLD(fp->recode[0],  0,  6) ;
      ei.opf     = GETFLD(fp->recode[1],  0,  9) ;

      ei.src1    = GETFLD(fp->recode[1], 26,  6) ;
      ei.src2    = GETFLD(fp->recode[1], 20,  6) ; 
      ei.src3    = GETFLD(fp->recode[1],  8,  6) ; 
      ei.dst     = GETFLD(fp->recode[1], 14,  6) ; 

      ei.i       = GETFLD(fp->recode[1],  0,  1) ;
      ei.fcn     = GETFLD(fp->recode[1], 14,  6) ; 
      ei.cc2     = GETFLD(fp->recode[1], 19,  1) ; 

      ei.bcond   = GETFLD(fp->recode[1], 28,  4) ; 

      ei.annul   = GETFLD(fp->recode[1],  1,  1) ; 
      ei.taken   = fp->taken ;
      ei.pred    = fp->pred ;

      ei.sqsh    = annul_next ;

/* This code snipped is here only until we have verified that the branch 
   predictor in the front end works correctly. 

      rand_no = random() ;
      if ((rand_no - 1000*(rand_no/1000)) >= 900)
	ei.pred    = FALSE ;
      else
	ei.pred	 = TRUE ;
*/

      /* recode the instruction */

      ri.pred       = TRUE ;
      ri.annul_next = FALSE ;
 
      recode (&ei, &ri) ;

      ri.recode[0] = fp->recode[0] ;
      ri.recode[1] = fp->recode[1] ;
      ri.inst      = fp->inst ;
      ri.taken     = fp->taken ;
      ri.pc        = fp->pc ;
      ri.ea        = fp->ea ;


      /* Intitialize the que entry fields. */

      qe = firstf_qslot->n_entry ;

      if (!annul_next)
	qe->state        = TRUE ;
      else
	qe->state	 = FALSE ;

      qe->rec_inst     = ri ;
      qe->rdst         = ri.dst ;

      qe->i_no	       = n_inst_no ;

      qe->src1_qslot   = NOT_VALID ;
      qe->src2_qslot   = NOT_VALID ;
      qe->src3_qslot   = NOT_VALID ;
      qe->ccr_qslot    = NOT_VALID ;

      qe->src1_cycle   = 0 ;
      qe->src2_cycle   = 0 ;
      qe->src3_cycle   = 0 ;
      qe->ccr_cycle    = 0 ;

      qe->start_cycle  = clk_cycle ;
      qe->fetch_cycle  = fp->fetch_cycle ;
      qe->issue_cycle  = fp->fetch_cycle + be_conf.issue_delay ;
      qe->qwait_cycle  = MAX_CYCLE ;
      qe->disp_cycle   = MAX_CYCLE ;
      qe->exec_cycle   = MAX_CYCLE ;
      qe->lswait_cycle = MAX_CYCLE ;
      qe->ewait_cycle  = MAX_CYCLE ;
      qe->compl_cycle  = MAX_CYCLE ;

      (global_stats.itype_stats[qe->rec_inst.type])++ ;

      /* Scan the instruction que for the last valid entries with result
         destination registers that match the source registers, or with modified
         ccr or fcc fields (if the instruction has an input dependency for ccr or
         fcc). The que slot numbers of the que entries with the last matching
         destination numbers are entered into the que entry source dependency
         fields. */

      for (qs = iq_head ; qs != firstf_qslot ; qs = qs->next_qslot) {
	if (!ANLD(qs->c_entry->state)) {

	  if (VALID(ri.src1) && (qs->c_entry->rdst == ri.src1)) {
	    qe->src1_qslot = qs->no ;
	    if (LS_OP(qs->c_entry->rec_inst.type))
	      qe->src1_cycle = qs->c_entry->ewait_cycle ;
	    else
	      qe->src1_cycle = qs->c_entry->exec_cycle ;
	  }

	  if (VALID(ri.src2) && (qs->c_entry->rdst == ri.src2)) {
	    qe->src2_qslot = qs->no ;
	    if (LS_OP(qs->c_entry->rec_inst.type))
	      qe->src2_cycle = qs->c_entry->ewait_cycle ;
	    else
	      qe->src2_cycle = qs->c_entry->exec_cycle ;
	  }

	  if (VALID(ri.src3) && (qs->c_entry->rdst == ri.src3)) {
	    qe->src3_qslot = qs->no ;
	    if (LS_OP(qs->c_entry->rec_inst.type))
	      qe->src3_cycle = qs->c_entry->ewait_cycle ;
	    else
	      qe->src3_cycle = qs->c_entry->exec_cycle ;
	  }

	  if (VALID(ri.spsrc) && (qs->c_entry->rec_inst.spdst == ri.spsrc)) {
	    qe->ccr_qslot = qs->no ;
	    if (LS_OP(qs->c_entry->rec_inst.type))
	      qe->ccr_cycle = qs->c_entry->ewait_cycle ;
	    else
	      qe->ccr_cycle = qs->c_entry->exec_cycle ;
	  }

	}
      }


      /* Copy the n_entry value to the c_entry value so that the instruction is
	 visible to other instrucitons that are to be issued after it in the
	 current cycle. */

      *(firstf_qslot->c_entry) = *qe ;


      /* Upate the que entry pointers. */

      lastv_qslot  = firstf_qslot ;
      if (firstf_qslot != NULL)
	firstf_qslot = firstf_qslot->next_qslot ;
      else 
	firstf_qslot = NULL ;


      /* force the instruction request count to zero if the issued instruction
	 is an incorrectly predicted control transfer instruction. */

      if (!annul_next && CT_OP(ri.type) && !ri.pred)
	fetch_r->cnt = 0 ;

      /* increment instruction sequence counter. */

      n_inst_no++ ;

      if (ANLD(qe->state))
	global_stats.no_of_annulled_insts++ ;

      (global_stats.no_of_instructions)++ ;
      fetch_cnt++ ;

      annul_next = ri.annul_next ;
    }

    fp++ ;

  }

  (global_stats.fetch_stats[fetch_cnt])++ ;

  if (lastv_qslot == NULL)
    iq_length = 0 ;
  else {
    iq_length = lastv_qslot->no - iq_head->no + 1;
    if (iq_length <= 0)
      iq_length = be_conf.no_of_qslots + iq_length;
  }

  (global_stats.iq_stats[iq_length])++ ;

} /* end issue_pipe */

/*==============================================================================*/

/* The dispatch pipe selects a set of executable instructions (ie instructions
   that have all their input dependencies resolved) and dispatches them to the
   available set of execution units. It is assumed that dispatching is only
   constrained by the number of available execution units (the processor will be
   designed such that support resources will exist for every execution unit).

   The first dispatch cycle of an instruction may overlap the last wait cycle of
   a dependent instruction (for non load store instructions this is equivalent to
   the last execute cycle).

   Load store instructions are dispatched to an execution unit capable of
   calculating the load store addresses.  Load store instructions are marked as
   executed when their address generation has completed. They will be marked
   completed when the load store has been completed in the load store unit. 

   Control transfer instructions are dispatched to an execution unit capable of
   determining the branch direction, and branch target. Control transfer
   instruction are marked as executed when the branch direction, and branch
   target have been determined. They will be marked completed when the target
   instructions have been fetched by the branch unit. */

void
disp_pipe (int   *ls_agen ,
	   lsreq *ls_r ) {

  qslot	 *qsd ;			/* dispatch que slot pointer			*/

  eunit  *eu ;			/* execution unit				*/

  lsreq  *ls_request ;		/* load store request pointer			*/

  int	  inst_type ;		/* instruction type				*/

  int	  eu_free ;		/* execution unit is free			*/
  int	  eu_mtch ;		/* execution unit matches the instruction	*/
  int	  ls_free ;		/* a free load store buffer entry is available	*/

  int	  iq_sync = 1 ;		/* indicates that the instruction que is in 
				   synch up to the current instruction.		*/
  int	  iq_compl = 1 ;	/* indicates that the instruciton que is completed 
				   up to the current instruction.		*/

  int	  src_ready ;		/* indicates that all source dependencies are 
				   resloved					*/
  int	  inst_ready ;		/* indicates that the current instruction can be
				   dispatched					*/

  int	  spec_src1_ready ;	/* source 1 is speculatively ready		*/
  int	  spec_src2_ready ;	/* source 1 is speculatively ready		*/
  int	  spec_src3_ready ;	/* source 1 is speculatively ready		*/
  int	  spec_ccr_ready ;	/* ccr source is speculatively ready		*/

  int	  lsr_cnt ;		/* load store request counter			*/

  int	  ls_undisp = FALSE ;	/* undispatched load store instructions         */
  int	  st_undisp = FALSE ;	/* undispatched store instructions              */
  int	  ls_undisp_now = FALSE ;	/* undispatched load store instructions */
  int	  st_undisp_now = FALSE ;	/* undispatched store instructions      */
  int	  agen_ok;              /* address generation ordering                  */

  /* Search through the instruction que for all instructions that have completed
     issue, have all their dependencies met, and have not yet started dispatch
     (ie are being qued). Source dependencies are considered as being met when
     the dependent instructions are in their last execution cycle.

     Also, check if instruction que is in synch. Instructions that require
     instruction que synch cannon be dispatched until all preceeding instructions
     have been completed. Instructions following the synched instruction cannot
     be dispatched until the synched instruction has been completed. */

  qsd = iq_head ;
  while (qsd != firstf_qslot) {

    if (!ANLD(qsd->c_entry->state)) {

      /* Check for any qued instructions. An instruction is considered as being
	 qued when it is in its last cycle of issue, but have not yet started
	 dispatch. */

      if ((qsd->c_entry->issue_cycle <= clk_cycle) && 
	  (qsd->c_entry->qwait_cycle == MAX_CYCLE)) {

	if(LS_OP(qsd->c_entry->rec_inst.type))
	  ls_undisp_now = TRUE;
	if(ST_OP(qsd->c_entry->rec_inst.type))
	  st_undisp_now = TRUE;

	/* Two cycle load store requires speculative dispatch of instructions.
	   Basically this means that all D0 accesses are assumed to hit, and
	   be available in the following cycle. Instructions that are dependent
	   on the D0 accesses will be speculatively dispatched. If the access
	   missed in the D0, then the speculatively dispatched instruction
	   will be cancelled. */

	spec_src1_ready = FALSE ;
	spec_src2_ready = FALSE ;
	spec_src3_ready = FALSE ;
	spec_ccr_ready  = FALSE ;
	ls_request = ls_r ;
	for (lsr_cnt = 0 ; lsr_cnt < MAX_LS_DISP ; lsr_cnt++) {
	  if (VALID(ls_request->valid)) {
	    if (ls_request->qslot_no == qsd->c_entry->src1_qslot)
	      spec_src1_ready = TRUE ;
	    if (ls_request->qslot_no == qsd->c_entry->src2_qslot)
	      spec_src2_ready = TRUE ;
	    if (ls_request->qslot_no == qsd->c_entry->src3_qslot)
	      spec_src3_ready = TRUE ;
	    if (ls_request->qslot_no == qsd->c_entry->ccr_qslot)
	      spec_ccr_ready = TRUE ;
	  }
	  ls_request++ ;
	}

	if (gl_conf.d0_lat == 1) {
	  src_ready   =  ((qsd->n_entry->src1_cycle  <= clk_cycle + 1)  ||
			   spec_src1_ready)                             &&
	                 ((qsd->n_entry->src2_cycle  <= clk_cycle + 1)  ||
			   spec_src2_ready)                             &&
			 ((qsd->n_entry->src3_cycle  <= clk_cycle + 1)  ||
			   spec_src3_ready                              ||
		           ST_OP(qsd->c_entry->rec_inst.type))          &&
	                 ((qsd->n_entry->ccr_cycle   <= clk_cycle + 1)  ||
			  spec_ccr_ready) ;
	}
	else {
	  src_ready   =  (qsd->n_entry->src1_cycle  <= clk_cycle + 1)  &&
	                 (qsd->n_entry->src2_cycle  <= clk_cycle + 1)  &&
			 ((qsd->n_entry->src3_cycle  <= clk_cycle + 1) ||
		           ST_OP(qsd->c_entry->rec_inst.type))         &&
	                 (qsd->n_entry->ccr_cycle   <= clk_cycle + 1)  ;
	}

	inst_ready  =  !(qsd->c_entry->rec_inst.sync && !iq_compl) ;

	if (iq_sync && src_ready && inst_ready) {

	  /* Search through execution unit list for a free execution unit that
	     is capable of executing the instruction. */

	  /* agen ordering support */

	  /*
	  agen_ok = (!(LD_OP(qsd->c_entry->rec_inst.type) && (st_undisp == TRUE)) && 
		     !(ST_OP(qsd->c_entry->rec_inst.type) && (ls_undisp == TRUE)));
	  */
	  agen_ok = TRUE;

	  eu = eu_head ;
	  while (eu != NULL) {

	    inst_type = qsd->c_entry->rec_inst.type ;

	    eu_free = (eu->free_cycle <= clk_cycle + be_conf.disp_delay) ;
	    eu_mtch = GETBIT(eu->type_vec, inst_type) ;
	    ls_free = (*ls_agen < be_conf.no_of_agen_bufs - 1) || 
	              (qsd == iq_head) ;


	    if (eu_free && eu_mtch && (!LS_OP(inst_type) || ls_free && agen_ok)) {

	      /* Reserve the execution unit by updating the execution unit list.
                 */

	      eu->free_cycle = clk_cycle + be_conf.disp_delay +
	                                   be_conf.i_type[inst_type].ithrp ;

	      (eu->use_stats)++ ;

	      /* Update the dispatch cycle of the instruction. */

	      qsd->n_entry->disp_cycle  = clk_cycle + be_conf.disp_delay ;
	      qsd->n_entry->qwait_cycle = clk_cycle ;

	      strcpy(qsd->n_entry->eu_name, eu->name) ;

	      if (LS_OP(inst_type)) 
		(*ls_agen)++ ;

	      /* reset undispatch flag 
		 if load/store has been dispatched sucessfully */
	      if(LS_OP(inst_type))
		ls_undisp_now = FALSE;
	      if(ST_OP(inst_type))
		st_undisp_now = FALSE;

	      break ;
	    }

	    eu = eu->next_eunit ;
	  }
	}

	/* set undispatch flag 
	   if load/store has not been dispatched sucessfully */
	if (ls_undisp_now == TRUE)
	  ls_undisp = TRUE;		
	if (st_undisp_now == TRUE)
	  st_undisp = TRUE;		

      }

      /* check whether the current instruction is completed. If it is not, then
	 reset the iq_compl flag. This will prevent any further synch
	 instructions in the instruction que from dispatching in the current
	 cycle.

	 Check whether the current instruction is a not completed instruction
	 that require sync. This will prevent any further instrucitons to be
	 dispatched in the current cycle.  */

      if (qsd->c_entry->compl_cycle > clk_cycle) {
	iq_compl = 0 ;
	if (qsd->c_entry->rec_inst.sync )
	  iq_sync = 0 ;
      }
    }

    qsd = qsd->next_qslot ;
  }


} /* end disp_pipe */

/*==============================================================================*/

/* The execute pipe searches the instruction que for instructions that have
   completed dispatch. It updates the execution cycle and the wait cycles of
   these instructions (based on the instruction latencies defined in the
   instruction latency table). It adds load store instructions to the load store
   request list. */

void
exec_pipe (int	  *ls_agen ,
	   int	  *ls_no ,
	   lsreq  *ls_r ,
	   lspkt  *ls_p )	

{
  qslot	  *qse ;		/* execute que slot pointer			*/
  qslot	  *qsu ;		/* update que slot pointer			*/

  qentry  *qe ;			/* que entry pointer				*/

  lspkt   *lsp ;		/* pointer to load store packet list		*/

  rinst   ri ;			/* recoded instruction				*/

  lsreq	 *ls_request ;		/* load store request				*/

  ulong	  exec_cycle ;		/* completion cyle for execute 			*/
  ulong	  ewait_cycle ;		/* completion cyle for wait 			*/

  int     lsr_cnt = 0 ;		/* load store request count			*/

  int	  iq_compl = 1 ;	/* indicates that the instruciton que is completed 
				   up to the current instruction.		*/
  int	  ls_exec = 1 ;		/* indicates that all load store instruction
				   addresses have been resolved			*/
  int	  st_exec = 1 ;		/* indicates that all store instruction addresses 
				   have been resolved				*/

  int	  ld_ready ;		/* indicates that a load instruction can be
				   dispatched					*/
  int	  st_ready ;		/* indicates that a store instruction can be
				   dispatched					*/

  int	  ls_confl ;		/* indicates that there is a load to store
				   address conflict				*/

  int     lsp_cnt = 0 ;		/* load store packet counter			*/

  int	  que_flag ;		/* indicates that the search through the 
				   instruction que has passed the qslot that
				   generated the load store request		*/

  int	  iq_compl_st = 1 ;	/* indicates that the instruciton que is completed 
				   or dispatched to LSP 
				   up to the current instruction                */

  /* Search the load store packet for load stores that are returned by the load
     store unit in the current cycle. Update the source completion cycles of
     any dependent sources. */

  lsp = ls_p ;
  while (lsp_cnt <  MAX_LS_COMPL) {

    if (VALID(lsp->valid)) {
      que_flag = 0 ;
      qsu = iq_head ;
      while (qsu != firstf_qslot) { 

	if (que_flag) {
	  if (qsu->c_entry->src1_qslot == lsp->qslot_no)
	    qsu->n_entry->src1_cycle =  lsp->compl_cycle ;
	  if (qsu->c_entry->src2_qslot == lsp->qslot_no)
	    qsu->n_entry->src2_cycle =  lsp->compl_cycle ;
	  if (qsu->c_entry->src3_qslot == lsp->qslot_no)
	    qsu->n_entry->src3_cycle =  lsp->compl_cycle ;
	  if (qsu->c_entry->ccr_qslot == lsp->qslot_no)
	    qsu->n_entry->ccr_cycle =   lsp->compl_cycle ;
	}
	else {
	  if (qsu->no == lsp->qslot_no)
	    que_flag = 1 ;
	}

	qsu = qsu->next_qslot ;
      }
    }

    lsp++ ;
    lsp_cnt++ ;
  }

  /* Search through the instruction que for any instructions that have been
     speculatively dispatched (ie on the assumption that one or more of the
     source operands will hit in the d0 cache). The instruction dispatch will be
     cancelled if any of the sources are not yet available. */

  qse = iq_head ;
  while (qse != firstf_qslot) {
    qe = qse->n_entry ;
    if (!ANLD(qe->state) && (qe->disp_cycle < MAX_CYCLE)) {
      if ((VALID(qe->src1_qslot) && (qe->src1_cycle > clk_cycle)) ||
	  (VALID(qe->src2_qslot) && (qe->src2_cycle > clk_cycle)) ||
	  (VALID(qe->src3_qslot) && (qe->src3_cycle > clk_cycle)  &&
	   !ST_OP(qe->rec_inst.type))                              ||
	  (VALID(qe->ccr_qslot)  && (qe->ccr_cycle  > clk_cycle))) { 
	qse->n_entry->qwait_cycle = MAX_CYCLE ;
        qse->n_entry->disp_cycle  = MAX_CYCLE ;
	qse->c_entry->qwait_cycle = MAX_CYCLE ;
        qse->c_entry->disp_cycle  = MAX_CYCLE ;

	if (LS_OP(qe->rec_inst.type)) 
	  (*ls_agen)-- ;
      }
    }
    qse = qse->next_qslot ;
  }

  qse = iq_head ;
  ls_request = ls_r ;

  /* Search through instruction que for all non control transfer and all non load
     store instructions that have completed dispatch. Update the execute and wait
     cycle of the found instructions. */

  while (qse != firstf_qslot) {

    if (!ANLD(qse->c_entry->state)) {

      ri = qse->c_entry->rec_inst ;
      if (qse->c_entry->disp_cycle == clk_cycle) {

	exec_cycle = clk_cycle + be_conf.i_type[ri.type].ilat ;
	qse->n_entry->exec_cycle = exec_cycle ;

	if (!LS_OP(ri.type)) {

	  /* Search through instruction que for any instructions that are
	     dependent on the found instructions. Update the source, or ccr
	     (wichever are applicable) cycles of the dependent instructions to
	     reflect the execute cycle of the found instructions. */

	  qsu = qse->next_qslot ;
	  while (qsu != firstf_qslot) { 
	    if (qsu->c_entry->src1_qslot == qse->no)
	      qsu->n_entry->src1_cycle =  exec_cycle ;
	    if (qsu->c_entry->src2_qslot == qse->no) 
	      qsu->n_entry->src2_cycle =  exec_cycle ;
	    if (qsu->c_entry->src3_qslot == qse->no)
	      qsu->n_entry->src3_cycle =  exec_cycle ;
	    if (qsu->c_entry->ccr_qslot == qse->no)
	      qsu->n_entry->ccr_cycle =  exec_cycle ;
	    qsu = qsu->next_qslot ;
	  }

	  /* Update the ewait cycle of non control transfer instructions. */

	  if (!CT_OP(ri.type)) {
	    ewait_cycle = clk_cycle + be_conf.i_type[ri.type].ilat ;
	    qse->n_entry->ewait_cycle = ewait_cycle ;
	  }

	  qse->n_entry->lswait_cycle = clk_cycle + be_conf.i_type[ri.type].ilat ;

	}
    
      }

      /* Check for load store instructions. Load store instructions will be
	 entered into the load store request list. */

      /*
      st_ready    = iq_compl ;
      */
      st_ready    = iq_compl_st && (qse->c_entry->src3_cycle <= clk_cycle + 1);

      if (gl_conf.mem_model == RMO)
	st_ready    = iq_compl || 
	              (st_exec && (qse->c_entry->src3_cycle <= clk_cycle + 1));

      if (gl_conf.mem_model == LSX)
	ld_ready  = iq_compl  ;
      if (gl_conf.mem_model == LSO)
	ld_ready  = ls_exec ;
      if (gl_conf.mem_model == STO) 
	ld_ready  = st_exec ;
      if (gl_conf.mem_model == RMO) 
	ld_ready  = st_exec ;

      if ((qse->c_entry->exec_cycle   <= clk_cycle) &&
	  (qse->c_entry->lswait_cycle == MAX_CYCLE) &&
	  (lsr_cnt < gl_conf.no_of_ls_req)) {

	if (LD_OP(ri.type) && ld_ready && (*ls_no < gl_conf.no_of_ls_bufs)) {
	  ls_confl = 0 ;
	  qsu = iq_head ;
	  while (qsu != qse) {
	    if ((ST_OP(qsu->c_entry->rec_inst.type))      &&
		(qsu->n_entry->lswait_cycle == MAX_CYCLE) &&
		((qsu->c_entry->rec_inst.ea & 0x1fff) == 
		 (qse->c_entry->rec_inst.ea & 0x1fff)))
	      ls_confl = 1 ;
	    qsu = qsu->next_qslot ;
	  }

	  if (!ls_confl) {
	    ls_request->valid = TRUE ;
	    ls_request->recode[0] = ri.recode[0] ;
	    ls_request->recode[1] = ri.recode[1] ;
	    ls_request->ea        = ri.ea ;
	    ls_request->qslot_no = qse->no ;

	    lsr_cnt++ ;
	    ls_request++ ;

	    (*ls_agen)-- ;
	    (*ls_no)++ ;
	    qse->n_entry->lswait_cycle = clk_cycle ;
	  }

	}

	if (ST_OP(ri.type) && st_ready && (*ls_no < gl_conf.no_of_ls_bufs)) {

	  ls_request->valid = TRUE ;
	  ls_request->recode[0] = ri.recode[0] ;
	  ls_request->recode[1] = ri.recode[1] ;
	  ls_request->ea        = ri.ea ;
	  ls_request->qslot_no = qse->no ;

	  lsr_cnt++ ;
	  ls_request++ ;

	  (*ls_agen)-- ;
	  (*ls_no)++ ;
	  qse->n_entry->lswait_cycle = clk_cycle ;
	}

      }


      /* check that all preceeding load store instructions have been dispatched
	 (this if for load store ordering). */

      if (LS_OP(qse->c_entry->rec_inst.type) &&
	  (qse->n_entry->exec_cycle > clk_cycle)) 
	ls_exec = 0 ;

      if (ST_OP(qse->c_entry->rec_inst.type) &&
	  (qse->n_entry->exec_cycle > clk_cycle)) 
	st_exec = 0 ;

      /* check whether the current instruction is completed. If it is not, then
	 reset the iq_compl flag. */


      if ((!ST_OP(qse->c_entry->rec_inst.type) &&
	   !LS_OP(qse->c_entry->rec_inst.type) &&
	   (qse->n_entry->ewait_cycle + 2 > clk_cycle)) ||
	  (ST_OP(qse->c_entry->rec_inst.type) &&
	   (qse->n_entry->lswait_cycle == MAX_CYCLE)) ||
	  (LS_OP(qse->c_entry->rec_inst.type) &&
	   (qse->n_entry->lswait_cycle == MAX_CYCLE)))
	iq_compl_st = 0 ;

      if (qse->c_entry->ewait_cycle > clk_cycle) 
	iq_compl = 0 ;

    }

    qse = qse->next_qslot ;

  }


  /* invalidate all non-used load store request list entries. */

  for (; lsr_cnt < MAX_LS_DISP ; lsr_cnt++) {
    ls_request->valid = NOT_VALID ;
    ls_request++ ;
  }

} /* end exec_pipe */

/*==============================================================================*/

/* The completion pipe searches the instruction que, the load store packet, and
   the fetch packet for instructions that complete in the current cycle. It
   updates the 'completion cycle' counts of the instructions.

   The completion pipe also updates the source cycle counts of any instructions
   that are dependent on a load or a store. It also enteres any control transfer
   instructions that complete execution in the current cycle into the completed
   branch list. The branch processor will use this list to update the branch
   history and branch target tables. */


void
compl_pipe (lspkt *ls_p , 	/* load store packet array			*/
	    int   *ls_no ,      /* number of outstanding load store instructions*/
	    ireq  *fetch_r ,	/* instruction fetch request			*/
	    int   last_instr )  /* last instruction has been fetched from trace */

{
  qslot	 *qsc ;			/* completion que slot pointer			*/
  qslot	 *qsu ;			/* update que slot pointer			*/

  lspkt   *lsp ;		/* pointer to load store packet list		*/
  brlis   *brl ;		/* pointer to branch resolution list		*/

  rinst    ri ;			/* recoded instruction				*/

  int	   brl_cnt = 0 ;       	/* number of branches completed			*/
  int      lsp_cnt = 0 ;	/* load store packet counter			*/


  /* Search instruction que for all instructions that are not control transfer,
     nor load store instructions, and which complete execution in the current
     cycle. 'Complete' these instructions by updating their 'completion cycle'
     counts. */

  qsc = iq_head ;
  while (qsc != firstf_qslot) {
    if (!ANLD(qsc->c_entry->state)) {
      ri = qsc->c_entry->rec_inst ;
      if (!CT_OP(ri.type) && !LS_OP(ri.type)) { 
	if (qsc->c_entry->exec_cycle == clk_cycle) 
	  qsc->n_entry->compl_cycle = clk_cycle + be_conf.compl_delay ;
      }
    }
    qsc = qsc->next_qslot ;
  }


  /* Search through the instruction que for any control transfer instructions
     that have completed execution and fetch. Update the completion cycle counts
     of the control transfer instructions.

     Update the fetch request 'completed branch' list for all control transfer
     instructions that have completed execution. */

  brl = fetch_r->compl ;
  qsc = iq_head ;
  while (qsc != firstf_qslot) { 

    if (!ANLD(qsc->c_entry->state)) {

      if (CT_OP(qsc->c_entry->rec_inst.type)) {

	if (qsc->c_entry->ewait_cycle == MAX_CYCLE) {
	  qsu = qsc->next_qslot ;
	  if (qsu != firstf_qslot) {
	    qsu = qsu->next_qslot ;
	    if (qsu != firstf_qslot) {
	      if ((qsc->c_entry->exec_cycle  <= clk_cycle) &&
		  (qsu->c_entry->fetch_cycle <= clk_cycle)) {
		qsc->n_entry->ewait_cycle = clk_cycle ;
		qsc->n_entry->compl_cycle = clk_cycle + be_conf.compl_delay ;
	      }
	    }
	    else if (last_instr && (qsc->c_entry->exec_cycle  <= clk_cycle)) {
	      qsc->n_entry->ewait_cycle = clk_cycle ;
	      qsc->n_entry->compl_cycle = clk_cycle + be_conf.compl_delay ;
	    }
	  }
	  else if (last_instr && (qsc->c_entry->exec_cycle  <= clk_cycle)) {
	    qsc->n_entry->ewait_cycle = clk_cycle ;
	    qsc->n_entry->compl_cycle = clk_cycle + be_conf.compl_delay ;
	  }
	
	  /* this would be the real code if there were no discontinuetes in
	     the trace.

	  qsu = qsc->next_qslot ;
	  while (qsu != firstf_qslot) {
	    if (qsc->c_entry->rec_inst.taken) {
	      if ((qsc->c_entry->rec_inst.ea   == qsu->c_entry->rec_inst.pc)) {
		if ((qsc->c_entry->exec_cycle  <= clk_cycle) &&
		    (qsu->c_entry->fetch_cycle <= clk_cycle)) {
		  qsc->n_entry->ewait_cycle = clk_cycle ;
		  qsc->n_entry->compl_cycle = clk_cycle + be_conf.compl_delay ;
		}
	      }
	    }
	    else {
	      if ((qsc->c_entry->rec_inst.pc+8 == qsu->c_entry->rec_inst.pc)) {
		if ((qsc->c_entry->exec_cycle  <= clk_cycle) &&
		    (qsu->c_entry->fetch_cycle <= clk_cycle)) {
		  qsc->n_entry->ewait_cycle = clk_cycle ;
		  qsc->n_entry->compl_cycle = clk_cycle + be_conf.compl_delay ;
		}
	      }
	    }

	    qsu = qsu->next_qslot ;
	  }

	  end of code */

	}

	if (qsc->c_entry->exec_cycle == clk_cycle) { 
	  brl->valid     = TRUE ;
	  brl->inst_no   = qsc->c_entry->i_no ;
	  brl->recode[0] = qsc->c_entry->rec_inst.recode[0] ;
	  brl->recode[1] = qsc->c_entry->rec_inst.recode[1] ;
	  brl->pc        = qsc->c_entry->rec_inst.pc ;
	  brl->ea        = qsc->c_entry->rec_inst.ea ;
	  brl_cnt++ ;
	  brl++ ;

	  /* update the global branch resolution statistics */

	  switch (GETFLD(qsc->c_entry->rec_inst.recode[0], 3, 5))  {

	  case 0x01 :	/* BPcc		*/
	  case 0x02 :	/* Bicc		*/
	  case 0x05 :	/* FBPfcc	*/
	  case 0x06 :	/* FBfcc	*/
	    if (GETFLD(qsc->c_entry->rec_inst.recode[1], 28, 3) != 0) {
	      global_stats.br_comp_refs++ ;
	      global_stats.br_res_comp = global_stats.br_res_comp + 
		                         (clk_cycle - qsc->c_entry->fetch_cycle) ;
	      if (!qsc->c_entry->rec_inst.pred) {
		global_stats.br_ip_comp_refs++ ;
		global_stats.br_ip_res_comp = global_stats.br_ip_res_comp + 
	    		 	         (clk_cycle - qsc->c_entry->fetch_cycle) ;
	      }
	    }
	    break ;

	  case 0x03 :	/* BPr		*/
	    global_stats.br_comp_refs++ ;
	    global_stats.br_res_comp = global_stats.br_res_comp + 
		                       (clk_cycle - qsc->c_entry->fetch_cycle) ;
	    if (!qsc->c_entry->rec_inst.pred) {
	      global_stats.br_ip_comp_refs++ ;
	      global_stats.br_ip_res_comp = global_stats.br_ip_res_comp + 
	    			         (clk_cycle - qsc->c_entry->fetch_cycle) ;
	    }
	    break ;
	    
	  default :
	    break ;
	  }

	}
      
      }

    }

    qsc = qsc->next_qslot ;

  }


  /* Invalidate rest of the fetch request 'completed branch' resolution list. */

  for (; brl_cnt < MAX_BR_COMPL ; brl_cnt++) {
    brl->valid = NOT_VALID ;
    brl++ ;
  }


  /* Search the load store packet for load stores that are returned by the load
     store unit in the current cycle. 'Complete' the load store instructions by
     updating their 'completion cycle' counts. */

  lsp = ls_p ;
  while (lsp_cnt <  MAX_LS_COMPL) {

    if (VALID(lsp->valid)) {
      qsu = iq_head ;
      while (qsu != firstf_qslot) { 
	if (qsu->no == lsp->qslot_no) {
	  qsu->n_entry->ewait_cycle = lsp->compl_cycle ;
	  qsu->n_entry->compl_cycle = lsp->compl_cycle + be_conf.compl_delay ;
	  (*ls_no)-- ;
	}

	qsu = qsu->next_qslot ;
      }
    }

    lsp++ ;
    lsp_cnt++ ;
  }


} /* end compl_pipe */

/*==============================================================================*/

/* The commit pipe removes completed que entries (commits) from the head of the
   instruction que. The maximum number of instructions that can be committed in
   one cycle is defined by the processor back end configuration. The removed que
   entries are invalidated, initialized, and moved to the end of the instruction
   que (wrapped around).

   All committed control transfer instructions are entered into the committed
   instruction list. The branch processor will use this list to update the branch
   history table and branch target table. */

int
commit_pipe (ireq	*fetch_r )	/* instruction fetch request		*/

{
 
  qslot   *qs ;			/* que slot					*/
  qentry  *qe ;			/* que entry					*/

  brlis   *brl ;		/* branch resolution list			*/

  int      commit_cnt = 0 ;	/* number of instructions committed		*/
  int	   brl_cnt    = 0 ;     /* number of branches committed			*/
  int	   type_cnt   = 0 ;	/* instruction type				*/

  brl = fetch_r->commit ;

  while ((iq_head != firstf_qslot) && (commit_cnt < be_conf.commit_cnt)) {

    qs = iq_head ;
    qe = qs->c_entry ;
    if (VALID(qe->state) && (qe->compl_cycle < clk_cycle) ||
	ANLD(qe->state)) {

      /* Update the commited control transfer instruction list if the instruction
	 to be committed is a control transfer instruction. */

      if (!ANLD(qe->state) && CT_OP(qe->rec_inst.type)) {
	brl->valid     = TRUE ;
	brl->inst_no   = qe->i_no ;
	brl->recode[0] = qe->rec_inst.recode[0] ;
	brl->recode[1] = qe->rec_inst.recode[1] ;
	brl->pc        = qe->rec_inst.pc ;
	brl->ea        = qe->rec_inst.ea ;
	brl_cnt++ ;
	brl++ ;

	/* update the global branch resolution statistics */

	switch (GETFLD(qe->rec_inst.recode[0], 3, 5))  {

	case 0x01 :	/* BPcc		*/
	case 0x02 :	/* Bicc		*/
	case 0x05 :	/* FBPfcc	*/
	case 0x06 :	/* FBfcc	*/
	  if (GETFLD(qe->rec_inst.recode[1], 28, 3) != 0) {
	    global_stats.br_cmmt_refs++ ;
	    global_stats.br_res_cmmt = global_stats.br_res_cmmt + 
	    			       (clk_cycle - qe->fetch_cycle) ;
	    if (!qe->rec_inst.pred) {
	      global_stats.br_ip_cmmt_refs++ ;
	      global_stats.br_ip_res_cmmt = global_stats.br_ip_res_cmmt + 
	    			            (clk_cycle - qe->fetch_cycle) ;
	    }
	  }
	  break ;

	case 0x03 :	/* BPr		*/
	  global_stats.br_cmmt_refs++ ;
	  global_stats.br_res_cmmt = global_stats.br_res_cmmt + 
	    			     (clk_cycle - qe->fetch_cycle) ;
	  if (!qe->rec_inst.pred) {
	    global_stats.br_ip_cmmt_refs++ ;
	    global_stats.br_ip_res_cmmt = global_stats.br_ip_res_cmmt + 
	                                 (clk_cycle - qe->fetch_cycle) ;
	  }
	  break ;
	    
	default :
	  break ;
	}

      }

      global_stats.cum_inst_life_time = global_stats.cum_inst_life_time +
      					clk_cycle - qe->fetch_cycle ;


      /* Move the qslot at the head of the instruction que to the tail of the
	 instruction que. Initialize all pointers and que entry states. */

      if (qs == lastv_qslot) 
	lastv_qslot = NULL ;

      qs->c_entry->state = NOT_VALID ;
      qs->n_entry->state = NOT_VALID ;
            
      iq_head = qs->next_qslot ;
      iq_head->prev_qslot = NULL ;

      iq_tail->next_qslot = qs ;
      qs->prev_qslot = iq_tail ;
      iq_tail = qs ;
      iq_tail->next_qslot = NULL ;

      if (firstf_qslot == NULL) 
	firstf_qslot = iq_tail ;

      commit_cnt++ ;

    }
    else
      break ;

  }


  /* Invalidate rest of branch resolution list. */

  for (; brl_cnt < MAX_BR_COMMIT ; brl_cnt++) {
    brl->valid = NOT_VALID ;
    brl++ ;
  }


  /* Check for empty instruction que */

  if (iq_head == firstf_qslot)
    return (1) ;
  else
    return (0) ;


} /* end commit_pipe */

/*==============================================================================*/

int
back_end (ireq	*fetch_r ,	/* instruction fetch request			*/
	  lsreq	*ls_r ,		/* load store request				*/
	  ipkt	*fetch_p ,	/* instruction fetch packet array		*/
	  lspkt	*ls_p , 	/* load store packet array			*/
	  int   last_instr )	/* last instruction has been fetched from trace */

{
  static int	ls_no = 0 ;	/* number of outstanding load store ops		*/
  static int	ls_agen = 0 ;	/* number of outstanding load store agen ops	*/
  int	        empty_ique ;	/* instruction que is empty 			*/

  fetch_pipe (fetch_r, fetch_p) ;
  issue_pipe (fetch_r, fetch_p) ;
  exec_pipe  (&ls_agen, &ls_no, ls_r, ls_p) ;
  compl_pipe (ls_p, &ls_no, fetch_r, last_instr) ;
  disp_pipe  (&ls_agen, ls_r) ;
  empty_ique = commit_pipe (fetch_r) ;
	
  return (empty_ique) ;

} /* end back_end */

/*==============================================================================*/
