/*
 *  $Id: back_end.h,v 1.6 1995/01/18 20:36:54 forssell Exp $
 *  
 *  Copyright (c) 1994 HAL Computer Systems International, Ltd.
 *  All rights reserved.  Unpublished -- rights reserved under
 *  the Copyright Laws of the United States.  USE OF A COPYRIGHT
 *  NOTICE IS PRECAUTIONARY ONLY AND DOES NOT IMPLY PUBLICATION
 *  OR DISCLOSURE.
 *  
 *  THIS SOFTWARE CONTAINS CONFIDENTIAL INFORMATION AND TRADE
 *  SECRETS OF HAL COMPUTER SYSTEMS INTERNATIONAL, LTD.  USE,
 *  DISCLOSURE, OR REPRODUCTION IS PROHIBITED WITHOUT THE
 *  PRIOR EXPRESS WRITTEN PERMISSION OF HAL COMPUTER SYSTEMS
 *  INTERNATIONAL, LTD.
 *  
 *                          RESTRICTED RIGHTS LEGEND
 *  Use, duplication, or disclosure by the Government is subject
 *  to the restrictions as set forth in subparagraph (c)(l)(ii)
 *  of the Rights in Technical Data and Computer Software clause
 *  at DFARS 252.227-7013.
 *  
 *           HAL COMPUTER SYSTEMS INTERNATIONAL, LTD.
 *                   1315 Dell Avenue
 *                   Campbell, CA  95008
 */
/*==============================================================================*/ 
/* Hal Denali Trace Driven Simulator						*/
/* Copyright (c) 1994 HaL Computer Systems, Inc.                      		*/
/*==============================================================================*/
/* back_end.h									*/
/*										*/
/*==============================================================================*/
/* Revision history								*/
/*										*/
/* Date	     Who	   	Description					*/ 
/* --------  -----------      	-----------					*/
/* 10/30/94  Per Forssell  	Initial defintion				*/
/*										*/
/*==============================================================================*/

/*==============================================================================*/
/* Instruction definitions							*/
/*==============================================================================*/

/* Instruction groups ----------------------------------------------------------*/

#define	ADDSUB	(0)  	/* 0x00		*/
#define	SHIFT	(1)	/* 0x01		*/
#define	LOGIC	(2)	/* 0x02		*/
#define	IMULT	(3)	/* 0x03		*/
#define	IDIV	(4)	/* 0x04		*/
#define	IMOVE	(5)	/* 0x05		*/

#define	FADD	(6)	/* 0x06		*/
#define	FMULT	(7)	/* 0x07		*/
#define	FMADD	(8)	/* 0x08		*/
#define	FDIV	(9)	/* 0x09		*/
#define	FSQRT	(10)	/* 0x0a		*/
#define	FMOVE	(11)	/* 0x0b		*/
#define	FCONV	(12)	/* 0x0c		*/

#define	BRANCH	(13)	/* 0x0d		*/
#define	JMPL	(14)	/* 0x0e		*/
#define	CALL	(15)	/* 0x0f		*/
#define	RETURN	(16)	/* 0x10		*/

#define	ILOAD	(17)	/* 0x11		*/
#define	ISTORE	(18)	/* 0x12		*/
#define	FLOAD	(19)	/* 0x13		*/
#define	FSTORE	(20)	/* 0x14		*/

#define SYS     (29)	/* 0x1d		*/

#define	OTHER	(30)	/* 0x1e		*/
#define	UNIMP	(31)	/* 0x1f		*/

/* Instruction types -----------------------------------------------------------*/

#define LS_OP(t)    ((0x001e0000 >> t) & 0x1)	/* load store op		*/
#define LD_OP(t)    ((0x000a0000 >> t) & 0x1)   /* load op			*/
#define ST_OP(t)    ((0x00140000 >> t) & 0x1)   /* store op			*/
#define CT_OP(t)    ((0x0001e000 >> t) & 0x1)	/* control transfer op		*/

/* Register types --------------------------------------------------------------*/

#define	FX	(1)	/* fixed point register					*/
#define FP	(2)	/* floating point register				*/
#define SR	(3)	/* processor state register				*/
#define PR	(4)	/* processor priviledged register			*/

/* State and privileged registers ----------------------------------------------*/

#define CCR     (2)		/* fixed point condition code register		*/
#define FCC	(15)		/* floating point condition code from FSR	*/
#define ASI	(3)		/* floating point condition code from FSR	*/
#define FSR	(31)		/* floating point condition code from FSR	*/

/* Instruction syncs -----------------------------------------------------------*/

/* Either the read synch or the write synch should be set for any status and
   privileged registers that require a synch. A register does NOT require a synch
   if

   (1) it is renamed by mapping into the fixed point register space, and it has
       no global effect.  

   (2) it is mapped into the instruction que entry, and it is set at issue time.

   The effect of 'un-synching' a register can be studied by resetting the read
   and write synchs of that register. The effect may however be missleadingly
   good unless a mechanism is added into the simulator code to correctly handle 
   the register (basically the simulator will ignore any register effects for
   the unsynched register).

*/

#define RSYNCH_Y  		(1)
#define RSYNCH_FPRS  		(1)
#define RSYNCH_TPC  		(1)
#define RSYNCH_TNPC  		(1)
#define RSYNCH_TSTATE  		(1)
#define RSYNCH_TT  		(1)
#define RSYNCH_TICK  		(1)
#define RSYNCH_TBA  		(1)
#define RSYNCH_PSTATE  		(0)	/* mapped into iq			*/
#define RSYNCH_TL  		(1)
#define RSYNCH_PIL  		(0)	/* got to do something about pil	*/
#define RSYNCH_CWP  		(0)	/* mapped into iq entry			*/
#define RSYNCH_CANSAVE  	(0)
#define RSYNCH_CANRESTORE  	(0)
#define RSYNCH_CLEANWIN  	(0)
#define RSYNCH_OTHERWIN  	(0)
#define RSYNCH_WSTATE  		(1)
#define RSYNCH_FSR              (1)

#define WSYNCH_Y  		(1)
#define WSYNCH_FPRS  		(1)
#define WSYNCH_TPC  		(1)
#define WSYNCH_TNPC  		(1)
#define WSYNCH_TSTATE  		(1)
#define WSYNCH_TT  		(1)
#define WSYNCH_TICK  		(1)
#define WSYNCH_TBA  		(1)
#define WSYNCH_PSTATE  		(0)	/* mapped into iq			*/
#define WSYNCH_TL  		(1)
#define WSYNCH_PIL  		(0)	/* got to do something about pil	*/
#define WSYNCH_CWP  		(0)	/* mapped into iq entry			*/
#define WSYNCH_CANSAVE  	(0)
#define WSYNCH_CANRESTORE  	(0)
#define WSYNCH_CLEANWIN  	(0)
#define WSYNCH_OTHERWIN  	(0)
#define WSYNCH_WSTATE  		(1)
#define WSYNCH_FSR              (1)

#define SYNCH_UNIMP		(1)	

/* Processor state register bit definitions ------------------------------------*/

#define AG      (0)	/* PSTATE alternate globals				*/

/*==============================================================================*/
/* Execution unit definitions							*/
/*==============================================================================*/

/* Default parameters for processor back end -----------------------------------*/

#define DEF_NO_OF_QSLOTS        (64)
#define DEF_NO_OF_WINDOWS        (4)
#define DEF_NO_OF_EUNITS        (11)
#define DEF_ISSUE_DELAY          (2)
#define DEF_DISP_DELAY           (1)
#define DEF_COMPL_DELAY          (1)
#define DEF_COMMIT_CNT           (4)
#define DEF_NO_OF_AGEN_BUFS     (16)

#define DEF_EU_TYPE 								\
/*                                 F   I   R     B                         A */ \
/*           U O                   S F S I E     R F F F   F F   I   I L S D */ \
/*           N T                   T L T L T C J A C M S F M M F M I M O H D */ \
/*           I H S                 O O O O U A M N O O Q D A U A O D U G I S */ \
/*           M E Y                 R A R A R L P C N V R I D L D V I L I F U */ \
/*           P R S                 E D E D N L L H V E T V D T D E V T C T B */ \
/*                                                                           */ \
{"ALS1       0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 1 0 0 1 1 1",  \
 "ALS2       0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 1 0 0 1 1 1",  \
 "ALS3       0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 1 0 0 1 1 1",  \
 "ALS4       0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 1 0 0 1 1 1",  \
 "MULDIV     0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 1 1 0 0 0",  \
 "FMA        0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 1 1 0 0 1 1 1 0 0 0 0 0 0",  \
 "FDIVSQRT   0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 1 1 0 0 0 0 0 0 0 0 0",  \
 "BRANCH     0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 1 1 1 1 0 0 0 0 0 0 0 0 0 0 0 0 0",  \
 "AGEN1      0 0 0 0 0 0 0 0 0 0 0 1 1 1 1 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0",  \
 "AGEN2      0 0 0 0 0 0 0 0 0 0 0 1 1 1 1 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0",  \
 "SYS1       1 1 1 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0"}

#define DEF_I_TYPE	  {  1,  1,  	/* ADDSUB             		*/ 	\
			     1,  1,   	/* SHIFT              		*/	\
			     1,  1,   	/* LOGIC              		*/	\
			     5,  5,  	/* IMULT              		*/	\
			    12, 12,  	/* IDIV               		*/	\
			     1,  1,   	/* IMOVE              		*/	\
			     4,  1,  	/* FADD                 	*/	\
			     4,  1,     /* FMULT               		*/	\
			     4,  1,  	/* FMADD                	*/	\
			    15, 15,  	/* FDIV               		*/	\
			    20, 20,  	/* FSQRT               		*/	\
			     1,  1,  	/* FMOVE              		*/	\
			     4,  1,  	/* FCONV              		*/	\
			     1,  1,  	/* BRANCH             		*/	\
			     1,  1,  	/* JMPL                   	*/	\
			     1,  1,  	/* CALL               		*/	\
			     1,  1,  	/* RETURN             		*/	\
			     1,  1,  	/* ILOAD              		*/	\
			     1,  1,  	/* ISTORE              		*/	\
			     1,  1,  	/* FLOAD                	*/	\
			     1,  1,  	/* FSTORE             		*/	\
			     2,  2,  	/* instruction type 21		*/	\
			     2,	 2,  	/* instruction type 22		*/	\
			     2,	 2,  	/* instruction type 23		*/	\
			     2,	 2,  	/* instruction type 24		*/	\
			     2,	 2,  	/* instruction type 25		*/	\
			     2,	 2,  	/* instruction type 26		*/	\
			     2,	 2,  	/* instruction type 27		*/	\
			     2,	 2,  	/* instruction type 28		*/	\
			     1,	 1,  	/* SYS                		*/	\
			     1,	 1,  	/* OTHER              		*/	\
			     1,  1  }	/* UNIMP              		*/

#define I_TYPE_NAMES   {"ADDSUB",	\
			"SHIFT",	\
			"LOGIC",	\
			"IMULT",	\
			"IDIV",		\
			"IMOVE",	\
					\
			"FADD",		\
			"FMULT",	\
			"FMADD",	\
			"FDIV",		\
			"FSQRT",	\
			"FMOVE",	\
			"FCONV",	\
					\
			"BRANCH",	\
			"JMPL",		\
			"CALL",		\
			"RETURN",	\
					\
			"ILOAD",	\
			"ISTORE",	\
			"FLOAD",	\
			"FSTORE",	\
					\
			"IT21",		\
			"IT22",		\
			"IT23",		\
			"IT24",		\
			"IT25",		\
			"IT26",		\
			"IT27",		\
			"IT28",		\
					\
			"SYS",		\
			"OTHER",	\
			"UNIMP" }


/*==============================================================================*/
/* Structure definitions         						*/
/*==============================================================================*/

/* The expanded instruction structure explodes the recoded instruction words into
   fields relevant to the back end instruction recoder. This structure is only
   for purpose of readability. It is only used within the backend issue_pipe
   routine. */

typedef	struct exploded_instruction {
  int	op ;			/* v9 instruction op field			*/
  int	op2 ;			/* v9 instruction op2 field			*/
  int	op3 ;			/* v9 instruction op3 field			*/
  int	opf ;			/* v9 instruction opf field			*/

  int	i ;			/* v9 instruction immediate bit			*/
  int	fcn ;			/* v9 instruction fcn field			*/
  int	cc2 ;			/* v9 instruction cc2 field			*/

  int	bcond ;			/* v9 instruction branch condition field	*/

  int	src1 ;			/* architected source 1 register		*/
  int	src2 ;			/* architected source 2 register		*/
  int	src3 ;			/* architected source 3 register		*/
  int	dst ;			/* architected destination register		*/

  int	pred ;			/* prediction bit for control transfer inst's	*/
  int	taken ;			/* taken bit for control transfer inst's	*/
  int	annul ;			/* annul bit for control transfer inst's	*/

  int	sqsh ;			/* instruction will be squashed due to annul	*/
} einst ;

/*------------------------------------------------------------------------------*/

/* The execution unit list is a linked list of all execution units defined for
   the current proccessor configuration. Each entry defines an available
   execution unit, the instructions that the execution unit can handle, and the
   earliest cycle that the execution unit is ready for dispatch of a new
   operation. */

typedef struct exec_unit {
  char	name[MAX_NAME] ;	/* execution unit type name			*/
  ulong	type_vec ;		/* vector of up to 32 instruction types executable
				   by the execution unit			*/
  ulong	free_cycle ;		/* cycle when execution unit is free to start next
				   operation					*/
  ulong	use_stats ;		/* execution unit usage times			*/
  struct exec_unit *next_eunit ;	/* next unit in the execution unit list	*/
} eunit ;


/*==============================================================================*/
/* Variable declarations       							*/
/*==============================================================================*/

/* configuration variables -----------------------------------------------------*/

bconf	be_conf ;		/* back end configuration			*/

/* execution unit configuration variables --------------------------------------*/

eunit	*eu_head ;		/* first entry in the execution unit list	*/

/*==============================================================================*/



