/*
 *  $Id: be_config.c,v 1.4 1995/01/10 20:13:29 forssell Exp $
 *  
 *  Copyright (c) 1994 HAL Computer Systems International, Ltd.
 *  All rights reserved.  Unpublished -- rights reserved under
 *  the Copyright Laws of the United States.  USE OF A COPYRIGHT
 *  NOTICE IS PRECAUTIONARY ONLY AND DOES NOT IMPLY PUBLICATION
 *  OR DISCLOSURE.
 *  
 *  THIS SOFTWARE CONTAINS CONFIDENTIAL INFORMATION AND TRADE
 *  SECRETS OF HAL COMPUTER SYSTEMS INTERNATIONAL, LTD.  USE,
 *  DISCLOSURE, OR REPRODUCTION IS PROHIBITED WITHOUT THE
 *  PRIOR EXPRESS WRITTEN PERMISSION OF HAL COMPUTER SYSTEMS
 *  INTERNATIONAL, LTD.
 *  
 *                          RESTRICTED RIGHTS LEGEND
 *  Use, duplication, or disclosure by the Government is subject
 *  to the restrictions as set forth in subparagraph (c)(l)(ii)
 *  of the Rights in Technical Data and Computer Software clause
 *  at DFARS 252.227-7013.
 *  
 *           HAL COMPUTER SYSTEMS INTERNATIONAL, LTD.
 *                   1315 Dell Avenue
 *                   Campbell, CA  95008
 */
/*==============================================================================*/ 
/* Hal Denali Trace Driven Simulator						*/
/* Copyright (c) 1994 HaL Computer Systems, Inc.                      		*/
/*==============================================================================*/
/* be_config.c									*/
/*										*/
/*==============================================================================*/
/* Revision history								*/
/*										*/
/* Date	     Who	   	Description					*/ 
/* --------  -----------      	-----------					*/
/* 10/30/94  Per Forssell  	Initial defintion				*/
/*										*/
/*==============================================================================*/

#include	<stdio.h>
#include	<stdlib.h>
#include	<string.h>

#include	"globdef.h"
#include	"back_end.h"

/*==============================================================================*/

FILE *open_file () ;

/*==============================================================================*/

extern FILE     *config_file ;	/* pointer to the configuration file		*/
extern char	config_file_name[] ; 	/* configuration file name		*/

extern qslot	*iq_head ;	/* first entry in the instruction que		*/
extern qslot	*iq_tail ;	/* last entry in the instruction que		*/
extern qslot	*lastv_qslot ;	/* last valid qslot				*/
extern qslot	*firstf_qslot ;	/* first free qslot				*/

extern bconf	be_conf ;	/* back end configuration structure		*/

/*==============================================================================*/

/* The back end configuration file parser reads the configuration file, it over
   writes the default back end configuration parameters with parameters found in
   the configuration file. */

void
parse_be_config_file ()
	
{
  char	line[MAX_LINE] ;	/* input line from configuration file		*/
  char	param[MAX_NAME] ;	/* configuration parameter name			*/

  char	eu_name[MAX_NAME] ;  	/* execution unit name				*/
  char	eu_type[MAX_LINE] ;	/* execution unit type 				*/
  char	eu_type_entry[1] ;	/* execution unit type entry (0/1)		*/
  ulong eu_type_vec ;		/* execution unit type packed into a
				   double word					*/

  char	*it_name_array[] = I_TYPE_NAMES ;	/* instruction type name array	*/
  char	it_name[MAX_NAME] ;	/* instruction type name			*/

  int	eu_def = 0 ;		/* indicates that execution units are defined	*/
  int	type_cnt ;		/* instruction type counter			*/

  config_file = open_file (config_file_name, "r") ;

  while (fgets (line, MAX_LINE, config_file) != NULL) {

    strcpy (param, strtok (line, WHITE_SPACE)) ;

    if (!strcmp (param, "NO_OF_QSLOTS")) 
      be_conf.no_of_qslots = atoi (strtok (NULL, WHITE_SPACE)) ;
    if (!strcmp (param, "NO_OF_WINDOWS"))
      be_conf.no_of_windows = atoi (strtok (NULL, WHITE_SPACE)) ;

    if (!strcmp (param, "ISSUE_DELAY"))
      be_conf.issue_delay = atoi (strtok (NULL, WHITE_SPACE)) ;
    if (!strcmp (param, "DISP_DELAY"))
      be_conf.disp_delay = atoi (strtok (NULL, WHITE_SPACE)) ;
    if (!strcmp (param, "COMPL_DELAY"))
      be_conf.compl_delay = atoi (strtok (NULL, WHITE_SPACE)) ;

    if (!strcmp (param, "COMMIT_CNT"))
      be_conf.commit_cnt = atoi (strtok (NULL, WHITE_SPACE)) ;

    if (!strcmp (param, "NO_OF_AGEN_BUFS"))
      be_conf.no_of_agen_bufs = atoi (strtok (NULL, WHITE_SPACE)) ;
	
    if (!strcmp (param, "EU")) {
      if (!eu_def) {
	be_conf.no_of_eunits = 1 ;
	eu_def = 1 ;
      }
      else
	be_conf.no_of_eunits++ ;

      strcpy(eu_name, strtok(NULL, WHITE_SPACE)) ;
      
      eu_type_vec = 0 ;
      for (type_cnt = 0; type_cnt < MAX_ITYPES; type_cnt++) {
	strcpy(eu_type_entry, strtok(NULL, WHITE_SPACE)) ;
	if (eu_type_entry == NULL) {
	  eu_type_vec = eu_type_vec >> (MAX_ITYPES - type_cnt) ;
	  break ;
	}
	eu_type_vec = eu_type_vec | 
	  (atoi(eu_type_entry) << ((MAX_ITYPES - 1) - type_cnt)) ;
      }
    
      strcpy(be_conf.eu_type[be_conf.no_of_eunits - 1].name, eu_name) ;
      be_conf.eu_type[be_conf.no_of_eunits - 1].type_vec = eu_type_vec ;
    }

    if (!strcmp (param, "IT")) {
      strcpy (it_name, strtok (NULL, WHITE_SPACE)) ;
      for (type_cnt = 0; type_cnt < MAX_ITYPES; type_cnt++) {
	if (!strcmp(it_name, it_name_array[type_cnt])) {
	  be_conf.i_type[type_cnt].ilat  = atoi (strtok (NULL, WHITE_SPACE)) ;
	  be_conf.i_type[type_cnt].ithrp = atoi (strtok (NULL, WHITE_SPACE)) ;
	}    
      }
    }
  }

  fclose (config_file) ;
 
}

/*==============================================================================*/

/* The back end configuration routine reads in all back end configuration
   parameters from the configuration file. Default parameter values are used for
   parameters not defined in the configuration file or if no configuration file
   is specified.

   The routine constructs and initializes the instruction que and the execution
   unit que. */

void
be_config (ireq	 *fetch_r ,	/* instruction fetch request			*/
	   lsreq *ls_r )	/* load store request				*/
	
{

  int	   qslot_cnt ;	/* que slot counter					*/
  int	   eunit_cnt ;	/* execution unit counter				*/
  int      itype_cnt ;	/* instruction type count				*/
  int      lsreq_cnt ;  /* load store request count				*/
  int      brres_cnt ;  /* branch resolution count			 	*/
  int      type_cnt ;   /* instruction type count				*/

  qslot	  *qs ;		/* que slot pointer					*/
  qentry  *c_qe ;	/* current que entry pointer				*/
  qentry  *n_qe ;	/* next que entry pointer				*/

  lsreq   *lsr ;	/* load store request					*/
  brlis	  *brr ;	/* branch resolution list entry				*/
  eunit   *eu ;		/* execution unit pointer				*/

  char    *eu_list[] = DEF_EU_TYPE ;	/* default execution unit type list	*/
  char	  eu_name[MAX_NAME] ;   	/* execution unit name			*/
  char	  eu_type[MAX_LINE] ;		/* execution unit type 			*/
  char	  eu_type_entry[1] ;		/* execution unit type entry (0/1)	*/
  ulong   eu_type_vec ;			/* execution unit type packed into a
					   double word				*/

  itype	  i_type[]   = DEF_I_TYPE ;	/* default instruction type list	*/

  /* Initialize all configuration parameters to their default values. */

  be_conf.no_of_qslots    = DEF_NO_OF_QSLOTS ;
  be_conf.no_of_windows   = DEF_NO_OF_WINDOWS ;
  be_conf.no_of_eunits    = DEF_NO_OF_EUNITS ;
  be_conf.issue_delay     = DEF_ISSUE_DELAY ;
  be_conf.disp_delay      = DEF_DISP_DELAY ;
  be_conf.compl_delay     = DEF_COMPL_DELAY ;
  be_conf.commit_cnt      = DEF_COMMIT_CNT ;
  be_conf.no_of_agen_bufs = DEF_NO_OF_AGEN_BUFS ;

  /* Initialize the configuration execution unit types to their defaults. */

  for (eunit_cnt = 0; eunit_cnt < DEF_NO_OF_EUNITS; eunit_cnt++) {

    strcpy(eu_type,  eu_list[eunit_cnt]) ;
    strcpy(eu_name, strtok(eu_type, " ")) ;
    
    eu_type_vec = 0 ;
    for (type_cnt = 0; type_cnt < MAX_ITYPES; type_cnt++) {
      strcpy(eu_type_entry, strtok(NULL, " ")) ;
      if (eu_type_entry == NULL) {
	eu_type_vec = eu_type_vec >> (MAX_ITYPES - type_cnt) ;
	break ;
      }
      eu_type_vec = eu_type_vec | 
	           (atoi(eu_type_entry) << ((MAX_ITYPES - 1) - type_cnt)) ;
    }
    
    strcpy(be_conf.eu_type[eunit_cnt].name, eu_name) ;
    be_conf.eu_type[eunit_cnt].type_vec = eu_type_vec ;

  }

  /* Initialize the configuration instruction types to their defaults. */

  for (itype_cnt = 0; itype_cnt < MAX_ITYPES; itype_cnt++) 
    be_conf.i_type[itype_cnt] = i_type[itype_cnt] ;

  /* Read parameters from configuration file. */

  if (strcmp (config_file_name, ""))
    parse_be_config_file () ;


  /* Construct and initialize the instruction que */

  for (qslot_cnt = 0; qslot_cnt < be_conf.no_of_qslots; qslot_cnt++) {

    c_qe = (qentry *) malloc (sizeof (qentry)) ;
    n_qe = (qentry *) malloc (sizeof (qentry)) ;
    qs   = (qslot *)  malloc (sizeof (qslot)) ;

    c_qe->state = NOT_VALID ;
    n_qe->state = NOT_VALID ;

    qs->no = qslot_cnt ;
    qs->c_entry = c_qe ;
    qs->n_entry = n_qe ;

    if (qslot_cnt == 0) {
      qs->prev_qslot = NULL ;
      qs->next_qslot = NULL ;

      iq_head      = qs ;
      iq_tail      = qs ;
      firstf_qslot = qs ;
      lastv_qslot  = NULL ;
    }
    else {
      iq_tail->next_qslot = qs ;
      qs->prev_qslot = iq_tail ;
      qs->next_qslot = NULL ;

      iq_tail      = qs ;
      lastv_qslot  = NULL ;
    }
  }
   

  /* Construct and initialize the execution unit que */

  for (eunit_cnt = 0; eunit_cnt < be_conf.no_of_eunits; eunit_cnt++) {

    if (eunit_cnt == 0) {
      eu_head = (eunit *) malloc (sizeof (eunit)) ;
      eu = eu_head ;
    }
    else {
      eu->next_eunit = (eunit *) malloc (sizeof (eunit)) ;
      eu = eu->next_eunit ;
    }

    strcpy (eu->name, be_conf.eu_type[eunit_cnt].name) ;
    eu->type_vec   = be_conf.eu_type[eunit_cnt].type_vec ;
    eu->free_cycle = 0 ;
    eu->use_stats  = 0 ;
    eu->next_eunit = NULL ;

  }


  /* Initialize the fetch and load store requests */

  fetch_r->cnt  = 0 ;

  brr = fetch_r->compl ;
  for (brres_cnt = 0; brres_cnt < MAX_BR_COMPL; brres_cnt++) {
    brr->valid = NOT_VALID ;
    brr++ ;
  }

  brr = fetch_r->commit ;
  for (brres_cnt = 0; brres_cnt < MAX_BR_COMMIT; brres_cnt++) {
    brr->valid = NOT_VALID ;
    brr++ ;
  }

  lsr = ls_r ;
  for (lsreq_cnt = 0; lsreq_cnt < MAX_LS_DISP; lsreq_cnt++) { 
    lsr->valid = NOT_VALID ;
    lsr++ ;
  }

} /* end be_config */

/*==============================================================================*/
