/*
 *  $Id: be_stats.c,v 1.15 1995/10/02 22:37:04 aiman Exp $
 *  
 *  Copyright (c) 1994 HAL Computer Systems International, Ltd.
 *  All rights reserved.  Unpublished -- rights reserved under
 *  the Copyright Laws of the United States.  USE OF A COPYRIGHT
 *  NOTICE IS PRECAUTIONARY ONLY AND DOES NOT IMPLY PUBLICATION
 *  OR DISCLOSURE.
 *  
 *  THIS SOFTWARE CONTAINS CONFIDENTIAL INFORMATION AND TRADE
 *  SECRETS OF HAL COMPUTER SYSTEMS INTERNATIONAL, LTD.  USE,
 *  DISCLOSURE, OR REPRODUCTION IS PROHIBITED WITHOUT THE
 *  PRIOR EXPRESS WRITTEN PERMISSION OF HAL COMPUTER SYSTEMS
 *  INTERNATIONAL, LTD.
 *  
 *                          RESTRICTED RIGHTS LEGEND
 *  Use, duplication, or disclosure by the Government is subject
 *  to the restrictions as set forth in subparagraph (c)(l)(ii)
 *  of the Rights in Technical Data and Computer Software clause
 *  at DFARS 252.227-7013.
 *  
 *           HAL COMPUTER SYSTEMS INTERNATIONAL, LTD.
 *                   1315 Dell Avenue
 *                   Campbell, CA  95008
 */
/*==============================================================================*/ 
/* Hal Denali Trace Driven Simulator						*/
/* Copyright (c) 1994 HaL Computer Systems, Inc.                      		*/
/*==============================================================================*/
/* be_stats.c									*/
/*										*/
/*==============================================================================*/
/* Revision history								*/
/*										*/
/* Date	     Who	   	Description					*/ 
/* --------  -----------      	-----------					*/
/* 10/30/94  Per Forssell  	Initial defintion				*/
/*										*/
/*==============================================================================*/

#include	<stdio.h>
#include	<string.h>
#include	<ctype.h>
#include	<time.h>

#include	"globdef.h"
#include	"back_end.h"
#include	"config_front_end.h"

/*==============================================================================*/

extern 	char	 stats_file_name[] ;
extern 	char	 trace_file_name[] ;
extern 	char	 config_file_name[] ;

extern	FILE	*stats_file ;

/*==============================================================================*/

extern	gstats   global_stats ;

extern  gconf	 gl_conf ;

/*==============================================================================*/

/* time_stamp returns a string with the current time and date. */

char*
current_time ()
	
{
  time_t c_time ;		/* current clock time				*/

  time(&c_time) ;
  return (strtok (ctime(&c_time), "\n")) ;
    
} /* end current_time */

/*==============================================================================*/

/* The back end statistics routine will print out statistics data for the back
   end portion of Denali. */

void
be_stats ()
	
{
  char	 str[9][85] ;		/* header strings				*/
  char   filler[9][2] ;		/* header filler characters			*/ 

  eunit *eu ;			/* pointer to execution unit list		*/

  char	 *itype_names[] = I_TYPE_NAMES ;

  int    str_cnt ;		/* string character counter			*/
  int	 eu_cnt ;		/* execution unit counter			*/
  int	 iq_cnt ;		/* instruction que length count			*/
  int	 fetch_cnt ;		/* instruction fetch count			*/
  int	 itype_cnt ;		/* instruction type count			*/

  char   *eu_header_str[] =  
  {"#                                     F   I   R     B                         A  #", 
   "#               U O                   S F S I E     R F F F   F F   I   I L S D  #", 
   "#               N T                   T L T L T C J A C M S F M M F M I M O H D  #", 
   "#               I H S                 O O O O U A M N O O Q D A U A O D U G I S  #", 
   "#               M E Y                 R A R A R L P C N V R I D L D V I L I F U  #", 
   "#               P R S                 E D E D N L L H V E T V D T D E V T C T B  #"} ; 

  sprintf (str[0],  "#=========================================================") ;
  sprintf (str[1],  "# HaL Denali Trace Driven Simulator v%s", TSIM_VERSION_NO) ;
  sprintf (str[2],  "#=========================================================") ;
  sprintf (str[3],  "# File   : %s", stats_file_name) ; 
  sprintf (str[4],  "# Date   : %s", current_time ()) ;
  sprintf (str[5],  "# Trace  : %s", trace_file_name) ;
  if (strcmp(config_file_name, ""))
    sprintf (str[6],  "# Config : %s", config_file_name) ;
  else
    sprintf (str[6],  "# Config : default") ;
  sprintf (str[7],  "#=========================================================") ;

  sprintf (filler[0],  "=") ;
  sprintf (filler[1],  " ") ;
  sprintf (filler[2],  "=") ;
  sprintf (filler[3],  " ") ;
  sprintf (filler[4],  " ") ;
  sprintf (filler[5],  " ") ;
  sprintf (filler[6],  " ") ;
  sprintf (filler[7],  "=") ;
  sprintf (filler[8],  "-") ;

  for (str_cnt = 0; str_cnt <= 7; str_cnt++) {
    while (strlen (str[str_cnt]) <= 80) {
      strcat (str[str_cnt], filler[str_cnt]) ;
    }
  }

  strcat (str[0], "#") ;
  strcat (str[1], "#") ;
  strcat (str[2], "#") ;
  strcat (str[3], "#") ;
  strcat (str[4], "#") ;
  strcat (str[5], "#") ;
  strcat (str[6], "#") ;
  strcat (str[7], "#") ;

  for (str_cnt = 0; str_cnt <= 7; str_cnt++) {
    fprintf (stats_file, "%s \n", str[str_cnt]) ;
  }

  fprintf (stats_file, "\n") ;

  sprintf (str[8], "# Processor front end configuration parameters ") ;
  while (strlen (str[8]) <= 80) 
    strcat (str[8], filler[8]) ;
  strcat (str[8], "#") ;
  fprintf (stats_file, "%s \n", str[8]) ;
  fprintf (stats_file, "\n") ;
  
  fprintf (stats_file, "ICACHE_SIZE      %8d Kbytes\n", fl_conf.icache_size*fl_conf.idline/512) ;
  fprintf (stats_file, "DCACHE_SIZE      %8d Kbytes\n", fl_conf.dcache_size*fl_conf.idline/512) ;
  fprintf (stats_file, "UCACHE_SIZE      %8d Mbytes\n", fl_conf.ucache_size*fl_conf.uline/4/256/1024) ;
  fprintf (stats_file, "UCACHE_BLOCKING  %8d \n", fl_conf.ucache_blocking);
  fprintf (stats_file, "UCACHE_PIPESTAGE %8d \n", fl_conf.ucache_pipestages);
  fprintf (stats_file, "UCACHE_PMISSES   %8d \n", fl_conf.ucache_pmisses);
  fprintf (stats_file, "ISTREAM_BUFFER   %8d \n", fl_conf.istream_buffer);
  fprintf (stats_file, "MTLB_SIZE        %8d \n", fl_conf.tlb_size) ;
  fprintf (stats_file, "mITLB_SIZE       %8d \n", fl_conf.mItlb_size) ;
  fprintf (stats_file, "mDTLB_SIZE       %8d \n", fl_conf.mDtlb_size) ;
  fprintf (stats_file, "D0_LAT           %8d \n", gl_conf.d0_lat) ;    
  fprintf (stats_file, "UNIFIED_LAT      %8d \n", fl_conf.unified_lat) ;
  fprintf (stats_file, "MEMORY_LAT       %8d \n", fl_conf.memory_lat) ;
  fprintf (stats_file, "MTLB_LAT         %8d \n", fl_conf.mtlb_lat) ;
  fprintf (stats_file, "MICRO_LAT        %8d \n", fl_conf.micro_lat) ;
  fprintf (stats_file, "NO_INT_LV        %8d \n", fl_conf.no_int_lv) ;
  fprintf (stats_file, "MAX_UT_BUF       %8d \n", fl_conf.max_ut_buf) ;
  fprintf (stats_file, "NO_DRTY_BUF      %8d \n", fl_conf.drty_buf) ;
  fprintf (stats_file, "ULINESIZE        %8d \n", fl_conf.uline) ;
  fprintf (stats_file, "IDLINESIZE       %8d \n", fl_conf.idline);
  fprintf (stats_file, "ULINESHFT        %8d \n", fl_conf.ushft) ;
  fprintf (stats_file, "IDLINESHFT       %8d \n", fl_conf.idshft);




  fprintf (stats_file, "\n") ;

  sprintf (str[8], "# Processor back end configuration parameters ") ;
  while (strlen (str[8]) <= 80) 
    strcat (str[8], filler[8]) ;
  strcat (str[8], "#") ;
  fprintf (stats_file, "%s \n", str[8]) ;
  fprintf (stats_file, "\n") ;
  
  fprintf (stats_file, "NO_OF_QSLOTS     %8d \n", be_conf.no_of_qslots) ;
  fprintf (stats_file, "NO_OF_WINDOWS    %8d \n", be_conf.no_of_windows) ;
  fprintf (stats_file, "NO_OF_LS_COMPL   %8d \n", gl_conf.no_of_ls_compl) ;
  fprintf (stats_file, "NO_OF_LS_REQ     %8d \n", gl_conf.no_of_ls_req) ;
  fprintf (stats_file, "NO_OF_LS_BUFS    %8d \n", gl_conf.no_of_ls_bufs) ;
  fprintf (stats_file, "NO_OF_AGEN_BUFS  %8d \n", be_conf.no_of_agen_bufs) ;

  fprintf (stats_file, "\n") ;

  fprintf (stats_file, "ISSUE_DELAY      %8d \n", be_conf.issue_delay) ;    
  fprintf (stats_file, "DISP_DELAY       %8d \n", be_conf.disp_delay) ;     
  fprintf (stats_file, "COMPL_DELAY      %8d \n", be_conf.compl_delay) ;    

  fprintf (stats_file, "\n") ;

  fprintf (stats_file, "COMMIT_CNT       %8d \n", be_conf.commit_cnt) ;    
  fprintf (stats_file, "IFETCH_CNT       %8d \n", gl_conf.ifetch_cnt) ;     

  fprintf (stats_file, "\n") ;

  if (gl_conf.mem_model == RMO)
    fprintf (stats_file, "MEM_MODEL            \"RMO\"\n") ;    
  if (gl_conf.mem_model == STO)
    fprintf (stats_file, "MEM_MODEL            \"STO\"\n") ;    
  if (gl_conf.mem_model == LSO)
    fprintf (stats_file, "MEM_MODEL            \"LSO\"\n") ;    
  if (gl_conf.mem_model == LSX)
    fprintf (stats_file, "MEM_MODEL            \"LSX\"\n") ;    

  fprintf (stats_file, "\n") ;

  sprintf (str[8], "# Execution unit definitions ") ;
  while (strlen (str[8]) <= 80) 
    strcat (str[8], filler[8]) ;
  strcat (str[8], "#") ;
  fprintf (stats_file, "%s \n", str[8]) ;
  fprintf (stats_file, "\n") ;
  for (str_cnt = 0; str_cnt <= 5; str_cnt++) {
    fprintf (stats_file, "%s \n", eu_header_str[str_cnt]) ;
  }

  fprintf (stats_file, "\n") ;

  for (eu_cnt = 0; eu_cnt < be_conf.no_of_eunits; eu_cnt++) {
    fprintf (stats_file, "EU %-12s", be_conf.eu_type[eu_cnt].name) ;
    for (itype_cnt = 0; itype_cnt < MAX_ITYPES; itype_cnt++)
      fprintf (stats_file, " %1d", 
           (be_conf.eu_type[eu_cnt].type_vec >> 
	   (MAX_ITYPES - itype_cnt - 1)) & 0x1) ;
  fprintf (stats_file, "\n") ;
  }

  fprintf (stats_file, "\n") ;

  sprintf (str[8], "# Instruction type parameters ") ;
  while (strlen (str[8]) <= 80) 
    strcat (str[8], filler[8]) ;
  strcat (str[8], "#") ;
  fprintf (stats_file, "%s \n", str[8]) ;
  fprintf (stats_file, "\n") ;
  fprintf (stats_file, "#                   LATENCY    THRUPUT \n") ;

  fprintf (stats_file, "\n") ;

  for (itype_cnt = 0; itype_cnt < MAX_ITYPES; itype_cnt++)
    fprintf (stats_file, "IT %-10s %10d %10d \n",
	     itype_names[itype_cnt],
	     be_conf.i_type[itype_cnt].ilat,
	     be_conf.i_type[itype_cnt].ithrp ) ;

  fprintf (stats_file, "\n") ;

  sprintf (str[8], "# Processor statistics ") ;
  while (strlen (str[8]) <= 80) 
    strcat (str[8], filler[8]) ;
  strcat (str[8], "#") ;
  fprintf (stats_file, "%s \n", str[8]) ;
  fprintf (stats_file, "\n") ;

  if (!global_stats.no_of_instructions) {
     printf ("ERROR: NO INSTRUCTIONS HAVE BEEN EXECUTED!\n");
     return;
  }
  fprintf (stats_file, "number_of_instructions          : %8ld \n", 
	   global_stats.no_of_instructions) ;
  fprintf (stats_file, "number_of_annulled_instructions : %8ld \n", 
	   global_stats.no_of_annulled_insts) ;
  fprintf (stats_file, "number_of_cycles                : %8ld \n", 
	   global_stats.no_of_cycles) ;
  fprintf (stats_file, "average_ipc (with annuled)      : %8.2f \n", 
	   (float) global_stats.no_of_instructions/
	   (float) global_stats.no_of_cycles) ;
  fprintf (stats_file, "average_ipc (without annuled)   : %8.2f \n", 
	   (float) (global_stats.no_of_instructions - 
		    global_stats.no_of_annulled_insts)/
	   (float) global_stats.no_of_cycles) ;
  fprintf (stats_file, "average_instruction_life_time   : %8ld \n", 
	   global_stats.cum_inst_life_time/global_stats.no_of_instructions) ;

  fprintf (stats_file, "\n") ;

  fprintf (stats_file, "cti_fetch_stalls                : %8ld \n",
	   global_stats.cti_fetch_stall) ;

  fprintf (stats_file, "\n") ;

  fprintf (stats_file, "average_branch_completion       : %8ld (all)\n",
	   global_stats.br_res_comp/
	   global_stats.br_comp_refs) ;

 if (global_stats.br_ip_comp_refs != 0)
  fprintf (stats_file, "average_branch_completion       : %8ld (incorrect pred)\n",
	   global_stats.br_ip_res_comp/
	   global_stats.br_ip_comp_refs) ;
 else
  fprintf (stats_file, "average_branch_completion 1.0\n");


  fprintf (stats_file, "average_branch_commit           : %8ld (all)\n",
	   global_stats.br_res_cmmt/
	   global_stats.br_cmmt_refs) ;
 if (global_stats.br_ip_cmmt_refs !=0)
  fprintf (stats_file, "average_branch_commit           : %8ld (incorrect pred)\n",
	   global_stats.br_ip_res_cmmt/
	   global_stats.br_ip_cmmt_refs) ;
 else
  fprintf (stats_file, "average_branch_commit  1.0\n");

  fprintf (stats_file, "\n") ;

  for (iq_cnt = 0; iq_cnt <= be_conf.no_of_qslots; iq_cnt++)
    fprintf (stats_file, "cti_fetch_stall_iq_length       : %8ld  (length %d) \n",
	     global_stats.cti_fetch_stall_iq_stats[iq_cnt] ,
	     iq_cnt) ;             

  fprintf (stats_file, "\n") ;

  for (fetch_cnt = 0; fetch_cnt <= gl_conf.ifetch_cnt; fetch_cnt++)
    fprintf (stats_file, "fetch_count                     : %8ld  (fetch %d) \n",
	     global_stats.fetch_stats[fetch_cnt] ,
	     fetch_cnt) ;

  fprintf (stats_file, "\n") ;

  for (iq_cnt = 0; iq_cnt <= be_conf.no_of_qslots; iq_cnt++)
    fprintf (stats_file, "instruction_que_length          : %8ld  (length %d) \n",
	     global_stats.iq_stats[iq_cnt] ,
	     iq_cnt) ;             

  fprintf (stats_file, "\n") ;

  eu = eu_head ;
  while (eu != NULL) {
    fprintf (stats_file, "execution_unit_use              : %8ld  (%s) \n", 
	     eu->use_stats ,
	     eu->name) ;
    eu = eu->next_eunit ;
  }
	   
  fprintf (stats_file, "\n\n") ;

  for (itype_cnt = 0; itype_cnt < MAX_ITYPES; itype_cnt++)
    fprintf (stats_file, "instruction_type_counts         : %8ld  (%s) \n",
	     global_stats.itype_stats[itype_cnt] ,
	     itype_names[itype_cnt]) ;             

  fprintf (stats_file, "\n") ;

} /* end be_stats */
