/*
 *  $Id: config_front_end.h,v 1.12 1995/10/02 22:37:06 aiman Exp $
 *  
 *  Copyright (c) 1994 HAL Computer Systems International, Ltd.
 *  All rights reserved.  Unpublished -- rights reserved under
 *  the Copyright Laws of the United States.  USE OF A COPYRIGHT
 *  NOTICE IS PRECAUTIONARY ONLY AND DOES NOT IMPLY PUBLICATION
 *  OR DISCLOSURE.
 *  
 *  THIS SOFTWARE CONTAINS CONFIDENTIAL INFORMATION AND TRADE
 *  SECRETS OF HAL COMPUTER SYSTEMS INTERNATIONAL, LTD.  USE,
 *  DISCLOSURE, OR REPRODUCTION IS PROHIBITED WITHOUT THE
 *  PRIOR EXPRESS WRITTEN PERMISSION OF HAL COMPUTER SYSTEMS
 *  INTERNATIONAL, LTD.
 *  
 *                          RESTRICTED RIGHTS LEGEND
 *  Use, duplication, or disclosure by the Government is subject
 *  to the restrictions as set forth in subparagraph (c)(l)(ii)
 *  of the Rights in Technical Data and Computer Software clause
 *  at DFARS 252.227-7013.
 *  
 *           HAL COMPUTER SYSTEMS INTERNATIONAL, LTD.
 *                   1315 Dell Avenue
 *                   Campbell, CA  95008
 */   


typedef struct {
  unsigned pc;
  unsigned iw;
  unsigned ea;
  unsigned opcode : 8;
  unsigned flags  : 8;    /* taken branch, annulled instr, etc.*/
  unsigned        : 16;   /* padding; reserved for future use */
} traceinfo;


typedef struct {
  unsigned set[2];
  unsigned valid;
  unsigned dirty;
  unsigned repl; /* The last bit indicates the LRU set */
} d0_cache;

typedef struct {
  unsigned set[2];
  unsigned valid;
  unsigned repl; /* The last bit indicates the LRU set */
} i0_cache;

typedef struct {
  unsigned set[4];
  unsigned valid;
  unsigned repl; /* The L S Byte packs the order of most recently to
		    least recently used set */
} main_tlb;

typedef struct {
  unsigned va;
  unsigned valid;
} micro_tlb;


/* NIRMAL: WE NEED TO FIX front_end.c SO IT WOULD WORK AFTER WE REMOVE THE
   VALUES FROM THE FOLLOWING ENUMERATIONS */

typedef enum {READ=0, WRITE=1, WRITEBACK=2, RETAG, RFILL, WFILL, SECOND_HALF,
	     SNOOP, SNOOP_WB, SNOOP_INV}  cmd_type ;

typedef enum {INVALID=0, VALID, INPROGRESS, COMPLETED} access_state ;

typedef enum {I0REQ=0, D0REQ=1, TRREQ, U2, SC} requestor;

typedef struct {
  unsigned va;        /* pc address of the instruction */
  unsigned ea;        /* effective address of ld/st    */

  requestor requestor;/* requesting resource */
  cmd_type  cmd;      /* access type */
  access_state state; /* request state */
  int      transfer_no; /* used to do subline transfers */

  int      multi_cycle; /* used by u2 only */
  unsigned qslot_no;
  unsigned time;
  int      cycles;
  unsigned iw;
  unsigned long recode[2];
} wait_buffer;

typedef struct {
  int max_ut_buf;
  int icache_size;
  int dcache_size;
  int ucache_size;
  int ucache_blocking;
  int ucache_pipestages;
  int ucache_pmisses;
  int ucache_trans_size;
  int istream_buffer;
  int mItlb_size;
  int mDtlb_size;
  int tlb_size;
  int no_int_lv;
  int unified_lat;
  int memory_lat;
  int mtlb_lat;
  int micro_lat;
  int drty_buf;
  int uline;
  int idline;
  int ushft;
  int idshft;
} fconf;


extern unsigned long clk_cycle;
extern cmd_type   command;
extern unsigned long  va;
extern unsigned       limit;
extern int            size;
extern unsigned       icache_miss;
extern unsigned       icache_refs;
extern unsigned       icache_repl;
extern unsigned       dcache_miss;
extern unsigned       dcache_refs;
extern unsigned       dcache_st;
extern unsigned       dcache_repl;
extern unsigned       dcache_drepl;

extern unsigned       ucache_repl;
extern unsigned       ucache_drepl;
extern unsigned       ucache_miss;
extern unsigned       ucache_refs;
extern unsigned       ucache_i_refs;
extern unsigned       ucache_d_refs;
extern unsigned       ucache_wb_refs;
extern unsigned       ucache_retag_refs;
extern unsigned       ucache_l1_snoops;
extern unsigned       ucache_l1_dsnoops;
extern unsigned       ucache_avg_refs;
extern unsigned       ucache_repl;
extern unsigned       ucache_drepl;
extern unsigned       ucache_avg_misses;
extern unsigned       cpu_stall_imiss;
extern unsigned       cpu_stall_dmiss;
extern unsigned       stream_buffer_hit;           
extern unsigned       stream_buffer_miss;           

extern unsigned       mItlb_miss;
extern unsigned       mItlb_refs;
extern unsigned       mItlb_repl;
extern unsigned       mDtlb_miss;
extern unsigned       mDtlb_refs;
extern unsigned       mDtlb_repl;
extern unsigned       tlb_miss;
extern unsigned       tlb_refs;
extern unsigned       tlb_repl;
extern unsigned       mItlb_pntr;
extern unsigned       mDtlb_pntr;
extern unsigned       drty_hits;
extern unsigned       drty_flshs;
extern unsigned         ln_brks;
extern wait_buffer     *wt_main_tlb;
extern wait_buffer     *wt_unifd_cache;
extern wait_buffer     *wt_snoop_cache;
extern wait_buffer     *wt_iu_cache;
extern wait_buffer     *wt_ldst_req;
extern unsigned        *interleave;
extern i0_cache        *icache;
extern d0_cache        *dcache;
extern main_tlb        *tlb;
extern micro_tlb       *mItlb;
extern micro_tlb       *mDtlb;
extern traceinfo       *tp;
extern traceinfo       *mrkr;
extern fconf           fl_conf;
extern int Imiss_flag, Dmiss_flag,
mImiss_flag, mDmiss_flag, Tmiss_flag, Umiss_flag;
extern int dirty_va, Dmiss_repl;
extern unsigned long jmpl_ret_refs;
extern unsigned long jmpl_retl_refs;
extern unsigned long jmpl_ret_np;
extern unsigned long jmpl_retl_np;
extern unsigned long jmpl_ret_mp;
extern unsigned long jmpl_retl_mp;
extern unsigned long call_refs;
extern unsigned long jmpl_call_refs;
extern unsigned long stack_spill;
extern unsigned long br_uncond_refs;
extern unsigned long br_cond_refs;
extern unsigned long br_cond_taken;
extern unsigned long br_corpred_on_mis;
extern unsigned long br_mispred_on_hit;
extern unsigned long br_mispred_on_mis;
extern unsigned warmup_state;
extern void init_l1();
