/*
 * $Id: dasm.c,v 1.6 1993/01/12 00:28:43 ae Exp $
 *
 * Copyright (c) 1991 HaL Computer Systems, Inc.  All rights reserved.
 * UNPUBLISHED -- rights reserved under the Copyright Laws of the United
 * States.  Use of a copyright notice is precautionary only and does not
 * imply publication or disclosure.
 * 
 * This software contains confidential information and trade secrets of HaL
 * Computer Systems, Inc.  Use, disclosure, or reproduction is prohibited
 * without the prior express written permission of HaL Computer Systems, Inc.
 * 
 *                         RESTRICTED RIGHTS LEGEND
 * Use, duplication, or disclosure by the Government is subject to
 * restrictions as set forth in subparagraph (c)(l)(ii) of the Rights in
 * Technical Data and Computer Software clause at DFARS 252.227-7013.
 *                        HaL Computer Systems, Inc.
 *                  1315 Dell Avenue, Campbell, CA  95008
 * 
 */

#include	<stdio.h>
#include	"syntax.h"
#include	"assert.h"
#include	"dasm.h"


#define	DEBUG_LOOKUP	(debug and 0x01)
#define	DEBUG_WALK	(debug and 0x02)
#define	TAB		"    "
#define	ARG_OFFSET	10

Local int	debug=	0;

#define	SYMBOLIC	1
#define	STATENO		7

typedef struct NSTATE {
    UINT	stateN[STATENO];
} NSTATE, *NSTATEp;

Local NSTATE state_size= { 4, 8, 64, 16, 512, 16, 16 };

typedef enum ACTION {
    NA=		-1,
    ADDR,
    IREGD,
    IREG1,
    IREG2,
    FREGD,
    FREGDX,	/* Xtended. Low order bits maps registers f32-f63 */
    FREG1,
    FREG1X,
    FREG2,
    FREG2X,
    FREG3,
    FREG3X,
    CREG0,
    CREG1,
    CREG2,
    FSR,
    CSR,
    YREG,
    ASTATEREGS,
    ASTATEREGD,
    STATEREGS,
    STATEREGD,
    PSR,
    FQ,
    CQ,
    ASR,
    WIM,
    TBR,
    ASI,
    IREGIMMED6,
    IREGIMMED10,
    IREGIMMED11,
    IREGIMMED13,
    LABEL22,	/*  22 bit displacement				  */
    LABEL19,	/*  19 bit displacement				  */
    BLABEL,	/*  branch on integer register label disp2+disp14 */
    BREG1,	/*  branch on integer register  value		  */
    CALL,
    SETHI,
    UNIMP,
    ANNUL,
    PREDICT,
    CC,
    FCC,
    DONERETRY,	/* return  from trap kludge	*/
    TRAP,
    TRAPCC,
    SHIFTX,
    IFXCC,
    FMOVECC
} ACTION;

#define STATE_OP      	state.stateN[0] 
#define STATE_OP2     	state.stateN[1]
#define STATE_OP3     	state.stateN[2] 
#define STATE_COND    	state.stateN[3]
#define STATE_OPF     	state.stateN[4] 
#define STATE_VARIATION	state.stateN[5] 
#define STATE_MCOND	state.stateN[6] 
#define	STATE(i)	state.stateN[(i)]

#define	ACTION_SIZE	4

typedef struct INSTRUCTION_TABLE {
    char	*name;
    ACTION	action[ACTION_SIZE];
    NSTATE	state;
} INSTRUCTION_TABLE,  *INSTRUCTION_TABLEp;
     
typedef struct NTREE NTREE, *NTREEp;

struct NTREE {
    NSTATE		state;
    INSTRUCTION_TABLEp	entry;
    NTREEp		next;
};

NTREE	instruction_tree;

#ifdef SPARC
typedef union TEMPLATE {

    UINT	instruction;

    struct {
	UINT	op:2;
	UINT	disp30:30;
    } Format1;

    struct {
	UINT	op:2;
	UINT	rd:5;
	UINT	op2:3;
	UINT	imm22:22;
    } Format2;

    struct {
	UINT	op:2;
	UINT	a:1;
	UINT	cond:4;
	UINT	op2:3;
	UINT	disp22:22;
    } Format3;

    struct {
	UINT	op:2;
	UINT	rd:5;
	UINT	op3:6;
	UINT	rs1:5;
	UINT	i:1;
	UINT	asi:8;
	UINT	rs2:5;
    } Format4;

    struct {
	UINT	op:2;
	UINT	rd:5;
	UINT	op3:6;
	UINT	rs1:5;
	UINT	i:1;
	UINT	simm13:13;
    } Format5;

    struct {
	UINT	op:2;
	UINT	rd:5;
	UINT	op3:6;
	UINT	rs1:5;
	UINT	opf:9;
	UINT	rs2:5;
    } Format6;

    struct {
	UINT	op:2;
	UINT	reserved:1;
	UINT	cond:4;
	UINT	op3:6;
	UINT	rs1:5;
	UINT	opf:9;
	UINT	rs2:5;
    } Format7;

    struct {
	UINT	op:2;
	UINT	:1;
	UINT	cond:4;
	UINT	op3:6;
	UINT	rs1:5;
	UINT	i:1;
	UINT	cc:1;
	UINT	:5;
	UINT	trap:7;
    } Format8;

    /*
     *	Specifically for HaL mad (multiply add)	
     */
    struct {  
	UINT	op:2;
	UINT	:5;
	UINT	op3:6;
	UINT	rs1:5;
	UINT	rs3:5;
	UINT	variation:4;
	UINT	rs2:5;
    } Format9;

    /*
     *	V9:	Branch with prediction
     */
    struct {
	UINT	op:2;
	UINT	a:1;
	UINT	cond:4;
	UINT	op2:3;
	UINT	cc:2;
	UINT	p:1;
	UINT	disp19:19;
    } Format10;

    /*
     *	V9:	Move Integer register on condition (1)
     */
    struct {
	UINT	op:2;
	UINT	rd:5;
	UINT	op3:6;
	UINT	ifcc:1;
	UINT	cond:4;
	UINT	i:1;
	UINT	cc:2;
	UINT	:6;
	UINT	rs2:5;
    } Format11;

    /*
     *	V9:	Move Integer register on condition (2)
     */
    struct {
	UINT	op:2;
	UINT	rd:5;
	UINT	op3:6;
	UINT	ifcc:1;
	UINT	cond:4;
	UINT	i:1;
	UINT	cc:2;
	UINT	simm11:11;
    } Format12;

    /*
     *	V7/9:	Shifts are overloaded !! Must look at the 'x' field
     */
    struct {
	UINT	op:2;
	UINT	rd:5;
	UINT	op3:6;
	UINT	rs1:5;
	UINT	i:1;
	UINT	x:1;
	UINT	:6;
	UINT	simm6:6;
    } Format13;

    /*
     *	V9:	Move Floating Point register on condition
     */
    struct {
	UINT	op:2;
	UINT	rd:5;
	UINT	op3:6;
	UINT	:1;
	UINT	cond:4;
	UINT	cc:3;
	UINT	opf:6;
	UINT	rs2:5;
    } Format14;

    /*
     *	V9:	Move Integer register on value of Integer register
     */
    struct {
	UINT	op:2;
	UINT	rd:5;
	UINT	op3:6;
	UINT	rs1:5;
	UINT	i:1;
	UINT	rcond:3;
	UINT	simm10:10;
    } Format15;

    /*
     *	V9:	Branch on Integer Register with prediction
     */
    struct {
	UINT	op:2;
	UINT	a:1;
	UINT	:1;
	UINT	rcond:3;
	UINT	op2:3;
	UINT	disp2:2;
	UINT	p:1;
	UINT	rs1:5;
	UINT	disp14:14;
    } Format16;


} TEMPLATE, *TEMPLATEp;
#endif

#ifdef INTEL
typedef union TEMPLATE {

    UINT	instruction;

    struct {
	UINT	disp30:30;
	UINT	op:2;
    } Format1;

    struct {
	UINT	imm22:22;
	UINT	op2:3;
	UINT	rd:5;
	UINT	op:2;
    } Format2;

    struct {
	UINT	disp22:22;
	UINT	op2:3;
	UINT	cond:4;
	UINT	a:1;
	UINT	op:2;
    } Format3;

    struct {
	UINT	rs2:5;
	UINT	asi:8;
	UINT	i:1;
	UINT	rs1:5;
	UINT	op3:6;
	UINT	rd:5;
	UINT	op:2;
    } Format4;

    struct {
	UINT	simm13:13;
	UINT	i:1;
	UINT	rs1:5;
	UINT	op3:6;
	UINT	rd:5;
	UINT	op:2;
    } Format5;

    struct {
	UINT	rs2:5;
	UINT	opf:9;
	UINT	rs1:5;
	UINT	op3:6;
	UINT	rd:5;
	UINT	op:2;
    } Format6;

    struct {
	UINT	rs2:5;
	UINT	opf:9;
	UINT	rs1:5;
	UINT	op3:6;
	UINT	cond:4;
	UINT	reserved:1;
	UINT	op:2;
    } Format7;

    struct {
	UINT	trap:7;
	UINT	:5;
	UINT	cc:1;
	UINT	i:1;
	UINT	rs1:5;
	UINT	op3:6;
	UINT	cond:4;
	UINT	:1;
	UINT	op:2;
    } Format8;

    /*
     *	Specifically for HaL mad (multiply add)	
     */
    struct {  
	UINT	rs2:5;
	UINT	variation:4;
	UINT	rs3:5;
	UINT	rs1:5;
	UINT	op3:6;
	UINT	:5;
	UINT	op:2;
    } Format9;

    /*
     *	V9:	Branch with prediction
     */
    struct {
	UINT	disp19:19;
	UINT	p:1;
	UINT	cc:2;
	UINT	op2:3;
	UINT	cond:4;
	UINT	a:1;
	UINT	op:2;
    } Format10;

    /*
     *	V9:	Move Integer register on condition (1)
     */
    struct {
	UINT	rs2:5;
	UINT	:6;
	UINT	cc:2;
	UINT	i:1;
	UINT	cond:4;
	UINT	ifcc:1;
	UINT	op3:6;
	UINT	rd:5;
	UINT	op:2;
    } Format11;

    /*
     *	V9:	Move Integer register on condition (2)
     */
    struct {
	UINT	simm11:11;
	UINT	cc:2;
	UINT	i:1;
	UINT	cond:4;
	UINT	ifcc:1;
	UINT	op3:6;
	UINT	rd:5;
	UINT	op:2;
    } Format12;

    /*
     *	V7/9:	Shifts are overloaded !! Must look at the 'x' field
     */
    struct {
	UINT	simm6:6;
	UINT	:6;
	UINT	x:1;
	UINT	i:1;
	UINT	rs1:5;
	UINT	op3:6;
	UINT	rd:5;
	UINT	op:2;
    } Format13;

    /*
     *	V9:	Move Floating Point register on condition
     */
    struct {
	UINT	rs2:5;
	UINT	opf:6;
	UINT	cc:3;
	UINT	cond:4;
	UINT	:1;
	UINT	op3:6;
	UINT	rd:5;
	UINT	op:2;
    } Format14;

    /*
     *	V9:	Move Integer register on value of Integer register
     */
    struct {
	UINT	simm10:10;
	UINT	rcond:3;
	UINT	i:1;
	UINT	rs1:5;
	UINT	op3:6;
	UINT	rd:5;
	UINT	op:2;
    } Format15;

    /*
     *	V9:	Branch on Integer Register with prediction
     */
    struct {
	UINT	disp14:14;
	UINT	rs1:5;
	UINT	p:1;
	UINT	disp2:2;
	UINT	op2:3;
	UINT	rcond:3;
	UINT	:1;
	UINT	a:1;
	UINT	op:2;
    } Format16;


} TEMPLATE, *TEMPLATEp;
#endif

#define	NONE			(NA),(NA),(NA),(NA)
#define	F0(x)			(x), (NA),(NA),(NA)
#define	F0_F1(x,y)		(x), (y), (NA),(NA)
#define	F0_F2(x,y)		(x), (NA),(y), (NA)

#define	F0_F1_F2(x,y,z)		(x), (y), (z), (NA)
#define	F0_F1_F2_F3(w,x,y,z)	(w), (x), (y), (z)

/*
**	The implicit order in the state is
**		OP,  OP2,  OP3,  COND,  OPF, VARIATION,  MCOND,
*/
#define	OP(x)				(x), (NA),(NA), (NA), (NA), (NA), (NA)
#define	OP_OP2(x,y)			(x), (y), (NA), (NA), (NA), (NA), (NA)
#define	OP_OP2_COND(x,y,z)		(x), (y), (NA), (z),  (NA), (NA), (NA)

#define	OP_OP3(x,y)			(x), (NA), (y), (NA), (NA), (NA), (NA)
#define	OP_OP3_COND(x,y,z)		(x), (NA), (y), (z),  (NA), (NA), (NA)
#define	OP_OP3_OPF(x,y,z)		(x), (NA), (y), (NA), (z),  (NA), (NA)

#define	OP_OP3_MCOND(x,y,z)		(x), (NA), (y), (NA), (NA), (NA), (z)
#define	OP_OP3_OPF_MCOND(w,x,y,z)	(w), (NA), (x), (NA), (y),  (NA), (z)

#define	OP_OP3_VARIATION(x,y,z)		(x), (NA), (y), (NA), (NA), (z), (NA)

INSTRUCTION_TABLE instruction_table[] = {

"sethi",	F0_F1(SETHI, IREGD),			OP_OP2(0x0, 0x4),
"unimp",	F0(UNIMP),				OP_OP2(0x0, 0x0),
"flush",	F0(ADDR),				OP_OP3(0x2, 0x3b),
"and",		F0_F1_F2(IREG1, IREGIMMED13, IREGD),	OP_OP3(0x2, 0x01),
"andcc",	F0_F1_F2(IREG1, IREGIMMED13, IREGD),	OP_OP3(0x2, 0x11),
"andn",		F0_F1_F2(IREG1, IREGIMMED13, IREGD),	OP_OP3(0x2, 0x05),
"andncc",	F0_F1_F2(IREG1, IREGIMMED13, IREGD),	OP_OP3(0x2, 0x15),
"or",		F0_F1_F2(IREG1, IREGIMMED13, IREGD),	OP_OP3(0x2, 0x02),
"orcc",		F0_F1_F2(IREG1, IREGIMMED13, IREGD),	OP_OP3(0x2, 0x12),
"orn",		F0_F1_F2(IREG1, IREGIMMED13, IREGD),	OP_OP3(0x2, 0x06),
"orncc",	F0_F1_F2(IREG1, IREGIMMED13, IREGD),	OP_OP3(0x2, 0x16),
"xor",		F0_F1_F2(IREG1, IREGIMMED13, IREGD),	OP_OP3(0x2, 0x03),
"xorcc",	F0_F1_F2(IREG1, IREGIMMED13, IREGD),	OP_OP3(0x2, 0x13),
"xnor",		F0_F1_F2(IREG1, IREGIMMED13, IREGD),	OP_OP3(0x2, 0x07),
"xnorcc",	F0_F1_F2(IREG1, IREGIMMED13, IREGD),	OP_OP3(0x2, 0x17),
"add",		F0_F1_F2(IREG1, IREGIMMED13, IREGD),	OP_OP3(0x2, 0x0),
"addcc",	F0_F1_F2(IREG1, IREGIMMED13, IREGD),	OP_OP3(0x2, 0x10),
"addx",		F0_F1_F2(IREG1, IREGIMMED13, IREGD),	OP_OP3(0x2, 0x08),
"addxcc",	F0_F1_F2(IREG1, IREGIMMED13, IREGD),	OP_OP3(0x2, 0x18),
"taddcc",	F0_F1_F2(IREG1, IREGIMMED13, IREGD),	OP_OP3(0x2, 0x20),
"taddcctv",	F0_F1_F2(IREG1, IREGIMMED13, IREGD),	OP_OP3(0x2, 0x22),
"sub",		F0_F1_F2(IREG1, IREGIMMED13, IREGD),	OP_OP3(0x2, 0x04),
"subcc",	F0_F1_F2(IREG1, IREGIMMED13, IREGD),	OP_OP3(0x2, 0x14),
"subx",		F0_F1_F2(IREG1, IREGIMMED13, IREGD),	OP_OP3(0x2, 0x0c),
"subxcc",	F0_F1_F2(IREG1, IREGIMMED13, IREGD),	OP_OP3(0x2, 0x1c),
"tsubcc",	F0_F1_F2(IREG1, IREGIMMED13, IREGD),	OP_OP3(0x2, 0x21),
"tsubcctv",	F0_F1_F2(IREG1, IREGIMMED13, IREGD),	OP_OP3(0x2, 0x23),
"mulscc",	F0_F1_F2(IREG1, IREGIMMED13, IREGD),	OP_OP3(0x2, 0x24),
"umul",		F0_F1_F2(IREG1, IREGIMMED13, IREGD),	OP_OP3(0x2, 0x0a),
"smul",		F0_F1_F2(IREG1, IREGIMMED13, IREGD),	OP_OP3(0x2, 0x0b),
"umulcc",	F0_F1_F2(IREG1, IREGIMMED13, IREGD),	OP_OP3(0x2, 0x1a),
"smulcc",	F0_F1_F2(IREG1, IREGIMMED13, IREGD),	OP_OP3(0x2, 0x1b),
"mulx",		F0_F1_F2(IREG1, IREGIMMED13, IREGD),	OP_OP3(0x2, 0x09),
"udivx",	F0_F1_F2(IREG1, IREGIMMED13, IREGD),	OP_OP3(0x2, 0x0d),
"sdivx",	F0_F1_F2(IREG1, IREGIMMED13, IREGD),	OP_OP3(0x2, 0x2d),
"save",		F0_F1_F2(IREG1, IREGIMMED13, IREGD),	OP_OP3(0x2, 0x3c),
"restore",	F0_F1_F2(IREG1, IREGIMMED13, IREGD),	OP_OP3(0x2, 0x3d),

"ba",		F0_F1(ANNUL, LABEL22),		OP_OP2_COND(0x0, 0x2, 0x8),
"bn",		F0_F1(ANNUL, LABEL22),		OP_OP2_COND(0x0, 0x2, 0x0),
"bne",		F0_F1(ANNUL, LABEL22),		OP_OP2_COND(0x0, 0x2, 0x9),
"be",		F0_F1(ANNUL, LABEL22),		OP_OP2_COND(0x0, 0x2, 0x1),
"bg",		F0_F1(ANNUL, LABEL22),		OP_OP2_COND(0x0, 0x2, 0xa),
"ble",		F0_F1(ANNUL, LABEL22),		OP_OP2_COND(0x0, 0x2, 0x2),
"bge",		F0_F1(ANNUL, LABEL22),		OP_OP2_COND(0x0, 0x2, 0xb),
"bl",		F0_F1(ANNUL, LABEL22),		OP_OP2_COND(0x0, 0x2, 0x3),
"bgu",		F0_F1(ANNUL, LABEL22),		OP_OP2_COND(0x0, 0x2, 0xc),
"bleu",		F0_F1(ANNUL, LABEL22),		OP_OP2_COND(0x0, 0x2, 0x4),
"bcc",		F0_F1(ANNUL, LABEL22),		OP_OP2_COND(0x0, 0x2, 0xd),
"bcs",		F0_F1(ANNUL, LABEL22),		OP_OP2_COND(0x0, 0x2, 0x5),
"bpos",		F0_F1(ANNUL, LABEL22),		OP_OP2_COND(0x0, 0x2, 0xe),
"bneg",		F0_F1(ANNUL, LABEL22),		OP_OP2_COND(0x0, 0x2, 0x6),
"bvc",		F0_F1(ANNUL, LABEL22),		OP_OP2_COND(0x0, 0x2, 0xf),
"bvs",		F0_F1(ANNUL, LABEL22),		OP_OP2_COND(0x0, 0x2, 0x7),

"fba",		F0_F1(ANNUL, LABEL22),		OP_OP2_COND(0x0, 0x6, 0x8),
"fbn",		F0_F1(ANNUL, LABEL22),		OP_OP2_COND(0x0, 0x6, 0x0),
"fbu",		F0_F1(ANNUL, LABEL22),		OP_OP2_COND(0x0, 0x6, 0x7),
"fbg",		F0_F1(ANNUL, LABEL22),		OP_OP2_COND(0x0, 0x6, 0x6),
"fbug",		F0_F1(ANNUL, LABEL22),		OP_OP2_COND(0x0, 0x6, 0x5),
"fbl",		F0_F1(ANNUL, LABEL22),		OP_OP2_COND(0x0, 0x6, 0x4),
"fbul",		F0_F1(ANNUL, LABEL22),		OP_OP2_COND(0x0, 0x6, 0x3),
"fblg",		F0_F1(ANNUL, LABEL22),		OP_OP2_COND(0x0, 0x6, 0x2),
"fbne",		F0_F1(ANNUL, LABEL22),		OP_OP2_COND(0x0, 0x6, 0x1),
"fbe",		F0_F1(ANNUL, LABEL22),		OP_OP2_COND(0x0, 0x6, 0x9),
"fbue",		F0_F1(ANNUL, LABEL22),		OP_OP2_COND(0x0, 0x6, 0xa),
"fbge",		F0_F1(ANNUL, LABEL22),		OP_OP2_COND(0x0, 0x6, 0xb),
"fbuge",	F0_F1(ANNUL, LABEL22),		OP_OP2_COND(0x0, 0x6, 0xc),
"fble",		F0_F1(ANNUL, LABEL22),		OP_OP2_COND(0x0, 0x6, 0xd),
"fbule",	F0_F1(ANNUL, LABEL22),		OP_OP2_COND(0x0, 0x6, 0xe),
"fbo",		F0_F1(ANNUL, LABEL22),		OP_OP2_COND(0x0, 0x6, 0xf),
"cba",		F0_F1(ANNUL, LABEL22),		OP_OP2_COND(0x0, 0x7, 0x8),
"cbn",		F0_F1(ANNUL, LABEL22),		OP_OP2_COND(0x0, 0x7, 0x0),
"cb3",		F0_F1(ANNUL, LABEL22),		OP_OP2_COND(0x0, 0x7, 0x7),
"cb2",		F0_F1(ANNUL, LABEL22),		OP_OP2_COND(0x0, 0x7, 0x6),
"cb23",		F0_F1(ANNUL, LABEL22),		OP_OP2_COND(0x0, 0x7, 0x5),
"cb1",		F0_F1(ANNUL, LABEL22),		OP_OP2_COND(0x0, 0x7, 0x4),
"cb13",		F0_F1(ANNUL, LABEL22),		OP_OP2_COND(0x0, 0x7, 0x3),
"cb12",		F0_F1(ANNUL, LABEL22),		OP_OP2_COND(0x0, 0x7, 0x2),
"cb123",	F0_F1(ANNUL, LABEL22),		OP_OP2_COND(0x0, 0x7, 0x1),
"cb0",		F0_F1(ANNUL, LABEL22),		OP_OP2_COND(0x0, 0x7, 0x9),
"cb03",		F0_F1(ANNUL, LABEL22),		OP_OP2_COND(0x0, 0x7, 0xa),
"cb02",		F0_F1(ANNUL, LABEL22),		OP_OP2_COND(0x0, 0x7, 0xb),
"cb023",	F0_F1(ANNUL, LABEL22),		OP_OP2_COND(0x0, 0x7, 0xc),
"cb01",		F0_F1(ANNUL, LABEL22),		OP_OP2_COND(0x0, 0x7, 0xd),
"cb013",	F0_F1(ANNUL, LABEL22),		OP_OP2_COND(0x0, 0x7, 0xe),
"cb012",	F0_F1(ANNUL, LABEL22),		OP_OP2_COND(0x0, 0x7, 0xf),

"call",		F0(CALL),			OP(0x1),

"jmpl",		F0_F2(ADDR, IREGD),		OP_OP3(0x2, 0x38),
"rett",		F0(ADDR),			OP_OP3(0x2, 0x39),

"ta",		F0_F1(TRAPCC, TRAP),		OP_OP3_COND(0x2, 0x3a, 0x8),
"tn",		F0_F1(TRAPCC, TRAP),		OP_OP3_COND(0x2, 0x3a, 0x0),
"tne",		F0_F1(TRAPCC, TRAP),		OP_OP3_COND(0x2, 0x3a, 0x9),
"te",		F0_F1(TRAPCC, TRAP),		OP_OP3_COND(0x2, 0x3a, 0x1),
"tg",		F0_F1(TRAPCC, TRAP),		OP_OP3_COND(0x2, 0x3a, 0xa),
"tle",		F0_F1(TRAPCC, TRAP),		OP_OP3_COND(0x2, 0x3a, 0x2),
"tge",		F0_F1(TRAPCC, TRAP),		OP_OP3_COND(0x2, 0x3a, 0xb),
"tl",		F0_F1(TRAPCC, TRAP),		OP_OP3_COND(0x2, 0x3a, 0x3),
"tgu",		F0_F1(TRAPCC, TRAP),		OP_OP3_COND(0x2, 0x3a, 0xc),
"tleu",		F0_F1(TRAPCC, TRAP),		OP_OP3_COND(0x2, 0x3a, 0x4),
"tcc",		F0_F1(TRAPCC, TRAP),		OP_OP3_COND(0x2, 0x3a, 0xd),
"tcs",		F0_F1(TRAPCC, TRAP),		OP_OP3_COND(0x2, 0x3a, 0x5),
"tpos",		F0_F1(TRAPCC, TRAP),		OP_OP3_COND(0x2, 0x3a, 0xe),
"tneg",		F0_F1(TRAPCC, TRAP),		OP_OP3_COND(0x2, 0x3a, 0x6),
"tvc",		F0_F1(TRAPCC, TRAP),		OP_OP3_COND(0x2, 0x3a, 0xf),
"tvs",		F0_F1(TRAPCC, TRAP),		OP_OP3_COND(0x2, 0x3a, 0x7),

"rd",		F0_F2(ASTATEREGS, IREGD),	OP_OP3(0x2, 0x28),
"rdpr",		F0_F2(STATEREGS, IREGD),	OP_OP3(0x2, 0x2a),

/*********************************************************
**	Conflict with rd TBR (priviledged V7 instruction)
"rd",		F0_F2(TBR, IREGD),		OP_OP3(0x2, 0x2b),
*********************************************************/
"flushw",	NONE,				OP_OP3(0x2, 0x2b),

"wr",		F0_F1_F2(IREG1, IREGIMMED13, ASTATEREGD), OP_OP3(0x2, 0x30),
"wr",		F0_F1_F2(IREG1, IREGIMMED13, PSR),	OP_OP3(0x2, 0x31),
"wrpr",		F0_F1_F2(IREG1, IREGIMMED13, STATEREGD), OP_OP3(0x2, 0x32),
"wr",		F0_F1_F2(IREG1, IREGIMMED13, TBR),	OP_OP3(0x2, 0x33),

"fitos",	F0_F1(FREG2, FREGD),		OP_OP3_OPF(0x2, 0x34, 0x0c4),
"fitod",	F0_F1(FREG2, FREGDX),		OP_OP3_OPF(0x2, 0x34, 0x0c8),
"fitoq",	F0_F1(FREG2, FREGDX),		OP_OP3_OPF(0x2, 0x34, 0x0cc),
"fstoi",	F0_F1(FREG2, FREGD),		OP_OP3_OPF(0x2, 0x34, 0x0d1),
"fdtoi",	F0_F1(FREG2X, FREGD),		OP_OP3_OPF(0x2, 0x34, 0x0d2),
"fqtoi",	F0_F1(FREG2X, FREGD),		OP_OP3_OPF(0x2, 0x34, 0x0d3),
"fstod",	F0_F1(FREG2, FREGDX),		OP_OP3_OPF(0x2, 0x34, 0x0c9),
"fstoq",	F0_F1(FREG2, FREGDX),		OP_OP3_OPF(0x2, 0x34, 0x0cd),
"fdtos",	F0_F1(FREG2X, FREGD),		OP_OP3_OPF(0x2, 0x34, 0x0c6),
"fdtoq",	F0_F1(FREG2X, FREGD),		OP_OP3_OPF(0x2, 0x34, 0x0ce),
"fqtos",	F0_F1(FREG2X, FREGD),		OP_OP3_OPF(0x2, 0x34, 0x0c7),
"fqtod",	F0_F1(FREG2X, FREGDX),		OP_OP3_OPF(0x2, 0x34, 0x0cb),
"fmovs",	F0_F1(FREG2, FREGD),		OP_OP3_OPF(0x2, 0x34, 0x001),
"fnegs",	F0_F1(FREG2, FREGD),		OP_OP3_OPF(0x2, 0x34, 0x005),
"fabss",	F0_F1(FREG2, FREGD),		OP_OP3_OPF(0x2, 0x34, 0x009),
"fsqrts",	F0_F1(FREG2, FREGD),		OP_OP3_OPF(0x2, 0x34, 0x029),
"fsqrtd",	F0_F1(FREG2X, FREGDX),		OP_OP3_OPF(0x2, 0x34, 0x02a),
"fsqrtx",	F0_F1(FREG2X, FREGDX),		OP_OP3_OPF(0x2, 0x34, 0x02b),
"fadds",	F0_F1_F2(FREG1, FREG2, FREGD),	OP_OP3_OPF(0x2, 0x34, 0x041),
"faddd",	F0_F1_F2(FREG1X, FREG2X, FREGDX),	OP_OP3_OPF(0x2, 0x34, 0x042),
"faddq",	F0_F1_F2(FREG1X, FREG2X, FREGDX),	OP_OP3_OPF(0x2, 0x34, 0x043),
"fsubs",	F0_F1_F2(FREG1, FREG2, FREGD),	OP_OP3_OPF(0x2, 0x34, 0x045),
"fsubd",	F0_F1_F2(FREG1X, FREG2X, FREGDX),	OP_OP3_OPF(0x2, 0x34, 0x046),
"fsubq",	F0_F1_F2(FREG1X, FREG2X, FREGDX),	OP_OP3_OPF(0x2, 0x34, 0x047),
"fmuls",	F0_F1_F2(FREG1, FREG2, FREGD),	OP_OP3_OPF(0x2, 0x34, 0x049),
"fmuld",	F0_F1_F2(FREG1X, FREG2X, FREGDX),	OP_OP3_OPF(0x2, 0x34, 0x04a),
"fmulq",	F0_F1_F2(FREG1X, FREG2X, FREGD),	OP_OP3_OPF(0x2, 0x34, 0x04b),
"fsmuld",	F0_F1_F2(FREG1X, FREG2X, FREGD),	OP_OP3_OPF(0x2, 0x34, 0x069),
"fdmulq",	F0_F1_F2(FREG1X, FREG2X, FREGD),	OP_OP3_OPF(0x2, 0x34, 0x06e),
"fdivs",	F0_F1_F2(FREG1, FREG2, FREGD),	OP_OP3_OPF(0x2, 0x34, 0x04d),
"fdivd",	F0_F1_F2(FREG1X, FREG2X, FREGDX),	OP_OP3_OPF(0x2, 0x34, 0x04e),
"fdivq",	F0_F1_F2(FREG1X, FREG2X, FREGDX),	OP_OP3_OPF(0x2, 0x34, 0x04f),
"fcmps",	F0_F1(FREG1, FREG2),		OP_OP3_OPF(0x2, 0x35, 0x051),
"fcmpd",	F0_F1(FREG1X, FREG2X),		OP_OP3_OPF(0x2, 0x35, 0x052),
"fcmpq",	F0_F1(FREG1X, FREG2X),		OP_OP3_OPF(0x2, 0x35, 0x053),
"fcmpes",	F0_F1(FREG1, FREG2),		OP_OP3_OPF(0x2, 0x35, 0x055),
"fcmped",	F0_F1(FREG1X, FREG2X),		OP_OP3_OPF(0x2, 0x35, 0x056),
"fcmpeq",	F0_F1(FREG1X, FREG2X),		OP_OP3_OPF(0x2, 0x35, 0x057),

"ldsb",		F0_F2(ADDR, IREGD),		OP_OP3(0x3, 0x09),
"ldsh",		F0_F2(ADDR, IREGD),		OP_OP3(0x3, 0x0a),
"ldub",		F0_F2(ADDR, IREGD),		OP_OP3(0x3, 0x01),
"lduw",		F0_F2(ADDR, IREGD),		OP_OP3(0x3, 0x02),
"ld",		F0_F2(ADDR, IREGD),		OP_OP3(0x3, 0x0),
"ldd",		F0_F2(ADDR, IREGD),		OP_OP3(0x3, 0x03),

"ldsba",	F0_F1_F2(ADDR, ASI, IREGD),	OP_OP3(0x3, 0x19),
"ldsha",	F0_F1_F2(ADDR, ASI, IREGD),	OP_OP3(0x3, 0x1a),
"lduba",	F0_F1_F2(ADDR, ASI, IREGD),	OP_OP3(0x3, 0x11),
"lduwa",	F0_F1_F2(ADDR, ASI, IREGD),	OP_OP3(0x3, 0x12),
"lda",		F0_F1_F2(ADDR, ASI, IREGD),	OP_OP3(0x3, 0x10),
"ldda",		F0_F1_F2(ADDR, ASI, IREGD),	OP_OP3(0x3, 0x13),

"ldf",		F0_F2(ADDR, FREGD),		OP_OP3(0x3, 0x20),
"lddf",		F0_F2(ADDR, FREGDX),		OP_OP3(0x3, 0x23),
"ldfsr",	F0_F2(ADDR, FSR),		OP_OP3(0x3, 0x21),
"ldc",		F0_F2(ADDR, CREG0),		OP_OP3(0x3, 0x30),
"lddc",		F0_F2(ADDR, CREG0),		OP_OP3(0x3, 0x33),
"ldcsr",	F0_F2(ADDR, CSR),		OP_OP3(0x3, 0x31),
"stb",		F0_F1(IREGD, ADDR),		OP_OP3(0x3, 0x05),
"sth",		F0_F1(IREGD, ADDR),		OP_OP3(0x3, 0x06),
"st",		F0_F1(IREGD, ADDR),		OP_OP3(0x3, 0x04),
"std",		F0_F1(IREGD, ADDR),		OP_OP3(0x3, 0x07),

"stba",		F0_F1_F2(IREGD, ADDR, ASI),	OP_OP3(0x3, 0x15),
"stha",		F0_F1_F2(IREGD, ADDR, ASI),	OP_OP3(0x3, 0x16),
"sta",		F0_F1_F2(IREGD, ADDR, ASI),	OP_OP3(0x3, 0x14),
"stda",		F0_F1_F2(IREGD, ADDR, ASI),	OP_OP3(0x3, 0x17),

"stf",		F0_F1(FREGD, ADDR),		OP_OP3(0x3, 0x24),
"stdf",		F0_F1(FREGDX, ADDR),		OP_OP3(0x3, 0x27),
"stfsr",	F0_F1(FSR, ADDR),		OP_OP3(0x3, 0x25),
"stdfq",	F0_F1(FQ, ADDR),		OP_OP3(0x3, 0x26),
"stc",		F0_F1(CREG0, ADDR),		OP_OP3(0x3, 0x34),
"stdc",		F0_F1(CREG0, ADDR),		OP_OP3(0x3, 0x37),
"stcsr",	F0_F1(CSR, ADDR),		OP_OP3(0x3, 0x35),
"stdcq",	F0_F1(CQ, ADDR),		OP_OP3(0x3, 0x36),
"ldstub",	F0_F2(ADDR, IREGD),		OP_OP3(0x3, 0x0d),
"ldstuba",	F0_F1_F2(ADDR, ASI, IREGD),	OP_OP3(0x3, 0x1d),
"swap",		F0_F2(ADDR, IREGD),		OP_OP3(0x3, 0x0f),
"swapa",	F0_F1_F2(ADDR, ASI, IREGD),	OP_OP3(0x3, 0x1f),

/*
**	Current V9 extensions and HaL specific Instructions
*/
"fmadds",	F0_F1_F2_F3(FREG1,FREG2,FREG3,FREGD),	OP_OP3_VARIATION(0x2, 0x36, 0x1),
"fmaddd",	F0_F1_F2_F3(FREG1X,FREG2X,FREG3X,FREGDX),	OP_OP3_VARIATION(0x2, 0x36, 0x2),
"fmsubs",	F0_F1_F2_F3(FREG1,FREG2,FREG3,FREGD),	OP_OP3_VARIATION(0x2, 0x36, 0x5),
"fmsubd",	F0_F1_F2_F3(FREG1X,FREG2X,FREG3X,FREGDX),	OP_OP3_VARIATION(0x2, 0x36, 0x6),
"fnmadds",	F0_F1_F2_F3(FREG1,FREG2,FREG3,FREGD),	OP_OP3_VARIATION(0x2, 0x36, 0xd),
"fnmaddd",	F0_F1_F2_F3(FREG1X,FREG2X,FREG3X,FREGDX),	OP_OP3_VARIATION(0x2, 0x36, 0xe),
"fnmsubs",	F0_F1_F2_F3(FREG1,FREG2,FREG3,FREGD),	OP_OP3_VARIATION(0x2, 0x36, 0x9),
"fnmsubd",	F0_F1_F2_F3(FREG1X,FREG2X,FREG3X,FREGDX),	OP_OP3_VARIATION(0x2, 0x36, 0xa),

"ba",		F0_F1_F2_F3(ANNUL, PREDICT, CC, LABEL19),		OP_OP2_COND(0x0, 0x1, 0x8),
"bn",		F0_F1_F2_F3(ANNUL, PREDICT, CC, LABEL19),		OP_OP2_COND(0x0, 0x1, 0x0),
"bne",		F0_F1_F2_F3(ANNUL, PREDICT, CC, LABEL19),		OP_OP2_COND(0x0, 0x1, 0x9),
"be",		F0_F1_F2_F3(ANNUL, PREDICT, CC, LABEL19),		OP_OP2_COND(0x0, 0x1, 0x1),
"bg",		F0_F1_F2_F3(ANNUL, PREDICT, CC, LABEL19),		OP_OP2_COND(0x0, 0x1, 0xa),
"ble",		F0_F1_F2_F3(ANNUL, PREDICT, CC, LABEL19),		OP_OP2_COND(0x0, 0x1, 0x2),
"bge",		F0_F1_F2_F3(ANNUL, PREDICT, CC, LABEL19),		OP_OP2_COND(0x0, 0x1, 0xb),
"bl",		F0_F1_F2_F3(ANNUL, PREDICT, CC, LABEL19),		OP_OP2_COND(0x0, 0x1, 0x3),
"bgu",		F0_F1_F2_F3(ANNUL, PREDICT, CC, LABEL19),		OP_OP2_COND(0x0, 0x1, 0xc),
"bleu",		F0_F1_F2_F3(ANNUL, PREDICT, CC, LABEL19),		OP_OP2_COND(0x0, 0x1, 0x4),
"bcc",		F0_F1_F2_F3(ANNUL, PREDICT, CC, LABEL19),		OP_OP2_COND(0x0, 0x1, 0xd),
"bcs",		F0_F1_F2_F3(ANNUL, PREDICT, CC, LABEL19),		OP_OP2_COND(0x0, 0x1, 0x5),
"bpos",		F0_F1_F2_F3(ANNUL, PREDICT, CC, LABEL19),		OP_OP2_COND(0x0, 0x1, 0xe),
"bneg",		F0_F1_F2_F3(ANNUL, PREDICT, CC, LABEL19),		OP_OP2_COND(0x0, 0x1, 0x6),
"bvc",		F0_F1_F2_F3(ANNUL, PREDICT, CC, LABEL19),		OP_OP2_COND(0x0, 0x1, 0xf),
"bvs",		F0_F1_F2_F3(ANNUL, PREDICT, CC, LABEL19),		OP_OP2_COND(0x0, 0x1, 0x7),

"fba",		F0_F1_F2_F3(ANNUL, PREDICT, FCC, LABEL19),	OP_OP2_COND(0x0, 0x5, 0x8),
"fbn",		F0_F1_F2_F3(ANNUL, PREDICT, FCC, LABEL19),	OP_OP2_COND(0x0, 0x5, 0x0),
"fbne",		F0_F1_F2_F3(ANNUL, PREDICT, FCC, LABEL19),	OP_OP2_COND(0x0, 0x5, 0x9),
"fbe",		F0_F1_F2_F3(ANNUL, PREDICT, FCC, LABEL19),	OP_OP2_COND(0x0, 0x5, 0x1),
"fbg",		F0_F1_F2_F3(ANNUL, PREDICT, FCC, LABEL19),	OP_OP2_COND(0x0, 0x5, 0xa),
"fble",		F0_F1_F2_F3(ANNUL, PREDICT, FCC, LABEL19),	OP_OP2_COND(0x0, 0x5, 0x2),
"fbge",		F0_F1_F2_F3(ANNUL, PREDICT, FCC, LABEL19),	OP_OP2_COND(0x0, 0x5, 0xb),
"fbl",		F0_F1_F2_F3(ANNUL, PREDICT, FCC, LABEL19),	OP_OP2_COND(0x0, 0x5, 0x3),
"fbgu",		F0_F1_F2_F3(ANNUL, PREDICT, FCC, LABEL19),	OP_OP2_COND(0x0, 0x5, 0xc),
"fbleu",	F0_F1_F2_F3(ANNUL, PREDICT, FCC, LABEL19),	OP_OP2_COND(0x0, 0x5, 0x4),
"fbcc",		F0_F1_F2_F3(ANNUL, PREDICT, FCC, LABEL19),	OP_OP2_COND(0x0, 0x5, 0xd),
"fbcs",		F0_F1_F2_F3(ANNUL, PREDICT, FCC, LABEL19),	OP_OP2_COND(0x0, 0x5, 0x5),
"fbpos",	F0_F1_F2_F3(ANNUL, PREDICT, FCC, LABEL19),	OP_OP2_COND(0x0, 0x5, 0xe),
"fbneg",	F0_F1_F2_F3(ANNUL, PREDICT, FCC, LABEL19),	OP_OP2_COND(0x0, 0x5, 0x6),
"fbvc",		F0_F1_F2_F3(ANNUL, PREDICT, FCC, LABEL19),	OP_OP2_COND(0x0, 0x5, 0xf),
"fbvs",		F0_F1_F2_F3(ANNUL, PREDICT, FCC, LABEL19),	OP_OP2_COND(0x0, 0x5, 0x7),

"brnz",		F0_F1_F2_F3(ANNUL, PREDICT, BREG1, BLABEL),	OP_OP2_COND(0x0, 0x3, 0x5),
"brz",		F0_F1_F2_F3(ANNUL, PREDICT, BREG1, BLABEL),	OP_OP2_COND(0x0, 0x3, 0x1),
"brpos",	F0_F1_F2_F3(ANNUL, PREDICT, BREG1, BLABEL),	OP_OP2_COND(0x0, 0x3, 0x7),
"brneg",	F0_F1_F2_F3(ANNUL, PREDICT, BREG1, BLABEL),	OP_OP2_COND(0x0, 0x3, 0x3),
"brnegz",	F0_F1_F2_F3(ANNUL, PREDICT, BREG1, BLABEL),	OP_OP2_COND(0x0, 0x3, 0x2),
"brposnz",	F0_F1_F2_F3(ANNUL, PREDICT, BREG1, BLABEL),	OP_OP2_COND(0x0, 0x3, 0x6),

"casx",		F0_F2(ADDR, IREGD),				OP_OP3(0x3, 0x2e),
"casxa",	F0_F1_F2(ADDR, ASI, IREGD),			OP_OP3(0x3, 0x3e),
"cas",		F0_F2(ADDR, IREGD),				OP_OP3(0x3, 0x2c),
"casa",		F0_F1_F2(ADDR, ASI, IREGD),			OP_OP3(0x3, 0x3c),

"done/retry",	F0(DONERETRY),					OP_OP3(0x2, 0x3e),

"ldx",		F0_F2(ADDR, IREGD),				OP_OP3(0x3, 0x0b),
"ldxa",		F0_F1_F2(ADDR, ASI, IREGD),			OP_OP3(0x3, 0x1b),
"stx",		F0_F1(IREGD, ADDR),				OP_OP3(0x3, 0x0e),
"sta",		F0_F1_F2(IREGD, ADDR, ASI),			OP_OP3(0x3, 0x1e),

"fxtos",	F0_F1(FREG2X, FREGD),				OP_OP3_OPF(0x2, 0x34, 0x084),
"fxtod",	F0_F1(FREG2X, FREGD),				OP_OP3_OPF(0x2, 0x34, 0x088),
"fxtoq",	F0_F1(FREG2X, FREGD),				OP_OP3_OPF(0x2, 0x34, 0x08c),

"fstox",	F0_F1(FREG2, FREGD),				OP_OP3_OPF(0x2, 0x34, 0x081),
"fdtox",	F0_F1(FREG2X, FREGD),				OP_OP3_OPF(0x2, 0x34, 0x082),
"fqtox",	F0_F1(FREG2X, FREGD),				OP_OP3_OPF(0x2, 0x34, 0x083),

"fmovd",	F0_F1(FREG2X, FREGDX),				OP_OP3_OPF(0x2, 0x34, 0x002),
"fnegd",	F0_F1(FREG2X, FREGDX),				OP_OP3_OPF(0x2, 0x34, 0x006),
"fabsd",	F0_F1(FREG2X, FREGDX),				OP_OP3_OPF(0x2, 0x34, 0x00a),
"fmovq",	F0_F1(FREG2X, FREGDX),				OP_OP3_OPF(0x2, 0x34, 0x003),
"fnegq",	F0_F1(FREG2X, FREGDX),				OP_OP3_OPF(0x2, 0x34, 0x007),
"fabsq",	F0_F1(FREG2X, FREGDX),				OP_OP3_OPF(0x2, 0x34, 0x00b),

"sll",		F0_F1_F2_F3(SHIFTX, IREG1, IREGIMMED6, IREGD),	OP_OP3(0x2, 0x25),
"srl",		F0_F1_F2_F3(SHIFTX, IREG1, IREGIMMED6, IREGD),	OP_OP3(0x2, 0x26),
"sra",		F0_F1_F2_F3(SHIFTX, IREG1, IREGIMMED6, IREGD),	OP_OP3(0x2, 0x27),

"mova",		F0_F1_F2(IFXCC, IREGIMMED11, IREGD),		OP_OP3_MCOND(0x2, 0x2c, 0x8),
"movn",		F0_F1_F2(IFXCC, IREGIMMED11, IREGD),		OP_OP3_MCOND(0x2, 0x2c, 0x0),
"movne",	F0_F1_F2(IFXCC, IREGIMMED11, IREGD),		OP_OP3_MCOND(0x2, 0x2c, 0x9),
"move",		F0_F1_F2(IFXCC, IREGIMMED11, IREGD),		OP_OP3_MCOND(0x2, 0x2c, 0x1),
"movg",		F0_F1_F2(IFXCC, IREGIMMED11, IREGD),		OP_OP3_MCOND(0x2, 0x2c, 0xa),
"movle",	F0_F1_F2(IFXCC, IREGIMMED11, IREGD),		OP_OP3_MCOND(0x2, 0x2c, 0x2),
"movge",	F0_F1_F2(IFXCC, IREGIMMED11, IREGD),		OP_OP3_MCOND(0x2, 0x2c, 0xb),
"movl",		F0_F1_F2(IFXCC, IREGIMMED11, IREGD),		OP_OP3_MCOND(0x2, 0x2c, 0x3),
"movgu",	F0_F1_F2(IFXCC, IREGIMMED11, IREGD),		OP_OP3_MCOND(0x2, 0x2c, 0xc),
"movleu",	F0_F1_F2(IFXCC, IREGIMMED11, IREGD),		OP_OP3_MCOND(0x2, 0x2c, 0x4),
"movcc",	F0_F1_F2(IFXCC, IREGIMMED11, IREGD),		OP_OP3_MCOND(0x2, 0x2c, 0xd),
"movcs",	F0_F1_F2(IFXCC, IREGIMMED11, IREGD),		OP_OP3_MCOND(0x2, 0x2c, 0x5),
"movpos",	F0_F1_F2(IFXCC, IREGIMMED11, IREGD),		OP_OP3_MCOND(0x2, 0x2c, 0xe),
"movneg",	F0_F1_F2(IFXCC, IREGIMMED11, IREGD),		OP_OP3_MCOND(0x2, 0x2c, 0x6),
"movvc",	F0_F1_F2(IFXCC, IREGIMMED11, IREGD),		OP_OP3_MCOND(0x2, 0x2c, 0xf),
"movvs",	F0_F1_F2(IFXCC, IREGIMMED11, IREGD),		OP_OP3_MCOND(0x2, 0x2c, 0x7),

"fmovas",	F0_F1_F2(FMOVECC, FREG2, FREGD),	OP_OP3_OPF_MCOND(0x2, 0x35, 0x1, 0x8),
"fmovns",	F0_F1_F2(FMOVECC, FREG2, FREGD),	OP_OP3_OPF_MCOND(0x2, 0x35, 0x1, 0x0),
"fmovnes",	F0_F1_F2(FMOVECC, FREG2, FREGD),	OP_OP3_OPF_MCOND(0x2, 0x35, 0x1, 0x9),
"fmoves",	F0_F1_F2(FMOVECC, FREG2, FREGD),	OP_OP3_OPF_MCOND(0x2, 0x35, 0x1, 0x1),
"fmovgs",	F0_F1_F2(FMOVECC, FREG2, FREGD),	OP_OP3_OPF_MCOND(0x2, 0x35, 0x1, 0xa),
"fmovles",	F0_F1_F2(FMOVECC, FREG2, FREGD),	OP_OP3_OPF_MCOND(0x2, 0x35, 0x1, 0x2),
"fmovges",	F0_F1_F2(FMOVECC, FREG2, FREGD),	OP_OP3_OPF_MCOND(0x2, 0x35, 0x1, 0xb),
"fmovls",	F0_F1_F2(FMOVECC, FREG2, FREGD),	OP_OP3_OPF_MCOND(0x2, 0x35, 0x1, 0x3),
"fmovgus",	F0_F1_F2(FMOVECC, FREG2, FREGD),	OP_OP3_OPF_MCOND(0x2, 0x35, 0x1, 0xc),
"fmovleus",	F0_F1_F2(FMOVECC, FREG2, FREGD),	OP_OP3_OPF_MCOND(0x2, 0x35, 0x1, 0x4),
"fmovccs",	F0_F1_F2(FMOVECC, FREG2, FREGD),	OP_OP3_OPF_MCOND(0x2, 0x35, 0x1, 0xd),
"fmovcss",	F0_F1_F2(FMOVECC, FREG2, FREGD),	OP_OP3_OPF_MCOND(0x2, 0x35, 0x1, 0x5),
"fmovposs",	F0_F1_F2(FMOVECC, FREG2, FREGD),	OP_OP3_OPF_MCOND(0x2, 0x35, 0x1, 0xe),
"fmovnegs",	F0_F1_F2(FMOVECC, FREG2, FREGD),	OP_OP3_OPF_MCOND(0x2, 0x35, 0x1, 0x6),
"fmovvcs",	F0_F1_F2(FMOVECC, FREG2, FREGD),	OP_OP3_OPF_MCOND(0x2, 0x35, 0x1, 0xf),
"fmovvss",	F0_F1_F2(FMOVECC, FREG2, FREGD),	OP_OP3_OPF_MCOND(0x2, 0x35, 0x1, 0x7),

"fmovad",	F0_F1_F2(FMOVECC, FREG2, FREGD),	OP_OP3_OPF_MCOND(0x2, 0x35, 0x2, 0x8),
"fmovnd",	F0_F1_F2(FMOVECC, FREG2, FREGD),	OP_OP3_OPF_MCOND(0x2, 0x35, 0x2, 0x0),
"fmovned",	F0_F1_F2(FMOVECC, FREG2, FREGD),	OP_OP3_OPF_MCOND(0x2, 0x35, 0x2, 0x9),
"fmoved",	F0_F1_F2(FMOVECC, FREG2, FREGD),	OP_OP3_OPF_MCOND(0x2, 0x35, 0x2, 0x1),
"fmovgd",	F0_F1_F2(FMOVECC, FREG2, FREGD),	OP_OP3_OPF_MCOND(0x2, 0x35, 0x2, 0xa),
"fmovled",	F0_F1_F2(FMOVECC, FREG2, FREGD),	OP_OP3_OPF_MCOND(0x2, 0x35, 0x2, 0x2),
"fmovged",	F0_F1_F2(FMOVECC, FREG2, FREGD),	OP_OP3_OPF_MCOND(0x2, 0x35, 0x2, 0xb),
"fmovld",	F0_F1_F2(FMOVECC, FREG2, FREGD),	OP_OP3_OPF_MCOND(0x2, 0x35, 0x2, 0x3),
"fmovgud",	F0_F1_F2(FMOVECC, FREG2, FREGD),	OP_OP3_OPF_MCOND(0x2, 0x35, 0x2, 0xc),
"fmovleud",	F0_F1_F2(FMOVECC, FREG2, FREGD),	OP_OP3_OPF_MCOND(0x2, 0x35, 0x2, 0x4),
"fmovccd",	F0_F1_F2(FMOVECC, FREG2, FREGD),	OP_OP3_OPF_MCOND(0x2, 0x35, 0x2, 0xd),
"fmovcsd",	F0_F1_F2(FMOVECC, FREG2, FREGD),	OP_OP3_OPF_MCOND(0x2, 0x35, 0x2, 0x5),
"fmovposd",	F0_F1_F2(FMOVECC, FREG2, FREGD),	OP_OP3_OPF_MCOND(0x2, 0x35, 0x2, 0xe),
"fmovnegd",	F0_F1_F2(FMOVECC, FREG2, FREGD),	OP_OP3_OPF_MCOND(0x2, 0x35, 0x2, 0x6),
"fmovvcd",	F0_F1_F2(FMOVECC, FREG2, FREGD),	OP_OP3_OPF_MCOND(0x2, 0x35, 0x2, 0xf),
"fmovvsd",	F0_F1_F2(FMOVECC, FREG2, FREGD),	OP_OP3_OPF_MCOND(0x2, 0x35, 0x2, 0x7),

"fmovaq",	F0_F1_F2(FMOVECC, FREG2, FREGD),	OP_OP3_OPF_MCOND(0x2, 0x35, 0x3, 0x8),
"fmovnq",	F0_F1_F2(FMOVECC, FREG2, FREGD),	OP_OP3_OPF_MCOND(0x2, 0x35, 0x3, 0x0),
"fmovneq",	F0_F1_F2(FMOVECC, FREG2, FREGD),	OP_OP3_OPF_MCOND(0x2, 0x35, 0x3, 0x9),
"fmoveq",	F0_F1_F2(FMOVECC, FREG2, FREGD),	OP_OP3_OPF_MCOND(0x2, 0x35, 0x3, 0x1),
"fmovgq",	F0_F1_F2(FMOVECC, FREG2, FREGD),	OP_OP3_OPF_MCOND(0x2, 0x35, 0x3, 0xa),
"fmovleq",	F0_F1_F2(FMOVECC, FREG2, FREGD),	OP_OP3_OPF_MCOND(0x2, 0x35, 0x3, 0x2),
"fmovgeq",	F0_F1_F2(FMOVECC, FREG2, FREGD),	OP_OP3_OPF_MCOND(0x2, 0x35, 0x3, 0xb),
"fmovlq",	F0_F1_F2(FMOVECC, FREG2, FREGD),	OP_OP3_OPF_MCOND(0x2, 0x35, 0x3, 0x3),
"fmovguq",	F0_F1_F2(FMOVECC, FREG2, FREGD),	OP_OP3_OPF_MCOND(0x2, 0x35, 0x3, 0xc),
"fmovleuq",	F0_F1_F2(FMOVECC, FREG2, FREGD),	OP_OP3_OPF_MCOND(0x2, 0x35, 0x3, 0x4),
"fmovccq",	F0_F1_F2(FMOVECC, FREG2, FREGD),	OP_OP3_OPF_MCOND(0x2, 0x35, 0x3, 0xd),
"fmovcsq",	F0_F1_F2(FMOVECC, FREG2, FREGD),	OP_OP3_OPF_MCOND(0x2, 0x35, 0x3, 0x5),
"fmovposq",	F0_F1_F2(FMOVECC, FREG2, FREGD),	OP_OP3_OPF_MCOND(0x2, 0x35, 0x3, 0xe),
"fmovnegq",	F0_F1_F2(FMOVECC, FREG2, FREGD),	OP_OP3_OPF_MCOND(0x2, 0x35, 0x3, 0x6),
"fmovvcq",	F0_F1_F2(FMOVECC, FREG2, FREGD),	OP_OP3_OPF_MCOND(0x2, 0x35, 0x3, 0xf),
"fmovvsq",	F0_F1_F2(FMOVECC, FREG2, FREGD),	OP_OP3_OPF_MCOND(0x2, 0x35, 0x3, 0x7),

"movrnz",	F0_F1_F2(IREG1, IREGIMMED10, IREGD),		OP_OP3_MCOND(0x2, 0x2f, 0x5),
"movrz",	F0_F1_F2(IREG1, IREGIMMED10, IREGD),		OP_OP3_MCOND(0x2, 0x2f, 0x1),
"movrpos",	F0_F1_F2(IREG1, IREGIMMED10, IREGD),		OP_OP3_MCOND(0x2, 0x2f, 0x7),
"movrneg",	F0_F1_F2(IREG1, IREGIMMED10, IREGD),		OP_OP3_MCOND(0x2, 0x2f, 0x3),
"movrnegz",	F0_F1_F2(IREG1, IREGIMMED10, IREGD),		OP_OP3_MCOND(0x2, 0x2f, 0x2),
"movrposnz",	F0_F1_F2(IREG1, IREGIMMED10, IREGD),		OP_OP3_MCOND(0x2, 0x2f, 0x6),

"fmovrnzs",	F0_F1_F2(FREG1, FREG2, FREGD),		OP_OP3_OPF_MCOND(0x2, 0x35, 0x5, 0x5),
"fmovrzs",	F0_F1_F2(FREG1, FREG2, FREGD),		OP_OP3_OPF_MCOND(0x2, 0x35, 0x5, 0x1),
"fmovrposs",	F0_F1_F2(FREG1, FREG2, FREGD),		OP_OP3_OPF_MCOND(0x2, 0x35, 0x5, 0x7),
"fmovrnegs",	F0_F1_F2(FREG1, FREG2, FREGD),		OP_OP3_OPF_MCOND(0x2, 0x35, 0x5, 0x3),
"fmovrnegzs",	F0_F1_F2(FREG1, FREG2, FREGD),		OP_OP3_OPF_MCOND(0x2, 0x35, 0x5, 0x2),
"fmovrposnzs",	F0_F1_F2(FREG1, FREG2, FREGD),		OP_OP3_OPF_MCOND(0x2, 0x35, 0x5, 0x6),

"fmovrnzd",	F0_F1_F2(FREG1, FREG2, FREGD),		OP_OP3_OPF_MCOND(0x2, 0x35, 0x6, 0x5),
"fmovrzd",	F0_F1_F2(FREG1, FREG2, FREGD),		OP_OP3_OPF_MCOND(0x2, 0x35, 0x6, 0x1),
"fmovrposd",	F0_F1_F2(FREG1, FREG2, FREGD),		OP_OP3_OPF_MCOND(0x2, 0x35, 0x6, 0x7),
"fmovrnegd",	F0_F1_F2(FREG1, FREG2, FREGD),		OP_OP3_OPF_MCOND(0x2, 0x35, 0x6, 0x3),
"fmovrnegzd",	F0_F1_F2(FREG1, FREG2, FREGD),		OP_OP3_OPF_MCOND(0x2, 0x35, 0x6, 0x2),
"fmovrposnzd",	F0_F1_F2(FREG1, FREG2, FREGD),		OP_OP3_OPF_MCOND(0x2, 0x35, 0x6, 0x6),

"fmovrnzq",	F0_F1_F2(FREG1, FREG2, FREGD),		OP_OP3_OPF_MCOND(0x2, 0x35, 0x7, 0x5),
"fmovrzq",	F0_F1_F2(FREG1, FREG2, FREGD),		OP_OP3_OPF_MCOND(0x2, 0x35, 0x7, 0x1),
"fmovrposq",	F0_F1_F2(FREG1, FREG2, FREGD),		OP_OP3_OPF_MCOND(0x2, 0x35, 0x7, 0x7),
"fmovrnegq",	F0_F1_F2(FREG1, FREG2, FREGD),		OP_OP3_OPF_MCOND(0x2, 0x35, 0x7, 0x3),
"fmovrnegzq",	F0_F1_F2(FREG1, FREG2, FREGD),		OP_OP3_OPF_MCOND(0x2, 0x35, 0x7, 0x2),
"fmovrposnzq",	F0_F1_F2(FREG1, FREG2, FREGD),		OP_OP3_OPF_MCOND(0x2, 0x35, 0x7, 0x6),

};

int instruction_table_size= sizeof(instruction_table) / sizeof(INSTRUCTION_TABLE);

char	*ireg[]= { 
	"%g0",  "%g1",  "%g2",  "%g3",  "%g4",  "%g5",  "%g6",  "%g7",
	"%o0",  "%o1",  "%o2",  "%o3",  "%o4",  "%o5",  "%sp",  "%o7",
	"%l0",  "%l1",  "%l2",  "%l3",  "%l4",  "%l5",  "%l6",  "%l7",
	"%i0",  "%i1",  "%i2",  "%i3",  "%i4",  "%i5",  "%fp",  "%i7"
};
int	ireg_table_size= sizeof(ireg) / sizeof(char *);

char	*freg[]= { 
	"%f0 ", "%f1 ", "%f2 ", "%f3 ", "%f4 ", "%f5 ", "%f6 ", "%f7 ",
	"%f8 ", "%f9 ", "%f10", "%f11", "%f12", "%f13", "%f14", "%f15",
	"%f16", "%f17", "%f18", "%f19", "%f20", "%f21", "%f22", "%f23",
	"%f24", "%f25", "%f26", "%f27", "%f28", "%f29", "%f30", "%f31",
	"%f32", "%f33", "%f34", "%f35", "%f36", "%f37", "%f38", "%f39",
	"%f40", "%f41", "%f42", "%f43", "%f44", "%f45", "%f46", "%f47",
	"%f48", "%f49", "%f50", "%f51", "%f52", "%f53", "%f54", "%f55",
	"%f56", "%f57", "%f58", "%f59", "%f60", "%f61", "%f62", "%f63"
};
int	freg_table_size= sizeof(freg) / sizeof(char *);

char	*fdreg[]= { 
	"%d0",  "%d2",  "%d4",  "%d6",  "%d8",  "%d10", "%d12", "%d14",
	"%d16", "%d18", "%d20", "%d22", "%d24", "%d26", "%d28", "%d30",
	"%d32", "%d34", "%d36", "%d38", "%d40", "%d42", "%d44", "%d46",
	"%d48", "%d50", "%d52", "%d54", "%d56", "%d58", "%d60",
};

char	*creg[]= { 
	"%c0",  "%c1",  "%c2",  "%c3",  "%c4",  "%c5",  "%c6",  "%c7",
	"%c8",  "%c9",  "%c10", "%c11", "%c12", "%c13", "%c14", "%c15",
	"%c16", "%c17", "%c18", "%c19", "%c20", "%c21", "%c22", "%c23",
	"%c24", "%c25", "%c26", "%c27", "%c28", "%c29", "%c30", "%c31"
};

char	*astatereg[]= { 
	"%y",        "%asr",       "%ccr",    "%asi",
	"%asr",      "%asr",       "%fprs",   "%asr",
	"%asr",      "%asr",       "%asr",    "%asr",
	"%asr",      "%asr",       "%asr",    "%asr",
	"%asr",      "%asr",       "%asr",    "%asr",
	"%asr",      "%asr",       "%asr",    "%asr",
	"%asr",      "%asr",       "%brk",    "%sir",
	"%faddr",    "%ftype",     "%asr",    "%scr"
};

char	*statereg[]= { 
	"%tpc",      "%tncp",     "%tstate",  "%tt", 
	"%tick",     "%tba",      "%pstate",  "%tl", 
	"%pil",      "%cwp",      "%cansave", "%canrestore",
	"%cleanwin", "%otherwin", "%wstate",  "%fpq",
	"%pr",       "%pr",       "%pr",      "%pr",
	"%pr",       "%pr",       "%pr",      "%pr",
	"%pr",       "%pr",       "%pr",      "%pr",
	"%pr",       "%pr",       "%pr",      "%pr"
};

typedef struct	REGISTER_USAGE { int		no_save;
    int		no_calls;
    int		ireg_source[32];
    int		ireg_dest[32];
    int		freg_source[64];
    int		freg_dest[64];
}   REGISTER_USAGE, *REGISTER_USAGEp;

REGISTER_USAGE	reguse;

#define MARK_IREG_SOURCE(reg)   reguse.ireg_source[(reg)]++
#define MARK_FREG_SOURCE(reg)   reguse.freg_source[(reg)]++

#define MARK_IREG_DEST(reg)	reguse.ireg_dest[(reg)]++
#define MARK_FREG_DEST(reg)	reguse.freg_dest[(reg)]++

#define	MARK_CALL_SEEN		reguse.no_calls++;
#define	MARK_SAVE_SEEN		reguse.no_save++;

Local BOOLEAN	init=	false;

#define	NEXT_CHAR(buff)		&((buff)[strlen(buff)])
#define	ALLOCATE(STRUCT, n)	(STRUCT *) allocate(sizeof(STRUCT), (n))

extern char	*Symbolic_Address();


Local PTR	allocate();
Local void	Format_Addr();
Local char	*Pad_Buffer();
Local void	Instruction_Walk_Init();
Local void	Instruction_Table_Print();
Local void	Instruction_Verify();
Local void	Print_State();

Local INSTRUCTION_TABLEp	Instruction_Lookup_Walk();



Export char *Instruction_Lookup(instruction, pc)
UINT	instruction;
UINT	pc;
{
    int			i, level, regno;
    UINT		value, bvalue;
    enum ACTION		action;
    NSTATE		state;
    TEMPLATE		template;
    INSTRUCTION_TABLEp	entry;
    Local char		BUFFER[100];
    char		*buffer= BUFFER;
    char		*symbol;

    ASSERT(init is true, "Instruction_Lookup not initialized");
    /*
     *	Based on op[31:30] start the tree walk
     */
    template.instruction=	instruction;
    STATE_OP=			template.Format1.op;
    STATE_OP2=			template.Format2.op2;
    STATE_OP3=			template.Format4.op3;
    STATE_COND=			template.Format7.cond;
    STATE_OPF=			template.Format6.opf;
    STATE_VARIATION=	 	template.Format9.variation;
    STATE_MCOND=	 	template.Format11.cond;

    if (DEBUG_LOOKUP) {
	fprintf(stderr,"Instruction @ pc 0x%x\n", pc);
	Print_State(0, state);
    }

    level=	-1;
    entry=	Instruction_Lookup_Walk(
			&instruction_tree,
			state,
			level
		);
    if (not entry) {
	strcpy(buffer,"invalid instruction");
	return buffer;
    }
    strcpy(buffer, entry->name);

    for(i=0; i<ACTION_SIZE; i++) {
	action=	entry->action[i];
	
	/*
	 *  Special formatting
	 */
	switch(action) {
	case ANNUL:
	    if (template.Format3.a)
		strcat(buffer, ",a");
	    continue;

	case PREDICT:
	    strcat(buffer, (template.Format10.p) ? ",pt" : ",pn");
	    Pad_Buffer(buffer, ARG_OFFSET, ' ');
	    continue;

	case SHIFTX:
	    if (template.Format13.x)
		strcat(buffer, "x");
	    break;
	}

	switch(action) {
	case NA:
	    continue;

	case DONERETRY:
	    switch(template.Format4.rd) {
	    case 0:
	    	strcpy(buffer, "done");
		break;
	    case 1:
	    	strcpy(buffer,"retry");
		break;
	    default:
	    	strcpy(buffer,"bad done/retry");
		break;
	    }
	    continue;

	case LABEL22:
	    Pad_Buffer(buffer, ARG_OFFSET, ' ');
	    break;

	default:
	    switch(i) {
	    case 0:
		Pad_Buffer(buffer, ARG_OFFSET, ' ');
		break;
	    case 1:
	    case 2:
	    case 3:
		switch(action) {
		case IREG1:
		case FREG1:
		case FREG1X:
		case CC:
		    break;
		default:
		    strcat(buffer, ", ");
		}
		break;
	    }
	}

	buffer=		NEXT_CHAR(buffer);

	switch(action) {
	case ADDR:
	    Format_Addr(buffer, &template);
	    break;

	case IREGD:
	    strcat(buffer, ireg[template.Format6.rd]);
	    break;
	
	case IREG1:
	    strcat(buffer, ireg[template.Format6.rs1]);
	    break;

	case IREG2:
	    strcat(buffer, ireg[template.Format6.rs2]);
	    break;

	case BREG1:
	    strcat(buffer, ireg[template.Format16.rs1]);
	    break;

	case FREGD:
	    strcat(buffer, freg[template.Format6.rd]);
	    break;

	case FREGDX:
	    regno= template.Format6.rd + ((template.Format6.rd&0x1) ? 32:0);
	    strcat(buffer, freg[regno]);
	    break;

	case FREG1:
	    strcat(buffer, freg[template.Format6.rs1]);
	    break;

	case FREG1X:
	    regno= template.Format6.rs1 + ((template.Format6.rs1&0x1) ? 32:0);
	    strcat(buffer, freg[regno]);
	    break;

	case FREG2:
	    strcat(buffer, freg[template.Format6.rs2]);
	    break;

	case FREG2X:
	    regno= template.Format6.rs2 + ((template.Format6.rs2&0x1) ? 32:0);
	    strcat(buffer, freg[regno]);
	    break;

	case FREG3:
	    strcat(buffer, freg[template.Format9.rs3]);
	    break;

	case FREG3X:
	    regno= template.Format9.rs3 + ((template.Format9.rs3&0x1) ? 32:0);
	    strcat(buffer, freg[regno]);
	    break;

	case CREG0:
	    strcat(buffer, creg[template.Format6.rd]);
	    break;

	case CREG1:
	    strcat(buffer, creg[template.Format6.rs1]);
	    break;

	case CREG2:
	    strcat(buffer, creg[template.Format6.rs2]);
	    break;

	case FSR:
	    strcat(buffer, "fsr");
	    break;

	case CSR:
	    strcat(buffer, "csr");
	    break;

	case FQ:
	    strcat(buffer, "fq");
	    break;

	case CQ:
	    strcat(buffer, "cq");
	    break;

	case ASR:
	    strcat(buffer, "asr");
	    break;

	case WIM:
	    strcat(buffer, "win");
	    break;

	case TBR:
	    strcat(buffer, "tbr");
	    break;

	case PSR:
	    strcat(buffer, "psr");
	    break;

	case YREG:
	    strcat(buffer, "y");
	    break;

	case ASTATEREGS:
	    strcat(buffer, astatereg[template.Format4.rs1]);
	    break;

	case ASTATEREGD:
	    strcat(buffer, astatereg[template.Format4.rd]);
	    break;

	case STATEREGS:
	    strcat(buffer, statereg[template.Format4.rs1]);
	    break;

	case STATEREGD:
	    strcat(buffer, statereg[template.Format4.rd]);
	    break;

	case CC:
	    switch(template.Format10.cc) {
	    case 0:
		strcat(buffer, "icc");
		break;
	    case 1:
		strcat(buffer, "xcc");
		break;
	    default:
		strcat(buffer, "badCC");
		break;
	    }
	    break;

	case FCC:
	    sprintf(buffer, "fcc%d", template.Format10.cc);
	    break;

	case TRAPCC:
	    switch(template.Format8.cc) {
	    case 0:
		strcat(buffer, "icc");
		break;
	    case 1:
		strcat(buffer, "xcc");
		break;
	    }
	    break;

	case IFXCC:
	    value=	(template.Format12.ifcc<<2) + template.Format12.cc;
	    switch(value) {
	    case 4:
		strcat(buffer, "icc");
		break;
	    case 6:
		strcat(buffer, "xcc");
		break;
	    default:
		if (value <= 3)
		    sprintf(buffer, "fcc%d", value);
		else
		    strcat(buffer, "badCC");
	    }
	    break;

	case FMOVECC:
	    switch(template.Format14.cc) {
	    case 4:
		strcat(buffer, "icc");
		break;
	    case 6:
		strcat(buffer, "xcc");
		break;
	    default:
		if (value <= 3)
		    sprintf(buffer, "fcc%d", value);
		else
		    strcat(buffer, "badCC");
	    }
	    break;

	case IREGIMMED6:
	case IREGIMMED10:
	case IREGIMMED11:
	case IREGIMMED13:
	    if (template.Format5.i) {
		switch(action) {
		case IREGIMMED6:
		    sprintf(buffer, "0x%x", template.Format13.simm6);
		    break;
		case IREGIMMED10:
		    sprintf(buffer, "0x%x", template.Format15.simm10);
		    break;
		case IREGIMMED11:
		    sprintf(buffer, "0x%x", template.Format12.simm11);
		    break;
		case IREGIMMED13:
		    sprintf(buffer, "0x%x", template.Format5.simm13);
		    break;
		}
	    }
	    else
		strcat(buffer, ireg[template.Format6.rs2]);
	    break;

	case ASI:
	    sprintf(buffer, "0x%x", template.Format4.asi);
	    break;

	case LABEL19:
	    value=	pc + (template.Format10.disp19<<2);
	    if (SYMBOLIC) {
		symbol=		Symbolic_Address(value);
		strcat(buffer, symbol);
	    }
	    else {
		sprintf(buffer, "(pc+4*0x%x)", template.Format10.disp19);
	    }
	    break;

	case LABEL22:
	    value=	pc + (template.Format2.imm22<<2);
	    if (SYMBOLIC) {
		symbol=		Symbolic_Address(value);
		strcat(buffer, symbol);
	    }
	    else {
		sprintf(buffer, "(pc+4*0x%x)", template.Format2.imm22);
	    }
	    break;

	case BLABEL:
	    bvalue=	(template.Format16.disp2 << 14) +template.Format16.disp14;
	    if (SYMBOLIC) {
		symbol=		Symbolic_Address(pc + (bvalue<<2));
		strcat(buffer, symbol);
	    }
	    else {
		sprintf(buffer, "(pc+4*0x%x)", bvalue);
	    }
	    break;

	case CALL:
	    value=	pc + (template.Format1.disp30<<2);
	    if (SYMBOLIC) {
		symbol=		Symbolic_Address(value);
		strcat(buffer, symbol);
	    }
	    else {
		sprintf(buffer, "(pc+4*0x%x)", template.Format1.disp30);
	    }
	    break;

	case SHIFTX:
	case ANNUL:
	    break;

	case SETHI:
	    /*
	     *	Nop instruction
	     */
	    if (template.Format2.imm22 == 0 and template.Format2.rd == 0) {
		buffer=	BUFFER;
		strcpy(buffer, "nop");
		return BUFFER;
	    }
	    else {
		strcat(buffer, "%hi(");
		value=	template.Format2.imm22 << 10;
		if (SYMBOLIC) {
		    symbol=	Symbolic_Address(value);
		    strcat(buffer, symbol);
		}
		else {
		    sprintf(buffer, "0x%x", value);
		}
		strcat(buffer, ")");
	    }
	    break;

	case UNIMP:
	    sprintf(buffer, "0x%x", template.Format2.imm22);
	    break;

	case TRAP:
	    sprintf(buffer, "0x%x", template.Format8.trap);
	    break;

	default:
	    ASSERT(0, "BAD CASE");
	}
    }
    return BUFFER;
}


Export void	Instruction_Register_Usage(instruction)
UINT	instruction;
{
    int			i, level, regno;
    UINT		value;
    enum ACTION		action;
    NSTATE		state;
    TEMPLATE		template;
    INSTRUCTION_TABLEp	entry;

    ASSERT(init is true, "Instruction_Lookup not initialized");
    /*
     *	Based on op[31:30] start the tree walk
     */
    template.instruction=	instruction;
    STATE_OP=			template.Format1.op;
    STATE_OP2=			template.Format2.op2;
    STATE_OP3=			template.Format4.op3;
    STATE_COND=			template.Format7.cond;
    STATE_OPF=			template.Format6.opf;
    STATE_VARIATION=		template.Format9.variation;
    STATE_MCOND=		template.Format11.cond;

    level=	-1;
    entry=	Instruction_Lookup_Walk(
			&instruction_tree,
			state,
			level
		);
    if (not entry) {
	return;
    }
    if (streq(entry->name,"save"))
	MARK_SAVE_SEEN;

    for(i=0; i<ACTION_SIZE; i++) {
	action=	entry->action[i];
	
	switch(action) {
	case NA:
	    break;

	case CALL:
	    MARK_CALL_SEEN;
	    break;

	case ASI:
	case LABEL22:
	case LABEL19:
	case ANNUL:
	case SETHI:
	case UNIMP:
	case TRAP:
	case CREG0:
	case CREG1:
	case CREG2:
	case CSR:
	case FSR:
	case FQ:
	case CQ:
	case ASR:
	case WIM:
	case TBR:
	case PSR:
	case YREG:
	case ASTATEREGS:
	case ASTATEREGD:
	case STATEREGS:
	case STATEREGD:
	    break;

	/*
	 *  Form [reg+immediate13], Form [reg+reg]
	 */
	case ADDR:
	    if (template.Format5.i)
		MARK_IREG_SOURCE(template.Format5.rs1);
	    else {
		MARK_IREG_SOURCE(template.Format6.rs1);
		MARK_IREG_SOURCE(template.Format6.rs2);
	    }
	    break;

	case IREGD:
	    MARK_IREG_DEST(template.Format6.rd);
	    break;
	
	case IREG1:
	    MARK_IREG_SOURCE(template.Format6.rs1);
	    break;

	case IREG2:
	    MARK_IREG_SOURCE(template.Format6.rs2);
	    break;

	case FREGD:
	    MARK_FREG_DEST(template.Format6.rd);
	    break;

	case FREGDX:
	    regno= template.Format6.rd + ((template.Format6.rd&0x1) ? 32:0);
	    MARK_FREG_DEST(regno);
	    break;

	case FREG1:
	    MARK_FREG_SOURCE(template.Format6.rs1);
	    break;

	case FREG1X:
	    regno= template.Format6.rs1 + ((template.Format6.rs1&0x1) ? 32:0);
	    MARK_FREG_DEST(regno);
	    break;

	case FREG2:
	    MARK_FREG_SOURCE(template.Format6.rs2);
	    break;

	case FREG2X:
	    regno= template.Format6.rs2 + ((template.Format6.rs2&0x1) ? 32:0);
	    MARK_FREG_DEST(regno);
	    break;

	case FREG3:
	    MARK_FREG_SOURCE(template.Format9.rs3);
	    break;

	case FREG3X:
	    regno= template.Format9.rs3 + ((template.Format9.rs3&0x1) ? 32:0);
	    MARK_FREG_DEST(regno);
	    break;

	case IREGIMMED13:
	    if (not template.Format5.i)
		MARK_IREG_SOURCE(template.Format6.rs2);
	    break;

	default:
	    ASSERT(0, "BAD CASE");
	}
    }
}


Export void	Instruction_Register_Usage_Init()
{
    int		i;
    int		*ireg_dest=	reguse.ireg_dest;
    int		*freg_dest=	reguse.freg_dest;
    int		*ireg_source=	reguse.ireg_source;
    int		*freg_source=	reguse.freg_source;

    for(i=0; i<ireg_table_size; i++) {
	ireg_source[i]=		0;
	ireg_dest[i]=		0;
    }

    for(i=0; i<freg_table_size; i++) {
	freg_source[i]=		0;
	freg_dest[i]=		0;
    }
    reguse.no_save= reguse.no_calls=	0;
}

Export void	Instruction_Register_Usage_Report(func, verbose)
char	*func;
BOOLEAN	verbose;
{
    int		i, icount, fcount;
    int		*ireg_dest=	reguse.ireg_dest;
    int		*freg_dest=	reguse.freg_dest;
    int		*ireg_source=	reguse.ireg_source;
    int		*freg_source=	reguse.freg_source;

    fprintf(stdout,"\nRegister Report for Function \"%s\"\n", func);
	
    fprintf(stdout,"\tSave= %d,  Calls = %d\n",
	reguse.no_save, reguse.no_calls
    );

    if (verbose) {
	fprintf(stdout,"\tInteger Registers: reg:dest= source\n\t");
	for(i=0; i<ireg_table_size; i++) {
	    fprintf(stdout,"%s:%d=%d, ",
		    ireg[i], ireg_dest[i], ireg_source[i]
	    );
	    if ((i%8) == 7)
		fprintf(stdout,"\n\t");
	}
	
	fprintf(stdout,"\n\tFloating Point Registers: reg:dest= source\n\t");
	for(i=0; i<ireg_table_size; i++) {
	    fprintf(stdout,"%s %d=%d, ",
		    freg[i], freg_dest[i], freg_source[i]
	    );
	    if ((i%8) == 7)
    		fprintf(stdout,"\n\t");
	}
    }
    else {
	icount=	 fcount=	0;
	for(i=0; i<ireg_table_size; i++) {
	    if (ireg_dest[i]  or ireg_source[i])
		icount++;
	    if (freg_dest[i]  or freg_source[i])
		fcount++;
	}
	if (icount>0) {
	    fprintf(stdout,"\tInteger Registers Used (%d): ", icount);
	    for(i=0; i<ireg_table_size; i++)
		if (ireg_dest[i] or ireg_source[i])
		    fprintf(stdout, "%s ", ireg[i]);
	    fprintf(stdout,"\n");
	
	}
	if (fcount>0) {
	    fprintf(stdout,"\tFloating Registers Used (%d): ", fcount);
	    for(i=0; i<ireg_table_size; i++)
		if (freg_dest[i] or freg_source[i])
		    fprintf(stdout, "%s ", freg[i]);
	    fprintf(stdout,"\n");
	}
	/********************************
	fprintf(stdout,"\nTotal Number Registers = %d\n", icount+fcount);
	********************************/
    }
    fprintf(stdout,"\n");
}


Local void Format_Addr(buffer, instr)
char		*buffer;
TEMPLATEp	instr;
{

    buffer=	&buffer[ strlen(buffer) ];

    /*
     *	Form [reg+immediate13]
     */
    if (instr->Format5.i) {
	if (instr->Format5.rs1 == 0) {
	    if (instr->Format5.simm13 == 0)
		sprintf(buffer, "[0]");
	    else
		sprintf(buffer, "[0x%x]", instr->Format5.simm13);
	}
	else {
	    sprintf(buffer, "[%s+0x%x]",
		    ireg[instr->Format5.rs1],
		    instr->Format5.simm13
	    );
	}
    }
    /*
     *	Form [reg+reg]
     */
    else {
	if (instr->Format6.rs2== 0)
	    sprintf(buffer, "[%s]", ireg[instr->Format6.rs1]);
	else
	    sprintf(buffer, "[%s+%s]",
		    ireg[instr->Format6.rs1],
		    ireg[instr->Format6.rs2]
	    );
    }
}



Local char	*Pad_Buffer(buffer, col, fill)
char	*buffer;
int	col;
char	fill;
{
    int		len=	strlen(buffer);
    char	*bp;

    bp=		&buffer[len];
    do {
	*bp++	=	fill;
	len++;
    } while(len < col);

    *bp=	'\0';
    return bp;
}


Local INSTRUCTION_TABLEp Instruction_Lookup_Walk(table, state, level)
NTREEp		table;
NSTATE		state;
int		level;
{
    int		i, n;
    NTREEp	next_table;

    level++;
    if (DEBUG_WALK) {
	fprintf(stderr,"Walk @level %d\n", level);
	Print_State("table->", table->state);
    }
    
    /*
     *	Table may contain instruction or further parse of table
     */
    if (table->entry) {
#ifdef	DEBUG
	for(i=0; i<level; i++)
	    ASSERT((table->STATE(i) == NA) or (table->STATE(i) == STATE(i)),
		   "Bad State"
	    );
#endif	DEBUG
	return table->entry;
    }

    if (next_table=  table->next) {
	for( ; level<STATENO; level++) {
	    n=		next_table->STATE(level);
	    if (n != NA)
		return( Instruction_Lookup_Walk( &(next_table[STATE(level)]), state, level) );
	}
    }
    return NULL;
}


Export void Instruction_Lookup_Init()
{
    INSTRUCTION_TABLEp	ip;
    NSTATE		vector;
    int			i;
    Local NSTATE	na_vector= { NA, NA, NA, NA, NA, NA};

#if	DEBUG_INSTRUCTION == 1
    debug |=	0x1;
#endif
#if	DEBUG_INSTRUCTION == 2
    debug |=	0x2;
#endif

    ASSERT(init is false, "Instruction_Lookup_Init: already initialized");

    instruction_tree.state=	na_vector;
    for(i=0; i<instruction_table_size; i++) {
	ip=		&instruction_table[i];
	vector=		na_vector;
	if (DEBUG_WALK) {
	    fprintf(stderr,"Enter Instruction \"%s\" ", ip->name);
	    Print_State("state", ip->state);
	}
	Instruction_Walk_Init(ip, &instruction_tree, vector);
    }
    init=	true;

    if (DEBUG_WALK)
	Instruction_Verify();
}


Local void Instruction_Verify()
{
    INSTRUCTION_TABLEp	ip;
    TEMPLATE		template;
    char		*instr;
    int			i;

    if (DEBUG_WALK)
	Instruction_Table_Print();

    for(i=0; i<instruction_table_size; i++) {
	ip=			&instruction_table[i];
	template.instruction=	0;
	if (ip->STATE_OP   != NA)
	    template.Format1.op=	ip->STATE_OP;
	if (ip->STATE_OP2  != NA)
	    template.Format2.op2=	ip->STATE_OP2;
	if (ip->STATE_OP3  != NA)
	    template.Format4.op3=	ip->STATE_OP3;
	if (ip->STATE_OPF  != NA)
	    template.Format6.opf=	ip->STATE_OPF;
	if (ip->STATE_COND != NA)
	    template.Format3.cond=	ip->STATE_COND;
	if (ip->STATE_VARIATION != NA)
	    template.Format9.variation=	ip->STATE_VARIATION;
	if (ip->STATE_MCOND != NA)
	    template.Format11.cond=	ip->STATE_MCOND;

	if (DEBUG_LOOKUP) {
	    fprintf(stderr,"looking up instruction \"%s\" 0x%08x",
		    ip->name,
		    template.instruction
	    );
	    Print_State("state", ip->state);
	    instr=	Instruction_Lookup(template.instruction, 0);
	    fprintf(stderr,"\t%s\n", instr);
	}
    }
}


Local void Instruction_Walk_Init(entry, tree, state)
INSTRUCTION_TABLEp	entry;
NTREEp			tree;
NSTATE			state;
{
    int		i, j, n;
    NSTATE	init;

    if (DEBUG_WALK) {
	Print_State("Walk State", state);
    }
    /*
     *	Allocate the sub-tree if necessary
     */
    for(i= 0; i<STATENO; i++) {
	n=	entry->STATE(i);
	if (n == NA)
	    continue;
	if (n != STATE(i)) {
	    STATE(i)=	entry->STATE(i);
	    if (not tree->next) {
	        if (DEBUG_WALK) {
		    fprintf(stderr,"\tInitialize @level %d, State [ ", i);
		    for(j=0; j<STATENO; j++)
			fprintf(stderr,"%d ", STATE(j));
		    fprintf(stderr,"]\n");
		}
		init=		state;
		tree->next=	ALLOCATE(NTREE, state_size.stateN[i]);
		for(j=0; j<state_size.stateN[i]; j++) {
		    init.stateN[i]=		j;
		    tree->next[j].state=	init;
		}
	    }
	    Instruction_Walk_Init(entry, &(tree->next[n]), state);
	    return;
	}
    }
    /*
     *	All states match. Install instruction
     */
    if (tree->entry) {
	if (DEBUG_WALK)
	    fprintf(stderr,"already busy with %s\n", tree->entry->name);
	ASSERT(0, "bad instruction init");
    }
    if (DEBUG_WALK)
	Print_State("Install at State", tree->state);
    tree->entry=	entry;
}


Local PTR	allocate(size, n)
int	size;
int	n;
{
    PTR		p=	(PTR)	malloc(size*n);
    ASSERT(p, "not enough memory from malloc");
    bzero(p, size*n);
    return (p);
}


Local void	Instruction_Table_Print()
{
    int		level;
    Local void	Instruction_Table_Print_Walk();

    fprintf(stderr,"\n\nINSTRUCTION TABLE\n");
    level=	0;
    Instruction_Table_Print_Walk(level, &instruction_tree);
    fflush(stderr);
    fprintf(stderr,"\n");
}


Local void	Instruction_Table_Print_Walk(level, tree)
int	level;
NTREEp	tree;
{
    int		i, j, n;
    NTREEp	next;

    for(j=0; j<level; j++)
	fprintf(stderr,"    ");
    fprintf(stderr,"Instruction @level %d [", level);
    for(j=0; j<STATENO; j++) {
	if (tree->STATE(j) == NA)
	    fprintf(stderr,"na ");
	else
	    fprintf(stderr,"%d ", tree->STATE(j));
    }
    fprintf(stderr, "]\t");

    /*
     *	Walk nary tree
     */
    if (next= tree->next) {
	while(next->STATE(level) == NA)
	    level++;
	n=	state_size.stateN[level];
	fprintf(stderr, "--> (%d)\n", n);
	for(i=0;  i<n; i++) { 
	    next= &(tree->next[i]);
	    Instruction_Table_Print_Walk(level+1, next);
	}
    }
    else {
	fprintf(stderr,"===\"%s\"\n", 
		(tree->entry) ? tree->entry->name : "NOTMAPPED");
    }
}


Local void	Print_State(msg, state)
char	*msg;
NSTATE	state;
{
    if (msg)
	fprintf(stderr,"%s: ", msg);

    fprintf(stderr,"(op %d  op2:%d  op3:%d  cond:%d  opf:%d  variation:%d  mcond:%d)\n",
		STATE_OP,
		STATE_OP2,
		STATE_OP3,
		STATE_OPF,
		STATE_COND,
		STATE_VARIATION,
		STATE_MCOND
    );
}
