/*
 *  $Id: decode.c,v 1.5 1995/02/03 19:25:22 melvin Exp $
 *  
 *  Copyright (c) 1994 HAL Computer Systems International, Ltd.
 *  All rights reserved.  Unpublished -- rights reserved under
 *  the Copyright Laws of the United States.  USE OF A COPYRIGHT
 *  NOTICE IS PRECAUTIONARY ONLY AND DOES NOT IMPLY PUBLICATION
 *  OR DISCLOSURE.
 *  
 *  THIS SOFTWARE CONTAINS CONFIDENTIAL INFORMATION AND TRADE
 *  SECRETS OF HAL COMPUTER SYSTEMS INTERNATIONAL, LTD.  USE,
 *  DISCLOSURE, OR REPRODUCTION IS PROHIBITED WITHOUT THE
 *  PRIOR EXPRESS WRITTEN PERMISSION OF HAL COMPUTER SYSTEMS
 *  INTERNATIONAL, LTD.
 *  
 *                          RESTRICTED RIGHTS LEGEND
 *  Use, duplication, or disclosure by the Government is subject
 *  to the restrictions as set forth in subparagraph (c)(l)(ii)
 *  of the Rights in Technical Data and Computer Software clause
 *  at DFARS 252.227-7013.
 *  
 *           HAL COMPUTER SYSTEMS INTERNATIONAL, LTD.
 *                   1315 Dell Avenue
 *                   Campbell, CA  95008
 */
/*
 * Editing the decode.c file from halsim simulator 1994/11/16 saxena
 *
 */

#include "opcodes.h"
void decode_float1(unsigned long instr, unsigned long recode[]);
void decode_float2(unsigned long instr, unsigned long recode[]);
void decode_impdep1(unsigned long instr, unsigned long recode[]);

void decode(unsigned long instr, unsigned long recode[])
{
  unsigned op, opcode;

  op = F_op(instr);
  switch (op) {
  case 0:
    opcode = (op << 6) | (F_op2(instr) << 3);
    break;
  case 1:
    opcode = CALL;
    break;
  case 2:
  case 3:
    opcode = (op << 6) | F_op3(instr);
    break;
  }

  /* format -> rsvd[31:8]|opcode[7:0] */
  recode[0] = opcode;

  switch (opcode) {
  case FPop1: 
    /* format ->
       rs1[31:26]|rs2[25:20]|rd[19:14]|rsvd[13:9]|opf[8:0] 
       OR
       rsvd[31:26]|rs2[25:20]|rd[19:14]|rsvd[13:9]|opf[8:0] 
       */
    decode_float1(instr,recode);
    break;
  case FPop2:
    /* format ->
       rs1[31:26]|rs2[25:20]|{000cc1cc0}|rsvd[13:9]|opf[8:0]
       OR
       cond[31:26]|rs2[25:20]|rd[19:14]|rsvd[13:9]|opf_cc[8:6]|opf_low[5:0] 
       OR
       rs1[31:26]|rs2[25:20]|rd[19:14]|rsvd[13:8]|rcond[7:5]|opf_low[4:0]
       */
    decode_float2(instr,recode);
    break;
  case IMPDEP1:
    /* format ->
       rs1[31:26]|rs2[25:20]|rd[19:14]|rs3[13:8]|rsvd[7:4]|opf[3:0]
       */
    decode_impdep1(instr,recode);
    break;
  case IMPDEP2:
    /* format ->
       rs1[31:26]|rs2[25:20]|rd[19:14]|rs3[13:8]|rsvd[7:4]|opf[3:0]
       */
    decode_impdep1(instr,recode);
    break;
  case LDF:
  case STF:
    /* format ->
       rs1[31:26]|rs2[25:20]|rd[19:14]|rsvd[13:0]
       OR
       rs1[31:26]|rsvd[25:20]|rd[19:14]|simm[13:1]|i[0:0];
       */
    recode[1] = F_rs1(instr)<<26;
    recode[1] =  recode[1]|(F_rd(instr)<<14);
    if (F_i(instr)) {
      recode[1] = recode[1]|(F_simm13(instr)<<1)|0x1; 
    } else {
      recode[1] = recode[1]|(F_rs2(instr)<<20);
    }
    break;
  case LDFA:
  case STFA:
    /* format ->
       rs1[31:26]|rs2[25:20]|rd[19:14]|rsvd[13:9]|imm_asi[8:1]|{0}
       OR
       rs1[31:26]|rsvd[25:20]|rd[19:14]|simm[13:1]|i[0:0]
       */
    recode[1] = F_rs1(instr)<<26;
    recode[1] =  recode[1]|(F_rd(instr)<<14);
    if (F_i(instr)) {
      recode[1] = recode[1]|(F_simm13(instr)<<1)|0x1; 
    } else {
      recode[1] = recode[1]|(F_rs2(instr)<<20)|(F_asi(instr)<<1);
    }
    break;
  case LDDF:
  case STDF:
    /* format ->
       rs1[31:26]|rs2[25:20]|rd[19:14]|rsvd[13:0]
       OR
       rs1[31:26]|rsvd[25:20]|rd[19:14]|simm[13:1]|i[0:0];
       */
    recode[1] = F_rs1(instr)<<26;
    recode[1] =  recode[1]|(DOUBLE(F_rd(instr))<<14);
    if (F_i(instr)) {
      recode[1] = recode[1]|(F_simm13(instr)<<1)|0x1; 
    } else {
      recode[1] = recode[1]|(F_rs2(instr)<<20);
    }

    break;
  case LDDFA:
  case STDFA:
    /* format ->
       rs1[31:26]|rs2[25:20]|rd[19:14]|rsvd[13:9]|imm_asi[8:1]|{0} 
       OR
       rs1[31:26]|rsvd[25:20]|rd[19:14]|simm[13:1]|i[0:0] */
    recode[1] = F_rs1(instr)<<26;
    recode[1] =  recode[1]|(DOUBLE(F_rd(instr))<<14);
    if (F_i(instr)) {
      recode[1] = recode[1]|(F_simm13(instr)<<1)|0x1; 
    } else {
      recode[1] = recode[1]|(F_rs2(instr)<<20)|(F_asi(instr)<<1);
    }
    break;
  case LDQF:
  case STQF:
    /* format ->
       rs1[31:26]|rs2[25:20]|rd[19:14]|rsvd[13:0]
       OR
       rs1[31:26]|rsvd[25:20]|rd[19:14]|simm[13:1]|i[0:0];
       */
    recode[1] = F_rs1(instr)<<26;
    recode[1] =  recode[1]|(QUAD(F_rd(instr))<<14);
    if (F_i(instr)) {
      recode[1] = recode[1]|(F_simm13(instr)<<1)|0x1; 
    } else {
      recode[1] = recode[1]|(F_rs2(instr)<<20);
    }
    break;
  case LDQFA:
  case STQFA:
    /* format ->
       rs1[31:26]|rs2[25:20]|rd[19:14]|rsvd[13:9]|imm_asi[8:1]|{0}
       OR
       rs1[31:26]|rsvd[25:20]|rd[19:14]|simm[13:1]|i[0:0]
       */
    recode[1] = F_rs1(instr)<<26;
    recode[1] =  recode[1]|(QUAD(F_rd(instr))<<14);
    if (F_i(instr)) {
      recode[1] = recode[1]|(F_simm13(instr)<<1)|0x1; 
    } else {
      recode[1] = recode[1]|(F_rs2(instr)<<20)|(F_asi(instr)<<1);
    }
    break;
  case SLL:
  case SRL:
  case SRA:
    /* format ->
       rs1[31:26]|rs2[25:20]|rd[19:14]|x[13:13]|rsvd[12:1]|{0}
       OR
       rs1[31:26]|rsvd[25:20]|rd[19:14]|x[13:13]|rsvd[12:7]|shcnt[6:1]|i[0:0];
       */
    recode[1] = F_rs1(instr)<<26;
    recode[1] =  recode[1]|(F_rd(instr)<<14);
    if (F_i(instr)) {
      recode[1] = recode[1]|(F_sh64(instr)<<13)|((instr&0x3f)<<1)|0x1; 
    } else {
      recode[1] = recode[1]|(F_rs2(instr)<<20)|(F_sh64(instr)<<13);
    }
    break;
  case Tcc:
    /* format ->
       rs1[31:26]|rs2[25:20]|cond[19:14]|cc10[13:12]|rsvd[11:1]|{0}
       OR
       rs1[31:26]|rsvd[25:20]|cond[19:14]|cc10[13:12]|rsvd[11:8]|swtrap[7:1]|{1};
       */
    recode[1] = F_rs1(instr)<<26;
    recode[1] = recode[1]|(F_cond(instr)<<14)|((instr&0x1800)<<1);
    if (F_i(instr)) {
      recode[1] = recode[1]|((F_trapno(instr))<<1)|0x1; 
    } else {
      recode[1] = recode[1]|(F_rs2(instr)<<20);
    }
    break;	
  case RDASR:
  case RDPR:
    /* format ->
       rs1[31:26]|rsvd[25:20]|rd[19:14]|rsvd[13:0];
       */
    recode[1] = F_rs1(instr)<<26;
    recode[1] =  recode[1]|(F_rd(instr)<<14);
    break;
  case WRASR:
  case WRPR:
  case POPC:
    /* format ->
       rs1[31:26]|rs2[25:20]|rd[19:14]|rsvd[13:0]
       OR
       rs1[31:26]|rsvd[25:20]|rd[19:14]|simm[13:1]|i[0:0];
       */
    recode[1] = F_rs1(instr)<<26;
    recode[1] =  recode[1]|(F_rd(instr)<<14);
    if (F_i(instr)) {
      recode[1] = recode[1]|(F_simm13(instr)<<1)|0x1; 
    } else {
      recode[1] = recode[1]|(F_rs2(instr)<<20);
    }
    break;
  case LDFSR:
  case STFSR:
    /* format -> rs1[31:26]|rs2[25:20]|rd[19:14]|rsvd[13:0] OR
       rs1[31:26]|rsvd[25:20]|rd[19:14]|simm[13:1]|i[0:0]; */
    recode[1] = F_rs1(instr)<<26;
    recode[1] =  recode[1]|(F_rd(instr)<<14);
    if (F_i(instr)) {
      recode[1] = recode[1]|(F_simm13(instr)<<1)|0x1; 
    } else {
      recode[1] = recode[1]|(F_rs2(instr)<<20);
    }
    break;
  case MOVcc:
    /* format ->
       rs1[31:26]|rs2[25:20]|cc2_cond[19:14]|cc10[13:12]|rsvd[10:1]|{0}
       OR
       rs1[31:26]|rsvd[25:20]|cc2_cond[19:14]|cc10[13:12]|simm11[11:1]|
       swtrap[7:1]|{1};
       */
    recode[1] = F_rs1(instr)<<26;
    recode[1] = recode[1]|(F_rd(instr)<<14)|((instr&0x1800)<<1);
    if (F_i(instr)) {
      recode[1] = recode[1]|((F_simm11(instr))<<1)|0x1; 
    } else {
      recode[1] = recode[1]|(F_rs2(instr)<<20);
    }
    break;
  case MOVR:
    /* format ->
       rs1[31:26]|rs2[25:20]|rd[19:14]|rcond[13:11]|rsvd[10:1]|{0}
       OR
       rs1[31:26]|rsvd[25:20]|rd[19:14]|rcond[13:11]|simm10[10:1]|{1};
       */
    recode[1] = F_rs1(instr)<<26;
    recode[1] = recode[1]|(F_rd(instr)<<14)|(F_rcond(instr)<<11);
    if (F_i(instr)) {
      recode[1] = recode[1]|((F_simm10(instr))<<1)|0x1; 
    } else {
      recode[1] = recode[1]|(F_rs2(instr)<<20);
    }
 
          
    break;
  case CAS:
  case CASX:
  case LDUW:
  case LDUB:
  case LDUH:
  case STW:
  case STB:
  case STH:
  case LDSW:
  case LDSB:
  case LDSH:
  case LDX:
  case LDSTUB:
  case STX:
  case SWAP:
  case LDD:
  case STD:
    /* format ->
       rs1[31:26]|rs2[25:20]|rd[19:14]|rsvd[13:0]
       OR
       rs1[31:26]|rsvd[25:20]|rd[19:14]|simm[13:1]|i[0:0];
       */
    recode[1] = F_rs1(instr)<<26;
    recode[1] =  recode[1]|(F_rd(instr)<<14);
    if (F_i(instr)) {
      recode[1] = recode[1]|(F_simm13(instr)<<1)|0x1; 
    } else {
      recode[1] = recode[1]|(F_rs2(instr)<<20);
    }
    break;
  case CASA:
  case CASXA:
  case LDUWA:
  case LDUBA:
  case LDUHA:
  case STWA:
  case STBA:
  case STHA:
  case LDSWA:
  case LDSBA:
  case LDSHA:
  case LDXA:
  case LDSTUBA:
  case STXA:
  case SWAPA:
  case LDDA:
  case STDA:
    /* format ->
     rs1[31:26]|rs2[25:20]|rd[19:14]|rsvd[13:9]|imm_asi[8:1]|{0}
     OR
     rs1[31:26]|rsvd[25:20]|rd[19:14]|simm[13:1]|i[0:0]
    */
    recode[1] = F_rs1(instr)<<26;
    recode[1] =  recode[1]|(F_rd(instr)<<14);
    if (F_i(instr)) {
      recode[1] = recode[1]|(F_simm13(instr)<<1)|0x1; 
    } else {
      recode[1] = recode[1]|(F_rs2(instr)<<20)|(F_asi(instr)<<1);
    }
    break;
  case CALL:
    /* format -> {00}|disp30[29:0] */
    recode[1] = F_disp30(instr);
    break;
  case BiccP:
  case FBfccP:
    /* format -> cond[31:28]|disp19[27:9]|cc10[8:7]|rsvd[6:2]|annul|predict */
    recode[1] = (F_cond(instr)<<28) |
      (((instr>>29)&0x1)<<0x1) | ((instr>>19)&0x1); 
    recode[1] = recode[1] | (F_disp19(instr)<<9)|(((instr>>20)&0x3)<<7);
    break;
  case Bicc:
  case FBfcc:
    /* format -> cond[31:28]|disp22[27:6]|rsvd[6:2]|annul|{0} */
    recode[1] = (F_cond(instr)<<28) |
      (((instr>>29)&0x1)<<0x1); 
    recode[1] = recode[1] | (F_disp22(instr)<<6);
    break;
  case BregP:
    /* format -> rs1[31:26]|rcond[25:23]|d16[22:7]|rsvd[6:2]|annul|predict */
    recode[1] = (F_rs1(instr)<<26)|(F_cond(instr)<<23);
    recode[1] = recode[1]|(F_disp16(instr)<<7);
    recode[1] = recode[1] | (((instr>>29)&0x1)<<0x1) | ((instr>>19)&0x1);
    break;
  case SETHI:
    /* format -> rsvd[31:20]|rd[19:14] */
    recode[1] = F_rd(instr)<<14; 
    break;
  case ADDcc:
  case ADD:
  case ANDcc:
  case AND:
  case ORcc:
  case OR:
  case XORcc:
  case XOR:
  case SUBcc:
  case SUB:
  case ANDNcc:
  case ANDN:
  case ORNcc:
  case ORN:
  case XNORcc:
  case XNOR:
  case ADDCcc:
  case ADDC:
  case UMULcc:
  case UMUL:
  case MULX:
  case SMULcc:
  case SMUL:
  case SUBCcc:
  case SUBC:
  case UDIVcc:
  case UDIV:
  case UDIVX:
  case SDIVcc:
  case SDIV:
  case SDIVX:
  case TADDcc:
  case TSUBcc:
  case TADDccTV:
  case TSUBccTV:
  case MULScc:
  case SAVE:
  case RESTORE:
  case FLUSH:
  case FLUSHW:
  case JMPL:
  case RETURN:
    /* format ->
     rs1[31:26]|rs2[25:20]|rd[19:14]|rsvd[13:0]
     OR
     rs1[31:26]|rsvd[25:20]|rd[19:14]|simm[13:1]|i[0:0];
    */
    recode[1] = F_rs1(instr)<<26;
    recode[1] =  recode[1]|(F_rd(instr)<<14);
    if (F_i(instr)) {
      recode[1] = recode[1]|(F_simm13(instr)<<1)|0x1; 
    } else {
      recode[1] = recode[1]|(F_rs2(instr)<<20);
    }
    break;
  case DONE:
  case SAVED:
    /* format -> rsvd[31:20]|fcn[19:14]|rsvd[13:0] */
    recode[1] = F_rd(instr)<<14;
    break;
  case PREFETCH:
    /* format ->
     rs1[31:26]|rs2[25:20]|fcn[19:14]|rsvd[13:0]
		   OR
     rs1[31:26]|rsvd[25:20]|fcn[19:14]|simm[13:1]|i[0:0];
    */
    recode[1] = F_rs1(instr)<<26;
    recode[1] =  recode[1]|(F_rd(instr)<<14);
    if (F_i(instr)) {
      recode[1] = recode[1]|(F_simm13(instr)<<1)|0x1; 
    } else {
      recode[1] = recode[1]|(F_rs2(instr)<<20);
    }
    break;
  case PREFETCHA:
    /* format ->
     rs1[31:26]|rs2[25:20]|fcn[19:14]|rsvd[13:9]|imm_asi[8:1]|{0}
		   OR
     rs1[31:26]|rsvd[25:20]|fcn[19:14]|simm[13:1]|i[0:0]
    */
    recode[1] = F_rs1(instr)<<26;
    recode[1] =  recode[1]|(F_rd(instr)<<14);
    if (F_i(instr)) {
      recode[1] = recode[1]|(F_simm13(instr)<<1)|0x1; 
    } else {
      recode[1] = recode[1]|(F_rs2(instr)<<20)|(F_asi(instr)<<1);
    }
    /* no ops as far as the simulator is concerned */
    break;
  default:
    break;
  }
  return;
}


void decode_float1(unsigned long instr, unsigned long recode[])
{
  int r1, r2, r3, rd; 
  unsigned opf;
 
  opf  = F_opf(instr);

  r1 = F_rs1(instr);
  r2 = F_rs2(instr);
  rd = F_rd(instr);

  switch (opf) {
  case FMOVs:
  case FNEGs:
  case FABSs:
  case FSQRTs:
    recode[1] = (r2<<20)|(rd<<14)|opf;
    break;
  case FMOVd:
  case FNEGd:
  case FABSd:
  case FSQRTd:
    recode[1] = (DOUBLE(r2)<<20)|(DOUBLE(rd)<<14)|opf;
    break;
  case FMOVq:
  case FNEGq:
  case FABSq:
  case FSQRTq:
    recode[1] = (QUAD(r2)<<20)|(QUAD(rd)<<14)|opf;
    break;
  case FMULs:
  case FDIVs:
  case FADDs:
  case FSUBs:
    recode[1] = (r1<<26)|(r2<<20)|(rd<<14)|opf;
    break;
  case FMULd:
  case FDIVd:
  case FADDd:
  case FSUBd:
    recode[1] = (DOUBLE(r1)<<26)|(DOUBLE(r2)<<20)|(DOUBLE(rd)<<14)|opf;
    break;
  case FMULq:
  case FDIVq:
  case FADDq:
  case FSUBq:
    recode[1] = (QUAD(r1)<<26)|(QUAD(r2)<<20)|(QUAD(rd)<<14)|opf;
    break;
  case FsMULd:
    recode[1] = (r1<<26)|(r2<<20)|(DOUBLE(rd)<<14)|opf;
    break;
  case FdMULq:
    recode[1] = (DOUBLE(r1)<<26)|(DOUBLE(r2)<<20)|(QUAD(rd)<<14)|opf;
    break;
  case FiTOs:
    recode[1] = (r2<<20)|(rd<<14)|opf;
    break;
  case FxTOs:
    recode[1] = (DOUBLE(r2)<<20)|(rd<<14)|opf;
    break;
  case FdTOs:
    recode[1] = (DOUBLE(r2)<<20)|(rd<<14)|opf;
    break;
  case FqTOs:
    recode[1] = (QUAD(r2)<<20)|(rd<<14)|opf;
    break;
  case FiTOd:
    recode[1] = (r2<<20)|(DOUBLE(rd)<<14)|opf;
    break;
  case FsTOd:
    recode[1] = (r2<<20)|(DOUBLE(rd)<<14)|opf;
    break;
  case FsTOi:
    recode[1] = (r2<<20)|(rd<<14)|opf;
    break;
  case FdTOi:
    recode[1] = (DOUBLE(r2)<<20)|(rd<<14)|opf;
    break;
  case FqTOi:
    recode[1] = (QUAD(r2)<<20)|(rd<<14)|opf;
    break;
  case FsTOx:
    recode[1] = (r2<<20)|(DOUBLE(rd)<<14)|opf;
    break;
  case FdTOx:
    recode[1] = (DOUBLE(r2)<<20)|(DOUBLE(rd)<<14)|opf;
    break;
  case FqTOx:
    break;
  case FxTOd:
    recode[1] = (DOUBLE(r2)<<20)|(DOUBLE(rd)<<14)|opf;
    break;
  case FqTOd:
    recode[1] = (QUAD(r2)<<20)|(DOUBLE(rd)<<14)|opf;
    break;
  case FiTOq:
    recode[1] = (r2<<20)|(QUAD(rd)<<14)|opf;
    break;
  case FsTOq:
    recode[1] = (r2<<20)|(QUAD(rd)<<14)|opf;
    break;
  case FdTOq:
    recode[1] = (DOUBLE(r2)<<20)|(QUAD(rd)<<14)|opf;
    break;
  case FxTOq:
    recode[1] = (DOUBLE(r2)<<20)|(QUAD(rd)<<14)|opf;
    break;
  default:
    break;
  }
}



void decode_float2(unsigned long instr, unsigned long recode[])
{
  int r1, r2, rd; 
  unsigned opf = F_opf(instr);

  r1 = F_rs1(instr); /* or cond field in FMOVcc instructions */
  r2 = F_rs2(instr);
  rd = F_rd(instr); /* or condition code registers fcc0-fcc3 */

  switch (opf) {
  case FCMPs:
    recode[1] = (r1<<26)|(r2<<20)|(rd<<14)|opf;
    break;
  case FCMPd:
    recode[1] = (DOUBLE(r1)<<26)|(DOUBLE(r2)<<20)|(rd<<14)|opf;
    break;
  case FCMPq:
    recode[1] = (QUAD(r1)<<26)|(QUAD(r2)<<20)|(rd<<14)|opf;
    break;
  case FCMPEs:
    recode[1] = (r1<<26)|(r2<<20)|(rd<<14)|opf;
    break;
  case FCMPEd:
    recode[1] = (DOUBLE(r1)<<26)|(DOUBLE(r2)<<20)|(rd<<14)|opf;
    break;
  case FCMPEq:
    recode[1] = (QUAD(r1)<<26)|(QUAD(r2)<<20)|(rd<<14)|opf;
    break;
  default:
    switch (opf & 0x3f) {
    case FMOVccS:
      recode[1] = (r1<<26)|(r2<<20)|(rd<<14)|opf;
      break;
    case FMOVccD:
      recode[1] = (r1<<26)|(DOUBLE(r2)<<20)|(DOUBLE(rd)<<14)|opf;
      break;
    case FMOVccQ:
      recode[1] = (r1<<26)|(QUAD(r2)<<20)|(QUAD(rd)<<14)|opf;
      break;
    case FMOVrcS1:
    case FMOVrcS2:
      recode[1] = (r1<<26)|(r2<<20)|(rd<<14)|opf;
      break;
    case FMOVrcD1:
    case FMOVrcD2:
      recode[1] = (r1<<26)|(DOUBLE(r2)<<20)|(DOUBLE(rd)<<14)|opf;
      break;
    case FMOVrcQ1:
    case FMOVrcQ2:
      recode[1] = (r1<<26)|(QUAD(r2)<<20)|(QUAD(rd)<<14)|opf;
      break;
    default:
      break;
    }
  }
}

void decode_impdep1(unsigned long instr, unsigned long recode[])
{
  int r1, r2, r3, rd;
  unsigned opf = F_opf(instr);

  r1 = F_rs1(instr);
  r2 = F_rs2(instr);
  r3 = F_rs3(instr);
  rd = F_rd(instr);

  switch (opf & 0xf) {
  case FMADDs:
  case FMSUBs:
  case FNMADDs:
  case FNMSUBs:
    recode[1] = (r1<<26)|(r2<<20)|(rd<<14)|(r3<<8)|(opf&0xf);
    break;
  case FMADDd:
  case FMSUBd:
  case FNMADDd:
  case FNMSUBd:
    recode[1] = (DOUBLE(r1)<<26)|
      (DOUBLE(r2)<<20)|
	(DOUBLE(rd)<<14)|
	  (DOUBLE(r3)<<8)|(opf&0xf);
  break;
  case FMADDq:
  case FMSUBq:
  case FNMADDq:
  case FNMSUBq:
    recode[1] = (QUAD(r1)<<26)|
      (QUAD(r2)<<20)|
	(QUAD(rd)<<14)|
	  (QUAD(r3)<<8)|(opf&0xf);
    break;
  default:
    break;
  }
  return;
}
