/*
 *  $Id: fl_config.c,v 1.8 1995/10/02 22:37:43 aiman Exp $
 *  
 *  Copyright (c) 1994 HAL Computer Systems International, Ltd.
 *  All rights reserved.  Unpublished -- rights reserved under
 *  the Copyright Laws of the United States.  USE OF A COPYRIGHT
 *  NOTICE IS PRECAUTIONARY ONLY AND DOES NOT IMPLY PUBLICATION
 *  OR DISCLOSURE.
 *  
 *  THIS SOFTWARE CONTAINS CONFIDENTIAL INFORMATION AND TRADE
 *  SECRETS OF HAL COMPUTER SYSTEMS INTERNATIONAL, LTD.  USE,
 *  DISCLOSURE, OR REPRODUCTION IS PROHIBITED WITHOUT THE
 *  PRIOR EXPRESS WRITTEN PERMISSION OF HAL COMPUTER SYSTEMS
 *  INTERNATIONAL, LTD.
 *  
 *                          RESTRICTED RIGHTS LEGEND
 *  Use, duplication, or disclosure by the Government is subject
 *  to the restrictions as set forth in subparagraph (c)(l)(ii)
 *  of the Rights in Technical Data and Computer Software clause
 *  at DFARS 252.227-7013.
 *  
 *           HAL COMPUTER SYSTEMS INTERNATIONAL, LTD.
 *                   1315 Dell Avenue
 *                   Campbell, CA  95008
 */
/*==============================================================================*/ 
/* Hal Denali Trace Driven Simulator						*/
/* Copyright (c) 1994 HaL Computer Systems, Inc.                      		*/
/*==============================================================================*/
/* fl_config.c									*/
/*										*/
/*==============================================================================*/
/* Revision history								*/
/*										*/
/* Date	     Who	   	Description					*/ 
/* --------  -----------      	-----------					*/
/* 10/30/94  Per Forssell  	Initial defintion				*/
/*										*/
/*==============================================================================*/

#include	<stdio.h>
#include	<stdlib.h>
#include	<string.h>

#include	"globdef.h"
#include        "config_front_end.h"

extern gconf    gl_conf;


/*==============================================================================*/

FILE *open_file () ;

/*==============================================================================*/

extern FILE     *config_file ;		/* pointer to the configuration file	*/
extern char	config_file_name[] ; 	/* configuration file name		*/

/*==============================================================================*/

/* The configuration file parser reads the configuration file, it over
   writes the default global configuration parameters with parameters found in
   the configuration file. */

void
parse_fl_config_file ()
	
{
  char	line[MAX_LINE] ;	/* input line from configuration file		*/
  char	param[MAX_NAME] ;	/* configuration parameter name			*/
  int i; 

  config_file = open_file (config_file_name, "r") ;
  while (fgets (line, MAX_LINE, config_file) != NULL) {

    strcpy (param, strtok (line, WHITE_SPACE)) ;

    if (!strcmp (param, "ICACHE_SIZE"))
      fl_conf.icache_size = 4*atoi (strtok (NULL, WHITE_SPACE)) ;
    if (!strcmp (param, "DCACHE_SIZE"))
      fl_conf.dcache_size = 4*atoi (strtok (NULL, WHITE_SPACE)) ;
    if (!strcmp (param, "UCACHE_SIZE"))
      fl_conf.ucache_size = 512*atoi (strtok (NULL, WHITE_SPACE)) ;
    if (!strcmp (param, "UCACHE_BLOCKING")) {
      fl_conf.ucache_blocking = atoi (strtok (NULL, WHITE_SPACE)) ;
      if (!(fl_conf.ucache_blocking == 1 || fl_conf.ucache_blocking ==0)) {
	 printf ("Unexpected value for UCACHE_BLOCKING. Legal values are 1,0");
	 exit(1);
      } 
    }
    if (!strcmp (param, "UCACHE_PIPESTAGES")) {
      fl_conf.ucache_pipestages = atoi (strtok (NULL, WHITE_SPACE)) ;
      if (!(fl_conf.ucache_pipestages >=1)) {
	 printf ("Unexpected value for UCACHE_PIPESTAGES. Legal values are >1 & <11");
	 exit(1);
      } 
    }
    if (!strcmp (param, "UCACHE_PMISSES")) {
      fl_conf.ucache_pmisses = atoi (strtok (NULL, WHITE_SPACE)) ;
      if (!(fl_conf.ucache_pmisses >=1 && fl_conf.ucache_pmisses <=10)) {
	 printf ("Unexpected value for UCACHE_PMISSES. Legal values are >1 & <11");
	 exit(1);
      } 
    }
    if (!strcmp (param, "ISTREAM_BUFFER")) {
      fl_conf.istream_buffer = atoi (strtok (NULL, WHITE_SPACE)) ;
      if (!(fl_conf.istream_buffer >=0 && fl_conf.istream_buffer <=10)) {
	 printf ("Unexpected value for ISTREAM_BUFFER. Legal values are >1 & <11");
	 exit(1);
      } 
    }
    if (!strcmp (param, "MTLB_SIZE"))
      fl_conf.tlb_size = atoi (strtok (NULL, WHITE_SPACE)) ;
    if (!strcmp (param, "mITLB_SIZE")) 
      fl_conf.mItlb_size = atoi (strtok (NULL, WHITE_SPACE)) ;
    if (!strcmp (param, "mDTLB_SIZE")) 
      fl_conf.mDtlb_size = atoi (strtok (NULL, WHITE_SPACE)) ;
    if (!strcmp (param, "NO_INT_LV")) 
      fl_conf.no_int_lv = atoi (strtok (NULL, WHITE_SPACE)) ;
    if (!strcmp (param, "UNIFIED_LAT")) 
      fl_conf.unified_lat = atoi (strtok (NULL, WHITE_SPACE)) ;
    if (!strcmp (param, "MEMORY_LAT")) 
      fl_conf.memory_lat = atoi (strtok (NULL, WHITE_SPACE)) ;
    if (!strcmp (param, "MTLB_LAT")) 
      fl_conf.mtlb_lat = atoi (strtok (NULL, WHITE_SPACE)) ;
    if (!strcmp (param, "MICRO_LAT")) 
      fl_conf.micro_lat = atoi (strtok (NULL, WHITE_SPACE)) ;
    if (!strcmp (param, "MAX_UT_BUF")) 
      fl_conf.max_ut_buf = atoi (strtok (NULL, WHITE_SPACE)) ;
    if (!strcmp (param, "NO_DRT_BUF")) 
      fl_conf.drty_buf = atoi (strtok (NULL, WHITE_SPACE)) ;
    if (!strcmp (param, "ULINESIZE")) { 
      fl_conf.uline = atoi (strtok (NULL, WHITE_SPACE)) ;
      for (i=0;i<32;i++) 
	if ((fl_conf.uline>>i)&0x1) {
	  fl_conf.ushft = i;
	  break;
	}
    }

    if (!strcmp (param, "IDLINESIZE")) { 
      fl_conf.idline = atoi (strtok (NULL, WHITE_SPACE)) ;
      for (i=0;i<32;i++) 
	if ((fl_conf.idline>>i)&0x1) {
	  fl_conf.idshft = i;
	  break;
	}
    }
    
  }
  fl_conf.icache_size = fl_conf.icache_size*(128/fl_conf.idline);
  fl_conf.dcache_size = fl_conf.dcache_size*(128/fl_conf.idline);
  fl_conf.ucache_size = fl_conf.ucache_size*(512/fl_conf.uline)*4;


  fclose (config_file) ;
 
} /* end parse_fl_config_file */

/*==============================================================================*/

/* The global configuration routine reads in all global configuration parameters
   from the configuration file. Default parameter values are used for parameters
   not defined in the configuration file or if no configuration file is
   specified. */

void
fl_config ()	
	
{
  int i;
  /* Initialize all configuration parameters to their default values. */

  fl_conf.max_ut_buf       = DEF_MAX_UT_BUF ;
  fl_conf.drty_buf         = DEF_NO_DRT_BUF ;
  fl_conf.icache_size      = 4*DEF_IENTRY ;
  fl_conf.dcache_size      = 4*DEF_DENTRY ;
  fl_conf.uline            = DEF_ULINESIZE;
  fl_conf.ucache_size      = 512*DEF_UENTRY*4;
  fl_conf.ucache_blocking  = 0;
  fl_conf.ucache_pipestages= 3;
  fl_conf.ucache_pmisses   = 3;
  fl_conf.istream_buffer   = 0;
  fl_conf.tlb_size         = DEF_TENTRY ;
  fl_conf.mItlb_size       = DEF_mITENTRY ;
  fl_conf.mDtlb_size       = DEF_mDTENTRY ;
  fl_conf.no_int_lv        = DEF_NO_INT_LV ;
  fl_conf.unified_lat      = DEF_UNIFIED_LAT ;
  fl_conf.memory_lat       = DEF_MEMORY_LAT ;
  fl_conf.micro_lat        = DEF_MICRO_LAT ;
  fl_conf.mtlb_lat         = DEF_MTLB_LAT ;
  fl_conf.idline           = DEF_IDLINESIZE;

  for (i=0;i<32;i++) 
    if ((fl_conf.uline>>i)&0x1) {
      fl_conf.ushft = i;
      break;
    }

  for (i=0;i<32;i++) 
    if ((fl_conf.idline>>i)&0x1) {
      fl_conf.idshft = i;
      break;
    }

  /* Read parameters from configuration file. */

  if (strcmp (config_file_name, ""))
    parse_fl_config_file () ;


} /* end fl_config */

/*==============================================================================*/




