/*
 *  $Id: front_end.c,v 1.16 1995/10/02 22:37:46 aiman Exp $
 *  
 *  Copyright (c) 1994 HAL Computer Systems International, Ltd.
 *  All rights reserved.  Unpublished -- rights reserved under
 *  the Copyright Laws of the United States.  USE OF A COPYRIGHT
 *  NOTICE IS PRECAUTIONARY ONLY AND DOES NOT IMPLY PUBLICATION
 *  OR DISCLOSURE.
 *  
 *  THIS SOFTWARE CONTAINS CONFIDENTIAL INFORMATION AND TRADE
 *  SECRETS OF HAL COMPUTER SYSTEMS INTERNATIONAL, LTD.  USE,
 *  DISCLOSURE, OR REPRODUCTION IS PROHIBITED WITHOUT THE
 *  PRIOR EXPRESS WRITTEN PERMISSION OF HAL COMPUTER SYSTEMS
 *  INTERNATIONAL, LTD.
 *  
 *                          RESTRICTED RIGHTS LEGEND
 *  Use, duplication, or disclosure by the Government is subject
 *  to the restrictions as set forth in subparagraph (c)(l)(ii)
 *  of the Rights in Technical Data and Computer Software clause
 *  at DFARS 252.227-7013.
 *  
 *           HAL COMPUTER SYSTEMS INTERNATIONAL, LTD.
 *                   1315 Dell Avenue
 *                   Campbell, CA  95008
 */

/**********************************************************************
  Program to evaluate benchmark programs on Denali. This portion of
  the simulator models the denali i-cache, d-cache, micro i-tlb,
  micro d-tlb, main tlb, the unified cache, and the branch unit. 
  
  Started 11/1/94, Nirmal Saxena 
  
  ***********************************************************************/
#include <stdio.h>
#include "globdef.h"
#include "config_front_end.h"
#include "opcodes.h"
#define READY 0
#define DTMISS 1
#define DTWAIT 2
#define DREQ   3
#define DCMISS 4
#define DCWAIT 5
#define debug 0
/* This function returns 0 if the trace file is not exhausted or the
   pending load store transaction buffers are not empty;  else
   this function returns a 1. However, the main() routine will
   continue calling this function until all of the outstanding
   ld/st unit requests are done.
   */
static unsigned count=0;       /* flag to detect first call           */
static unsigned trace_count=0; /* to count the number of records read */
static unsigned dirty_buffer[8] = {0,0,0,0,0,0,0,0};
int access_icache(unsigned);
int fill_icache(unsigned);
int access_dcache(unsigned,unsigned);
int fill_dcache(unsigned,unsigned);
int access_mItlb(unsigned);
int fill_mItlb(unsigned);
int access_mDtlb(unsigned);
int fill_mDtlb(unsigned);
int access_tlb(unsigned);
unsigned access_drtybuf(unsigned);
void fill_drtybuf(unsigned);
void dump_state(lsreq *ls_r);
extern unsigned long clk_cycle;
extern FILE *trace_file;
extern char warmup_file_name[];
extern gconf gl_conf;

int front_end(ireq  *fetch_r ,   /* instruction fetch request         */
	      lsreq *ls_r ,      /* load store request                */
	      ipkt  *fetch_p ,   /* instruction fetch packet array    */
	      lspkt *ls_p )      /* load store packet array           */
{
  int i, i1, j, k, m, flag, flagu;
  int cnt;			/* instruction count requested */
  unsigned temp, temp1;
  unsigned long	index;
  unsigned long recode[2];      /* recoded instruction         */
  static unsigned trace_done=0; /* flag to remember trace done */
  static unsigned howmany=0;
  
  howmany++;
  
  cnt = fetch_r->cnt;



  /*
  printf("fetch count %u\n",cnt);
  */

  if (count == 0) {
    fill_tracebuf();           /* filling the trace buffer on first call */
  }
  
  if (count == 0)
    trace_done = 1;            /* no more records in the trace file      */
  
  for (i=0;i<MAX_IFETCH;i++)
    fetch_p[i].valid = -1;     /* initialize to invalid                  */
  /* The instruction references are handled as follows:
     
     1. Check if there is any outstanding micro_tlb miss that is
     being handled.  If yes go to 2. If NO go to 4.
     
     2. Check if the miss handling is about to complete. If NO
     return invalid packet information. If YES then go to 3.
     
     3. Access i0cache to determine if HIT occurs. If HIT then
     return data requested up to line break, or cnt, or
     first occurance of branch (with delay slot); whichever
     comes first. If MISS install a request to the Unified
     cache (in the ucache wait buffer) and send invalid 
     packet information.
     
     4. Check to see if there is an outstanding i0cache miss.
     If yes go to 5. If no, go to 6.
     
     5. Check if i0cache miss handling is about to complete. If yes,
     return instructions requested up to line break, or cnt,
     or first branch (with delay slot); whichever comes first.
     else, return invalid data.
     
     6. Access micro itlb. If hits then  GO TO 3, else install
     main tlb request in the main_tlb wait buffer and send
     invalid packet information.
     
     */
  
  if (trace_done == 0) {
    
    flag = 0;
    
    /* check to see if there is an outstanding micro itlb miss.
       The flagu value signals true and also indicates the number of remaining
       cycles for completion */
    
    for (i=0;i<(fl_conf.max_ut_buf+1);i++) {
      if ((wt_main_tlb[i].state==VALID)&&(wt_main_tlb[i].requestor==I0REQ)) {
	if (i==0)
	  flag = wt_main_tlb[i].cycles; /* currently active */
	else
	  flag = 1000; 
	break;
      }
    }
    
    switch (flag) {
    case 0:
      /* now check to see if there is any outstanding i0cache miss */
      flagu = 0;    /* indicator for outstanding i0cache misses */
      for (i=0;i<2;i++) {
	if ((wt_iu_cache[i].state==COMPLETED)) {
          flagu = 1;
	  wt_iu_cache[i].state = INVALID;
	  break;
	}
        if ((wt_iu_cache[i].state==INPROGRESS) || 
	    (wt_iu_cache[i].state==VALID)) {
          flagu = 1000; 
	  break;
	}

      }
      if ((flagu == 1)|| (flagu==0)) { /* i0cache miss completing in this cycle or
					  no outstanding i0cache misses */
	if (flagu==0)
	  access_mItlb(mrkr->pc);
	if ((flagu==0)&&mImiss_flag) { /* install main_tlb_request */
	  for (i=0;i<(fl_conf.max_ut_buf+1);i++) {
	    if (wt_main_tlb[i].state==INVALID) {
	      if (i==0) {
		access_tlb(mrkr->pc);
		if (Tmiss_flag)
		  wt_main_tlb[i].cycles = fl_conf.mtlb_lat+fl_conf.micro_lat;
		else
		  wt_main_tlb[i].cycles = fl_conf.micro_lat;
  	      }
	      wt_main_tlb[i].va = mrkr->pc;
	      wt_main_tlb[i].iw = mrkr->iw;
	      wt_main_tlb[i].state = VALID;
	      wt_main_tlb[i].requestor = I0REQ;
	      flag = 1;
	      break;
	    }
	  }
	  if (flag==0) { /* no room to install main_tlb request, exit */
            printf("cycle count %u\n", howmany);
	    printf("main tlb request over run for i0tlb\n");
	    exit(-1);
	  }
	  for (i=0;i<MAX_IFETCH;i++) /* do not return any inst */
	    fetch_p[i].valid = -1;
       }   /* no miss in the i0 micro tlb or no need to access micro tlb */
	else {       /* micro tlb hit or access not needed */
	  if (flagu==0) 
	    access_icache(mrkr->pc); /* access i0cache now */
	  else
	    fill_icache(mrkr->pc);   /* fill i0cache */
	  
	  if ((flagu==0)&&Imiss_flag) { /* install unified cache request */
	     /* NIRMAL/AIMAN: change 2 to a define */
	    for (i=0;i<(2);i++) {
	      if (wt_iu_cache[i].state==INVALID) {
		wt_iu_cache[i].va = mrkr->pc;
		wt_iu_cache[i].iw = mrkr->iw;
		wt_iu_cache[i].cmd = READ; 
		wt_iu_cache[i].state = VALID;
		wt_iu_cache[i].cycles = 0;
		wt_iu_cache[i].requestor = I0REQ;
		if (i==0) {
/*  		  access_cache(mrkr->pc,0, &wt_unifd_cache[i]);  */
/* NIRMAL:access_cache() does not do anything unless you define OLD_U2 in u2.c
   we should remove the call once we test the final changes.
*/
/* 		  if (Umiss_flag) */
/* 		    wt_unifd_cache[i].cycles = fl_conf.memory_lat+fl_conf.unified_lat; */
/* 		  else */
/* 		    wt_unifd_cache[i].cycles = fl_conf.unified_lat; */
		}
		flag = 1;
		break;
	      }
	    }
	    if (flag==0) { /* no room to install unified cache request, exit */
              printf("cycle count %u\n", howmany);
	      printf("unified cache request over run for i0\n");
	      exit(-1);
	    }
	    for (i=0;i<MAX_IFETCH;i++) /* do not return any inst */
	      fetch_p[i].valid = -1;
	  }   /* no i0 cache miss */
	  else { /* both micro tlb and i0 cache hit */
	    
	    /* now send the instructions back to the back_end unit
	       with the following stopping rules whichever comes first:
	       
	       1. send as many as cnt (<=MAX_IFETCH) instructions
	       2. send up to cache line break
	       3. send up to branch with delay slot
	       4. send up to end of trace
	       */
	    
	    if (cnt >= ((fl_conf.idline -(mrkr->pc%fl_conf.idline))/4)) {
	      ln_brks++;
	      cnt =(fl_conf.idline -(mrkr->pc%fl_conf.idline))/4;
	    }
	    
	    i = 0;
	    flag = 0; /* now use it to signal branch or tcc detection */
	    
	    while ((trace_count < count)&&(i<cnt)&&(flag==0)) {
	      fetch_p[i].valid = 1;
	      fetch_p[i].inst  = mrkr->iw;
	      decode(mrkr->iw,recode);
	      fetch_p[i].recode[0] = recode[0];
	      fetch_p[i].recode[1] = recode[1];
	      fetch_p[i].fetch_cycle = clk_cycle + 1;
	      fetch_p[i].taken     = (mrkr->flags>0x1)&0x1;
	      fetch_p[i].pred      =
		BrPred(recode,mrkr->pc,fetch_p[i].taken) 
		  | rpStack(recode,mrkr->pc,mrkr->ea);
	      fetch_p[i].pc        = mrkr->pc;
	      fetch_p[i].ea        = mrkr->ea;
	      switch (recode[0]) { /* detect branches */
	      case Tcc:
	      case DONE:
	      case CALL:
	      case BiccP:
	      case FBfccP:
	      case Bicc:
	      case FBfcc:
	      case BregP:
	      case JMPL:
	      case RETURN:
		if ((fetch_p[i].taken==1)||(fetch_p[i].pred==0))
		  flag = 1;
		break;
	      default:
		break;
	      }
	      
	      trace_count++;
              if (trace_count == count)
		{
		  fill_tracebuf();
		}
              else
		mrkr++;
	      i++;
	    }
	    
	    /* now add the delay slot if possible */
	    
	    if (i<cnt)
	      if (trace_count < count) {
		fetch_p[i].valid = 1;
		fetch_p[i].inst  = mrkr->iw;
		decode(mrkr->iw,recode);
		fetch_p[i].recode[0] = recode[0];
		fetch_p[i].recode[1] = recode[1];
		fetch_p[i].fetch_cycle = clk_cycle+1;
		fetch_p[i].taken     = (mrkr->flags>0x1)&0x1;
		fetch_p[i].pred      =
		  BrPred(recode,mrkr->pc,fetch_p[i].taken)
		    | rpStack(recode,mrkr->pc,mrkr->ea);
		fetch_p[i].pc        = mrkr->pc;
		fetch_p[i].ea        = mrkr->ea;
		trace_count++;
                if (trace_count == count)
		  fill_tracebuf();
                else
		  mrkr++;
		i++;
	      }
	      else {
		fill_tracebuf();
		if (count == 0)
		  trace_done = 1;
		else {
		  fetch_p[i].valid = 1;
		  fetch_p[i].inst  = mrkr->iw;
		  decode(mrkr->iw,recode);
		  fetch_p[i].recode[0] = recode[0];
		  fetch_p[i].recode[1] = recode[1];
		  fetch_p[i].fetch_cycle = clk_cycle+1;
		  fetch_p[i].taken     = (mrkr->flags>0x1)&0x1;
		  fetch_p[i].pred      = 
		    BrPred(recode,mrkr->pc,fetch_p[i].taken)
		      | rpStack(recode,mrkr->pc,mrkr->ea);
		  fetch_p[i].pc        = mrkr->pc;
		  fetch_p[i].ea        = mrkr->ea;
		  trace_count++;
                  mrkr++;
		  i++;
		}
	      }
	    
	    for (j=i;j<MAX_IFETCH;j++)
	      fetch_p[j].valid=-1;  /* invalidate the rest */
	  }
	}
      }
      break;
    case 1:
      /* make sure that mrkr->pc is same as wt_main_tlb[0].va */
      if (mrkr->pc != wt_main_tlb[0].va) {
	printf("error! main_tlb va request not the same\n");
	exit(-1);
      }
      fill_mItlb(wt_main_tlb[0].va);
      access_icache(wt_main_tlb[0].va); /* access i0cache now */
      flag = 0;
      if (Imiss_flag) { /* install unified cache request */
	 /* NIRMAL/AIMAN: make 2 a define */
	for (i=0;i<(2);i++) {
	  if (wt_iu_cache[i].state==INVALID) {
	    wt_iu_cache[i].va = mrkr->pc;
	    wt_iu_cache[i].iw = mrkr->iw;
	    wt_iu_cache[i].state = VALID;
	    wt_iu_cache[i].cycles = 0;
	    wt_iu_cache[i].cmd = READ; 
	    wt_iu_cache[i].requestor = I0REQ;
	    if (i==0) {
/* 	      access_cache(mrkr->pc,0,&wt_unifd_cache[i]); */
/* NIRMAL:access_cache() does not do anything unless you define OLD_U2 in u2.c
   we should remove the call once we test the final changes.
*/
/* 	      if (Umiss_flag) */
/* 		wt_unifd_cache[i].cycles = fl_conf.memory_lat+fl_conf.unified_lat; */
/* 	      else */
/* 		wt_unifd_cache[i].cycles = fl_conf.unified_lat; */
	    }
	    flag = 1;
	    break;
	  }
	}
	if (flag==0) { /* no room to install unified_cache request, exit */
	  printf("cycle count %u\n", howmany);
	  printf("unified cache request over run from i0 cache\n");
	  exit(-1);
	}
	for (i=0;i<MAX_IFETCH;i++) /* do not return any inst */
	  fetch_p[i].valid = -1;
      }   /* no i0 cache miss */
      else { /* both micro tlb and i0 cache hit */
	
	/* now send the instructions back to the back_end unit
	   with the following stopping rules whichever comes first:
	   
	   1. send as many as cnt (<=MAX_IFETCH) instructions
	   2. send up to cache line break
	   3. send up to branch with delay slot
	   4. send up to end of trace
	   */
	
	if (cnt >= ((fl_conf.idline -(mrkr->pc%fl_conf.idline))/4)) {
	  ln_brks++;
	  cnt =(fl_conf.idline -(mrkr->pc%fl_conf.idline))/4;
	}
	
	i = 0;
	flag = 0; /* now use it to signal branch or tcc detection */
	
	while ((trace_count < count)&&(i<cnt)&&(flag==0)) {
	  fetch_p[i].valid = 1;
	  fetch_p[i].inst  = mrkr->iw;
	  decode(mrkr->iw,recode);
	  fetch_p[i].recode[0] = recode[0];
	  fetch_p[i].recode[1] = recode[1];
	  fetch_p[i].taken     = (mrkr->flags>0x1)&0x1;
	  fetch_p[i].fetch_cycle = clk_cycle+1;
	  fetch_p[i].pred      = 
	    BrPred(recode,mrkr->pc,fetch_p[i].taken)
	      | rpStack(recode,mrkr->pc,mrkr->ea);
	  fetch_p[i].pc        = mrkr->pc;
	  fetch_p[i].ea        = mrkr->ea;
          switch (recode[0]) { /* detect branches */
	  case Tcc:
	  case CALL:
	  case BiccP:
	  case FBfccP:
	  case Bicc:
	  case FBfcc:
	  case BregP:
	  case JMPL:
	  case RETURN:
	    if ((fetch_p[i].taken==1)||(fetch_p[i].pred==0))
	      flag = 1;
	    break;
	  default:
	    break;
	  }
	  trace_count++;
          if (trace_count == count)
	    fill_tracebuf();
          else
            mrkr++;
	  i++;
	}
	
	/* now add the delay slot if possible */
	
	if (i<cnt)
	  if (trace_count < count) {
	    fetch_p[i].valid = 1;
	    fetch_p[i].inst  = mrkr->iw;
	    decode(mrkr->iw,recode);
	    fetch_p[i].recode[0] = recode[0];
	    fetch_p[i].recode[1] = recode[1];
	    fetch_p[i].fetch_cycle = clk_cycle+1;
	    fetch_p[i].taken     = (mrkr->flags>0x1)&0x1;
	    fetch_p[i].pred      = 
	      BrPred(recode,mrkr->pc,fetch_p[i].taken)
		| rpStack(recode,mrkr->pc,mrkr->ea);
	    fetch_p[i].pc        = mrkr->pc;
	    fetch_p[i].ea        = mrkr->ea;
	    trace_count++;
            if (trace_count == count)
	      fill_tracebuf();
            else
	      mrkr++;
	    i++;
	  }
	  else {
	    fill_tracebuf();
	    if (count == 0)
	      trace_done = 1;
	    else {
	      fetch_p[i].valid = 1;
	      fetch_p[i].inst  = mrkr->iw;
	      decode(mrkr->iw,recode);
	      fetch_p[i].recode[0] = recode[0];
	      fetch_p[i].recode[1] = recode[1];
	      fetch_p[i].fetch_cycle = clk_cycle+1;
	      fetch_p[i].taken     = (mrkr->flags>0x1)&0x1;
	      fetch_p[i].pred      = 
		BrPred(recode,mrkr->pc,fetch_p[i].taken)
		  | rpStack(recode,mrkr->pc,mrkr->ea);
	      fetch_p[i].pc        = mrkr->pc;
	      fetch_p[i].ea        = mrkr->ea;
	      trace_count++;
              if (trace_count == count)
		fill_tracebuf();
              else
		mrkr++;
	      i++;
	    }
	  }
	
	for (j=i;j<MAX_IFETCH;j++)
	  fetch_p[j].valid=-1;  /* invalidate the rest */
      }
      
      break;
      
    default:
      for (i=0;i<MAX_IFETCH;i++)
	fetch_p[i].valid = -1; /* return none */
      break;
    }
  }
  
  /* The following portion of the code handles data cache load/store
   * requests.
   */


  /* Snoop invalidations */
  for (k=0;k<1;k++)    {
     if (wt_snoop_cache[k].state == VALID) {
	wt_snoop_cache[k].state = INVALID;
	invalidate_d0(wt_snoop_cache[k].ea);
	break;
     }
  }
    


  for (k=0;k<gl_conf.no_of_ls_bufs;k++)    
    ls_p[k].valid = -1;

  
  /* First receive new requests from the back_end */

  flag = 0;
  for (i=0;i<gl_conf.no_of_ls_bufs;i++) {
    if (ls_r[i].valid==1) {
      flag = 0;
      for (j=0;j<gl_conf.no_of_ls_bufs;j++) {
	if ( wt_ldst_req[j].state == INVALID ) {
	  wt_ldst_req[j].state = VALID;
	  wt_ldst_req[j].ea = ls_r[i].ea;
	  wt_ldst_req[j].qslot_no = ls_r[i].qslot_no;
	  wt_ldst_req[j].recode[0] = ls_r[i].recode[0];
	  wt_ldst_req[j].recode[1] = ls_r[i].recode[1];
	  wt_ldst_req[j].va = READY; /* this is being used to track ready */
				 /* status, this is not virt. address */
	  flag=1;
	  break;
	}
      }
	if (flag==0) { /* no room to install, exit */
	  printf(" request over run for ld/st unit \n");
	  dump_state(ls_r);
	  exit(-1);
	} 
    }
  }

  /* now process the ready requests with interleave factor in account */

  for (i=0; i<gl_conf.no_of_ls_bufs; i++) {
    if ((wt_ldst_req[i].state == VALID) &&
	(wt_ldst_req[i].va == READY) &&
	(interleave[(wt_ldst_req[i].ea>>3)%fl_conf.no_int_lv] == 0)) {
      interleave[(wt_ldst_req[i].ea>>3)%fl_conf.no_int_lv] = 1;
      for (j=0;j<gl_conf.no_of_ls_bufs;j++) {
	if ((wt_ldst_req[j].state == VALID) &&
	    (j != i) &&
	    ((wt_ldst_req[j].ea&PAGE)==(wt_ldst_req[i].ea&PAGE)) &&
	    (wt_ldst_req[j].va == DTMISS)) {
	     wt_ldst_req[i].va = DTWAIT;
	     break;
	   }
      }
      if (wt_ldst_req[i].va == READY) {
	wt_ldst_req[i].va = DREQ;
	access_mDtlb(wt_ldst_req[i].ea);
	if (mDmiss_flag) {
	  wt_ldst_req[i].va = DTMISS;
	  for (j=0;j<fl_conf.max_ut_buf;j++) {
	    if (wt_main_tlb[j].state == INVALID) {
	      if (j==0) {
		access_tlb(wt_ldst_req[i].ea);
		if (Tmiss_flag)
		  wt_main_tlb[j].cycles = fl_conf.mtlb_lat+fl_conf.micro_lat;
		else
		  wt_main_tlb[j].cycles = fl_conf.micro_lat;
	      }
	      wt_main_tlb[j].ea = wt_ldst_req[i].ea;
	      wt_main_tlb[j].state = VALID;
	      wt_main_tlb[j].qslot_no = wt_ldst_req[i].qslot_no;
	      wt_main_tlb[j].recode[0] = wt_ldst_req[i].recode[0];
	      wt_main_tlb[j].requestor = D0REQ;
	      break;
	    }
	  }
	}
      }
    }
  }
  
  /* check to see if there is an outstanding micro dtlb miss that is about to
     finish. flag value signals true and also indicates the number of remaining
     cycles for completion */
  
  flag = 0;
  if ((wt_main_tlb[0].state==VALID)&&(wt_main_tlb[0].requestor==D0REQ))
    flag = wt_main_tlb[0].cycles; /* currently active */
  else
    flag = 1000; 
  
  k = 0; /* pointer to index return packet array */
  
  if (flag==1) { /* process tlb miss */
    fill_mDtlb(wt_main_tlb[0].ea);
    for (i=0;i<gl_conf.no_of_ls_bufs;i++) {
      if ((wt_main_tlb[0].ea == wt_ldst_req[i].ea ) &&
	  (wt_main_tlb[0].qslot_no == wt_ldst_req[i].qslot_no) &&
	  (wt_ldst_req[i].va==DTMISS) &&
	  (wt_ldst_req[i].state==VALID)) {
	wt_ldst_req[i].va = DREQ; /* data cache request */
	break;
      }
    }
    for (i=0;i<gl_conf.no_of_ls_bufs;i++) {
      if ((wt_ldst_req[i].state==VALID) &&
	  ((wt_ldst_req[i].ea&PAGE)==(wt_main_tlb[0].ea&PAGE)) &&
	  (wt_ldst_req[i].va == DTWAIT)) {
	access_mDtlb(wt_ldst_req[i].ea);
	if (mDmiss_flag) {
	  printf("micro d-tlb reload error\n");
	  exit(-1);
	}
	wt_ldst_req[i].va = DREQ;
      }
    }
  }

 /* now process the data cache requests that are result of tlb refills */

  for (i=0;i<gl_conf.no_of_ls_bufs;i++) {
    if ((wt_ldst_req[i].state==VALID) &&
	(wt_ldst_req[i].va == DREQ)) {
      for (j=0;j<gl_conf.no_of_ls_bufs;j++) {
	if ((wt_ldst_req[j].state==VALID) &&
	    (i != j) &&
	    ((wt_ldst_req[j].ea >> fl_conf.idshft) == (wt_ldst_req[i].ea >>fl_conf.idshft)) &&
	    (wt_ldst_req[j].va == DCMISS)) {
	  wt_ldst_req[i].va = DCWAIT;
	  break;
	}
      }
      if (wt_ldst_req[i].va == DREQ) {
	access_dcache(wt_ldst_req[i].ea,wt_ldst_req[i].recode[0]);
	if (Dmiss_flag==1)
	  if (access_drtybuf(wt_ldst_req[i].ea)) {
	    Dmiss_flag = 0;
	    fill_dcache(wt_ldst_req[i].ea,wt_ldst_req[i].recode[0]);
	    if (Dmiss_repl)
	      fill_drtybuf(dirty_va);
	  }
     
	if ((Dmiss_flag)||(command == RETAG)) {
	  wt_ldst_req[i].va = DCMISS;
	  flagu = 0;
	  for (j=0;j<(fl_conf.max_ut_buf+1);j++) {
	    if (wt_unifd_cache[j].state==INVALID) {
	      wt_unifd_cache[j].state = VALID;
              wt_unifd_cache[j].cmd = command;
	      wt_unifd_cache[j].ea = wt_ldst_req[i].ea;
	      wt_unifd_cache[j].recode[0] = wt_ldst_req[i].recode[0];
	      wt_unifd_cache[j].qslot_no = wt_ldst_req[i].qslot_no;
	      wt_unifd_cache[j].requestor = D0REQ;
	      if (j==0) {
/* 		access_cache(wt_ldst_req[i].ea,1,&wt_unifd_cache[j]);  */
/* NIRMAL:access_cache() does not do anything unless you define OLD_U2 in u2.c
   we should remove the call once we test the final changes.
*/
/* 		if (Umiss_flag) */
/* 		  wt_unifd_cache[j].cycles = fl_conf.memory_lat+fl_conf.unified_lat; */
/* 		else */
/* 		  wt_unifd_cache[j].cycles = fl_conf.unified_lat; */
	      }
	      flagu = 1;
	      break;
	    }
	  }
	  if (flagu==0) { /* no room to install data_cache request */
	    printf("cycle count %u\n", howmany);
	    printf("unified cache request over run from d0 cache\n");
	    exit(-1);
	  }
	}
	else { /* hit case send packet */
	  wt_ldst_req[i].state = INVALID;
	  interleave[(wt_ldst_req[i].ea>>3)%fl_conf.no_int_lv] = 0;
	  ls_p[k].valid = 1;
	  ls_p[k].qslot_no = wt_ldst_req[i].qslot_no;
	  ls_p[k].compl_cycle = clk_cycle+gl_conf.d0_lat;
	  k++;
	}
      }
    }
  }
    

 
  /* check to see if there is any data cache request that is about to
     complete */
  flag = 0;

 /* variable i1 is being used to register the transaction that completed */

 for (i1=0; i1 <(fl_conf.max_ut_buf+1); i1++) {
  if ((wt_unifd_cache[i1].state==COMPLETED)&&(wt_unifd_cache[i1].requestor==D0REQ)) {
    flag = 1;
    temp1 = wt_unifd_cache[i1].ea ;
    wt_unifd_cache[i1].state = INVALID;
    break;
  }
  else
    flag = 1000;
}
  
  if (flag == 1) {
    for (i=0;i<gl_conf.no_of_ls_bufs;i++)
      if (((wt_unifd_cache[i1].ea) == (wt_ldst_req[i].ea) ) &&
	  (wt_unifd_cache[i1].qslot_no == wt_ldst_req[i].qslot_no) &&
	  (wt_ldst_req[i].va==DCMISS) &&
	  (wt_ldst_req[i].state==VALID)) {
	ls_p[k].valid = 1;
	ls_p[k].qslot_no = wt_unifd_cache[i1].qslot_no;
	interleave[(wt_unifd_cache[i1].ea>>3)%fl_conf.no_int_lv] = 0;
	ls_p[k].compl_cycle = clk_cycle + 1;
	wt_ldst_req[i].state = INVALID; /* not valid */
        if (wt_unifd_cache[i1].cmd != RETAG) {
	 fill_dcache(wt_unifd_cache[i1].ea,wt_unifd_cache[i1].recode[0]);
	if (Dmiss_repl)
	  fill_drtybuf(dirty_va);
        }
	break;
      }
    k++;
    /* this will satisfy other outstanding misses in the same cache line*/
    for (i=0;i<gl_conf.no_of_ls_bufs;i++) {
      if ((wt_ldst_req[i].state==VALID)
	  &&(wt_ldst_req[i].va == DCWAIT)
	  &&((temp1>>fl_conf.idshft)==(wt_ldst_req[i].ea>>fl_conf.idshft)))
	{
	  access_dcache(wt_ldst_req[i].ea,wt_ldst_req[i].recode[0]);
	  if (Dmiss_flag) {
	    printf("data cache miss reload error\n");
	    exit(-1);
	  }
	  ls_p[k].valid = 1;
	  interleave[(wt_ldst_req[i].ea>>3)%fl_conf.no_int_lv]=0;
	  ls_p[k].qslot_no = wt_ldst_req[i].qslot_no;
	  ls_p[k].compl_cycle = clk_cycle+1;
	  wt_ldst_req[i].state = INVALID;
	  k++;
	}
    }
 }

  /* now update the wait queues */

  for (i=0;i<(gl_conf.no_of_ls_bufs-1);i++)
    for (j=(i+1);j<gl_conf.no_of_ls_bufs;j++) {
      if ((wt_ldst_req[j].state==VALID) &&(wt_ldst_req[i].state==INVALID)) {
	wt_ldst_req[i].state = wt_ldst_req[j].state;
	wt_ldst_req[i].ea = wt_ldst_req[j].ea;
	wt_ldst_req[i].qslot_no = wt_ldst_req[j].qslot_no;
	wt_ldst_req[i].recode[0] = wt_ldst_req[j].recode[0];
	wt_ldst_req[i].va = wt_ldst_req[j].va;
	wt_ldst_req[j].state = INVALID;
	break;
      }
    }
  flag = 0; /* now use this to signal completion to back_end */  
  
  for (i=0;i<(fl_conf.max_ut_buf+1);i++) {
   if (wt_unifd_cache[i].state != INVALID)
       {
	  flag=1;
	  break;
       }
  }
  
  for (i=0;i<2;i++) {
   if (wt_iu_cache[i].state != INVALID)
       {
	  flag=1;
	  break;
       }
  }
  
  flagu=flag;
  flag=0;


  if (wt_main_tlb[0].cycles==1) {
    for (i=0;i<fl_conf.max_ut_buf;i++) 
      for (j=(i+1);j<(fl_conf.max_ut_buf+1);j++) 
	/* BUG:                        ^^^^  */
	if (wt_main_tlb[j].state==VALID) {
	  wt_main_tlb[i].state = wt_main_tlb[j].state;
	  wt_main_tlb[i].cycles = wt_main_tlb[j].cycles;
	  wt_main_tlb[i].qslot_no = wt_main_tlb[j].qslot_no;
	  wt_main_tlb[i].va = wt_main_tlb[j].va;
	  wt_main_tlb[i].ea = wt_main_tlb[j].ea;
	  wt_main_tlb[i].iw = wt_main_tlb[j].iw;
	  wt_main_tlb[i].requestor = wt_main_tlb[j].requestor;
	  wt_main_tlb[i].recode[0] = wt_main_tlb[j].recode[0];
	  wt_main_tlb[i].recode[1] = wt_main_tlb[j].recode[1];
	  wt_main_tlb[j].state = INVALID; /* invalidate the shifted request */
	  flag = 1;
	  break;
	}
    if(flag==1) {
      if (wt_main_tlb[0].requestor==I0REQ)
	access_tlb(wt_main_tlb[0].va);
      else
	access_tlb(wt_main_tlb[0].ea);
      if (Tmiss_flag) 
	wt_main_tlb[0].cycles = fl_conf.micro_lat+fl_conf.mtlb_lat;
      else
	wt_main_tlb[0].cycles = fl_conf.micro_lat;
    }
    else
      wt_main_tlb[0].state = INVALID;
  }
  else 
    if (wt_main_tlb[0].state==VALID)
      wt_main_tlb[0].cycles--;
  
  flag = flagu | flag; 

  if (debug)
    dump_state(ls_r);

  if ((flag==0) && (trace_done==1))
    return(1);
  else 
    return(0);
}

void dump_state(lsreq *ls_r)
{
 int i;

 printf("clock cycles %u\n", clk_cycle);

 for (i=0;i<gl_conf.no_of_ls_bufs;i++) {
   if (ls_r[i].valid==1) {
     printf("ea %8x qslot %8u opcode %8x\n", ls_r[i].ea, ls_r[i].qslot_no,
	    ls_r[i].recode[0]);
   }
 }

 for (i=0;i<gl_conf.no_of_ls_bufs;i++) {
    printf("ea %8x slot %u valid %8d st %8x \n", wt_ldst_req[i].ea,
	   (wt_ldst_req[i].ea>>3)%fl_conf.no_int_lv,
	   wt_ldst_req[i].state, wt_ldst_req[i].va);
  }
  printf("\n\n");
 
  printf ("\n\n");
  printf("TLB STATE\n");
  
  for (i=0;i<(gl_conf.no_of_ls_bufs+1);i++)
    printf("%8d ",wt_main_tlb[i].state);
  printf("\n");
  for (i=0;i<(gl_conf.no_of_ls_bufs+1);i++)
    printf("%8x ",wt_main_tlb[i].iw);
  printf("\n");
  for (i=0;i<(gl_conf.no_of_ls_bufs+1);i++)
    printf("%8x ",wt_main_tlb[i].va);
  printf("\n");
  for (i=0;i<(gl_conf.no_of_ls_bufs+1);i++)
    printf("%8x ",wt_main_tlb[i].ea);
  printf("\n");
  for (i=0;i<(gl_conf.no_of_ls_bufs+1);i++)
    printf("%8u ",wt_main_tlb[i].cycles);
  printf("\n");
  for (i=0;i<(gl_conf.no_of_ls_bufs+1);i++)
    printf("%8x ",wt_main_tlb[i].requestor);
  printf("\n\n");
  
  printf("CACHE STATE\n");
  
  for (i=0;i<(gl_conf.no_of_ls_bufs+1);i++)
    printf("%8d ",wt_unifd_cache[i].state);
  printf("\n");
  for (i=0;i<(gl_conf.no_of_ls_bufs+1);i++)
    printf("%8x ",wt_unifd_cache[i].iw);
  printf("\n");
  for (i=0;i<(gl_conf.no_of_ls_bufs+1);i++)
    printf("%8x ",wt_unifd_cache[i].va);
  printf("\n");
  for (i=0;i<(gl_conf.no_of_ls_bufs+1);i++)
    printf("%8x ",wt_unifd_cache[i].ea);
  printf("\n");
  for (i=0;i<(gl_conf.no_of_ls_bufs+1);i++)
    printf("%8u ",wt_unifd_cache[i].cycles);
  printf("\n");
  for (i=0;i<(gl_conf.no_of_ls_bufs+1);i++)
    printf("%8x ",wt_unifd_cache[i].requestor);
  printf("\n\n");
  
}

int fill_tracebuf()
{
  count = (unsigned) fread(tp, size, TRACE_BUF_SIZE, trace_file);
  mrkr = tp;
  trace_count = 0;
  return(count);
}

unsigned lru_repl(unsigned x, unsigned y)
{
  int i;
  unsigned tmp;
  
  tmp = x << 6;
  
  if (((y >> 6) % 4) != x)
    tmp = tmp | ((y & 0xc0) >> 2);
  else
    {
      tmp = tmp | (y & 0x3f);
      return (tmp);
    }
  
  if (((y >> 4) % 4) != x)
    tmp = tmp | ((y & 0x30) >> 2);
  else
    
    {
      tmp = tmp | (y & 0xf);
      return (tmp);
    }
  
  if (((y >> 2) % 4) != x)
    tmp = tmp | ((y & 0xc) >> 2);
  else
    tmp = tmp | (y & 0x3);
  return(tmp);
}


access_mItlb(unsigned va)
{
  int i;
  mItlb_refs++;
  mImiss_flag = 1;
  
  for (i = 0; i < fl_conf.mItlb_size; i++)
    if (((mItlb[i].va & PAGE) == (va & PAGE))
	&& (mItlb[i].valid == 1))
      mImiss_flag = 0;
  return(0);
}

fill_mItlb(unsigned va)
{
  mItlb_miss++;
  if (mItlb[mItlb_pntr].valid == 1)
    mItlb_repl++;
  mItlb[mItlb_pntr].va = va;
  mItlb[mItlb_pntr].valid = 1;
  mItlb_pntr = (mItlb_pntr + 1) % fl_conf.mItlb_size;
  return(0);
}


access_mDtlb(unsigned va)
{
  int i;
  mDtlb_refs++;
  mDmiss_flag = 1;
  
  for (i = 0; i < fl_conf.mDtlb_size; i++)
    if (((mDtlb[i].va & PAGE) == (va & PAGE))
	&& (mDtlb[i].valid == 1))
      mDmiss_flag = 0;
  return(0);
}

fill_mDtlb(unsigned va)
{
  mDtlb_miss++;
  if (mDtlb[mDtlb_pntr].valid == 1)
    mDtlb_repl++;
  mDtlb[mDtlb_pntr].va = va;
  mDtlb[mDtlb_pntr].valid = 1;  
  mDtlb_pntr = (mDtlb_pntr + 1) % fl_conf.mDtlb_size;
  return(0);
}


access_tlb(unsigned va)
{
  unsigned temp, index, flag;
  int i;
  index = (va >> 14) % fl_conf.tlb_size; /* indexing into the tlb */
  tlb_refs++;
  
  Tmiss_flag = 1;
  
  for (i = 0; ( i < 4 ) && ( Tmiss_flag == 1 ); i++) {
    if (((tlb[index].set[i] >> 14) == (va >> 14))
	&& (((tlb[index].valid >> i) % 2) == 1)) {
      Tmiss_flag = 0;
      tlb[index].repl = lru_repl((unsigned)i, tlb[index].repl);
      break;
    }
  }
  
  
  if (Tmiss_flag == 1) /* miss case */ {
    flag = 0;
    tlb_miss++;
    for (i = 0; (i < 4) && (flag == 0); i++) {
      if (((tlb[index].valid >> i) % 2) == 0) {
	flag = 1;
	tlb[index].set[i] = va;
	tlb[index].valid = tlb[index].valid | (0x1 << i);
	tlb[index].repl = lru_repl((unsigned)i, tlb[index].repl);
      }
    }
    
    if (flag == 0) /* lru replacement case */ {
      tlb[index].set[tlb[index].repl%4] = va;
      tlb[index].repl = lru_repl(tlb[index].repl % 4, tlb[index].repl);
      tlb_repl++;
    }
  }
}


access_icache(unsigned va)
{
  static unsigned prev_va;
  int	i, flag;
  unsigned index;
  index = (va >> fl_conf.idshft) % fl_conf.icache_size; /* indexing into the icache */
  Imiss_flag = 1;
  /* Nirmal: This is a hack to allow us to figure out what the real i0_ref
     rate is.  The counter is being incremented every cycle in some instances,
     leading to bogus L1 miss rates Aiman 9/19/95 */

  if (prev_va != va) {
     icache_refs++;
     prev_va = va;
  }

  flag = 0;
  
  /* finding a matching entry in two sets */
  
  for (i = 0; (i < 2) && (flag == 0); i++) {
    if (((icache[index].set[i] >> fl_conf.idshft) == (va >> fl_conf.idshft))
	&& (((icache[index].valid >> i) % 2) != 0)) {
      flag = 1;
      Imiss_flag = 0;
      icache[index].repl = (i % 2) ^ 0x1;
    }
  }
  return(0);
}

fill_icache(unsigned va)
{
  int	i, flag;
  unsigned index;
  index = (va >> fl_conf.idshft) % fl_conf.icache_size; /* indexing into the icache */
  icache_miss++;
  flag = 0;
  for (i = 0; (i < 2) && (flag == 0); i++) 
    if (((icache[index].valid >> i) % 2) == 0) {
      flag = 1;
      icache[index].set[i] = va;
      icache[index].valid = icache[index].valid | (0x1 << i);
      icache[index].repl = (i % 2) ^ 0x1;
    }
  
  if (flag == 0) /* lru replacement case */ {
    icache_repl++;
    icache[index].set[icache[index].repl%2] = va;
    icache[index].repl = icache[index].repl ^ 0x1;
  }
  return(0);
}



access_dcache(unsigned va, unsigned opcode)
{     
  int i, flag;
  unsigned index;
  index = (va >> fl_conf.idshft) % fl_conf.dcache_size; /* indexing into the dcache */
  
  Dmiss_flag = 1;
  
  dcache_refs++;

  flag = 0;
  command = READ; 
    switch (opcode) {
      case STW: 
      case STB: 
      case STH: 
      case STD: 
      case STX: 
      case SWAP:
      case STWA: 
      case STBA: 
      case STHA: 
      case STDA: 
      case STXA: 
      case SWAPA:
      case STF: 
      case STFSR: 
      case STQF: 
      case STDF: 
      case STFA: 
      case STQFA:
      case STDFA: 
      case CASA: 
      case CASXA: 
      case LDSTUB: 
      case LDSTUBA:
	  command = WRITE;
	break;
	
      default:
	break;
      }

  for (i = 0; (i < 2) && (flag == 0); i++) {
    if (((dcache[index].set[i] >> fl_conf.idshft) == (va >> fl_conf.idshft))
	&& (((dcache[index].valid >> i) % 2) == 1)) {
      Dmiss_flag = 0;
      flag = 1;
      switch (command) {
      case WRITE: 
        if ((dcache[index].dirty & (0x1 << i))==0)
	  command = RETAG;
	dcache_st++;
	dcache[index].dirty =  dcache[index].dirty | (0x1 << i);
	break;
	
      default:
	break;
      }
      dcache[index].repl = (i % 2) ^ 0x1;
    }
  }
}    

implant_in_unifd_cache (unsigned ea, cmd_type cmd) 
{
   int j;
   int flag=0;
   
   /* we need to writeback the line */
   for (j=0;j<(fl_conf.max_ut_buf+1);j++) {
      if (wt_unifd_cache[j].state==INVALID) {
	 wt_unifd_cache[j].state = VALID;
	 wt_unifd_cache[j].ea = ea;
	 wt_unifd_cache[j].cmd = cmd;
	 wt_unifd_cache[j].requestor = D0REQ;
	 flag =1;		/* we found a spot to flush a dirty line */
	 break;
      }
   }
   if (flag != 1) {
      printf ("Unified LDST queue is full\n"); exit(-1);
   }

}

invalidate_d0 (unsigned ea) 
{
  int i, flag=0;
  unsigned index;
  index = (ea >> fl_conf.idshft) % fl_conf.dcache_size; /* indexing into the dcache */
  
  
  for (i = 0; (i < 2) && (flag == 0); i++) {
     if (((dcache[index].set[i] >> fl_conf.idshft) == (ea >> fl_conf.idshft))
	 && (((dcache[index].valid >> i) % 2) == 1)) {

	flag =1;   /* we found a line to be snooped */
	if ((dcache[index].dirty >> i) %2) {
	   implant_in_unifd_cache(ea, SNOOP_WB);
	}
	dcache[index].valid =  dcache[index].valid & (~(1<<i));
	break;
     }
  }

  if(flag==0) {   /* we could not find the line, look in wb queue */
     for (i=0;i<fl_conf.drty_buf;i++) {
	 if (dirty_buffer[i]==ea) {
	    implant_in_unifd_cache(ea, SNOOP_WB);
	    dirty_buffer[i]=0;

	    flag=2;
	    break;
	 }
      }
  }
#ifdef NOT_NOW
  /* L1 needs to inform U2 when it invalidate its lines so we reset the
     inclusion bit */
  if (flag == 0) {
     printf("Invalidation to a non-existant line in L1\n");
     exit(-1);
  }
#endif
}


fill_dcache(unsigned va, unsigned opcode)
{
  int i, flag;
  unsigned index;
  index = (va >> fl_conf.idshft) % fl_conf.dcache_size;
  
  /* indexing into the dcache */
  
  dcache_miss++;
  flag = 0;
  Dmiss_repl = 0;
  for (i = 0; (i < 2) && (flag == 0); i++) {
    if (((dcache[index].valid >> i) % 2) == 0) {
      flag = 1;
      switch (opcode) {
      case STW: 
      case STB: 
      case STH: 
      case STD: 
      case STX: 
      case SWAP:
      case STWA: 
      case STBA: 
      case STHA: 
      case STDA: 
      case STXA: 
      case SWAPA:
      case STF: 
      case STFSR: 
      case STQF: 
      case STDF: 
      case STFA: 
      case STQFA:
      case STDFA: 
      case CASA: 
      case CASXA: 
      case LDSTUB: 
      case LDSTUBA:
	
	dcache_st++;
	dcache[index].dirty =  dcache[index].dirty | (0x1 << i);
	break;
	
      default:
	dcache[index].dirty = dcache[index].dirty & (~(0x1 << i));
	break;
      }
      dcache[index].valid = dcache[index].valid | (0x1 << i);
      dcache[index].set[i] = va;
      dcache[index].repl = (i % 2) ^ 0x1;
    }
  }
    
  if (flag == 0) /* lru replacement case */ {
    if (dcache[index].dirty & (0x1 << (dcache[index].repl % 2))) {
      dirty_va = dcache[index].set[dcache[index].repl%2];
      Dmiss_repl = 1;
      dcache_drepl++;
    }
    dcache_repl++;
    dcache[index].set[dcache[index].repl%2] = va;
    switch (opcode) {
    case STW: 
    case STB: 
    case STH: 
    case STD: 
    case STX: 
    case SWAP:
    case STWA: 
    case STBA: 
    case STHA: 
    case STDA: 
    case STXA: 
    case SWAPA:
    case STF: 
    case STFSR: 
    case STQF: 
    case STDF: 
    case STFA: 
    case STQFA:
    case STDFA: 
    case CASA: 
    case CASXA: 
    case LDSTUB: 
    case LDSTUBA:
      dcache_st++;
      
      dcache[index].dirty =  dcache[index].dirty | 
	  (0x1 << (dcache[index].repl % 2));
      break;
    default:
      dcache[index].dirty = dcache[index].dirty & 
	(~(0x1 << (dcache[index].repl % 2)));
      break;
    }
    dcache[index].repl = (dcache[index].repl % 2) ^ 0x1;
  }
    
  return(0);
}



unsigned access_drtybuf(unsigned ea)
{
  int i;

  for (i=0;i<fl_conf.drty_buf;i++)
    if (dirty_buffer[i]==ea) {
      drty_hits++;
      return(1);
    }

  return(0);
}

void fill_drtybuf (unsigned ea)
{
  int i,j;
  int flag;
  unsigned static victim=0;

  flag = 0;
  for (i=0;i<fl_conf.drty_buf;i++) {
    if (dirty_buffer[i]==0) {
      dirty_buffer[i] = ea;
      flag = 1;
      break;
    }
  }
  if (flag==0) {
    drty_flshs++;
    for (j=0;j<(fl_conf.max_ut_buf+1);j++) {
      if (wt_unifd_cache[j].state==INVALID) {
	wt_unifd_cache[j].state = VALID;
	wt_unifd_cache[j].ea = dirty_buffer[victim] ;
	dirty_buffer[victim] = ea;
	wt_unifd_cache[j].cmd = WRITEBACK;
	victim = (victim+1)%fl_conf.drty_buf;
	wt_unifd_cache[j].requestor = D0REQ; /* NIRMAL: REQUESTOR SHOULD BE 
						  D0REQ and WRITEBACK used in the 
                                                  command field instead
                                               */
	if (j==0) {
/*  	  access_cache(dirty_buffer[victim],2,&wt_unifd_cache[j]);  */
/* NIRMAL:access_cache() does not do anything unless you define OLD_U2 in u2.c
   we should remove the call once we test the final changes.
*/
/* 	  if (Umiss_flag) */
/* 	    wt_unifd_cache[j].cycles = fl_conf.memory_lat+fl_conf.unified_lat; */
/* 	  else */
/* 	    wt_unifd_cache[j].cycles = fl_conf.unified_lat; */
	}
	flag = 1;
	break;
      }
    }
    if (flag==0) { /* no room to install dirty buffer flush request */
      printf("cycle count %u\n", clk_cycle);
      printf("unified cache request over run from dirty flush\n");
      exit(-1);
    }
  }
}
    
void
init_l1()
{
   int i;
   char file[MAX_FNAME];
   FILE *fp;
 
   if (warmup_state) {
      sprintf(file, "%s.l1i", warmup_file_name);

      if (( fp = fopen(file, "r")) == (FILE *) NULL) {
	 perror("Failed to open l1 warmup file");
	 exit(1);
      }

      for (i=0;i<fl_conf.icache_size;i++) {
	 if (!fread((char *)&icache[i], sizeof(i0_cache), 1, fp)) {
	    perror("Warmup state size for I0 does not match expected size");
	    exit(1);
	 }
      }
      close(fp);

  

      sprintf(file, "%s.l1d", warmup_file_name);

      if (( fp = fopen(file, "r")) == (FILE *) NULL) {
	 perror("Failed to open l1 warmup file");
	 exit(1);
      }

      for (i=0;i<fl_conf.dcache_size;i++) {
	 if (!fread((char *)&dcache[i], sizeof(d0_cache), 1, fp)) {
	    perror("Warmup state size for D0 does not match expected size");
	    exit(1);
	 }
      }
      close(fp);
      
   }
   else{
      /* initialize icache */
      
      for (i=0;i<fl_conf.icache_size;i++) {
	 icache[i].set[0] =  0;
	 icache[i].set[1] =  0;
	 icache[i].valid  =  0;
	 icache[i].repl   =  0;
      } 
      
      /* initialize dcache */
      
      for (i=0;i<fl_conf.dcache_size;i++) {
	 dcache[i].set[0] =  0;
	 dcache[i].set[1] =  0;
	 dcache[i].dirty  =  0;
	 dcache[i].valid  =  0;
	 dcache[i].repl   =  0;
      } 
   }
}



















