/*
 *  $Id: gl_config.c,v 1.7 1995/01/23 18:53:13 forssell Exp $
 *  
 *  Copyright (c) 1994 HAL Computer Systems International, Ltd.
 *  All rights reserved.  Unpublished -- rights reserved under
 *  the Copyright Laws of the United States.  USE OF A COPYRIGHT
 *  NOTICE IS PRECAUTIONARY ONLY AND DOES NOT IMPLY PUBLICATION
 *  OR DISCLOSURE.
 *  
 *  THIS SOFTWARE CONTAINS CONFIDENTIAL INFORMATION AND TRADE
 *  SECRETS OF HAL COMPUTER SYSTEMS INTERNATIONAL, LTD.  USE,
 *  DISCLOSURE, OR REPRODUCTION IS PROHIBITED WITHOUT THE
 *  PRIOR EXPRESS WRITTEN PERMISSION OF HAL COMPUTER SYSTEMS
 *  INTERNATIONAL, LTD.
 *  
 *                          RESTRICTED RIGHTS LEGEND
 *  Use, duplication, or disclosure by the Government is subject
 *  to the restrictions as set forth in subparagraph (c)(l)(ii)
 *  of the Rights in Technical Data and Computer Software clause
 *  at DFARS 252.227-7013.
 *  
 *           HAL COMPUTER SYSTEMS INTERNATIONAL, LTD.
 *                   1315 Dell Avenue
 *                   Campbell, CA  95008
 */
/*==============================================================================*/ 
/* Hal Denali Trace Driven Simulator						*/
/* Copyright (c) 1994 HaL Computer Systems, Inc.                      		*/
/*==============================================================================*/
/* be_config.c									*/
/*										*/
/*==============================================================================*/
/* Revision history								*/
/*										*/
/* Date	     Who	   	Description					*/ 
/* --------  -----------      	-----------					*/
/* 10/30/94  Per Forssell  	Initial defintion				*/
/*										*/
/*==============================================================================*/

#include	<stdio.h>
#include	<stdlib.h>
#include	<string.h>

#include	"globdef.h"

/*==============================================================================*/

FILE *open_file () ;

/*==============================================================================*/

extern FILE     *config_file ;		/* pointer to the configuration file	*/
extern char	config_file_name[] ; 	/* configuration file name		*/

extern gconf	gl_conf ;		/* global configuration structure	*/

/*==============================================================================*/

/* The global configuration file parser reads the configuration file, it over
   writes the default global configuration parameters with parameters found in
   the configuration file. */

void
parse_gl_config_file ()
	
{
  char	line[MAX_LINE] ;	/* input line from configuration file		*/
  char	param[MAX_NAME] ;	/* configuration parameter name			*/
  char	mem_model[MAX_NAME] ;	/* memory model type				*/

  config_file = open_file (config_file_name, "r") ;
  while (fgets (line, MAX_LINE, config_file) != NULL) {

    strcpy (param, strtok (line, WHITE_SPACE)) ;

    if (!strcmp (param, "NO_OF_LS_COMPL"))
      gl_conf.no_of_ls_compl = atoi (strtok (NULL, WHITE_SPACE)) ;
    if (!strcmp (param, "NO_OF_LS_REQ"))
      gl_conf.no_of_ls_req = atoi (strtok (NULL, WHITE_SPACE)) ;
    if (!strcmp (param, "NO_OF_LS_BUFS"))
      gl_conf.no_of_ls_bufs = atoi (strtok (NULL, WHITE_SPACE)) ;
    if (!strcmp (param, "IFETCH_CNT"))
      gl_conf.ifetch_cnt = atoi (strtok (NULL, WHITE_SPACE)) ;
    if (!strcmp (param, "MEM_MODEL")) {
      strcpy (mem_model, strtok (NULL, WHITE_SPACE)) ;
      if (!strcmp(mem_model, "STO")) 
	gl_conf.mem_model = STO ;
      if (!strcmp(mem_model, "LSO")) 
	gl_conf.mem_model = LSO ;
      if (!strcmp(mem_model, "LSX")) 
	gl_conf.mem_model = LSX ;
      if (!strcmp(mem_model, "RMO")) 
	gl_conf.mem_model = RMO ;
    }
    if (!strcmp (param, "D0_LAT"))
      gl_conf.d0_lat = atoi (strtok (NULL, WHITE_SPACE)) ;
  }
  fclose (config_file) ;
 
} /* end parse_gl_config_file */

/*==============================================================================*/

/* The global configuration routine reads in all global configuration parameters
   from the configuration file. Default parameter values are used for parameters
   not defined in the configuration file or if no configuration file is
   specified. */

void
gl_config ()	
	
{
  /* Initialize all configuration parameters to their default values. */

  gl_conf.no_of_ls_compl   = DEF_NO_OF_LS_COMPL ;
  gl_conf.no_of_ls_req     = DEF_NO_OF_LS_REQ ;
  gl_conf.no_of_ls_bufs    = DEF_NO_OF_LS_BUFS ;
  gl_conf.ifetch_cnt       = DEF_IFETCH_CNT ;
  gl_conf.mem_model        = DEF_MEM_MODEL ;
  gl_conf.d0_lat           = DEF_D0_LAT ;


  /* Read parameters from configuration file. */

  if (strcmp (config_file_name, ""))
    parse_gl_config_file () ;


} /* end gl_config */

/*==============================================================================*/
