/*
 *  $Id: globdef.h,v 1.30 1995/08/15 19:44:51 aiman Exp $
 *  
 *  Copyright (c) 1994 HAL Computer Systems International, Ltd.
 *  All rights reserved.  Unpublished -- rights reserved under
 *  the Copyright Laws of the United States.  USE OF A COPYRIGHT
 *  NOTICE IS PRECAUTIONARY ONLY AND DOES NOT IMPLY PUBLICATION
 *  OR DISCLOSURE.
 *  
 *  THIS SOFTWARE CONTAINS CONFIDENTIAL INFORMATION AND TRADE
 *  SECRETS OF HAL COMPUTER SYSTEMS INTERNATIONAL, LTD.  USE,
 *  DISCLOSURE, OR REPRODUCTION IS PROHIBITED WITHOUT THE
 *  PRIOR EXPRESS WRITTEN PERMISSION OF HAL COMPUTER SYSTEMS
 *  INTERNATIONAL, LTD.
 *  
 *                          RESTRICTED RIGHTS LEGEND
 *  Use, duplication, or disclosure by the Government is subject
 *  to the restrictions as set forth in subparagraph (c)(l)(ii)
 *  of the Rights in Technical Data and Computer Software clause
 *  at DFARS 252.227-7013.
 *  
 *           HAL COMPUTER SYSTEMS INTERNATIONAL, LTD.
 *                   1315 Dell Avenue
 *                   Campbell, CA  95008
 */
/*==============================================================================*/ 
/* Hal Denali Trace Driven Simulator						*/
/* Copyright (c) 1994 HaL Computer Systems, Inc.                      		*/
/*==============================================================================*/
/* globdef.h									*/
/*										*/
/*==============================================================================*/
/* Revision history								*/
/*										*/
/* Date	     Who	   	Description					*/ 
/* --------  -----------      	-----------					*/
/* 10/30/94  Per Forssell  	Initial defintion				*/
/*										*/
/*==============================================================================*/

#define TSIM_VERSION_NO		("1.20")	/* version number		*/

/*==============================================================================*/
/* General definitions							     	*/
/*==============================================================================*/

#define	TRUE		(1)
#define FALSE		(0)
#define NOT_VALID      (-1)

#define MAX_NAME       (32) /* max number of characters in names		*/
#define MAX_FNAME     (128) /* max number of characters in file names		*/
#define MAX_LINE      (512) /* max number of characters in a file input line	*/

#define WHITE_SPACE   (" ,:\t\"") /* white space character string		*/

#define ulong	unsigned long	/* just another shorter name for unsigned long	*/

/*==============================================================================*/
/* Global macros								*/
/*==============================================================================*/

#define	VALID(x)		 (x != -1)
#define	ANLD(x)		         (x ==  0)

#define	GETBIT(field, bit)	((field >> bit) & 0x1)
#define SETBIT(field, bit)	 (field |  (0x1 << bit))
#define	CLRBIT(field, bit)	 (field & ~(0x1 << bit))
#define	DEFBIT(field, bit, val)	((field & ~(0x1 << bit)) | (val << bit)) 

#define GETFLD(field, bit, no)  ((field >> bit) & ~(0xffffffff << no))

/*==============================================================================*/
/* Global definitions								*/
/*==============================================================================*/

/* Memory models ---------------------------------------------------------------*/

#define STO	(0)		/* Store ordered memory model			*/
#define LSO	(1)		/* Store ordered memory model. Agen in order	*/
#define LSX	(2)		/* Load store ordered memory model		*/
#define RMO	(9)		/* Relaxed ordered memory model		        */

/* Maximum processor configuration variables -----------------------------------*/

#define	MAX_IFETCH    (8)  /* max instructions fetched per cycle		*/
#define MAX_LS_DISP   (8)  /* max load store instructions executed per cycle	*/
#define MAX_LS_COMPL  (8)  /* max load store instructions completed per cycle	*/
#define MAX_BR_COMPL  (8)  /* max branch instructions completed per cycle	*/
#define MAX_BR_COMMIT (8)  /* max branch instructions committed per cycle	*/
#define MAX_EUNITS   (64)  /* max number of execution units			*/
#define MAX_ITYPES   (32)  /* max number of instruction types			*/
#define MAX_QSLOTS  (512)  /* max number of que slots in instruction que	*/

#define	MAX_CYCLE    	(0x7fffffff)

#define SIM_SEGMENT_SIZE  (10000) 

/* Default global configuration parameters -------------------------------------*/

#define DEF_NO_OF_LS_COMPL	(2)	/* default number of load stores that 
					   can complete in one cycle		*/
#define DEF_NO_OF_LS_REQ	(2)	/* default number of load stores that 
					   can be requested in one cycle	*/
#define DEF_NO_OF_LS_BUFS	(8)	/* default number of buffers entries in
					   the load store processor		*/
#define DEF_IFETCH_CNT		(6)	/* default instruction fetch count	*/
#define DEF_MEM_MODEL	      (STO)     /* default memory model			*/

/* Processor state register init values ----------------------------------------*/

#define INIT_PSTATE	(0x0000)
#define INIT_CWP	(0x0000)


/* Front End cache and tlb parameters ------------------------------------------*/

#define PAGE            0xffffc000
#define LINEMSK         0xffffff80
#define DEF_IENTRY   	        64
#define DEF_DENTRY   	       128
#define DEF_UENTRY   	         4
#define DEF_TENTRY   	       128
#define DEF_IDLINESIZE 	       128
#define DEF_mITENTRY 	        32
#define DEF_mDTENTRY 	        32
#define TRACE_BUF_SIZE        4096
#define DEF_MAX_LDST_CNT         8
#define DEF_MAX_UT_BUF           8
#define DEF_NO_DRT_BUF           2      /* do not exceed 8 */
#define DEF_NO_INT_LV            2
#define DEF_D0_LAT               2      /* default d0 access time in cycles     */
#define DEF_UNIFIED_LAT          5
#define DEF_MEMORY_LAT   	50
#define DEF_MTLB_LAT           100
#define DEF_MICRO_LAT    	 4
/* #define I0REQ        	         0 */
/* #define D0REQ       	         1 */
#define DRTYREP      	         2
#define DEF_ULINESIZE          512
#define DEF_UCACHE_SETS          4


/*==============================================================================*/
/* Global structure definitions							*/
/*==============================================================================*/

/* Instructions are requested with a fetch request. The fetch request defines the
   max number of instructions requested, and branch resolution information about
   previously fetched insturctions. The branch resolution information lists all
   branches that complete and all branches that commit during the current clock
   cycle. The branch unit will update the branch prediction tables based on the
   complete and the commit cycle information. */

typedef struct branch_resolution_list {
  int	valid ;			/* branch resolution is valid			*/
  ulong	inst_no ;		/* branch instruction number			*/

  ulong	recode[2] ;		/* recoded instruction word			*/

  ulong	pc ;			/* program counter of this instruction		*/ 
  ulong	ea ;			/* branch target address			*/ 
} brlis ;

typedef struct instruction_request {
  int	cnt ;			/* requested number of instructions		*/
  brlis	compl[MAX_BR_COMPL] ;	/* completed branch instructions		*/
  brlis	commit[MAX_BR_COMMIT] ;	/* committed branch instructions		*/
} ireq ;

/*------------------------------------------------------------------------------*/

/* The fetched instruction list is an array of instruction packets received from
   the instruction unit in one cycle. The entries are ordered in the array in
   instruction order. Each entry holds the recoded instruction op code, and the
   completion cycle for the fetch. The number of valid entries is less than or
   equal to the maximum fetch bandwidth of the current processor configuration. */

typedef struct instruction_packet {
  int	valid ;			/* fetched instruction packet is valid		*/

  ulong	fetch_cycle ;		/* completion cycle for the instruction fetch	*/

  ulong	inst ;			/* original instruction opcode			*/
  ulong	recode[2] ;		/* recoded instruction word 			*/

  int   taken ;			/* taken branch 				*/
  int	pred ;			/* correctly predicted branch			*/

  ulong	pc ;			/* program counter of this instruction		*/
  ulong	ea ;			/* effective address of load store or branch	*/
} ipkt ;

/*------------------------------------------------------------------------------*/

/* The load store packet list is an array of all load store instructions that
   complete address calculation in the current cycle. The entries are ordered in
   the list in instruction order. Each entry holds the recoded instruction
   opcode, and the que slot number of the load store instruction. The number of
   valid entries in the list is less than or equal to the max load store dispatch
   bandwidth of the current processor configuration, and less than the max number
   of load store requests that the lsu can receive in the current cycle. */

typedef struct load_store_request {
  int	valid ;		        /* load store packet is valid			*/
  int	qslot_no ;		/* load store instruction que slot number	*/

  ulong	recode[2] ;		/* recoded instruction word  			*/

  ulong	ea ;			/* effective address of load store		*/
} lsreq ;

/*------------------------------------------------------------------------------*/

/* The completed load store list is an array of all load store instructions that
   complete execution in the load store unit in the current clock cycle. The
   entries are not ordered in any particular order. Each entry holds the que
   slot number of the load store instruction, and the completion cycle of the
   load store instruction. The number of entries in the list is less than or
   equal to the max load store completion bandwidth of the current processor
   configuration. */

typedef struct load_store_packet {
  int	valid ;		        /* load store completion packet is valid	*/
  int	qslot_no ;		/* load store instruction que slot number	*/
  ulong	compl_cycle ;		/* load store completion cycle			*/
} lspkt ;

/*------------------------------------------------------------------------------*/

/* The recoded instruction structure holds all information about an instruction
   in a format amenible for fast execution by the simulator. No new info, beyond
   what is allready available in the original instruction trace entry, is
   added. */

typedef struct	recoded_instruction {
  ulong	inst ;		/* original instruction 				*/

  ulong	recode[2] ;	/* recoded instruction word (see decode.c for format)	*/

  int	type ;		/* instruction type					*/

  int	src1 ;		/* src1 register (or -1 if not appl)			*/
  int	src2 ;		/* src2 register (or -1 if not appl)			*/
  int	src3 ;		/* src3 register (or -1 if not appl)			*/
	
  int	dst ;		/* destination register (or -1 if not appl)		*/
	
  int	spsrc ;		/* state or priv reg source (for example ccr, fcc)	
			   (or -1 if not appl)					*/
  int	spdst ;		/* state or priv reg destination (for example ccr, fcc)	
			   (or -1 if not appl)					*/

  int   taken ;		/* taken branch 					*/
  int	pred ;		/* correctly predicted branch				*/

  ulong	pc ;		/* program counter for this instruction			*/
  ulong	ea ;		/* effective address of load store or branch target	*/

  int	sync ;		/* requires pipe synch				        */

  int	annul_next ;	/* will be annuled due to preceeding cti instruction	*/
} rinst ;

/*------------------------------------------------------------------------------*/

/* A que entry holds all data related to an issued instruction. This data
   includes the recoded instruction opcode, its dependencies to older
   instructions, the completion cycles of the dependencies, and the completion
   cycles of the various pipe line stages. */

typedef struct que_entry { 
  int	state ;			/* que entry state				*/

  rinst	rec_inst ;		/* recoded instruction				*/
	
  int	src1_qslot ;		/* src1 dependency				*/
  int	src2_qslot ;		/* src2 dependency				*/
  int	src3_qslot ;		/* src3 dependency				*/
  int	ccr_qslot ;		/* ccr dependancy				*/

  int	rdst ;			/* remapped destination register (-1 if not 
				   applicable )					*/

  ulong	src1_cycle ;		/* completion cycle for src1			*/
  ulong	src2_cycle ;		/* completion cycle for src2			*/
  ulong	src3_cycle ;		/* completion cycle for src3			*/
  ulong	ccr_cycle ;		/* completion cycle for ccr			*/
	
  ulong	start_cycle ;		/* start cycle for instruction fetch		*/
  ulong	fetch_cycle ;		/* completion cycle for fetch			*/
  ulong	issue_cycle ;		/* completion cycle for issue			*/
  ulong	disp_cycle ;		/* completion cycle for dispatch		*/
  ulong	qwait_cycle ;		/* completion cycle for que			*/
  ulong	exec_cycle ;		/* completion cycle for execute			*/
  ulong	lswait_cycle ;		/* completion cycle for load store dispatch	*/
  ulong	ewait_cycle ;		/* completion cycle for load store		*/
  ulong	compl_cycle ;		/* completion cycle for instruction		*/
	
  ulong	i_no ;			/* instruction sequence number			*/

  char  eu_name[MAX_NAME] ;	/* execution unit name				*/

} qentry ;	

/*------------------------------------------------------------------------------*/

/* The instruction que is composed of a forward-backward linked list of que
   slots. A que slot is identified by a physical que slot number. A que slot
   points to a que entry that is valid for the current cycle (ie whose data was
   generated in past cycles), and to a que entry that will be valid in the next
   cycle (ie whose data is being generated in the current cycle). */

typedef struct que_slot { 
  int	no ;			/* physical que slot number			*/ 
		
  qentry *c_entry ;		/* que entry for current cycle			*/ 
  qentry *n_entry ;		/* que entry for next cycle			*/ 
	
  struct que_slot *next_qslot ;	/* next slot in the que				*/ 
  struct que_slot *prev_qslot ;	/* previous slot in the que			*/ 
} qslot ;

/*------------------------------------------------------------------------------*/

/* All parameters that define the processor back end configuration are read from
   the processor configurtion file. The paramaeter values are kept in a global
   back end configuration file. */

typedef struct eu_type_configuration {
  char	name[MAX_NAME] ;	/* execution unit name				*/
  ulong	type_vec ;	        /* execution unit type vector			*/
} eutype ;

typedef struct i_type_configuration {
  ulong	ilat ;			/* instruction latency				*/
  ulong	ithrp ;			/* instruction thruput				*/
} itype ;

typedef struct back_end_configuration {
  int	 no_of_qslots ; 	/* number of que slots in the instruction que	*/
  int  	 no_of_windows ;	/* number of register windows (NWINDOWS)	*/
  int  	 no_of_eunits ;		/* number of execution units			*/
  int  	 issue_delay ;		/* number of cycles needed for instructions to be 
				   issued into the instruction que (ie before
				   they are available for dispatch)		*/
  int  	 disp_delay ;		/* number of cycles needed for instructions to be 
				   dispatched from the instruction que (ie before
				   they are available for execute)		*/
  int  	 compl_delay ;		/* number of cycles needed between execution 
				   completion of an instruction, and before the
				   results of the instruction can be used for a 
				   subsequent instruction 			*/
  int  	 commit_cnt ;		/* maximum number of instructions that can be 
				   committed in one clock cycle 		*/
  int  	 no_of_agen_bufs ;	/* number of agen buffers			*/

  eutype eu_type[MAX_EUNITS] ;	/* execution unit configuration			*/
  itype	 i_type[MAX_ITYPES] ;	/* instruction type latencies and thruputs; ie 
				   how ulong is the execution delay of each 
				   instruction, and how ulong does each instruction 
				   reserve an execution unit before a subsequent 
				   operation may be dispatched to that execution 
				   unit						*/
} bconf ;

typedef struct global_configuration {
  int  	 no_of_ls_compl ;	/* number of load store instructions that can 
				   completed in one cycle 			*/
  int  	 no_of_ls_req ;		/* number of load store transactions that can 
				   requested in one cycle 			*/
  int  	 no_of_ls_bufs ;	/* number of load store buffers			*/
  int  	 ifetch_cnt ;		/* maximum number of instructions that can be 
				   fetched in one clock cycle 			*/
  int  	 mem_model ;		/* memory model (TSO, or LSO)			*/
  int  	 d0_lat ;		/* d0 access time				*/
} gconf ;

/*------------------------------------------------------------------------------*/

/* Performance statistics are collected globally for each processor cycle.  The
   statistics are summarized in a global structure. */

typedef struct global_statistics {
  ulong	no_of_cycles ;
  ulong	no_of_instructions ;

  ulong	no_of_iq_full ;
  ulong	no_of_iq_empty ;

  ulong cum_inst_life_time ;

  ulong iq_stats[MAX_QSLOTS+1] ;

  ulong fetch_stats[MAX_IFETCH+1] ;

  ulong	itype_stats[MAX_ITYPES] ;

  ulong	cti_fetch_stall ;
  ulong cti_fetch_stall_iq_stats[MAX_QSLOTS+1] ;

  ulong no_of_annulled_insts ;

  ulong br_comp_refs ;
  ulong br_res_comp ;

  ulong br_ip_comp_refs ;
  ulong br_ip_res_comp ;

  ulong br_cmmt_refs ;
  ulong br_res_cmmt ;

  ulong br_ip_cmmt_refs ;
  ulong br_ip_res_cmmt ;

} gstats ;

/*==============================================================================*/
