/*
 * $Id: sym.c,v 1.1 1992/02/26 00:34:10 weiss Exp $
 *
 * Copyright (c) 1991 HaL Computer Systems, Inc.  All rights reserved.
 * UNPUBLISHED -- rights reserved under the Copyright Laws of the United
 * States.  Use of a copyright notice is precautionary only and does not
 * imply publication or disclosure.
 * 
 * This software contains confidential information and trade secrets of HaL
 * Computer Systems, Inc.  Use, disclosure, or reproduction is prohibited
 * without the prior express written permission of HaL Computer Systems, Inc.
 * 
 *                         RESTRICTED RIGHTS LEGEND
 * Use, duplication, or disclosure by the Government is subject to
 * restrictions as set forth in subparagraph (c)(l)(ii) of the Rights in
 * Technical Data and Computer Software clause at DFARS 252.227-7013.
 *                        HaL Computer Systems, Inc.
 *                  1315 Dell Avenue, Campbell, CA  95008
 * 
 */

/*
**
**	char *Symbolic_Address(PTR address)
**		address:	The address of a variable or function
**
**		returns the symbolic address of address (if possible)
**		Function must be initialized by
**
**	BOOLEAN Symbolic_Address_Init(file);
**		file:		pathname to a.out
**		Initialize symbol table information necessary to make
**		Sym_Address work.
**
*/

/*#include	<sys/core.h>*/
/*#include	<sys/vmparam.h>*/
#include	<sys/param.h>
#include	<sys/file.h>
#include	<sys/unistd.h>
#include	<a.out.h>
#include	<stab.h>
#include	<stdio.h>
#include	<string.h>

#include	"assert.h"
#include	"syntax.h"
#include	"sym.h"

#ifdef	MAIN
#include	"dasm.h"
#include	<strings.h>
#endif	MAIN

#define	MAXSYMBOLSIZE	256
#define	MAXBUFFERSIZE	512

#define	INSTRUCTION_COLUMN	16
#define	COMMENT_COLUMN		58
#define	PRINTFILE	0


#define	VERBOSE		0
#define	UNKNOWN_NAME	NULL
#define	NA		-1

#define	ALLOCATE_TYPE(type, n)	((type *) Allocate(sizeof(type), n))

typedef	struct nlist	NLIST, *NLISTp;
#define	STRX		n_un.n_strx
#define	NAME		n_un.n_name

typedef struct TRANSLATE {
    char		*file;
    BOOLEAN		init;
    char		buffer[MAXSYMBOLSIZE];
    int			nosymbols;
    struct exec		hdr;
    NLISTp		symbol_table;
    NLISTp		lookup;
    char		*string_table;
} TRANSLATE, *TRANSLATEp;

Local TRANSLATE translate;

Export void Symbolic_Print_All();
Export char *Symbolic_Address();
Export BOOLEAN Symbolic_Address_Init();


Local PTR Allocate();
Local char *Build_Address();
Local char *Symbolic_Ntype();
Local void Print_Symbol_Table();
Local void Print_Lookup_Table();
Local BOOLEAN Seek_Read();
Local BOOLEAN Build_Table();
Local BOOLEAN Read_Header();
Local BOOLEAN Read_String_Table();
Local BOOLEAN Read_Symbol_Table();

#ifdef	MAIN

Local void	Print_Instruction();
Local char	*Pad_Buffer();
Local char	*Extract_Name();
Local BOOLEAN	Disassemble_Ascii_Trace();
Local BOOLEAN	Disassemble_Binary_Trace();
Local BOOLEAN	Disassemble_Executable();
Local BOOLEAN   Track_Register_Usage();

Local BOOLEAN	vflag;
Local BOOLEAN	sflag;
Local BOOLEAN	rflag;
Local BOOLEAN	xflag;
Local BOOLEAN	tflag;
Local BOOLEAN	lookup_init= false;

struct TRACE {
    UINT		pc;
    UINT		iw;
    UINT		ea;
    unsigned char	a;
    unsigned char	flags;
    unsigned char	b,c;
}   trace;


Extern main (argc,argv)
int argc; char **argv;
{
    char	*arg, *file;
    BOOLEAN	error;

    argv++;

    vflag= rflag= sflag= xflag= tflag= error=	false;

    if (not lookup_init) {
	Instruction_Lookup_Init();
	lookup_init=	true;
    }

    while(argc > 1) {
	arg=	argv[0];
	if (arg[0] == '-') {
	    switch(arg[1]) {
	    case 'v':
		vflag=		true;
		break;

	    case 'r':
		rflag=		true;
		Track_Register_Usage(argv[1]);
		argv++;
		argc--;
		break;
		break;

	    case 's':	
		file=		argv[1];
		sflag=		Symbolic_Address_Init(file);
		if (vflag)
		    fprintf(stderr,"symbol table %s initialized for \"%s\"\n",
			    (sflag) ? "" : "not", file
		    );
		argv++;
		argc--;
		break;

	    case 'x':
		Disassemble_Executable(argv[1]);
		argv++;
		argc--;
		break;

	    case 't':
		switch(arg[2]) {
		case 'a':
		    Disassemble_Ascii_Trace(argv[1]);
		    break;
		case 'b':
		    Disassemble_Binary_Trace(argv[1]);
		    break;
		default:
		    error=	true;	
		    break;
		}
		argv++;
		argc--;
		break;

	    default:
		error=	true;
	    }
	}
	else {
	    error=	true;
	}
	argc--;
	argv++;
    }

    if (error) {
	printf(	"Please specify one or more (order does count)\n");
	printf(	"\t-v          verbose (comments/etc)\n");
	printf(	"\t-r  file    register usage report\n");
	printf(	"\t-s  file    symbol file\n");
	printf(	"\t-x  file    omagic,zmagic,nmagic\n");
	printf(	"\t-ta file    HaL trace file (ascii)\n");
	printf(	"\t-tb file    HaL trace file (binary)\n");
	exit(1);
    }
}


Local BOOLEAN Disassemble_Executable(file)
char	*file;
{
    int		fd, size, offset, nread;
    UINT	instruction, pc;
    char	*instr, *label;

    if ((fd=  open(file, O_RDONLY,0)) < 0) {
	if (DEBUG)
	    fprintf(stderr, "Cannot open file \"%s\"\n",file);
	return false;
    }
    if (vflag)
	fprintf(stderr, "Disassemble of \"%s\"\n",file);

    if (rflag)
	Instruction_Register_Usage_Init();

    sflag=	Symbolic_Address_Init(file);
    if (not sflag)
	if (vflag)
	    fprintf(stderr, "--- Symbolic Information Not Available ---\n");

    switch(translate.hdr.a_magic&0777) {
    case OMAGIC:
    case NMAGIC:
	offset=	N_TXTOFF(translate.hdr);
	size=	translate.hdr.a_text;
	break;
    case ZMAGIC:
	/*
	 *  for zmagic file the header is mapped at the
	 *  beginning of text
	 */
	offset=	N_TXTOFF(translate.hdr) + sizeof(struct exec);
	size=	translate.hdr.a_text - sizeof(struct exec);
	break;
    default:
	if (vflag)
	    fprintf(stderr, "File not OMAGIC//NMAGIC or ZMAGIC\n");
	return false;
    }

    if (lseek(fd, offset, 0) == -1) {
	fprintf(stderr,"bad seek\n");
	return false;
    }

    pc=		translate.hdr.a_entry;

    

    while(size > 0) {
	nread=	read(fd, &instruction, sizeof(int));
	if (nread  isnt sizeof(int)) {
	    fprintf(stderr,"bad read\n");
	    return false;
	}

	Print_Instruction(instruction, pc, NA, NA);

	size -=		nread;
	pc   +=		nread;
    }
    return true;
}


Local BOOLEAN Track_Register_Usage(file)
char	*file;
{
    int		fd, size, offset, nread;
    UINT	instruction, pc;
    char	*lp; 
    char	function[MAXBUFFERSIZE];

    if ((fd=  open(file, O_RDONLY,0)) < 0) {
	if (DEBUG)
	    fprintf(stderr, "Cannot open file \"%s\"\n",file);
	return false;
    }
    if (vflag)
	fprintf(stderr, "Register Usage Report for \"%s\"\n",file);


    sflag=	Symbolic_Address_Init(file);
    if (not sflag) {
	fprintf(stderr, "--- Symbolic Information Not Available ---\n");
	return;
    }

    switch(translate.hdr.a_magic&0777) {
    case OMAGIC:
    case NMAGIC:
	offset=	N_TXTOFF(translate.hdr);
	size=	translate.hdr.a_text;
	break;
    case ZMAGIC:
	/*
	 *  for zmagic file the header is mapped at the
	 *  beginning of text
	 */
	offset=	N_TXTOFF(translate.hdr) + sizeof(struct exec);
	size=	translate.hdr.a_text - sizeof(struct exec);
	break;
    default:
	if (vflag)
	    fprintf(stderr, "File not OMAGIC//NMAGIC or ZMAGIC\n");
	return false;
    }

    if (lseek(fd, offset, 0) == -1) {
	fprintf(stderr,"bad seek\n");
	return false;
    }

    pc=		translate.hdr.a_entry;
    strcpy(function, Symbolic_Address(pc));
    Instruction_Register_Usage_Init();
    
    while(size > 0) {
	nread=	read(fd, &instruction, sizeof(int));
	if (nread  isnt sizeof(int)) {
	    fprintf(stderr,"bad read\n");
	    return false;
	}
	lp=	Extract_Name( Symbolic_Address(pc) );

	if (not streq(function, lp)) {
	    /***********************************************
	    if (DEBUG) {
		fprintf(stdout,"Change of function from %s to %s at pc=0x%x\n",
			function, lp, pc
		);
	    }
	    ***********************************************/
	    Instruction_Register_Usage_Report(function, vflag);

	    Instruction_Register_Usage_Init();
	    strcpy(function, lp);
	}

	Instruction_Register_Usage(instruction);

	size -=		nread;
	pc   +=		nread;
    }

    Instruction_Register_Usage_Report(function, vflag);
    return true;
}


/*
**	Extract name. SName hould be of the  form
**		func+0xnnnn
*/
Local char	*Extract_Name(name)
char	*name;
{
    char	*plus= index(name,'+');

    if (not plus)
	return name;

    *plus=	'\0';
    return name;
}


Local BOOLEAN Disassemble_Ascii_Trace(file)
char	*file;
{
    int		nread,  iw, pc, ea, flags, a, b, c;
    FILE	*stream;
    BOOLEAN	eof= false;

    if (streq(file,"") or streq(file,"-"))
	stream= 	stdin;
    else
	stream=		fopen(file, "r");
    if (not stream) {
	perror("cannot open file");
	return false;
    }
	    
    while(not eof) {
	nread=	fscanf(stream, "%i %i %i %i %i %i %i",
		       &pc, &iw, &ea, &flags, &a, &b, &c);
	if (nread isnt 7) {
	    if (nread is EOF)
		break;
	    fprintf(stderr,"premature eof\n");
	    return false;
	}
	Print_Instruction(iw, pc, (ea) ? ea : NA, flags);
    }
    return true;
}


Local BOOLEAN Disassemble_Binary_Trace(file)
char	*file;
{
    int		fd, nread;
    BOOLEAN	eof= false;

    if (streq(file,"") or streq(file,"-"))
	fd=	STDIN_FILENO;
    else
	fd=  open(file, O_RDONLY,0);

    if (fd < 0) {
	perror("cannot open file");
	return false;
    }
	    
    while(not eof) {
	nread=	read(fd, &trace, sizeof(trace));
	if (nread isnt sizeof(trace)) {
	    if (nread == 0)
		break;
	    fprintf(stderr,"premature eof\n");
	    return false;
	}
	Print_Instruction(trace.iw,
			  trace.pc,
			  (trace.ea) ? trace.ea : NA,
			  trace.flags
	);
    }
    return true;
}


Local void Print_Instruction(instruction, pc, ea, flags)
UINT	instruction;
UINT	pc;
UINT	ea;
UINT	flags;
{
    char	buffer[MAXBUFFERSIZE];
    char	*instr, *label, *bp;

    label=	Symbolic_Address(pc);
    bp=		strcpy(buffer, label);
    strcat(buffer, ":");


    bp=		Pad_Buffer(buffer, INSTRUCTION_COLUMN, ' ');
    instr=	Instruction_Lookup(instruction, pc);
    strcat(buffer, instr);

    bp=		Pad_Buffer(buffer, COMMENT_COLUMN, ' ');

    bp=		Pad_Buffer(buffer, 1, ';');
    /*
     *	If there is no symbolic information the pc is already printed
     */
    if (sflag)
	sprintf(bp, " pc=0x%x", pc);

    if (ea isnt NA)  {
	bp=	Pad_Buffer(buffer, 1, ',');
	sprintf(bp, " ea=0x%x", ea);
    }

    bp=		Pad_Buffer(buffer, 1, ' ');

    switch(flags) {
    case NA:
	break;
    case 1:
	strcat(buffer, ", annulled");
	break;
    case 2:
	strcat(buffer, ", taken");
	break;
    }

    fprintf(stdout, "%s\n",  buffer);
}


Local char	*Pad_Buffer(buffer, col, fill)
char	*buffer;
int	col;
char	fill;
{
    int		len=	strlen(buffer);
    char	*bp;

    bp=		&buffer[len];
    do {
	*bp++	=	fill;
	len++;
    } while(len < col);

    *bp=	'\0';
    return bp;
}

#endif	MAIN


Export BOOLEAN Symbolic_Address_Init(file)
char	*file;
{
    int		fd;

    /*
     *	Clean up from last file
     */
    if (translate.init) {
	if (translate.string_table)
	    free(translate.string_table);
	if (translate.symbol_table)
	    free(translate.symbol_table);
	if (translate.lookup)
	    free(translate.lookup);
    }

    translate.init=		false;

    if ((fd=  open(file, O_RDONLY,0)) < 0) {
	if (DEBUG)
	    fprintf(stderr, "Cannot open file \"%s\"\n",file);
	return false;
    }
    translate.file=	strdup(file);

    if (Read_Header(fd)       is true and
	Read_Symbol_Table(fd) is true and
	Read_String_Table(fd) is true and
	Build_Table() 	      is true
    ) {
	translate.init=		true;
    }
    else {
	if (DEBUG)
	    fprintf(stderr, "cannot process \"%s\"\n",file);
	return false;
    }
    close(fd);

    return translate.init;
}



#define	INVALID_REQ_ID		-1
#define	VALIDREQ(req)		(req isnt INVALID_REQ_ID)

Export char *Symbolic_Address(address)
UINT	address;
{
    NLISTp	lookup=		translate.lookup;
    static int	lastreqid=	INVALID_REQ_ID;
    int		ub, lb, mid;

    if (translate.init is false)
	return (Build_Address( 0, address, translate.buffer));

    /*
     *	Requests will be clumped together. Look at last buffered request
     */
    if (VALIDREQ(lastreqid)) {
	ASSERT(lastreqid <= translate.nosymbols, "bad lastrequest");
	if (lookup[lastreqid].n_value <= address and
	    address < lookup[lastreqid+1].n_value
	) {
	    return (Build_Address( &lookup[lastreqid], address, translate.buffer));
	}
    }

    /*
     *	Since the symbol table is already sorted we can do a
     *	binary search. The table has been padded with entries 0 and N
     *  so that we MUST be able to find an entry. The "last" entry
     *  should never be used
     */
    lb=		0;
    ub=		translate.nosymbols-1;
    while(lb <= ub) {
	mid=	(ub+lb) / 2;
	ASSERT(mid <= translate.nosymbols, "Bad lookup");

	if (lookup[mid].n_value <= address) {
	    if (address < lookup[mid+1].n_value) {
		lastreqid=	mid;
		return (Build_Address( &lookup[mid], address, translate.buffer));
	    }
	    else {
		lb=	mid+1;
	    }
	}
	else {
		ub=	mid-1;
	}
    }
    return (Build_Address( &lookup[translate.nosymbols-1], address, translate.buffer));
}

Local char *Build_Address(sym, pc, buffer)
NLISTp	sym;
UINT		pc;
char		*buffer;
{
    UINT	offset;

    buffer[0]=	'\0';

    if (PRINTFILE) {
	strcat(buffer, translate.file);
	strcat(buffer, ":");
    }
    if (sym and sym->NAME) {
	offset=		pc - sym->n_value;

	strcat(buffer, sym->NAME);

	if (offset) {
	    strcat(buffer, "+");
	    sprintf( &buffer[ strlen(buffer) ], "0x%x", offset);
	}
    }
    else {
	sprintf( &buffer[ strlen(buffer) ], "0x%x", pc);
    }
    return (buffer);
}

/*
**	Read and verify magic number
*/
Local BOOLEAN Read_Header(fd)
int	fd;
{
    int		magic;

    if (Seek_Read(fd, 0, &magic, sizeof(int)) is false) {
	if (DEBUG)
	    fprintf(stderr, "Cannot read magic number\n");
	return false;
    }

    switch(magic&0777) {
    case OMAGIC:
    case NMAGIC:
    case ZMAGIC:
	break;
    default:
	return false;
    }

    /*
     *	Read exec structure into memory
     */
    if (Seek_Read(fd, 0, &translate.hdr, sizeof(struct exec)) is false)
	return false;

    return true;
}

/*
**	Read symbol table (if there is one)
*/
Local BOOLEAN Read_Symbol_Table(fd)
int	fd;
{
    int		nosymbols;

    if (translate.hdr.a_syms == 0) {
	if (DEBUG)
	    fprintf(stderr, "stripped file\n");
	return false;
    }

    nosymbols=			translate.hdr.a_syms / sizeof(NLIST);
    translate.symbol_table=	ALLOCATE_TYPE(NLIST, nosymbols);
    /*
     *	Read symbol table into memory
     */
    if (Seek_Read(  fd,
		    N_SYMOFF(translate.hdr),
		    translate.symbol_table,
		    translate.hdr.a_syms
		 ) is false
    ) {
	if (DEBUG)
	    fprintf(stderr, "Cannot read symbol table\n");
	free(translate.symbol_table);
	return false;
    }

    return true;
}


/*
**	Read string table (if there is one)
*/
Local BOOLEAN Read_String_Table(fd)
int	fd;
{
    int		strsize;

    if (translate.hdr.a_syms == 0)
	return false;

    /*
     *	Read string table into memory
     *	    first get strsize from file
     */
    if (Seek_Read(fd, N_STROFF((translate.hdr)), &strsize, sizeof(int)) is false
    ) {
	if (DEBUG)
	    fprintf(stderr, "Cannot read string size\n");
	return false;
    }
    translate.string_table=	ALLOCATE_TYPE(char, strsize);
    /*
     *	strsize includes itself so adjust for bytes aready read
     */
    if (Seek_Read(  fd,
		    N_STROFF((translate.hdr)),
		    translate.string_table,
		    strsize
		 ) is false
    ) {
	if (DEBUG)
	    fprintf(stderr, "Cannot read strings table\n");
	free(translate.string_table);
	return false;
    }

    if (VERBOSE)
	Print_Symbol_Table("After Symbol Table Read", translate.hdr.a_syms / sizeof(NLIST));
    return true;
}


Local BOOLEAN Build_Table()
{
    NLIST	 	temp;
    BOOLEAN		sorted=  false;
    int			i, nosymbols, nsym;
    NLISTp		lookup;
    NLISTp		symtab=		translate.symbol_table;

    /*
     *	Leave space for two dummy entries in table
     */
    nosymbols=		translate.hdr.a_syms / sizeof(NLIST);
    lookup=		ALLOCATE_TYPE(NLIST, nosymbols+2);

    /*
     *	Only table "useful" symbols
     *	Change string index into pointer references
     */
    nsym=	1;
    for(i= 0; i < nosymbols; i++, symtab++) {
	switch (symtab->n_type&(N_STAB+N_TYPE)) {
	case N_TEXT:
	case N_DATA:
	case N_BSS:
	case N_COMM:
	    lookup[nsym]=	*symtab;
	    lookup[nsym].NAME=	(symtab->STRX == -1) ?
			    "?" : translate.string_table + symtab->STRX;
	    nsym++;
        }
    }
    lookup[0].NAME=		UNKNOWN_NAME;
    lookup[0].n_value=		0;
    lookup[nsym].NAME=		UNKNOWN_NAME;
    lookup[nsym].n_value=	0xffffffff;
    nsym++;

    translate.nosymbols=	nsym;

    while(not sorted) {
	sorted=		true;
	for(i=1; i<translate.nosymbols-1; i++) {
	    if (lookup[i].n_value > lookup[i+1].n_value) {
		temp=		lookup[i];
		lookup[i]=	lookup[i+1];
		lookup[i+1]=	temp;
		sorted=		false;
	    }
	    /*
	     *	We are not keeping seperate tables for data/bss/text.
	     *	This will keep symbols of the same "type" together
	     */
	    else if (1 < i                                    and
		     lookup[i].n_value == lookup[i+1].n_value and
		     ((lookup[i-1].n_type&(N_STAB+N_TYPE))    isnt
		      (lookup[i].n_type&(N_STAB+N_TYPE)))
		) {
		temp=		lookup[i];
		lookup[i]=	lookup[i+1];
		lookup[i+1]=	temp;
	    }
	}
    }
    translate.lookup=	lookup;
    if (VERBOSE)
	Print_Lookup_Table("After Build Table", translate.nosymbols);

    return true;
}


Local BOOLEAN Seek_Read(fd, offset, storage, size)
int	fd;
off_t	offset;
PTR	storage;
int	size;
{
    int		nread;

    if (lseek(fd, offset, 0) == -1) {
	if (DEBUG)
	    fprintf(stderr,"bad seek\n");
	return false;
    }
    if ((nread= read(fd, storage, size)) != size) {
	if (DEBUG)
	    fprintf(stderr,"bad read\n");
	return false;
    }
    return true;
}

Local void Print_Symbol_Table(msg, nsyms)
char	*msg;
int	nsyms;
{
    int			count;
    char		*type, *name;
    NLISTp	 	sym=	translate.symbol_table;

    printf("\nSYMBOL TABLE: \"%s\"\n", (msg) ? msg : "");
    printf("Sym No. StrOff   Type     Other Desc  Value      Symbol\n");

    for (count=0; count < nsyms; count++, sym++) {
	type=		Symbolic_Ntype(sym->n_type&(N_STAB+N_TYPE), sym->n_value);
	name=		(sym->STRX == -1) ?
				"" : (translate.string_table+sym->STRX);
	printf("%-7d %-7d %s%s %-5d %-5d 0x%08x \"%s\"\n",
		count,
		sym->STRX,
		type,
		(sym->n_type & N_EXT) ? ",EXT" : "    ",
		sym->n_other,
		sym->n_desc,
		sym->n_value,
		name
	);
    }
}


Local void Print_Lookup_Table(msg, nsyms)
char	*msg;
int	nsyms;
{
    int			count;
    char		*type;
    NLISTp	 	sym=	translate.lookup;

    printf("\nSYMBOL TABLE: \"%s\"\n", (msg) ? msg : "");
    printf("Sym No.  Type     Value      Symbol\n");

    for (count=0; count < nsyms; count++, sym++) {
	type=		Symbolic_Ntype(sym->n_type&(N_STAB+N_TYPE), sym->n_value);
	printf("%-7d %s%s 0x%08x \"%s\"\n",
		count,
		type,
		(sym->n_type & N_EXT) ? ",EXT" : "    ",
		sym->n_value,
		(sym->NAME) ? sym->NAME : "<null>"
	);
    }
}

Local char *Symbolic_Ntype(type, value)
int	type, value;
{
    switch (type) {
    case N_UNDF:	return((value) ? " COMM" : " UNDF");
    case N_ABS:		return("  ABS");
    case N_TEXT:	return(" TEXT");
    case N_DATA:	return(" DATA");
    case N_BSS:		return("  BSS");
    case N_COMM:	return(" COMM");
    case N_FN:		return(" file");
    case N_GSYM:	return(" gsym");
    case N_FNAME:	return("fname");
    case N_FUN:		return("  fun");
    case N_STSYM:	return("stsym");
    case N_LCSYM:	return("lcsym");
    case N_MAIN:	return(" main");
    case N_RSYM:	return(" rsym");
    case N_SLINE:	return("sline");
    case N_SSYM:	return(" ssym");
    case N_SO:		return("   so");
    case N_LSYM:	return(" lsym");
    case N_BINCL:	return("bincl");
    case N_SOL:		return("  sol");
    case N_PSYM:	return(" psym");
    case N_EINCL:	return("eincl");
    case N_ENTRY:	return("entry");
    case N_LBRAC:	return("lbrac");
    case N_EXCL:	return(" excl");
    case N_RBRAC:	return("rbrac");
    case N_BCOMM:	return("bcomm");
    case N_ECOMM:	return("ecomm");
    case N_ECOML:	return("ecoml");
    case N_LENG:	return(" leng");
    case N_PC:		return("   pc");
  /*case N_M2C:		return("  m2c");*/
  /*case N_SCOPE:	return("scope");*/
  /*case N_BROWS:	return("brows");*/
    }
    return("BADVAL");
}

Local PTR Allocate(size, n)
int	size;
int	n;
{
    PTR		p;

    p=		(PTR)	malloc(n*size);
    if (p)
	return p;

    fprintf(stderr, "allocation failed for Symbol Translation\n");
    exit(1);
} 
