/*
 *  $Id: tsim.c,v 1.15 1995/10/02 22:37:52 aiman Exp $
 *  
 *  Copyright (c) 1994 HAL Computer Systems International, Ltd.
 *  All rights reserved.  Unpublished -- rights reserved under
 *  the Copyright Laws of the United States.  USE OF A COPYRIGHT
 *  NOTICE IS PRECAUTIONARY ONLY AND DOES NOT IMPLY PUBLICATION
 *  OR DISCLOSURE.
 *  
 *  THIS SOFTWARE CONTAINS CONFIDENTIAL INFORMATION AND TRADE
 *  SECRETS OF HAL COMPUTER SYSTEMS INTERNATIONAL, LTD.  USE,
 *  DISCLOSURE, OR REPRODUCTION IS PROHIBITED WITHOUT THE
 *  PRIOR EXPRESS WRITTEN PERMISSION OF HAL COMPUTER SYSTEMS
 *  INTERNATIONAL, LTD.
 *  
 *                          RESTRICTED RIGHTS LEGEND
 *  Use, duplication, or disclosure by the Government is subject
 *  to the restrictions as set forth in subparagraph (c)(l)(ii)
 *  of the Rights in Technical Data and Computer Software clause
 *  at DFARS 252.227-7013.
 *  
 *           HAL COMPUTER SYSTEMS INTERNATIONAL, LTD.
 *                   1315 Dell Avenue
 *                   Campbell, CA  95008
 */
/*==============================================================================*/ 
/* Hal Denali Trace Driven Simulator						*/
/* Copyright (c) 1994 HaL Computer Systems, Inc.                      		*/
/*==============================================================================*/
/* tsim.c									*/
/*										*/
/*==============================================================================*/
/* Revision history								*/
/*										*/
/* Date	     Who	   	Description					*/ 
/* --------  -----------      	-----------					*/
/* 10/30/94  Per Forssell  	Initial defintion				*/
/*										*/
/*==============================================================================*/

#include	<stdio.h>
#include	<string.h>
#include	<signal.h>

#include	"globdef.h"
#include	"tsim.h"

/*==============================================================================*/

void gl_config () ;
void fe_config () ;
void be_config () ;

void be_stats () ;
void fe_stats () ;

int  front_end () ;
int  back_end () ;
void  u2 ();

int done = 0;
void sigusr1(x) { done = 1; }

/*==============================================================================*/

/* The read arguments routine reads in the command line arguments. The command
   line is as follows :

 tsim [-h] [-c <config file>] [-d <dump file>] [-s <stats file>] [-t <trace file>]

      -c <config file>	: name of the target processor configuration file
      -s <stats file>	: name of the simulation statistics output file
      -d <dump file>	: prints out cycle by cycle trace dump
      -t <trace file>	: name of the input trace file; will be stdin if omitted
      -w <warmup file>	: name of the file to be used to warm up tsim.
*/

int
read_arg
(int	argc ,
 char	**argv )
	
{

  int	arg_no ;

  strcpy (config_file_name, "") ;
  strcpy (stats_file_name, "tsim.stats") ;
  strcpy (dump_file_name,  "stdout") ;
  strcpy (trace_file_name, "stdin") ;

  arg_no = 1 ;
  while (arg_no <= argc) {

    if (!strcmp (*argv, "-d")) {
      dump_mode = 1 ;
      if (arg_no+1 <= argc)
	if (strncmp(*(argv+1), "-", 1)) {
	  strcpy (dump_file_name, *(argv+1)) ;
	  arg_no++ ;
	  argv++ ;
	}
    }
  
    else if (!strcmp (*argv, "-c")) {
      strcpy (config_file_name, *(argv+1)) ;
      arg_no++ ;
      argv++ ;
    }

    else if (!strcmp (*argv, "-s")) {
      strcpy (stats_file_name, *(argv+1)) ;
      arg_no++ ;
      argv++ ;
    }

    else if (!strcmp (*argv, "-t")) {
      strcpy (trace_file_name, *(argv+1)) ;
      arg_no++ ;
      argv++ ;
    }

    else if (!strcmp (*argv, "-w")) {
      strcpy (warmup_file_name, *(argv+1)) ;
      warmup_state = TRUE;
      arg_no++ ;
      argv++ ;
    }

    else if (!strcmp (*argv, "-h")) {
      printf ("\n") ;
      printf ("tsim ") ;
      printf ("[-h] ") ;
      printf ("[-c <config file>] ") ;
      printf ("[-d [<dump file>]] ") ;
      printf ("[-s <stats file>] ") ;
      printf ("[-t <trace file>] \n") ;
      printf ("\n") ;
      printf ("  -c <config file> : processor configuration file             \n") ;
      printf ("  -s <stats file>  : simulation statistics file               \n") ;
      printf ("  -d [<dump file>] : dump iq state; stdout if file omitted    \n") ;
      printf ("  -t <trace file>  : input trace file; stdin if omitted       \n") ;
      printf ("\n") ;

      return (1) ;
    }
			
    arg_no++ ;
    argv++ ;
  }

  return (0) ;
	
} /* end read_arg */

/*==============================================================================*/

/* open_file is a generic routine to open a file, and to write an appropriate
   error message if the file open did not succeed. */

FILE*
open_file (char *file_name, char *access_type)
	
{
  FILE  *fp ;		/* generic file pointer 				*/

  if (!strcmp(file_name, "stdin"))
    fp = stdin ;
  else if (!strcmp(file_name, "stdout"))
    fp = stdout ;
  else {
    fp = fopen (file_name, access_type) ;
    if (fp == NULL) {
      if (!strncmp (access_type, "r", 1))
	printf ("TSIM: ERROR: file %s could not be opened for read \n\n", 
		file_name) ;
      else if (!strncmp (access_type, "w", 1))
	printf ("TSIM: ERROR: file %s could not be opened for write \n\n", 
		file_name) ;
      else if (!strncmp (access_type, "a", 1))
	printf ("TSIM: ERROR: file %s could not be opened for append \n\n", 
		file_name) ;
      exit (1) ;
    }
  }

  return (fp) ;
    
} /* end open_file */

/*==============================================================================*/

/* The dump graph routine will write a cycle by cycle dump of the instruction que
   state.  The dump will be processed and displayed using the tsim viewer.  The
   trace dump is composed of a set of records. The record types are as follows:

   INST cycle_no inst_no instr_op_code 

   The INST record defines the next instruction in the instruction sequence. The
   cycle_no field defines the cycle number when the instruction starts fetch. The
   inst_no field defines the instruction sequence number. the instr_op_code
   defines the SPARC v9 opcode of the instruction.

   IQUE cycle_no first_inst_no qslot0 qslot1 .. qslotn   

   The IQUE record defines the state of the instruction que at cycle cycle_no.
   The state is defined using a string of characters, one character for each que
   slot. Each possible instruction state has a specific character associated with
   it :

	'>'    : que slot instruction is being fetched
	'V'    : que slot instruction is being issued
	'='    : que slot instruction is sitting in the que
	'&'    : que slot instruction is being dispatched
	'#'    : que slot instruction is being executed
	'~'    : que slot instruction is waiting for load store memory model
	'$'    : que slot instruction is wating for load store or branch unit
	'-'    : que slot instruction is executed but awaiting completion
	'.'    : que slot instruction is completed but not committed
   	' '    : que slot instruction is committed; que slot is free
   	'a'    : que slot instruction is annulled

   */

void
dump_graph () 
	
{
  qslot	  *qs ;			/* que slot pointer				*/

  qs = iq_head ;
  while (qs != NULL) {
    if (VALID(qs->c_entry->state) && (clk_cycle == qs->c_entry->start_cycle))
      fprintf (dump_file, "INST %07ld %07ld %08lx \n", 
			  clk_cycle , 
	                  qs->c_entry->i_no, 
	       		  qs->c_entry->rec_inst.inst) ;
    qs = qs->next_qslot ;
  }

  qs = iq_head ;
  while (qs != NULL) {
    if (VALID(qs->c_entry->state) && (clk_cycle == qs->c_entry->disp_cycle))
      fprintf (dump_file, "EUNIT %07ld %s \n", 
			  qs->c_entry->i_no , 
	                  qs->c_entry->eu_name) ;
    qs = qs->next_qslot ;
  }


  if (VALID(iq_head->c_entry->state))
    fprintf (dump_file, "IQUE %07ld %07ld ", 
	                  clk_cycle , 
			  iq_head->c_entry->i_no) ;
  else
    fprintf (dump_file, "IQUE %07ld %07ld ", 
	                  clk_cycle , 
			  (c_inst_no - 1) ) ;

  qs = iq_head ;
  while (qs != NULL) {

    if (VALID(qs->c_entry->state) && !ANLD(qs->c_entry->state)) {

      if ((clk_cycle >= qs->c_entry->start_cycle) && 	/* fetch pipe		*/
	  (clk_cycle <= qs->c_entry->fetch_cycle))
	fprintf (dump_file, ">") ;

      if ((clk_cycle >  qs->c_entry->fetch_cycle) && 	/* issue pipe		*/
	  (clk_cycle <= qs->c_entry->issue_cycle))
	fprintf (dump_file, "v") ;
      
      if ((clk_cycle >  qs->c_entry->issue_cycle) && 	/* que wait 		*/
	  (clk_cycle <= qs->c_entry->qwait_cycle))
	fprintf (dump_file, "=") ;

      if ((clk_cycle >  qs->c_entry->qwait_cycle) && 	/* dispatch pipe	*/
	  (clk_cycle <= qs->c_entry->disp_cycle))
	fprintf (dump_file, "&") ;

      if ((clk_cycle >  qs->c_entry->disp_cycle)  && 	/* execute pipe		*/
	  (clk_cycle <= qs->c_entry->exec_cycle))
	fprintf (dump_file, "#") ;

      if ((clk_cycle >  qs->c_entry->exec_cycle)  && 	/* execute wait		*/
	  (clk_cycle <= qs->c_entry->ewait_cycle)) {
	if (clk_cycle <= qs->c_entry->lswait_cycle)
	  fprintf (dump_file, "~") ;
	 else
	   fprintf (dump_file, "$") ;
      }

      if ((clk_cycle >  qs->c_entry->ewait_cycle) && 	/* complete pipe	*/
	  (clk_cycle <= qs->c_entry->compl_cycle))
	fprintf (dump_file, "-") ;

      if ((clk_cycle >  qs->c_entry->compl_cycle))	/* commit pipe		*/
	fprintf (dump_file, ".") ;
    
    }

    if (ANLD(qs->c_entry->state)) {

      if ((clk_cycle >= qs->c_entry->start_cycle) && 	/* fetch pipe		*/
	  (clk_cycle <= qs->c_entry->fetch_cycle))
	fprintf (dump_file, ">") ;

      if ((clk_cycle >  qs->c_entry->fetch_cycle) && 	/* issue pipe		*/
	  (clk_cycle <= qs->c_entry->issue_cycle))
	fprintf (dump_file, "v") ;
      
      if ((clk_cycle >  qs->c_entry->issue_cycle))
	fprintf (dump_file, "a") ;
    
    }

    qs = qs->next_qslot ;

  }

  fprintf (dump_file, "\n") ;


} /* end dump_graph */

/*==============================================================================*/

/* The update state function updates all current state variables with next state
 */

void
update_state ()
	
{
  qslot   *qs ;			/* que slot					*/

  qs = iq_head ;
  while (qs != NULL) {
    *qs->c_entry = *qs->n_entry ;
    qs = qs->next_qslot ;
  }

  c_inst_no = n_inst_no ;

} /* end update_state */

/*==============================================================================*/

int
main (int argc ,
      char **argv)

{

  int	last_instr = 0 ;	/* set when last instruction has been fetched	*/
  int	empty_ique = 1 ;	/* set when instruction que is empty		*/

  ireq 	fetch_r ;		/* instruction fetch request 			*/
  ipkt 	fetch_p[MAX_IFETCH] ;	/* instruction fetch packet 			*/

  lsreq	ls_r[MAX_LS_DISP] ;	/* load store request				*/
  lspkt	ls_p[MAX_LS_COMPL] ;	/* load store packet				*/

  ulong	sim_segment = 0 ;	/* simulation segment number			*/
  ulong inst_no ;		/* non annuled instruction number		*/


  int i ;

  signal(SIGUSR1, sigusr1);

  if (!read_arg (argc, argv)) {

    gl_config () ;
    be_config (&fetch_r,    &ls_r[0]) ;	
    fe_config (&fetch_p[0], &ls_p[0]) ;	

    trace_file = open_file (trace_file_name, "rb") ;
    if (dump_mode)
      dump_file = open_file (dump_file_name, "w") ;

    /* Write simulation progress report header. */

    printf ("\nDenali Trace Driven Simulator v%s\n\n", TSIM_VERSION_NO) ;
    printf ("  Trace file         : %s\n", trace_file_name) ;
    if (strcmp(config_file_name, ""))
      printf ("  Configuration file : %s\n", config_file_name) ;
    else
      printf ("  Configuration      : default\n") ;
    printf ("  Statistics file    : %s\n", stats_file_name) ;
    if (dump_mode)
      printf ("  Dump file          : %s\n", dump_file_name) ;
    if (warmup_state) {
       printf("  Warmup state file  : %s\n", warmup_file_name);
    }

    printf ("\n") ;

    while (!done && !(last_instr && empty_ique)) {

      empty_ique = back_end  (&fetch_r, &ls_r[0], &fetch_p[0], &ls_p[0],
                              last_instr) ;
      last_instr = front_end (&fetch_r, &ls_r[0], &fetch_p[0], &ls_p[0]) ;
      
      u2 ();

      update_state () ;

      if (dump_mode)
	dump_graph () ;	

      /* the following code is for debugging purposes 

      if ((clk_cycle == 7240)) {
	printf ("clk_cycle %d \n", clk_cycle) ;

	for (i = 0; i < MAX_LS_DISP; i++) {
	  if (ls_r[i].valid != -1)
	    printf ("request %d %d %lx \n", clk_cycle, ls_r[i].qslot_no, 
		                                 ls_r[i].ea) ;
	}
	for (i = 0; i < MAX_LS_COMPL; i++) {
	  if (VALID(ls_p[i].valid))
	    printf ("packet %d %d %d\n", clk_cycle, ls_p[i].qslot_no, 
		    ls_p[i].compl_cycle) ;
	}

	printf ("\n") ;

	printf ("qslot no %d \n",     iq_head->no) ;
	printf ("src1_cycle %d \n",   iq_head->c_entry->src1_cycle) ; 
	printf ("src2_cycle %d \n",   iq_head->c_entry->src2_cycle) ; 
	printf ("src3_cycle %d \n",   iq_head->c_entry->src3_cycle) ; 
	printf ("ccr_cycle %d \n",    iq_head->c_entry->ccr_cycle) ; 
	printf ("\n") ;
	printf ("src1_qslot %d \n",   iq_head->c_entry->src1_qslot) ; 
	printf ("src2_qslot %d \n",   iq_head->c_entry->src2_qslot) ; 
	printf ("src3_qslot %d \n",   iq_head->c_entry->src3_qslot) ; 
	printf ("ccr_qslot %d \n",    iq_head->c_entry->ccr_qslot) ; 
	printf ("\n") ;
	printf ("fetch_cycle %d \n",  iq_head->c_entry->fetch_cycle) ;
	printf ("issue_cycle %d \n",  iq_head->c_entry->issue_cycle) ;
	printf ("qwait_cycle %d \n",  iq_head->c_entry->qwait_cycle) ;
	printf ("disp_cycle %d \n",   iq_head->c_entry->disp_cycle) ;
	printf ("exec_cycle %d \n",   iq_head->c_entry->exec_cycle) ;
	printf ("ewait_cycle %d \n",  iq_head->c_entry->ewait_cycle) ;
	printf ("lswait_cycle %d \n", iq_head->c_entry->lswait_cycle) ;
	printf ("compl_cycle %d \n",  iq_head->c_entry->compl_cycle) ;

	printf ("\n") ;
      }

      end debugging code */

      /* Write simulation progress to stdout. */

      inst_no = n_inst_no - global_stats.no_of_annulled_insts ;

      if (sim_segment != (inst_no/SIM_SEGMENT_SIZE)) {
	sim_segment = inst_no/SIM_SEGMENT_SIZE ;
	printf ("  Instruction number %07d at clock cycle %07d \n", 
			sim_segment * SIM_SEGMENT_SIZE, 
		        clk_cycle) ;
      }

      clk_cycle++ ;
      global_stats.no_of_cycles++ ;

    }
	
    fclose (trace_file) ;
    if (dump_mode)
      fclose (dump_file) ;

    stats_file = open_file (stats_file_name, "w") ;
    be_stats () ;	
    fe_stats () ;	
    fclose (stats_file) ;
  }
	
  exit (0) ;
	
} /* end main */


/*==============================================================================*/
