/*
 *  $Id: tview.c,v 1.2 1995/01/17 01:14:57 forssell Exp $
 *  
 *  Copyright (c) 1994 HAL Computer Systems International, Ltd.
 *  All rights reserved.  Unpublished -- rights reserved under
 *  the Copyright Laws of the United States.  USE OF A COPYRIGHT
 *  NOTICE IS PRECAUTIONARY ONLY AND DOES NOT IMPLY PUBLICATION
 *  OR DISCLOSURE.
 *  
 *  THIS SOFTWARE CONTAINS CONFIDENTIAL INFORMATION AND TRADE
 *  SECRETS OF HAL COMPUTER SYSTEMS INTERNATIONAL, LTD.  USE,
 *  DISCLOSURE, OR REPRODUCTION IS PROHIBITED WITHOUT THE
 *  PRIOR EXPRESS WRITTEN PERMISSION OF HAL COMPUTER SYSTEMS
 *  INTERNATIONAL, LTD.
 *  
 *                          RESTRICTED RIGHTS LEGEND
 *  Use, duplication, or disclosure by the Government is subject
 *  to the restrictions as set forth in subparagraph (c)(l)(ii)
 *  of the Rights in Technical Data and Computer Software clause
 *  at DFARS 252.227-7013.
 *  
 *           HAL COMPUTER SYSTEMS INTERNATIONAL, LTD.
 *                   1315 Dell Avenue
 *                   Campbell, CA  95008
 */
/*==============================================================================*/ 
/* Hal Denali Trace Driven Simulator						*/
/* Copyright (c) 1994 HaL Computer Systems, Inc.                      		*/
/*==============================================================================*/
/* tview.c									*/
/*										*/
/*==============================================================================*/
/* Revision history								*/
/*										*/
/* Date	     Who	   	Description					*/ 
/* --------  -----------      	-----------					*/
/* 10/30/94  Per Forssell  	Initial defintion				*/
/*										*/
/*==============================================================================*/

#include	<stdio.h>
#include	<string.h>
#include	<ctype.h>
#include	<time.h>
#include	"tview.h"

/*==============================================================================*/

/* The read arguments routine reads in the command line arguments. The command
   line is as follows :

   tview [-h] [-n <cycles per page>] [-v <view file>] [-d <dump file>]

      -n <cycles per page> : number of clock cycles per page
      -v <view file>	   : name of the view output file; stdout if omitted
      -d <dump file>	   : name of the input dump file;  stdin  if omitted
*/

int
read_arg
(int	argc ,
 char	**argv,
 int	*cycles_per_page,
 char	*view_file_name,
 char	*dump_file_name)
	
{

  int	arg_no ;

  *cycles_per_page = DEF_CYCLES_PER_PAGE ;

  strcpy (view_file_name, "stdout") ;
  strcpy (dump_file_name, "stdin") ;

  arg_no = 1 ;
  while (arg_no <= argc) {

    if (!strcmp (*argv, "-n")) {
      *cycles_per_page = atoi(*(argv+1)) ;
      if (*cycles_per_page > MAX_CYCLES_PER_PAGE) {
	printf ("TVIEW: ERROR: Number of cycles per page must be less than %d. \n",
		*cycles_per_page) ;
	return (1) ;
      }
      arg_no++ ;
      argv++ ;
    }

    else if (!strcmp (*argv, "-v")) {
      strcpy (view_file_name, *(argv+1)) ;
      arg_no++ ;
      argv++ ;
    }

    else if (!strcmp (*argv, "-d")) {
      strcpy (dump_file_name, *(argv+1)) ;
      arg_no++ ;
      argv++ ;
    }

    else if (!strcmp (*argv, "-h")) {
      printf ("\n") ;
      printf ("tview ") ;
      printf ("[-h] ") ;
      printf ("[-n <cycles per page>] ") ;
      printf ("[-v <view file>] ") ;
      printf ("[-d <dump file>] \n") ;
      printf ("\n") ;
      printf ("  -n <cycles per page> : no of cycles per page		    \n") ;
      printf ("  -v <view file>       : output view file; stdout if omitted \n") ;
      printf ("  -d <dump file>       : input dump file ; stdin  if omitted \n") ;
      printf ("\n") ;

      return (1) ;
    }
			
    arg_no++ ;
    argv++ ;

  }

  return (0) ;
	
} /* end read_arg */

/*==============================================================================*/

/* open_file is a generic routine to open a file, and to write an appropriate
   error message if the file open did not succeed. */

FILE*
open_file (char *file_name, char *access_type)
	
{
  FILE  *fp ;		/* generic file pointer 				*/

  if (!strcmp(file_name, "stdin"))
    fp = stdin ;
  else if (!strcmp(file_name, "stdout"))
    fp = stdout ;
  else {
    fp = fopen (file_name, access_type) ;
    if (fp == NULL) {
      if (!strcmp (access_type, "r"))
	printf ("TVIEW: ERROR: file %s could not be opened for read \n\n", 
		file_name) ;
      else if (!strcmp (access_type, "w"))
	printf ("TVIEW: ERROR: file %s could not be opened for write \n\n", 
		file_name) ;
      else if (!strcmp (access_type, "a"))
	printf ("TVIEW: ERROR: file %s could not be opened for append \n\n", 
		file_name) ;
      exit (1) ;
    }
  }

  return (fp) ;
    
} /* end open_file */

/*==============================================================================*/

/* time_stamp returns a string with the current time and date. */

char*
current_time ()
	
{
  time_t c_time ;		/* current clock time				*/

  time(&c_time) ;
  return (strtok (ctime(&c_time), "\n")) ;
    
} /* end current_time */

/*==============================================================================*/

/* The read_dumpfile routine reads one full view page worth of data from the dump
   file. It composes the view lines within the view page from the data. */

int
read_dumpfile (FILE   *dump_file, 
	       vline  **page, 
	       long   *first_cycle, 
	       int     cycles_per_page)
	
{
  vline *last_vline ;		/* last valid view line in the page		*/
  vline *current_vline ;	/* current view line				*/

  char	df_line[MAX_LINE] ;	/* a line of data from the dump file		*/

  char  df_rec_type[MAX_NAME] ;	/* dump file record type			*/
  long	df_inst ;		/* dump file instruction opcode			*/
  char	df_ique[MAX_IQUE] ;	/* dump file instruction que state		*/
  char	df_eu_name[MAX_NAME] ;	/* dump file execution unit name		*/
  long  df_inst_no ;		/* dump file instruction number			*/
  long  df_cycle_no ;		/* dump file cycle number			*/

  int   c_cycle = 0 ;		/* current cycle number				*/

  int   cycle_cnt ;		/* clock cycle count				*/
  int   iq_cnt ;		/* instruction que entry count			*/

  int	iq_end ;		/* instruction que end flag			*/

  /* Search for last view line in the view page. Set last_vline to point to the
     last view line. */

  last_vline = *page ;
  if (last_vline != NULL) {
    while (last_vline->next_vline != NULL) {
      last_vline = last_vline->next_vline ;
    }
  }

  /* Read records from simulator dump file for all clock cycles within the 
     current page. Convert the records into view lines. Basically this entails
     turning the IQUE reocords in the dump file '90 degres'. */

  while (c_cycle < cycles_per_page) {

    /* Read next line from dump file. Return if all lines have been read. Read
       record type from line. */

    if (fgets (df_line, MAX_LINE, dump_file) == NULL) 
      return (1) ;
    sscanf (df_line, "%s \n", df_rec_type) ;


    /* For "INST" records, add new view line to page.  Assign the instruction
       number to the view line inst_no field.  Look up the dissasembled opcode
       and assign it to the view line inst_op field.  Initialize the new view
       line instruction history field to 'empty' up to the current cycle time. */

    if (!strcmp(df_rec_type, "INST")) {
      sscanf (df_line, "%*s %d %d %x \n",  &df_cycle_no, 
	                                   &df_inst_no, 
	                                   &df_inst) ;

      if (last_vline != NULL) { 
	last_vline->next_vline = (vline *) malloc (sizeof (vline)) ;
	last_vline = last_vline->next_vline ;
      }
      else {
	last_vline = (vline *) malloc (sizeof (vline)) ;
	*page = last_vline ;
      }

      last_vline->next_vline = NULL ;
	

      last_vline->inst_no = df_inst_no ;
      strcpy(last_vline->inst_op, Instruction_Lookup(df_inst, 0)) ;
      strcpy(last_vline->eu_name, EOL) ;

      strcpy(last_vline->inst_hist, EOL) ;
      for (cycle_cnt = 0; cycle_cnt < c_cycle; cycle_cnt++)
	strcat (last_vline->inst_hist, " ") ;
    }

    /* For "EUNIT" records, search the view page for the view line that matches
       the "EUNIT" record instruction number. Update the execution unit name
       field of the found view line. */

    if (!strcmp(df_rec_type, "EUNIT")) {
        sscanf (df_line, "%*s %d %s \n", &df_inst_no, 
	                                  df_eu_name) ;
    
      current_vline = *page ;
      while (current_vline != NULL) {
	if (current_vline->inst_no == df_inst_no) {
	  strcpy (current_vline->eu_name, df_eu_name) ;
	  break ;
	}
	current_vline = current_vline->next_vline ;
      }
    }


     /* For "IQUE" records, update all view lines with the instruction que state
	data from the "IQUE" record. */
  
    if (!strcmp(df_rec_type, "IQUE")) {
      strcpy (df_ique, "") ;
      sscanf (df_line, "%*s %d %d %s \n", &df_cycle_no, 
	                                  &df_inst_no, 
	                                   df_ique) ;
    
      current_vline = *page ;

      /* Clean up view page if the current clock cycle is the first clock cycle
	 for the current view page. */

      if (c_cycle == 0) {

	/* View lines whose instructions have been committed in previous view
	   pages are deleted from the current view page. We know which view lines
	   to delete by the instruction number in the first "IQUE" entry of the
	   view page. */

	while (current_vline != NULL) {
	  if (current_vline->inst_no < df_inst_no) {
	    *page = current_vline->next_vline ;
	    free (current_vline) ;
	    current_vline = *page ;
	  }
	  else
	    break ;
	}

	/* The instruction history of the view lines that remain in the page are
	   initialized. */

	while (current_vline != NULL) {
	  strcpy(current_vline->inst_hist, EOL) ;
	  current_vline = current_vline->next_vline ;
	} 

	/* Set the first cycle of the page to the cycle number in the "IQUE"
	   record. */

	*first_cycle = df_cycle_no ;
      }

      /* Update all view lines in the view page with the "IQUE" record
	 instruction history field. */

      current_vline = *page ;
      iq_cnt = 0 ;
      if (current_vline != NULL) {
	while (current_vline->inst_no < df_inst_no) {
	  strcat (current_vline->inst_hist, " ") ;
	  current_vline = current_vline->next_vline ;
	}
	
	iq_end = ((df_ique[iq_cnt] == ' ') || (df_ique[iq_cnt] == '\0')); 
	for (iq_cnt = 0; !iq_end && (current_vline != NULL); iq_cnt++) {
	  sprintf (current_vline->inst_hist, "%s%c", current_vline->inst_hist, 
		                                     df_ique[iq_cnt]) ;
	  current_vline = current_vline->next_vline ;
	  iq_end = isspace(df_ique[iq_cnt]) ;
	}
	while (current_vline != NULL) {
	  strcat (current_vline->inst_hist, " ");
	  current_vline = current_vline->next_vline ;
	}
	
      }
      c_cycle++ ;
    }

  }

  return (0) ;
    
} /* end read_dumpfile */

/*==============================================================================*/

/* The write_viewfile routine writes out one view page worth of view lines.  */

void
write_viewfile (FILE   *view_file, 
		vline  **page, 
		char   *dump_file_name, 
		long    first_cycle, 
		int     cycles_per_page,
		char   *time_stamp )
	
{
  static int  page_no = 0 ;	/* page number					*/

  vline *current_vline ;	/* current view line				*/

  char	str[12][MAX_CYCLES_PER_PAGE + 60] ;	/* header strings		*/
  char	tick_str[MAX_CYCLES_PER_PAGE] ;		/* tick marker string		*/
  char	cycle_str[MAX_CYCLES_PER_PAGE] ;	/* cycle count string		*/

  char  filler[10][2] ;		/* header filler characters			*/ 
  char	cycle_cnt_str[10] ;	/* cycle count string				*/

  int    str_cnt ;		/* string character counter			*/
  int    cycle_cnt ;		/* clock cycle counter				*/

  /* Write view page header.  The format for the view page is as follows :

     +===========================================================..========+
     | HaL Denali Trace Viewer vx.xx                                       |
     +===========================================================..========+
     | File : <dump_file_name>                                             | 
     | Page : <page_no>                                                    |
     | Date : <date_stamp>                                                 |
     +==========+==========================+============+========..========+
     | INST NO  | INST OP                  | EU         |                  |
     +----------+--------------------------+------------+--------..--------+
     | 00000001 | add     %r1, %r2, %r3    | ALS1       |                  |
     .          .                          .            .                  .
     +==========+==========================+============+==================+
                                                          
     */

  page_no++ ;

  sprintf (str[0],  "+=========================================================") ;
  sprintf (str[1],  "| HaL Denali Trace Viewer v%s", TVIEW_VERSION_NO) ;
  sprintf (str[2],  "+=========================================================") ;
  sprintf (str[3],  "| File : %s", dump_file_name) ;  
  sprintf (str[4],  "| Page : %-5d", page_no) ;
  sprintf (str[5],  "| Date : %s", time_stamp) ;
  sprintf (str[6],  "+==========+================================+============+") ;
  sprintf (str[7],  "| INST NO  | INST OP                        | EU         |") ;
  sprintf (str[8],  "+----------+--------------------------------+------------+") ;
  sprintf (str[9],  "+==========+================================+============+") ;
  sprintf (str[10], "+----------+--------------------------------+------------+") ;
  sprintf (str[11], "|          |                                |            |") ;

  sprintf (filler[0],  "=") ;
  sprintf (filler[1],  " ") ;
  sprintf (filler[2],  "=") ;
  sprintf (filler[3],  " ") ;
  sprintf (filler[4],  " ") ;
  sprintf (filler[5],  " ") ;
  sprintf (filler[6],  "=") ;
  sprintf (filler[7],  " ") ;
  sprintf (filler[8],  "-") ;
  sprintf (filler[9],  "=") ;
  
  for (str_cnt = 0; str_cnt <= 9; str_cnt++) {
    while (strlen (str[str_cnt]) <= (59 + cycles_per_page)) {
      strcpy (str[str_cnt], strcat (str[str_cnt], filler[str_cnt])) ;
    }
  }

  strcat (str[0], "+") ;
  strcat (str[1], "|") ;
  strcat (str[2], "+") ;
  strcat (str[3], "|") ;
  strcat (str[4], "|") ;
  strcat (str[5], "|") ;
  strcat (str[6], "+") ;
  strcat (str[7], "|") ;
  strcat (str[8], "+") ;
  strcat (str[9], "+") ;


  strcpy (tick_str, "") ;
  cycle_cnt = first_cycle ;
  while (strlen (tick_str) < cycles_per_page) {
    if (((cycle_cnt % 10) == 0) && 
	((cycle_cnt - first_cycle) < cycles_per_page)) 
      strcat (tick_str, "+") ;
    else
      strcat (tick_str, "-") ;
    cycle_cnt++ ;
  }

  strcpy (cycle_str, "") ;
  cycle_cnt = first_cycle ;
  while (strlen (cycle_str) < cycles_per_page) {
    if (((cycle_cnt % 10) == 0) && 
	((cycle_cnt - first_cycle) < cycles_per_page - 10)) {
      sprintf (cycle_cnt_str, "%d", cycle_cnt) ;
      if (strlen (cycle_cnt_str) < 8) {
	strcat (cycle_str, cycle_cnt_str) ;
	cycle_cnt = cycle_cnt + strlen (cycle_cnt_str) ;
      }
      else {
	strcat (cycle_str, " ") ;
	cycle_cnt++ ;
      }
    }
    else {
      strcat (cycle_str, " ") ;
      cycle_cnt++ ;
    }
  }


  /* Write out page header. */
  
  for (str_cnt = 0; str_cnt <= 8; str_cnt++) 
    fprintf (view_file, "%s \n", str[str_cnt]) ;

  /* Write out the full view page. */

  current_vline = *page ;
  while (current_vline != NULL) {

    fprintf (view_file, "| %-8d | %-30s | %-10s | %s | \n",
	     current_vline->inst_no ,
	     current_vline->inst_op ,
	     current_vline->eu_name ,
	     current_vline->inst_hist) ;
    current_vline = current_vline->next_vline ;

  }

  /* Write out cycle count bar. */

  fprintf (view_file, "%s-%s-+ \n", str[10], tick_str) ;
  fprintf (view_file, "%s %s | \n", str[11], cycle_str) ;
  fprintf (view_file, "%s \n\n\n", str[9]) ;


} /* end write_viewfile */

/*==============================================================================*/

int
main (int argc ,
      char **argv)

{
  FILE	*dump_file ;			/* simulator dump file			*/
  FILE	*view_file ;			/* view file				*/

  char	 view_file_name[MAX_FNAME] ;	/* name of the view file		*/
  char   dump_file_name[MAX_FNAME] ;	/* name of the simulator dump file	*/

  vline	 *page[1024] ;			/* page of view lines			*/

  long   first_cycle = 0 ;		/* first clock cycle in current page 	*/

  int	 cycles_per_page ;		/* number of cycles printed per page 	*/
  int	 last_record  = 0 ;		/* last dump file record flag		*/

  char	 time_stamp[MAX_NAME] ;		/* time stamp string			*/

  if (!read_arg (argc, argv, &cycles_per_page, view_file_name, dump_file_name)) {

    dump_file = open_file (dump_file_name, "r") ;
    view_file = open_file (view_file_name, "w") ;

    *page = NULL ;
    Instruction_Lookup_Init();

    strcpy(time_stamp, current_time()) ;

    while (!last_record) {

      last_record = read_dumpfile  (dump_file, 
				    page, 
				    &first_cycle, 
				     cycles_per_page) ;
      write_viewfile (view_file, 
		      page, 
		      dump_file_name, 
		      first_cycle,
		      cycles_per_page,
		      time_stamp) ;

    }
	
    fclose (dump_file) ;
    fclose (view_file) ;

  }
	
  exit (0) ;
	
} /* end main */

/*==============================================================================*/


