/*
 *  $Id: u2.c,v 1.4 1995/10/02 22:37:57 aiman Exp $
 *  
 *  Copyright (c) 1994 HAL Computer Systems International, Ltd.
 *  All rights reserved.  Unpublished -- rights reserved under
 *  the Copyright Laws of the United States.  USE OF A COPYRIGHT
 *  NOTICE IS PRECAUTIONARY ONLY AND DOES NOT IMPLY PUBLICATION
 *  OR DISCLOSURE.
 *  
 *  THIS SOFTWARE CONTAINS CONFIDENTIAL INFORMATION AND TRADE
 *  SECRETS OF HAL COMPUTER SYSTEMS INTERNATIONAL, LTD.  USE,
 *  DISCLOSURE, OR REPRODUCTION IS PROHIBITED WITHOUT THE
 *  PRIOR EXPRESS WRITTEN PERMISSION OF HAL COMPUTER SYSTEMS
 *  INTERNATIONAL, LTD.
 *  
 *                          RESTRICTED RIGHTS LEGEND
 *  Use, duplication, or disclosure by the Government is subject
 *  to the restrictions as set forth in subparagraph (c)(l)(ii)
 *  of the Rights in Technical Data and Computer Software clause
 *  at DFARS 252.227-7013.
 *  
 *           HAL COMPUTER SYSTEMS INTERNATIONAL, LTD.
 *                   1315 Dell Avenue
 *                   Campbell, CA  95008
 */

/**********************************************************************
  Program to evaluate benchmark programs on Denali. This portion of
  the simulator models the denali unified level 2 cache.
  
  Started 8/7/95, Aiman Kabakibo
  
 ***********************************************************************/
#include <stdio.h>
#include "globdef.h"
#include "config_front_end.h"
#include "opcodes.h"
#include "u2.h"



boolean tag_blocked, bus_blocked;
static wait_buffer *request, *request1, *request2;
static boolean blocked_max_misses;
static boolean   wqueue_full, u2_blocked;
static tag_queue *tag_wqueue, *tag_rqueue, *tag_pipe, *miss_queue;
void do_stream_buffer();


boolean stall_tag, max_misses, stall_pipe, debug;
unsigned last_pc, last_stall;

wait_buffer stream_buffer[STREAM_BUFFER_SIZE];

void u2()
{
   int i;
   int addr;
   static wait_buffer *completed_req=0;
   req_state lookup_state;
   boolean retry=FALSE;
   unsigned tmp;



#ifdef DEBUG
   if (clk_cycle>=DEBUG) {
      print_pre_status();
      debug=TRUE;
   }
#endif

   /* Update stall statistics */
   if (mrkr->pc == last_pc) {

      /* CPU might be blocked */
      if (i_wt_match(mrkr->pc)) {
	 if (last_stall != mrkr->pc) {
	    last_stall = mrkr->pc;
	    cpu_stall_imiss++;
	    if(debug)
		printf("\tI fetch stall\n");
	 }
      }
      /* Data fetch stall numbers are not really correct due to the way
         we measure them.  Use this just as a feel for data stall versus
	 a real number */
      else if ((tmp=d_wt_match(mrkr->ea)) &&  tmp != last_stall) {
	 last_stall = tmp;
	 cpu_stall_dmiss++;
	 if (debug) 
	     printf("\tD fetch stall\n");
      }
   }
   else {
      last_pc = mrkr->pc;
   }
 

   /* Updated miss count for pending misses */
   update_misses();

   /* Check if stream buffer need to do any work */
   if (fl_conf.istream_buffer) {
      do_stream_buffer();
   }
 
   /* Get next request to the tag/data unit */
   if (request) {
      /* Request was not serviced last cycle, we need to retry */
      retry = TRUE;
   }
   else if (request2) {
      request = request2;
      request2= 0;
   }
   else  if ((request = is_completed_miss())) {
      blocked_max_misses = FALSE;

      if (fl_conf.ucache_blocking) {
	 u2_blocked = FALSE;
      }

      shift_queue(miss_queue, fl_conf.ucache_pmisses);
   }
   else if (blocked_max_misses == FALSE) {

      if (fl_conf.ucache_blocking) {
	 if (u2_blocked == FALSE) {
	    if ((request = get_request())) {
	       u2_blocked = TRUE;
	    }
	 }
      }
      else {
	 request = get_request();
      }
   }


   /* Send request to tag/data unit */
   if (blocked_max_misses == FALSE) {

      if (request) {

	 if (request->cmd == SNOOP_WB) {
	    /* we have to write this to the bus. We will just ack request
               for now */
	    ucache_l1_dsnoops++;
	    request->state = INVALID;
	    request = 0;
	    if (fl_conf.ucache_blocking) {
	       u2_blocked = FALSE;
	    }
	 }
	 else if (request->cmd == WRITEBACK ) {
	    if (request->multi_cycle == FALSE && retry == FALSE) {

	       /* Will ack L1 and do writeback in background */
	       request1 = copy_request(request);
	       request2 = copy_request(request);
	    
	       request1->multi_cycle = FALSE;
	       request2->multi_cycle = TRUE;

	       request->state = INVALID;
	       request = request1;

	       if (fl_conf.ucache_blocking) {
		  u2_blocked = FALSE;
	       }
	    }

	    /* We will do the writeback in background */
	    if (tag_unit(request, TAG_WRITE, &completed_req) == TRUE) {
	       if (request->multi_cycle == FALSE) {
		  ucache_wb_refs++;
		  ucache_avg_refs += count_ucache_wt_reqs();
	       }
	       request = 0;
	    }
	 }

	 else if (request->cmd == RFILL || request->cmd == WFILL) {

	    /* Signal L1 that data is available, and write data to U2*/
	    if (request->multi_cycle == FALSE && retry == FALSE) {
	       request1 = copy_request(request);
	       request2 = copy_request(request);

	       request1->multi_cycle = FALSE;
	       request2->multi_cycle = TRUE;

	       request->state = COMPLETED;
	       request = request1;
	    }

	    /* Need to update the tags right away since we have sent 
               data to L1, we also do not want to hold on to the data
               so we will do right away */

	    if (tag_unit(request, TAG_READ, &completed_req) == TRUE) {
	       request = 0;
	    }
	 }

	 else if (request->cmd == RETAG) {
	    request->multi_cycle = FALSE;
	    if (tag_unit(request, TAG_READ, &completed_req) == TRUE) {
	       if (request->multi_cycle == FALSE) {
		  ucache_retag_refs++;
		  ucache_avg_refs += count_ucache_wt_reqs();
	       }
	       request = 0;
	    }
	 }

	 else if(request->cmd == READ || request->cmd == WRITE) {
	    
	    /* Normally, most U2 requests will hit in U2, requiring 2
               pipeline cycles to get data.  In case of a U2 miss, we
               only now the in the first half of the second pipeline
               cycle, so if we decided to cancel the request or send a
	       new one, we will only save 1/2 pipeline cycle or 1 cpu
	       cycle. We are ignoring this for now */
	    if (request->multi_cycle == FALSE && retry == FALSE) {
	       request2 = copy_request(request);
	       request->multi_cycle = FALSE;
	       request2->multi_cycle = TRUE;
	    }

	    if (tag_unit(request, TAG_READ, &completed_req) == TRUE) {
	       if (request->multi_cycle == FALSE) {
		  if (request->requestor == I0REQ) {
		     ucache_i_refs++;
		  }
		  if (request->requestor == D0REQ) {
		     ucache_d_refs++;
		  }
		  ucache_avg_refs += count_ucache_wt_reqs();
	       }
	       request = 0;
	    }
	 }

	 else {
	    assertion_error("I am not suppose to get here!");
	 }
      }
      else {
	 /* No request, update the tag unit state */
	 tag_unit((wait_buffer *)0, TAG_INVALID, &completed_req);
      }
   }


   if (completed_req) {
      if ((completed_req->multi_cycle == TRUE) || 
	  (access_u2(completed_req) == TRUE)) {
	 completed_req = (wait_buffer *) 0;
      }
   }
      
   /*      bus_unit(); */


#ifdef DEBUG
   if (clk_cycle>=DEBUG) {
      print_post_status();
      debug=TRUE;
   }
#endif

}


/*-----------------------------------------------------------------------*/
void
update_misses()
{

   int i, cnt=0;
   wait_buffer *req=0;


   for (i=fl_conf.ucache_pmisses-1; i>=0; i--) {
      if (miss_queue[i].request) {
	 --miss_queue[i].request->cycles;
      }
   }
}
/*-----------------------------------------------------------------------*/
wait_buffer *
is_completed_miss ()
{

   int i, cnt=0;
   wait_buffer *req=0;


   for (i=fl_conf.ucache_pmisses-1; i>=0; i--) {
      if (miss_queue[i].request && (miss_queue[i].request->cycles < 0)) {

	 /* Service first access */
	    
	 if (debug)
	     printf("\tCMPMISS     va=%08x cmd=%d t=%-5d \n",
		    miss_queue[i].addr, miss_queue[i].type, 
		    clk_cycle);
	 
	 
	 req = miss_queue[i].request;
	 max_misses = FALSE;
	 
	 /* Clean up miss queue state */
	 miss_queue[i].request=0;
	 miss_queue[i].addr = 0;
	 miss_queue[i].type = TAG_INVALID;
	 break;
      }
   }
 

   return(req);
}

/*-----------------------------------------------------------------------*/
/* return TRUE to indicate it has accepted the request or false to retry */
boolean
tag_unit (wait_buffer *request, tag_req_type type, wait_buffer **comp_request)
{


   static unsigned pending_requests=0;
   static unsigned SRAM_CLK=0;

   int i;
   int snoop_hit = FALSE;


   /* assertion: can not be blocked because of wqueue_full and get a new
      request at same time */
   if (request && (wqueue_full == TRUE || tag_blocked == TRUE)) {
      assertion_error ("Tag got a new request when it was blocked");
   }


   /* Algoirthm:
      1. We advance pipeline if we have requests
      2. Do a wqueue request if:
            - queue is full
	    - no new read request 
         Otherwise take a new request
    */



   /* Tag unit process reques on even SRAM_CLK values and block on odd values */
   /* This is done because SRAM pipeline cycle is 2 cpu cycles */
   if (++SRAM_CLK % 2) {
      if (pending_requests)  {
	 return (FALSE);
      }
   }


   /* Advance SRAM pipeline */
   if (pending_requests) {

      /* send completed request to controller */
      *comp_request = pop_queue(&tag_pipe[0], fl_conf.ucache_pipestages);
      if (*comp_request) {
	 pending_requests--;
	 tag_blocked = FALSE;
	 if ((*comp_request)->multi_cycle || 
	     (*comp_request)->cmd == WRITEBACK) {

	    (*comp_request)->state = INVALID;
	    (*comp_request) = 0;
	 }
      }
      shift_queue(&tag_pipe[0], fl_conf.ucache_pipestages);
   }
   shift_queue(&tag_rqueue[0], TAG_RQUEUE_SIZE);
   


   if (request && (wqueue_full == FALSE)) {

      /* Push new request into the queue */
      if (!pending_requests) {SRAM_CLK=0;}


      switch(type) {
       case TAG_WRITE:
	 pending_requests++;

	 push_queue(request,TAG_WRITE,tag_wqueue,TAG_WQUEUE_SIZE);
	 if (is_queue_full(tag_wqueue,TAG_WQUEUE_SIZE)) {
	    wqueue_full = TRUE;
	 }
	 else {
	    shift_queue(tag_wqueue, TAG_WQUEUE_SIZE); 
	 }
	 request->state = INPROGRESS;

	 break;
	 
       case TAG_READ:
	 if (request->multi_cycle == FALSE && 
	     (snoop_queue(request, tag_rqueue, TAG_RQUEUE_SIZE)== TRUE ||
	      snoop_queue(request, tag_wqueue, TAG_WQUEUE_SIZE)== TRUE)) {
	    snoop_hit = TRUE;
	 }
	 else {
	    pending_requests++;
	    push_queue(request, TAG_READ, tag_rqueue, TAG_RQUEUE_SIZE);
	    push_queue(request, TAG_READ, tag_pipe,   fl_conf.ucache_pipestages);
	    request->state = INPROGRESS;
	 }
	 break;

       default:
	 assertion_error("Unexpected command type received in tag unit");

      }
   }

   /* If write queue is full or we do not have a new request, do a write 
      request from the write queue */
   if (wqueue_full || !request || snoop_hit) {
      
      for (i=TAG_WQUEUE_SIZE-1; i>=0; i--) {
	 if (tag_wqueue[i].request) {
	    push_queue(tag_wqueue[i].request, TAG_WRITE, tag_pipe, fl_conf.ucache_pipestages);
	    tag_wqueue[i].request=0;
	    wqueue_full = FALSE;
	    shift_queue(tag_wqueue, TAG_WQUEUE_SIZE); 
	    break;
	 }
      }
   }

   return (snoop_hit? FALSE : TRUE);
}

/*-----------------------------------------------------------------------*/
#ifdef XXXX
void
access_cache (unsigned va, unsigned type, wait_buffer * request) 
{
  int i, index, flag;

#ifndef OLD_U2
  return;
#endif

  printf("t=%5d servicing a request to va=%08x cmd=%d ", 
	 clk_cycle, va, type);
  

  request->pending = 1;

  index = (va >> fl_conf.ushft) % fl_conf.ucache_size;
                               /* indexing into the ucache */
  
  Umiss_flag = 1;
  ucache_refs++;
  flag = 0;
  
  /* If we have a U2 hit update tag */
  for (i = 0; (i < fl_conf.ucache_sets) && (flag == 0); i++) {
    if (((ucache[index].set[i] >> fl_conf.ushft) == (va >> fl_conf.ushft))
	&& (((ucache[index].valid >> i) % 2) == VALID)) {
      Umiss_flag = 0;
      request->cycles = fl_conf.unified_lat;

      flag = 1;
      if (type == WRITEBACK)
	ucache[index].dirty =  ucache[index].dirty | (1 << i);
      
      ucache[index].repl = u2_lru_repl((unsigned)i, ucache[index].repl);
    }
  }
  
  if (Umiss_flag) /* miss case */ {
    request->cycles =  fl_conf.memory_lat + fl_conf.unified_lat;

    ucache_miss++;
    
    flag = 0; /* I DO NOT THINK THIS LINE IS NEEDED */
    

    for (i = 0; (i < fl_conf.ucache_sets) && (flag == 0); i++) {

      /* U2 miss, look for an invalid entry to replace */
      if (((ucache[index].valid >> i) % 2) == 0) {
	flag = 1;
	
	if (type == WRITEBACK)
	  ucache[index].dirty =  ucache[index].dirty | (1 << i);
	else
	  ucache[index].dirty = ucache[index].dirty & (~(1 << i));
	
	ucache[index].valid = ucache[index].valid | (0x1 << i);
	ucache[index].set[i] = va;
	ucache[index].repl = u2_lru_repl((unsigned)i, ucache[index].repl);

      }
    }

    if (flag == 0) /* lru replacement case */ {
      if (ucache[index].dirty & (0x1 << 
				(ucache[index].repl % fl_conf.ucache_sets))) {
	ucache_drepl++;
      }
      ucache_repl++;

      ucache[index].set[ucache[index].repl%fl_conf.ucache_sets] = va;

      if (type == WRITEBACK) 
	ucache[index].dirty =  ucache[index].dirty | 
	  (1 << (ucache[index].repl % fl_conf.ucache_sets));
      else
	ucache[index].dirty = ucache[index].dirty & 
	  (~(1 << (ucache[index].repl % fl_conf.ucache_sets)));

      ucache[index].repl = u2_lru_repl((unsigned)i, ucache[index].repl);
    }
  }

}
#endif
/*-----------------------------------------------------------------------*/
int 
clear_inclusion (unsigned ea)
{
  int index;

  index = (ea >> fl_conf.ushft) % fl_conf.ucache_size;
                               /* indexing into the ucache */
  if (!(((ucache[index].va >> fl_conf.ushft) == (ea >> fl_conf.ushft)) &&
	ucache[index].valid == TRUE) ) {

     ucache[index].inclusion = NOINCL;
  }
   
}
/*-----------------------------------------------------------------------*/
boolean
access_u2 (wait_buffer * request)
{
  int i, index;
  unsigned va;
  unsigned writeback = FALSE;

  va = request->requestor==I0REQ ? request->va : request->ea;
  index = (va >> fl_conf.ushft) % fl_conf.ucache_size;
                               /* indexing into the ucache */



  if ((request->cmd == RFILL) || (request->cmd == WFILL)) {
     if (request->multi_cycle == FALSE) {
	ucache[index].valid = TRUE;
	ucache[index].va = va;
	

	/* set inclusion of D side only */
	ucache[index].inclusion = (request->requestor == D0REQ)?
	  D0INCL:I0INCL;

	if (request->cmd == WFILL) {
	   ucache[index].dirty = TRUE;
	}
     } 

     /* Free up space for temp buffer */
     request->state = INVALID;
     request = 0;
     return(TRUE);
  }

  if (!(((ucache[index].va >> fl_conf.ushft) == (va >> fl_conf.ushft)) &&
	ucache[index].valid == TRUE) && (max_misses == TRUE) ) {

     /* Reached max outstanding misses.  Need to block */
     blocked_max_misses = TRUE;
     return(FALSE);
  }
  
  ucache_refs++;
   
  if (((ucache[index].va >> fl_conf.ushft) == (va >> fl_conf.ushft)) &&
      ucache[index].valid == TRUE) {

     if (debug)
	 printf("\tACCESS  va=%08x cmd=%d t=%-5d i=%02x t=%06x U2 HIT\n",
		va, request->cmd, clk_cycle, (va>>6)& 0xff, (va>>14));

     /*               */
     /* WE HAVE A HIT */
     /*               */

     if (fl_conf.ucache_blocking) {
	u2_blocked = FALSE;
     }

     switch(request->cmd) {

      case WRITEBACK:
      case SNOOP_WB:
	if (ucache[index].dirty != TRUE) {
	   assertion_error("Writeback to a line that is not dirty in U2");
	}

	request->state = INVALID;
	ucache[index].inclusion = NOINCL;
 	break;

      case RETAG:
	ucache[index].dirty = TRUE;
	request->state = COMPLETED;
	break;

      case WRITE:
	ucache[index].inclusion = D0INCL;
	if (request->requestor == I0REQ) 
	    assertion_error("I0 is making a write request to u2!??");
	ucache[index].dirty = TRUE;
	request->state = COMPLETED;
	break;

      case READ:
	ucache[index].inclusion = (request->requestor == D0REQ)?
	  D0INCL:I0INCL;
	request->state = COMPLETED;
	break;

      default:
	assertion_error("I am not suppose to get here (access_u2)!");
     }
  } 

  else {               

     /*                */
     /* WE HAVE A MISS */
     /*                */



     ucache_miss++;

     if (request->cmd == WRITEBACK) {
	assertion_error("U2 miss for the L1 writeback line");
     }

     if (ucache[index].valid == TRUE) {
	ucache_repl++;
	if (ucache[index].dirty == TRUE) {

	   /*                                 */
	   /* WE HAVE A MISS WITH A WRITEBACK */
	   /*                                 */
	   ucache_drepl++;
	   writeback = TRUE;
	   /* U2 NEEDS TO WRITBACK THE OLD LINE */


	   /* FOR NOW WE WILL JUST READ LINE OUT INTO WB 
	      BUFFER AND INVALIDATE */
	}
	if (ucache[index].inclusion == D0INCL) {
	   ucache_l1_snoops++;
	   wt_snoop_cache[0].ea    = ucache[index].va;
	   wt_snoop_cache[0].state = VALID;
	   wt_snoop_cache[0].cmd   = SNOOP;
	   wt_snoop_cache[0].time  = clk_cycle;
	}
	ucache[index].valid = FALSE;

     }
 

     if (debug) {
	if (writeback) 
	    printf("\tACCESS  va=%08x cmd=%d t=%-5d i=%02x t=%06x U2 MISS, WB %08x\n", 
		   va, request->cmd, clk_cycle, (va>>6)&0xff, (va>>14),
		   ucache[index].va);
	else
	    printf("\tACCESS  va=%08x cmd=%d t=%-5d i=%02x t=%06x U2 MISS\n", 
		   va, request->cmd, clk_cycle, (va>>6)&0xff, (va>>14));
     }
     
     blocked_max_misses = FALSE;	/* reset flag */
     /* We will remap the comman to a refill */
     if (request->cmd == WRITE) {
	request->cmd = WFILL;
     }
     else {
	request->cmd = RFILL;
     }
     
     push_queue(request, request->cmd, miss_queue, fl_conf.ucache_pmisses);
     ucache_avg_misses += count_ucache_miss_reqs ();

     if (is_queue_full(miss_queue, fl_conf.ucache_pmisses)) {
	max_misses = TRUE;
     }
     else {
	shift_queue(miss_queue, fl_conf.ucache_pmisses); 
     }
     request->cycles = fl_conf.memory_lat;
     
  }  

  return(TRUE);
}

/*-----------------------------------------------------------------------*/
#define TEMP_BUFF_SIZE 100

wait_buffer *
copy_request(wait_buffer * request)
{
   static wait_buffer temp_buffer[TEMP_BUFF_SIZE];
   int i;

   for (i=0; i<TEMP_BUFF_SIZE; i++) {
      if (temp_buffer[i].state == INVALID) {
	 temp_buffer[i] = *request;
	 return(&temp_buffer[i]);
      }
   }
   assertion_error("No room in temp buffer to copy the new request");
   return(0);
}

/*-----------------------------------------------------------------------*/
void
shift_queue (tag_queue * queue, unsigned size)
{
   int i;

   for (i=size-1;i>0;i--) {
      queue[i] = queue[i-1];
   }

   queue[0].request = 0;
   queue[0].type = INVALID;

}
/*-----------------------------------------------------------------------*/
void
push_queue (wait_buffer * request, tag_req_type  type, 
	    tag_queue * queue, unsigned size)
{

   if (queue[0].request) {
      assertion_error("Attempting to push a request into a full queue");
   }
   queue[0].request = request;
   queue[0].type    = type;
   queue[0].addr    = request->requestor==I0REQ?
                      request->va:
                      request->ea;
}

/*-----------------------------------------------------------------------*/
wait_buffer *
pop_queue (tag_queue * queue, unsigned size) 
{
   wait_buffer * entry;

   entry = queue[size-1].request;
   queue[size-1].request = 0;
   queue[size-1].type = INVALID;
   
   return(entry);
}
/*-----------------------------------------------------------------------*/
boolean
is_queue_full (tag_queue * queue, unsigned size) 
{
   int i;

   return(queue[size-1].request?TRUE:FALSE);
}


/*-----------------------------------------------------------------------*/
boolean
snoop_queue (wait_buffer *request, tag_queue * queue, unsigned size) {
   int i;
   /* AIMAN: would like to merge va and ea into one */
   for (i=size-1;i>0;i--) {
      if (queue[i].request) {
	 if (((request->requestor == I0REQ) && 
	      (request->va == queue[i].addr)) ||
	     ((request->requestor == D0REQ) &&
	      (request->ea == queue[i].addr))){
	    return(TRUE);
	 }
      }
   }
   return(FALSE);
}


/*-----------------------------------------------------------------------*/
#ifdef XXXX
unsigned 
u2_lru_repl(unsigned x, unsigned y)
{
   int i;
   unsigned tmp;
  
   if (fl_conf.ucache_sets == 1) {
      return(0xe4);
   } 
  
   tmp = x << 6;
  
   if (((y >> 6) % 4) != x)
       tmp = tmp | ((y & 0xc0) >> 2);
   else
       {
	  tmp = tmp | (y & 0x3f);
	  return (tmp);
       }
  
   if (((y >> 4) % 4) != x)
       tmp = tmp | ((y & 0x30) >> 2);
   else
    
       {
	  tmp = tmp | (y & 0xf);
	  return (tmp);
       }
  
   if (((y >> 2) % 4) != x)
       tmp = tmp | ((y & 0xc) >> 2);
   else
       tmp = tmp | (y & 0x3);
   return(tmp);
}

#endif
/*-----------------------------------------------------------------------*/
wait_buffer *
get_request()
{
   int i;
   wait_buffer *req = FALSE;
   int sb_hit = FALSE;


   
   /* requests to the tag unit can come from one of the following places
      1. L1 Instr cache
      2. L1 Data cache
      3. stream buffer

    */


   /* If last request was retried because of a snoop match in tag unit
      we retry */
   if (!req) {

      /* Get the earliest valid request from the instruction queue */
      for (i=0;i<WT_IU_CACHE_SIZE;i++) { 

	 if (wt_iu_cache[i].state == VALID) { 
	    wt_iu_cache[i].time = clk_cycle;
	    if (req) {
	       /* Get earliest request */
	       if (wt_iu_cache[i].time < req->time ) {
		  req = &wt_iu_cache[i];
	       }
	    } 
	    else { req = &wt_iu_cache[i]; }
	 }
      }

      /* if we found a new request, look it up in stream buffer */
      if (fl_conf.istream_buffer && req) {
	 for (i=0;i<(fl_conf.istream_buffer);i++) { 
	    if (((stream_buffer[i].state == COMPLETED) ||
		(stream_buffer[i].state == INPROGRESS))&& 
		((stream_buffer[i].va >> fl_conf.ushft) ==
		(req->va >> fl_conf.ushft))) { 

	       /* we found a match in streaming buffer */
	       if (stream_buffer[i].state == COMPLETED) {
		  sb_hit = TRUE;
		  req->state = COMPLETED;
		  stream_buffer[i].state = INVALID;
		  stream_buffer_hit++;
#ifdef DEBUG
		  printf("STREAMHIT\n");
#endif
	       }
	       else {
		  /* Retry since request is already being processed */
		  req = 0;
	       }

	       break;
	    }
	 }
	 if (! sb_hit && req) {
	    /* need to install stream_buffer requests */
	    stream_buffer_miss++;
	    for (i=0;i<(fl_conf.istream_buffer);i++) { 
	       stream_buffer[i] = *req;
	       stream_buffer[i].va += (i+1)*fl_conf.uline;
	    }
	 }
      }
   }
 



   if (!req) {
      /* Get the earliest valid request from the ldst queue */
      for (i=0;i<(fl_conf.max_ut_buf+1);i++) { 
	 if (wt_unifd_cache[i].state == VALID) { 
	    wt_unifd_cache[i].time = clk_cycle;

	    if (req) {
	       /* Get earliest request */
	       if (wt_unifd_cache[i].time < req->time ) {
		  req = &wt_unifd_cache[i];
	       }
	    } 
	    else { req = &wt_unifd_cache[i]; }
	 }
      }
   }

   if (fl_conf.istream_buffer && !req) {
      /* Get the earliest valid request from the stream buffer */
      for (i=0;i<(fl_conf.istream_buffer);i++) { 
	 if (stream_buffer[i].state == VALID) { 
	    stream_buffer[i].time = clk_cycle;
	    req = &stream_buffer[i];
	    break;
	 }
      }
   }

   return(req);
}
/*-----------------------------------------------------------------------*/

void init_u2 () 
{
   int i;
   char file[MAX_FNAME];
   FILE *fp;


   ucache = (unified_cache *) 
       malloc(fl_conf.ucache_size*sizeof(unified_cache));


   tag_pipe = (tag_queue *) 
       malloc(fl_conf.ucache_pipestages*sizeof(tag_queue));

   tag_rqueue = (tag_queue *) 
       malloc(TAG_RQUEUE_SIZE*sizeof(tag_queue));

   tag_wqueue = (tag_queue *) 
       malloc(TAG_WQUEUE_SIZE*sizeof(tag_queue));

   miss_queue = (tag_queue *) 
       malloc(fl_conf.ucache_pmisses*sizeof(tag_queue));

   if (warmup_state) {
      /* we need to slam the state */
      sprintf(file, "%s.u2", warmup_file_name);

      if (( fp = fopen(file, "r")) == (FILE *) NULL) {
	 perror("Failed to open U2 warmup file");
	 exit(1);
      }

      for (i=0;i<fl_conf.ucache_size;i++) {
	 if (!fread((char *)&ucache[i], sizeof(unified_cache), 1, fp)) {
	    perror("Warmup state size for U2 does not match expected size");
	    exit(1);
	 }

      }
      close(fp);

   }
   else {
         /* initialize unified cache */
      for (i=0;i<fl_conf.ucache_size;i++) {
	 ucache[i].va =  0;
	 ucache[i].inclusion = NOINCL;
	 ucache[i].dirty  =  0;
	 ucache[i].valid  =  0;
      } 
   }

   /* initialize pipeline queue */
  for (i=0;i<fl_conf.ucache_pipestages;i++) {
     tag_pipe[i].request = (wait_buffer *) 0;
     tag_pipe[i].addr = 0;
     tag_pipe[i].type = TAG_INVALID;
  }
   /* initialize tag write queue */
  for (i=0;i<TAG_WQUEUE_SIZE;i++) {
     tag_wqueue[i].request = (wait_buffer *) 0;
     tag_wqueue[i].addr = 0;
     tag_wqueue[i].type = TAG_INVALID;
  }
   /* initialize tag read queue */
  for (i=0;i<TAG_RQUEUE_SIZE;i++) {
     tag_rqueue[i].request = (wait_buffer *) 0;
     tag_rqueue[i].addr = 0;
     tag_rqueue[i].type = TAG_INVALID;
  }
   /* initialize miss queue */
  for (i=0;i<fl_conf.ucache_pmisses;i++) {
     miss_queue[i].request = (wait_buffer *) 0;
     miss_queue[i].addr = 0;
     miss_queue[i].type = TAG_INVALID;
  }

}
/*-----------------------------------------------------------------------*/

void
assertion_error (char *str) 
{
   fprintf(stderr, "ASSERTION ERROR @%d: %s\n", clk_cycle, str);
   exit(1);
}

/*-----------------------------------------------------------------------*/
print_pre_status()
{
   extern gconf gl_conf;
   int i;



   printf ("%5d __________________________________________________________\n", clk_cycle);

   printf("\tTRACE       pc=%08x iw=%08x ea=%08x\n", 
	  mrkr->pc, mrkr->iw, mrkr->ea);
#ifdef XXXX
   for (i=0;i<gl_conf.no_of_ls_bufs;i++) {
	if ( wt_ldst_req[i].state != INVALID ) {
	   printf ("\tLSQUEUE s=%d ea=%08x state=%d sn=%2d is=%d\n",
		   i,
		   wt_ldst_req[i].ea,
		   wt_ldst_req[i].state,
		   wt_ldst_req[i].qslot_no,
		   wt_ldst_req[i].va);

	}
     }
#endif

   /* print instr wait queue */
   for (i=0;i<WT_IU_CACHE_SIZE;i++) { 
      if (wt_iu_cache[i].state != INVALID) {
	  printf("\tINST_WQ s=%d va=%08x cmd=%d state=%d t=%-5d req=%d\n",
	     i,
	     wt_iu_cache[i].va, 
	     wt_iu_cache[i].cmd,
	     wt_iu_cache[i].state, 
	     wt_iu_cache[i].time,
	     wt_iu_cache[i].requestor);
       }
   }

   if (fl_conf.istream_buffer) {
      /* print stream buffer wait queue */
      for (i=0;i<WT_IU_CACHE_SIZE;i++) { 
	 if (stream_buffer[i].state != INVALID) {
	    printf("\tSTREAMQ s=%d va=%08x cmd=%d state=%d t=%-5d req=%d\n",
		   i,
		   stream_buffer[i].va, 
		   stream_buffer[i].cmd,
		   stream_buffer[i].state, 
		   stream_buffer[i].time,
		   stream_buffer[i].requestor);
	 }
      }
   }

   /* print data wait queue */
   for (i=0;i<(fl_conf.max_ut_buf+1);i++) { 
      if (wt_unifd_cache[i].state != INVALID) {
	 printf("\tDATA_WQ s=%d va=%08x cmd=%d state=%d t=%-5d req=%d\n",
	     i,
	     wt_unifd_cache[i].ea, 
	     wt_unifd_cache[i].cmd,
	     wt_unifd_cache[i].state, 
	     wt_unifd_cache[i].time,
	     wt_unifd_cache[i].requestor);
      }
   }


   for (i=0;i<fl_conf.ucache_pipestages;i++) { 

      if (tag_pipe[i].request) {
	 printf("\tPIPE    s=%d va=%08x cmd=%d state=%d t=%-5d req=%d m=%d\n",
	     i,
	     tag_pipe[i].addr,
	     tag_pipe[i].request->cmd,
	     tag_pipe[i].request->state, 
	     tag_pipe[i].request->time,
	     tag_pipe[i].request->requestor,
	     tag_pipe[i].request->multi_cycle);
      }
   }

#ifdef XXX
   for (i=0;i<TAG_RQUEUE_SIZE;i++) { 

      if (tag_rqueue[i].request) {
	 printf("\tRQUEUE  s=%d va=%08x\n",
	     i,
	     tag_rqueue[i].addr);
      }
   }
#endif
   for (i=0;i<TAG_WQUEUE_SIZE;i++) { 

      if (tag_wqueue[i].request) {
	 printf("\tWQUEUE  s=%d va=%08x\n",
	     i,
	     tag_wqueue[i].addr);
      }
   }

   for (i=fl_conf.ucache_pmisses-1; i>=0; i--) {
      if (miss_queue[i].request && miss_queue[i].type != TAG_INVALID) {

	 printf("\tMISSES  s=%d va=%08x cmd=%d state=%d t=%-5d req=%d c=%d\n",
		i,
		miss_queue[i].addr,
		miss_queue[i].request->cmd,
		miss_queue[i].request->state,
		miss_queue[i].request->time,
		miss_queue[i].request->requestor,
		miss_queue[i].request->cycles);
      }
   }

}
/*-----------------------------------------------------------------------*/

print_post_status()
{
   extern gconf gl_conf;
   int i;


   /* print snoop queue */
   for (i=0;i<(1);i++) { 
      if (wt_snoop_cache[i].state != INVALID) {
	 printf("\tSNOOPWQ s=%d va=%08x cmd=%d state=%d t=%-5d req=%d\n",
	     i,
	     wt_snoop_cache[i].ea, 
	     wt_snoop_cache[i].cmd,
	     wt_snoop_cache[i].state, 
	     wt_snoop_cache[i].time,
	     wt_snoop_cache[i].requestor);
      }
   }
 
   if(blocked_max_misses || wqueue_full || max_misses) {
      printf("\tMMiss_Blocked=%c  WQueue_full=%c  Max_misses=%c Blocked=%c\n",
	     blocked_max_misses==TRUE?  'Y' :'N',
	     wqueue_full==TRUE? 'Y' :'N',
	     max_misses==TRUE?  'Y' :'N',
	     u2_blocked==TRUE?  'Y' :'N'
	     );
   }
}

/*-----------------------------------------------------------------------*/
int
count_ucache_wt_reqs()
{
   int i, cnt=0;

   for (i=0;i<(fl_conf.max_ut_buf+1);i++) { 
      if (wt_unifd_cache[i].state != INVALID) {
	 cnt++;
      }
   }
   for (i=0;i<WT_IU_CACHE_SIZE;i++) { 
      if (wt_iu_cache[i].state != INVALID) {
	 cnt++;
      }
   }

   if (fl_conf.istream_buffer) {
      for (i=0;i<fl_conf.istream_buffer;i++) { 
	 if ((stream_buffer[i].state == VALID) || 
	     (stream_buffer[i].state == INPROGRESS)) {
	    cnt++;
	 }
      }
   }

   return(cnt); 
}

/*-----------------------------------------------------------------------*/
int
count_ucache_miss_reqs()
{
   int i, cnt=0;

   for (i=0;i<(fl_conf.ucache_pmisses);i++) { 
      if (miss_queue[i].request) {
	 cnt++;
      }
   }
   return(cnt);
}
/*-----------------------------------------------------------------------*/

boolean 
i_wt_match (unsigned va) 
{
   int i;
   for (i=0;i<WT_IU_CACHE_SIZE;i++) { 
      if (wt_iu_cache[i].state != INVALID && va == wt_iu_cache[i].va) {
	 return(TRUE);
      }
   }
   return(FALSE);
}
/*-----------------------------------------------------------------------*/
boolean 
d_wt_match (unsigned ea) 
{
   int i;
   for (i=0;i<(fl_conf.max_ut_buf+1);i++) { 
      if (wt_unifd_cache[i].state != INVALID && 
	  (wt_unifd_cache[i].cmd  == READ ||
	   wt_unifd_cache[i].cmd  == WRITE)) {
	 return(wt_unifd_cache[i].ea);
      }
   }
   return(FALSE);
}
/*-----------------------------------------------------------------------*/
void 
do_stream_buffer()
{
   wait_buffer * req = 0; 
   int i;
   
   
   /* Get the earliest valid request from the instruction queue */
   for (i=0;i<WT_IU_CACHE_SIZE;i++) { 

      if (wt_iu_cache[i].state == VALID) { 
	 wt_iu_cache[i].time = clk_cycle;
	 if (req) {
	    /* Get earliest request */
	    if (wt_iu_cache[i].time < req->time ) {
	       req = &wt_iu_cache[i];
	    }
	 } 
	 else { req = &wt_iu_cache[i]; }
      }
   }
   
   /* if we found a new request, look it up in stream buffer */
   if (req) {
      for (i=0;i<(fl_conf.istream_buffer);i++) { 
	 if (((stream_buffer[i].state == COMPLETED) ||
	      (stream_buffer[i].state == INPROGRESS))&& 
	     ((stream_buffer[i].va >> fl_conf.ushft) ==
	      (req->va >> fl_conf.ushft))) { 

	    /* we found a match in streaming buffer */
	    if (stream_buffer[i].state == COMPLETED) {
	       req->state = COMPLETED;
	       stream_buffer[i].state = INVALID;
	       stream_buffer_hit++;
#ifdef DEBUG
	       printf("STREAMHIT\n");
#endif 
	    }

	    break;
	 }
      }
   }
}
/*-----------------------------------------------------------------------*/
